@@ -17,9 +17,9 @@
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
-    You should have received a copy of the GNU Lesser General Public
+    You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -60,6 +60,8 @@ correct_goal(Module:Goal, _, _, Module:Goal) :-
 	(   var(Module)
 	;   var(Goal)
 	), !.
+correct_goal(Vars^Goal0, M, Bindings, Vars^Goal) :- !, % setof/bagof
+	correct_goal(Goal0, M, Bindings, Goal).
 correct_goal(Module:Goal0, _, Bindings, Module:Goal) :-
 	current_predicate(_, Module:Goal0), !,
 	correct_meta_arguments(Goal0, Module, Bindings, Goal).
@@ -94,13 +96,27 @@ dwim_existence_error(fail, _) :- !.
 dwim_existence_error(Unknown, PredSpec) :-
 	'$module'(TypeIn, TypeIn),
 	unqualify_if_context(TypeIn, PredSpec, Spec),
-	Error = error(existence_error(procedure, Spec),
-		     context(toplevel, 'DWIM could not correct goal')),
+	(   no_context(Spec)
+	->  true
+	;   Context = context(toplevel, 'DWIM could not correct goal')
+	),
+	Error = error(existence_error(procedure, Spec), Context),
 	(   Unknown == error
 	->  throw(Error)
 	;   print_message(warning, Error)
 	).
 
+%%	no_context(+PI) is semidet.
+%
+%	True if we should omit the DWIM message because messages.pl
+%	gives an additional explanation.
+
+no_context((^)/2).
+no_context((:-)/2).
+no_context((:-)/1).
+no_context((?-)/1).
+
+
 %%	correct_meta_arguments(:Goal, +Module, +Bindings, -Final) is det.
 %
 %	Correct possible meta-arguments. This performs a recursive check
@@ -135,20 +151,21 @@ correct_margs(A, Arity, MHead, GoalIn, GoalOut, M, Bindings) :-
 %	replace the head in Goal.
 
 correct_literal(Goal, Bindings, [Dwim], DwimGoal) :-
-	strip_module(Goal, _, G1),
+	strip_module(Goal, CM, G1),
 	strip_module(Dwim, DM, G2),
 	functor(G1, _, Arity),
 	functor(G2, Name, Arity), !,	% same arity: we can replace arguments
 	G1 =.. [_|Arguments],
 	G2 =.. [Name|Arguments],
-	'$module'(TypeIn, TypeIn),
-	(   (   current_predicate(TypeIn:Name/Arity)
-	    ->	ConfirmGoal = G2
-	    ;	'$prefix_module'(DM, TypeIn, G2, ConfirmGoal)
+	(   (   current_predicate(CM:Name/Arity)
+	    ->	ConfirmGoal = G2,
+		DwimGoal = CM:G2
+	    ;	'$prefix_module'(DM, CM, G2, ConfirmGoal),
+		DwimGoal = ConfirmGoal
 	    ),
 	    goal_name(ConfirmGoal, Bindings, String),
 	    '$confirm'(dwim_correct(String))
-	->  DwimGoal = DM:G2
+	->  true
 	;   DwimGoal = Goal
 	).
 correct_literal(Goal, Bindings, Dwims, NewGoal) :-
@@ -337,11 +354,11 @@ find_definitions([H0|T0], M, [H|T]) :-
 find_definition(Head, _, Def) :-
 	strip_module(Head, _, Plain),
 	callable(Plain),
-	functor(Plain, Name, Arity),
-	(   current_predicate(system:Name/Arity)
-	->  Def = system:Plain
-	;   predicate_property(Head, imported_from(Module))
-	->  Def = Module:Plain
+	(   predicate_property(Head, imported_from(Module))
+	->  (   predicate_property(system:Plain, imported_from(Module))
+	    ->	Def = system:Plain
+	    ;	Def = Module:Plain
+	    )
 	;   Def = Head
 	).
 
@@ -365,12 +382,10 @@ dwim_predicate_list(M:Head, DWIMs) :-
 	setof(DWIM, dwim_pred(M:Head, DWIM), DWIMs), !.
 dwim_predicate_list(Head, DWIMs) :-
 	setof(DWIM, '$similar_module'(Head, DWIM), DWIMs), !.
-dwim_predicate_list(M:Goal, DWIMs) :-
+dwim_predicate_list(_:Goal, DWIMs) :-
 	setof(Module:Goal,
-	      ( current_module(Module),
-		default_module(Module, M),
-		current_predicate(_, Module:Goal)
-	      ), DWIMs).
+	      current_predicate(_, Module:Goal),
+	      DWIMs).
 
 %%	dwim_pred(:Head, -DWIM) is nondet.
 %
