@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        J.Wielemaker@uva.nl
+    E-mail:        J.Wielemaker@ca.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2009, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -17,9 +16,9 @@
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
-    You should have received a copy of the GNU Lesser General Public
+    You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -39,6 +38,7 @@
 	prolog:message/3,		% entire message
 	prolog:error_message/3,		% 1-st argument of error term
 	prolog:message_context/3,	% Context of error messages
+	prolog:message_location/3,	% (File) location of error messages
 	prolog:message_line_element/2.	% Extend printing
 :- discontiguous
 	prolog_message/3.
@@ -57,9 +57,16 @@
 %	    * flush
 %	    Used only as last element of the list.   Simply flush the
 %	    output instead of producing a final newline.
+%	    * at_same_line
+%	    Start the messages at the same line (instead of using ~N)
 
 translate_message(Term) -->
 	translate_message2(Term), !.
+translate_message(Term) -->
+	{ Term = error(_, _) },
+	[ 'Unknown exception: ~p'-[Term] ].
+translate_message(Term) -->
+	[ 'Unknown message: ~p'-[Term] ].
 
 translate_message2(Term) -->
 	{var(Term)}, !,
@@ -73,7 +80,7 @@ translate_message2(error(resource_error(stack), Name)) -->
 translate_message2(error(resource_error(Missing), _)) -->
 	[ 'Not enough resources: ~w'-[Missing] ].
 translate_message2(error(ISO, SWI)) -->
-	swi_context(SWI),
+	swi_location(SWI),
 	term_message(ISO),
 	swi_extra(SWI).
 translate_message2('$aborted') -->
@@ -82,11 +89,6 @@ translate_message2(message_lines(Lines), L, T) :- % deal with old C-warning()
 	make_message_lines(Lines, L, T).
 translate_message2(format(Fmt, Args)) -->
 	[ Fmt-Args ].
-translate_message2(Term) -->
-	{ Term = error(_, _) },
-	[ 'Unknown exception: ~p'-[Term] ].
-translate_message2(Term) -->
-	[ 'Unknown message: ~p'-[Term] ].
 
 make_message_lines([], T, T) :- !.
 make_message_lines([Last],  ['~w'-[Last]|T], T) :- !.
@@ -130,6 +132,11 @@ iso_message(permission_error(Action, built_in_procedure, Pred)) -->
 	    ]
 	;   []
 	).
+iso_message(permission_error(import_into(Dest), procedure, Pred)) -->
+	[ 'No permission to import ~p into ~w'-[Pred, Dest] ].
+iso_message(permission_error(Action, static_procedure, Proc)) -->
+	[ 'No permission to ~w static procedure `~p'''-[Action, Proc] ],
+	defined_definition('Defined', Proc).
 iso_message(permission_error(Action, Type, Object)) -->
 	[ 'No permission to ~w ~w `~p'''-[Action, Type, Object] ].
 iso_message(evaluation_error(Which)) -->
@@ -141,6 +148,8 @@ iso_message(existence_error(Type, Object)) -->
 	[ '~w `~p'' does not exist'-[Type, Object] ].
 iso_message(busy(Type, Object)) -->
 	[ '~w `~p'' is busy'-[Type, Object] ].
+iso_message(syntax_error(swi_backslash_newline)) -->
+	[ 'Deprecated ... \\<newline><white>*.  Use \\c' ].
 iso_message(syntax_error(Id)) -->
 	[ 'Syntax error: ' ],
 	syntax_error(Id).
@@ -204,6 +213,8 @@ syntax_error(cannot_start_term) -->
 	[ 'Illegal start of term' ].
 syntax_error(punct(Punct, End)) -->
 	[ 'Unexpected `~w\' before `~w\''-[Punct, End] ].
+syntax_error(undefined_char_escape(C)) -->
+	[ 'Undefined character escape in quoted atom or string: `\\~w\''-[C] ].
 syntax_error(Message) -->
 	[ '~w'-[Message] ].
 
@@ -263,8 +274,8 @@ swi_message(context_error(nodirective, Goal)) -->
 	[ 'Wrong context: ~p can only be used in a directive'-[PI] ].
 swi_message(context_error(edit, no_default_file)) -->
 	(   { current_prolog_flag(windows, true) }
-	->  [ 'Edit/0 can only be used after opening a \
-	      Prolog file by double-clicking it' ]
+	->  [ 'Edit/0 can only be used after opening a \c
+	       Prolog file by double-clicking it' ]
 	;   [ 'Edit/0 can only be used with the "-s file" commandline option'
 	    ]
 	).
@@ -272,22 +283,36 @@ swi_message(format_argument_type(Fmt, Arg)) -->
 	[ 'Illegal argument to format sequence ~~~w: ~p'-[Fmt, Arg] ].
 swi_message(format(Msg)) -->
 	[ 'Format error: ~w'-[Msg] ].
+swi_message(conditional_compilation_error(unterminated, Where)) -->
+	[ 'Unterminated conditional compilation from '-[] ],
+	cond_location(Where).
+swi_message(conditional_compilation_error(no_if, What)) -->
+	[ ':- ~w without :- if'-[What] ].
 
-swi_context(X) -->
+cond_location(File:Line) -->
+	{ file_base_name(File, Base) },
+	[ '~w:~d'-[Base, Line] ].
+
+swi_location(X) -->
 	{ var(X)
 	}, !,
 	[].
-swi_context(context(Caller, _Msg)) -->
+swi_location(Context) -->
+	prolog:message_location(Context), !.
+swi_location(context(Caller, _Msg)) -->
 	{ ground(Caller)
 	}, !,
 	caller(Caller).
-swi_context(file(Path, Line, -1, _CharNo)) --> !,
+swi_location(file(Path, Line, -1, _CharNo)) --> !,
 	[ '~w:~d: '-[Path, Line] ].
-swi_context(file(Path, Line, LinePos, _CharNo)) -->
+swi_location(file(Path, Line, LinePos, _CharNo)) -->
 	[ '~w:~d:~d: '-[Path, Line, LinePos] ].
-swi_context(stream(Stream, Line, LinePos, _CharNo)) -->
-	[ 'Stream ~w:~d:~d '-[Stream, LinePos, Line] ].
-swi_context(_) -->
+swi_location(stream(Stream, Line, LinePos, CharNo)) -->
+	(   { stream_property(Stream, file_name(File)) }
+	->  swi_location(file(File, Line, LinePos, CharNo))
+        ;   [ 'Stream ~w:~d:~d '-[Stream, LinePos, Line] ]
+	).
+swi_location(_) -->
 	[].
 
 caller(system:'$record_clause'/3) --> !,
@@ -313,13 +338,25 @@ swi_extra(context(_, Msg)) -->
 	{ nonvar(Msg),
 	  Msg \== ''
 	}, !,
-	[ ' (~w)'-[Msg] ].
+	swi_comment(Msg).
 swi_extra(string(String, CharPos)) -->
 	{ sub_string(String, 0, CharPos, _, Before),
 	  sub_string(String, CharPos, _, 0, After)
 	},
 	[ nl, '~w'-[Before], nl, '** here **', nl, '~w'-[After] ].
 swi_extra(_) -->
+	[].
+
+swi_comment(already_from(Module)) --> !,
+	[ ' (already imported from ~q)'-[Module] ].
+swi_comment(Msg) -->
+	[ ' (~w)'-[Msg] ].
+
+
+thread_context -->
+	{ thread_self(Me), Me \== main }, !,
+	['[Thread ~w] '-[Me]].
+thread_context -->
 	[].
 
 		 /*******************************
@@ -346,7 +383,10 @@ prolog_message(initialization_exception(E)) -->
 	translate_message(E).
 prolog_message(unhandled_exception(E)) -->
 	[ 'Unhandled exception: ' ],
-	translate_message(E).
+	(   translate_message2(E)
+	->  []
+	;   [ '~p'-[E] ]
+	).
 prolog_message(goal_failed(Goal, Context)) -->
 	[ 'Goal (~w) failed: ~p'-[Goal, Context] ].
 prolog_message(no_current_module(Module)) -->
@@ -383,6 +423,12 @@ prolog_message(interrupt(end)) -->
 	[ 'continue' ].
 prolog_message(interrupt(trace)) -->
 	[ 'continue (trace mode)' ].
+prolog_message(unknown_in_module_user) -->
+	[ 'Using a non-error value for unknown in the global module', nl,
+	  'causes most of the development environment to stop working.', nl,
+	  'Please use :- dynamic or limit usage of unknown to a module.', nl,
+	  'See http://www.swi-prolog.org/howto/database.html'
+	].
 
 
 		 /*******************************
@@ -404,11 +450,22 @@ prolog_message(cannot_redefine_comma) -->
 prolog_message(illegal_autoload_index(Dir, Term)) -->
 	[ 'Illegal term in INDEX file of directory ~w: ~w'-[Dir, Term] ].
 prolog_message(redefined_procedure(Type, Proc)) -->
-	[ 'Redefined ~w procedure ~p'-[Type, Proc] ].
+	[ 'Redefined ~w procedure ~p'-[Type, Proc] ],
+	defined_definition('Previously defined', Proc).
 prolog_message(declare_module(Module, abolish(Predicates))) -->
 	[ 'Loading module ~w abolished: ~p'-[Module, Predicates] ].
+prolog_message(import_private(Module, Private)) -->
+	[ 'import/1: ~p is not exported (still imported into ~q)'-
+	  [Private, Module]
+	].
+prolog_message(ignored_weak_import(Into, From:PI)) -->
+	[ 'Local definition of ~p overrides weak import from ~q'-
+	  [Into:PI, From]
+	].
 prolog_message(undefined_export(Module, PI)) -->
 	[ 'Exported procedure ~q:~q is not defined'-[Module, PI] ].
+prolog_message(no_exported_op(Module, Op)) -->
+	[ 'Operator ~q:~q is not exported (still defined)'-[Module, Op] ].
 prolog_message(discontiguous((-)/2)) -->
 	prolog_message(minus_in_identifier).
 prolog_message(discontiguous(Proc)) -->
@@ -417,12 +474,19 @@ prolog_message(load_file(start(Level, File))) -->
 	[ '~|~t~*+Loading '-[Level] ],
 	load_file(File),
 	[ ' ...' ].
-prolog_message(load_file(done(Level, File, Action, Module, Time, Heap))) -->
+prolog_message(include_file(start(Level, File))) -->
+	[ '~|~t~*+include '-[Level] ],
+	load_file(File),
+	[ ' ...' ].
+prolog_message(include_file(done(Level, File))) -->
+	[ '~|~t~*+included '-[Level] ],
+	load_file(File).
+prolog_message(load_file(done(Level, File, Action, Module, Time, Clauses))) -->
 	[ '~|~t~*+'-[Level] ],
 	load_file(File),
 	[ ' ~w'-[Action] ],
 	load_module(Module),
-	[ ' ~2f sec, ~D bytes'-[Time, Heap] ].
+	[ ' ~2f sec, ~D clauses'-[Time, Clauses] ].
 prolog_message(dwim_undefined(Goal, Alternatives)) -->
 	{ goal_to_predicate_indicator(Goal, Pred)
 	},
@@ -449,7 +513,18 @@ prolog_message(redefine_module(Module, OldFile, File)) -->
 	].
 prolog_message(redefine_module_reply) -->
 	[ 'Please answer y(es), n(o) or a(bort)' ].
+prolog_message(reloaded_in_module(Absolute, OldContext, LM)) -->
+	[ '~w was previously loaded in module ~w'-[Absolute, OldContext], nl,
+	  '\tnow it is reloaded into module ~w'-[LM] ].
 
+defined_definition(Message, Spec) -->
+	{ strip_module(user:Spec, M, Name/Arity),
+	  functor(Head, Name, Arity),
+	  predicate_property(M:Head, file(File)),
+	  predicate_property(M:Head, line_count(Line))
+	}, !,
+	[ nl, '~w at ~w:~d'-[Message, File,Line] ].
+defined_definition(_, _) --> [].
 
 used_search([]) -->
 	[].
@@ -496,6 +571,7 @@ prolog_message(file_search(tried(Spec, Cond), Path)) -->
 		 *******************************/
 
 prolog_message(gc(start)) -->
+	thread_context,
 	[ 'GC: ', flush ].
 prolog_message(gc(done(G, T, Time, UG, UT, RG, RT))) -->
 	[ at_same_line,
@@ -503,6 +579,7 @@ prolog_message(gc(done(G, T, Time, UG, UT, RG, RT))) -->
 	  [G, T, Time, UG, UT, RG, RT]
 	].
 prolog_message(shift_stacks(start(_L,_G,_T))) -->
+	thread_context,
 	[ 'Stack-shift: ', flush ].
 prolog_message(shift_stacks(done(Time, L, G, T))) -->
 	{ LKB is L//1024,
@@ -514,9 +591,8 @@ prolog_message(shift_stacks(done(Time, L, G, T))) -->
 	  [LKB, GKB, TKB, Time]
 	].
 prolog_message(agc(start)) -->
-	{ thread_self(Me)
-	},
-	[ 'AGC: [~w] '-[Me], flush ].
+	thread_context,
+	[ 'AGC: ', flush ].
 prolog_message(agc(done(Collected, Remaining, Time))) -->
 	[ at_same_line,
 	  'reclaimed ~D atoms in ~2f sec. (remaining: ~D)'-
@@ -538,7 +614,8 @@ prolog_message(make(done(_Files))) -->
 prolog_message(make(library_index(Dir))) -->
 	[ 'Updating index for library ~w'-[Dir] ].
 prolog_message(autoload(Pred, File)) -->
-	[ 'Autoloading ~p from ~w'-[Pred, File] ].
+	thread_context,
+	[ 'autoloading ~p from ~w'-[Pred, File] ].
 prolog_message(autoload(read_index(Dir))) -->
 	[ 'Loading autoload index for ~w'-[Dir] ].
 
@@ -568,7 +645,7 @@ prolog_message(threads) -->
 prolog_message(threads) -->
 	[].
 prolog_message(copyright) -->
-	[ 'Copyright (c) 1990-2011 University of Amsterdam, VU Amsterdam', nl,
+	[ 'Copyright (c) 1990-2012 University of Amsterdam, VU Amsterdam', nl,
 	  'SWI-Prolog comes with ABSOLUTELY NO WARRANTY. This is free software,', nl,
 	  'and you are welcome to redistribute it under certain conditions.', nl,
 	  'Please visit http://www.swi-prolog.org for details.'
@@ -596,10 +673,10 @@ prolog_message(about) -->
 	prolog_message(copyright).
 prolog_message(halt) -->
 	[ 'halt' ].
-prolog_message(break(enter(Level))) -->
+prolog_message(break(begin, Level)) -->
 	[ 'Break level ~d'-[Level] ].
-prolog_message(break(exit(Level))) -->
-	[ nl, 'Exit break level ~d'-[Level] ].
+prolog_message(break(end, Level)) -->
+	[ 'Exit break level ~d'-[Level] ].
 prolog_message(var_query(_)) -->
 	[ '... 1,000,000 ............ 10,000,000 years later', nl, nl,
 	  '~t~8|>> 42 << (last release gives the question)'
@@ -611,10 +688,10 @@ prolog_message(query(QueryResult)) -->
 	query_result(QueryResult).
 
 query_result(no) -->		% failure
-	[ 'false.'-[] ],
+	[ ansi([bold,fg(red)], 'false.', []) ],
 	extra_line.
 query_result(yes([])) --> !,	% prompt_alternatives_on: groundness
-	[ 'true.'-[] ],
+	[ ansi(bold, 'true.', []) ],
 	extra_line.
 query_result(yes(Residuals)) -->
 	result([], Residuals),
@@ -650,13 +727,13 @@ prompt(Answer, _, _) --> !,
 	prompt(Answer, non_empty).
 
 prompt(yes, empty) --> !,
-	[ 'true.'-[] ],
+	[ ansi(bold, 'true.', []) ],
 	extra_line.
 prompt(yes, _) --> !,
 	[ full_stop ],
 	extra_line.
 prompt(more, empty) --> !,
-	[ 'true '-[], flush ].
+	[ ansi(bold, 'true ', []), flush ].
 prompt(more, _) --> !,
 	[ ' '-[], flush ].
 
@@ -728,9 +805,9 @@ extra_line -->
 extra_line -->
 	[].
 
-prolog_message(if_tty(Text)) -->
+prolog_message(if_tty(Message)) -->
 	(   {current_prolog_flag(tty_control, true)}
-	->  [ at_same_line, '~w'-[Text] ]
+	->  [ at_same_line | Message ]
 	;   []
 	).
 prolog_message(halt(Reason)) -->
@@ -826,7 +903,7 @@ tracing_list([trace(Head, Ports)|T]) -->
 prolog_message(frame(Frame, backtrace, _PC)) --> !,
 	{ prolog_frame_attribute(Frame, level, Level)
 	},
-	[ '~t[~D] ~10|'-[Level] ],
+	[ ansi(bold, '~t[~D] ~10|', [Level]) ],
 	frame_context(Frame),
 	frame_goal(Frame).
 prolog_message(frame(Frame, choice, PC)) --> !,
@@ -888,16 +965,16 @@ frame_flags(Frame) -->
 	[ '~w~w '-[T, S] ].
 
 port(Port) -->
-	{ port_name(Port, Name)
+	{ port_name(Port, Colour, Name)
 	}, !,
-	[ '~w: '-[Name] ].
+	[ ansi([bold,fg(Colour)], '~w: ', [Name]) ].
 
-port_name(call,	     'Call').
-port_name(exit,	     'Exit').
-port_name(fail,	     'Fail').
-port_name(redo,	     'Redo').
-port_name(unify,     'Unify').
-port_name(exception, 'Exception').
+port_name(call,	     green,   'Call').
+port_name(exit,	     green,   'Exit').
+port_name(fail,	     red,     'Fail').
+port_name(redo,	     yellow,  'Redo').
+port_name(unify,     blue,    'Unify').
+port_name(exception, magenta, 'Exception').
 
 clean_goal(M:Goal, Goal) :-
 	hidden_module(M), !.
@@ -925,8 +1002,22 @@ prolog_message(abnormal_thread_completion(Goal, exception(Ex))) --> !,
 	translate_message(Ex).
 prolog_message(abnormal_thread_completion(Goal, fail)) -->
 	[ 'Thread running "~p" died due to failure'-[Goal] ].
-prolog_message(threads_not_died(Count)) -->
-	[ '~D threads wouldn\'t die'-[Count] ].
+prolog_message(threads_not_died(Running)) -->
+	[ 'The following threads wouldn\'t die: ~p'-[Running] ].
+
+
+		 /*******************************
+		 *	       PACKS		*
+		 *******************************/
+
+prolog_message(pack(attached(Pack, BaseDir))) -->
+	[ 'Attached package ~w at ~q'-[Pack, BaseDir] ].
+prolog_message(pack(duplicate(Entry, OldDir, Dir))) -->
+	[ 'Package ~w already attached at ~q.'-[Entry,OldDir], nl,
+	  '\tIgnoring version from ~q'- [Entry, OldDir, Dir]
+	].
+prolog_message(pack(no_arch(Entry, Arch))) -->
+	[ 'Package ~w: no binary for architecture ~w'-[Entry, Arch] ].
 
 
 		 /*******************************
@@ -937,8 +1028,10 @@ prolog_message(threads_not_died(Count)) -->
 	user:message_hook/3.
 :- dynamic
 	user:message_hook/3.
+:- thread_local
+	user:thread_message_hook/3.
 
-%	print_message(+Kind, +Term)
+%%	print_message(+Kind, +Term)
 %
 %	Print an error message using a term as generated by the exception
 %	system.
@@ -947,7 +1040,10 @@ print_message(Level, Term) :-
 	(   must_print(Level, Term)
 	->  (   translate_message(Term, Lines, [])
 	    ->  (   nonvar(Term),
-		    notrace(user:message_hook(Term, Level, Lines))
+		    (	notrace(user:thread_message_hook(Term, Level, Lines))
+		    ->  true
+		    ;   notrace(user:message_hook(Term, Level, Lines))
+		    )
 		->  true
 		;   print_system_message(Term, Level, Lines)
 		)
@@ -955,7 +1051,7 @@ print_message(Level, Term) :-
 	;   true
 	).
 
-%	print_system_message(+Term, +Level, +Lines)
+%%	print_system_message(+Term, +Kind, +Lines)
 %
 %	Print the message if the user did not intecept the message.
 %	The first is used for errors and warnings that can be related
@@ -967,85 +1063,129 @@ print_system_message(_, informational, _) :-
 	current_prolog_flag(verbose, silent), !.
 print_system_message(_, banner, _) :-
 	current_prolog_flag(verbose, silent), !.
-print_system_message(Term, Level, Lines) :-
+print_system_message(_, _, []) :- !.
+print_system_message(Term, Kind, Lines) :-
 	flush_output(user_output),
 	source_location(File, Line),
 	Term \= error(syntax_error(_), _),
-	prefix(Level, Prefix, LinePrefix, PostFix, Wait, Stream), !,
-	format(Stream, Prefix, [File, Line]),
-	print_message_lines(Stream, LinePrefix, Lines),
-	format(Stream, PostFix, []),
-	(   Wait > 0
+	msg_property(Kind, location_prefix(File:Line, LocPrefix, LinePrefix)), !,
+	insert_prefix(Lines, LinePrefix, PrefixLines),
+	'$append'([ begin(Kind, Ctx),
+		    LocPrefix,
+		    nl
+		  | PrefixLines
+		  ],
+		  [ end(Ctx)
+		  ],
+		  AllLines),
+	msg_property(Kind, stream(Stream)),
+	print_message_lines(Stream, AllLines),
+	(   msg_property(Kind, wait(Wait)),
+	    Wait > 0
 	->  sleep(Wait)
 	;   true
 	).
-print_system_message(_, Level, Lines) :-
-	flush_output(user_output),
-	prefix(Level, LinePrefix, Stream), !,
-	print_message_lines(Stream, LinePrefix, Lines).
-print_system_message(_, Level, _) :-
-	\+ prefix(Level, _, _),
-	throw(error(domain_error(message_kind, Level), _)).
+print_system_message(_, Kind, Lines) :-
+	msg_property(Kind, stream(Stream)),
+	print_message_lines(Stream, kind(Kind), Lines).
 
-prefix(error,	      'ERROR: ~w:~d:~n',   '\t', '', 0.5, user_error).
-prefix(warning,	      'Warning: ~w:~d:~n', '\t', '', 0,   user_error).
+:- multifile
+	user:message_property/2.
 
-prefix(help,	      '',          user_error).
-prefix(query,	      '',          user_error).
-prefix(debug,	      '',          user_output).
-prefix(warning,	      Prefix,      user_error) :-
+msg_property(Kind, Property) :-
+	user:message_property(Kind, Property), !.
+msg_property(Kind, prefix(Prefix)) :-
+	msg_prefix(Kind, Prefix), !.
+msg_property(_, prefix('~N')) :- !.
+msg_property(_, stream(user_error)) :- !.
+msg_property(error,
+	     location_prefix(File:Line,
+			     '~NERROR: ~w:~d:'-[File,Line], '~N\t')) :- !.
+msg_property(warning,
+	     location_prefix(File:Line,
+			     '~NWarning: ~w:~d:'-[File,Line], '~N\t')) :- !.
+msg_property(error,   wait(0.1)) :- !.
+
+msg_prefix(debug(_),      '~N% ').
+msg_prefix(warning,	      Prefix) :-
 	thread_self(Id),
 	(   Id == main
-	->  Prefix = 'Warning: '
-	;   atomic_list_concat(['Warning: [Thread ', Id, '] '], Prefix)
+	->  Prefix = '~NWarning: '
+	;   Prefix = '~NWarning: [Thread ~w] '-Id
 	).
-prefix(error,	      Prefix,   user_error) :-
+msg_prefix(error,	      Prefix) :-
 	thread_self(Id),
 	(   Id == main
-	->  Prefix = 'ERROR: '
-	;   atomic_list_concat(['ERROR: [Thread ', Id, '] '], Prefix)
+	->  Prefix = '~NERROR: '
+	;   Prefix = '~NERROR: [Thread ~w] '-Id
 	).
-prefix(banner,	      '',	   user_error).
-prefix(informational, '% ',        user_error).
+msg_prefix(informational, '~N% ').
+msg_prefix(information,   '~N% ').
 
-%	print_message_lines(+Stream, +Prefix, +Lines)
+%%	print_message_lines(+Stream, +PrefixOrKind, +Lines)
 %
 %	Quintus compatibility predicate to print message lines using
 %	a prefix.
 
-print_message_lines(_, _, []) :- !.
-print_message_lines(S, P, [at_same_line|Lines]) :- !,
-	print_message_line(S, Lines, Rest),
-	print_message_lines(S, P, Rest).
-print_message_lines(S, P, Lines) :-
-	atom_concat('~N', P, Prefix),
-	format(S, Prefix, []),
-	print_message_line(S, Lines, Rest),
-	print_message_lines(S, P, Rest).
+print_message_lines(Stream, kind(Kind), Lines) :- !,
+	msg_property(Kind, prefix(Prefix)),
+	insert_prefix(Lines, Prefix, PrefixLines),
+	'$append'([ begin(Kind, Ctx)
+		  | PrefixLines
+		  ],
+		  [ end(Ctx)
+		  ],
+		  AllLines),
+	print_message_lines(Stream, AllLines).
+print_message_lines(Stream, Prefix, Lines) :-
+	insert_prefix(Lines, Prefix, PrefixLines),
+	print_message_lines(Stream, PrefixLines).
 
-print_message_line(S, [flush], []) :- !,
-	flush_output(S).
-print_message_line(S, [], []) :- !,
-	nl(S).
-print_message_line(S, [nl|T], T) :- !,
-	nl(S).
-print_message_line(S, [H|T0], T) :- !,
+%%	insert_prefix(+Lines, +Prefix, -PrefixedLines)
+
+insert_prefix([at_same_line|Lines0], Prefix, Lines) :- !,
+	prefix_nl(Lines0, Prefix, Lines).
+insert_prefix(Lines0, Prefix, [prefix(Prefix)|Lines]) :-
+	prefix_nl(Lines0, Prefix, Lines).
+
+prefix_nl([], _, [nl]).
+prefix_nl([nl], _, [nl]) :- !.
+prefix_nl([flush], _, [flush]) :- !.
+prefix_nl([nl|T0], Prefix, [nl, prefix(Prefix)|T]) :- !,
+	prefix_nl(T0, Prefix, T).
+prefix_nl([H|T0], Prefix, [H|T]) :-
+	prefix_nl(T0, Prefix, T).
+
+%%	print_message_lines(+Stream, +Lines)
+
+print_message_lines(_, []) :- !.
+print_message_lines(S, [H|T]) :-
 	line_element(S, H),
-	print_message_line(S, T0, T).
+	print_message_lines(S, T).
 
+line_element(S, E) :-
+	notrace(prolog:message_line_element(S, E)), !.
 line_element(S, full_stop) :- !,
 	'$put_token'(S, '.').		% insert space if needed.
+line_element(S, nl) :- !,
+	nl(S).
+line_element(S, prefix(Fmt-Args)) :- !,
+	format(S, Fmt, Args).
+line_element(S, prefix(Fmt)) :- !,
+	format(S, Fmt, []).
+line_element(S, flush) :- !,
+	flush_output(S).
 line_element(S, Fmt-Args) :- !,
 	format(S, Fmt, Args).
-line_element(S, E) :-
-	prolog:message_line_element(S, E), !.
 line_element(S, ansi(_, Fmt, Args)) :- !,
 	format(S, Fmt, Args).
+line_element(_, begin(_Level, _Ctx)) :- !.
+line_element(_, end(_Ctx)) :- !.
 line_element(S, Fmt) :-
 	format(S, Fmt, []).
 
 
-%	message_to_string(+Term, -String)
+%%	message_to_string(+Term, -String)
 %
 %	Translate an error term into a string
 
@@ -1061,10 +1201,13 @@ actions_to_format([Term, nl], Fmt, Args) :- !,
 actions_to_format([nl|T], Fmt, Args) :- !,
 	actions_to_format(T, Fmt0, Args),
 	atom_concat('~n', Fmt0, Fmt).
+actions_to_format([Skip|T], Fmt, Args) :-
+	action_skip(Skip), !,
+	actions_to_format(T, Fmt, Args).
 actions_to_format([Fmt0-Args0|Tail], Fmt, Args) :- !,
         actions_to_format(Tail, Fmt1, Args1),
         atom_concat(Fmt0, Fmt1, Fmt),
-        '$append'(Args0, Args1, Args).
+	append_args(Args0, Args1, Args).
 actions_to_format([Term|Tail], Fmt, Args) :-
 	atomic(Term), !,
         actions_to_format(Tail, Fmt1, Args),
@@ -1072,7 +1215,19 @@ actions_to_format([Term|Tail], Fmt, Args) :-
 actions_to_format([Term|Tail], Fmt, Args) :-
         actions_to_format(Tail, Fmt1, Args1),
         atom_concat('~w', Fmt1, Fmt),
-        '$append'([Term], Args1, Args).
+	append_args([Term], Args1, Args).
+
+action_skip(at_same_line).
+action_skip(ansi(_Attrs, _Fmt, _Args)).
+action_skip(begin(_Level, _Ctx)).
+action_skip(end(_Ctx)).
+
+append_args(M:Args0, Args1, M:Args) :- !,
+	strip_module(Args1, _, A1),
+	'$append'(Args0, A1, Args).
+append_args(Args0, Args1, Args) :-
+	strip_module(Args1, _, A1),
+	'$append'(Args0, A1, Args).
 
 
 		 /*******************************
@@ -1082,13 +1237,13 @@ actions_to_format([Term|Tail], Fmt, Args) :-
 :- dynamic
 	printed/2.
 
-%	print_once(Message, Level)
+%%	print_once(Message, Level)
 %
 %	True for messages that must be printed only once.
 
 print_once(compatibility(_), _).
 
-%	must_print(+Level, +Message)
+%%	must_print(+Level, +Message)
 %
 %	True if the message must be printed.
 
