@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
-    WWW:           http://www.swi.psy.uva.nl/projects/xpce/
-    Copyright (C): 1985-2006, University of Amsterdam
+    E-mail:        J.Wielemaker@cs.vu.nl
+    WWW:           http://www.swi-prolog.org/projects/xpce/
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -32,6 +31,7 @@
 :- module(prolog_xref,
 	  [ xref_source/1,		% +Source
 	    xref_called/3,		% ?Source, ?Callable, ?By
+	    xref_called/4,		% ?Source, ?Callable, ?By, ?Cond
 	    xref_defined/3,		% ?Source. ?Callable, -How
 	    xref_definition_line/2,	% +How, -Line
 	    xref_exported/2,		% ?Source, ?Callable
@@ -54,9 +54,8 @@
 	    xref_defined_class/3	% ?Source, ?ClassName, -How
 	  ]).
 :- use_module(library(debug), [debug/3, debugging/1]).
-:- use_module(library(lists), [append/3, member/2]).
-:- use_module(library(operators),
-	      [pop_operators/0, push_op/3, push_operators/1]).
+:- use_module(library(lists), [append/3, member/2, select/3]).
+:- use_module(library(operators), [push_op/3]).
 :- use_module(library(shlib), [current_foreign_library/2]).
 :- use_module(library(prolog_source)).
 :- use_module(library(option)).
@@ -65,7 +64,7 @@
 :- use_module(library(debug)).
 
 :- dynamic
-	called/3,			% Head, Src, From
+	called/4,			% Head, Src, From, Cond
 	(dynamic)/3,			% Head, Src, Line
 	(thread_local)/3,		% Head, Src, Line
 	(multifile)/3,			% Head, Src, Line
@@ -96,11 +95,15 @@ This code is used in two places:
     goals and predicates depending on their references.
 
 @bug	meta_predicate/1 declarations take the module into consideration.
-	Predicates that are both avalaible as meta-predicate and normal
+	Predicates that are both available as meta-predicate and normal
 	(in different modules) are handled as meta-predicate in all
 	places.
 */
 
+:- predicate_options(xref_source_file/4, 4,
+		     [ file_type(oneof([txt,prolog,directory]))
+		     ]).
+
 
 		 /*******************************
 		 *	      HOOKS		*
@@ -125,18 +128,17 @@ This code is used in two places:
 :- multifile
 	prolog:called_by/2,		% +Goal, -Called
 	prolog:meta_goal/2,		% +Goal, -Pattern
-	prolog:hook/1.			% +Callable
+	prolog:hook/1,			% +Callable
+	prolog:generated_predicate/1.	% :PI
+
+:- meta_predicate
+	prolog:generated_predicate(:).
 
 :- dynamic
 	meta_goal/2.
 
-called_by(Goal, Called) :-
-	prolog:called_by(Goal, Called), !.
-called_by(on_signal(_,_,New), [New+1]) :-
-	(   new == throw
-	;   new == default
-	), !, fail.
-
+:- meta_predicate
+	process_predicates(2, +, +).
 
 		 /*******************************
 		 *	     BUILT-INS		*
@@ -159,6 +161,10 @@ system_predicate(Goal) :-
 verbose :-
 	debugging(xref).
 
+:- thread_local
+	xref_input/2.			% File, Stream
+
+
 %%	xref_source(+Source) is det.
 %
 %	Generate the cross-reference data  for   Source  if  not already
@@ -183,9 +189,10 @@ xref_source(Source) :-
 	).
 
 do_xref(Src) :-
-	setup_call_cleanup(xref_setup(Src, In, State),
-			   collect(Src, In),
-			   xref_cleanup(State)).
+	setup_call_cleanup(
+	    xref_setup(Src, In, State),
+	    collect(Src, Src, In),
+	    xref_cleanup(State)).
 
 last_modified(Source, Modified) :-
 	prolog:xref_source_time(Source, Modified), !.
@@ -194,12 +201,11 @@ last_modified(Source, Modified) :-
 	exists_file(Source),
 	time_file(Source, Modified).
 
-:- thread_local
-	xref_stream/1.			% input stream
-
-xref_setup(Src, In, state(In, Xref, [SRef|HRefs])) :-
+xref_setup(Src, In, state(In, Dialect, Xref, [SRef|HRefs])) :-
+	current_prolog_flag(emulated_dialect, Dialect),
 	prolog_open_source(Src, In),
-	asserta(xref_stream(In), SRef),
+	set_initial_mode(In),
+	asserta(xref_input(Src, In), SRef),
 	set_xref(Xref),
 	(   verbose
 	->  HRefs = []
@@ -207,8 +213,9 @@ xref_setup(Src, In, state(In, Xref, [SRef|HRefs])) :-
 	    HRefs = [Ref]
 	).
 
-xref_cleanup(state(In, Xref, Refs)) :-
+xref_cleanup(state(In, Dialect, Xref, Refs)) :-
 	prolog_close_source(In),
+	set_prolog_flag(emulated_dialect, Dialect),
 	set_prolog_flag(xref, Xref),
 	maplist(erase, Refs).
 
@@ -216,13 +223,31 @@ set_xref(Xref) :-
 	current_prolog_flag(xref, Xref),
 	set_prolog_flag(xref, true).
 
+%%	set_initial_mode(+Stream) is det.
+%
+%	Set  the  initial  mode  for  processing    this   file  in  the
+%	cross-referencer. If the file is loaded, we use information from
+%	the previous load context, setting   the  appropriate module and
+%	dialect.
+
+set_initial_mode(Stream) :-
+	stream_property(Stream, file_name(Path)),
+	source_file_property(Path, load_context(M, _, Opts)), !,
+	'$set_source_module'(_, M),
+	(   option(dialect(Dialect), Opts)
+	->  expects_dialect(Dialect)
+	;   true
+	).
+set_initial_mode(_) :-
+	'$set_source_module'(_, user).
+
 
 %%	xref_input_stream(-Stream) is det.
 %
 %	Current input stream for cross-referencer.
 
 xref_input_stream(Stream) :-
-	xref_stream(Var), !,
+	xref_input(_, Var), !,
 	Stream = Var.
 
 %%	xref_push_op(Source, +Prec, +Type, :Name)
@@ -247,7 +272,7 @@ xref_push_op(Src, P, T, N0) :- !,
 
 xref_clean(Source) :-
 	prolog_canonical_source(Source, Src),
-	retractall(called(_, Src, _Origin)),
+	retractall(called(_, Src, _Origin, _Cond)),
 	retractall(dynamic(_, Src, Line)),
 	retractall(multifile(_, Src, Line)),
 	retractall(public(_, Src, Line)),
@@ -287,32 +312,39 @@ xref_done(Source, Time) :-
 
 
 %%	xref_called(?Source, ?Called, ?By) is nondet.
+%%	xref_called(?Source, ?Called, ?By, ?Cond) is nondet.
 %
 %	Enumerate the predicate-call relations. Predicate called by
 %	directives have a By '<directive>'.
 
 xref_called(Source, Called, By) :-
+	xref_called(Source, Called, By, _).
+
+xref_called(Source, Called, By, Cond) :-
 	(   ground(Source)
 	->  prolog_canonical_source(Source, Src)
 	;   Source = Src
 	),
-	called(Called, Src, By).
+	called(Called, Src, By, Cond).
 
 
 %%	xref_defined(?Source, +Goal, ?How) is nondet.
 %
-%	Test if Goal is accessible in Source. If this is the case, How
-%	specifies the reason why the predicate is accessible. Note that
+%	Test if Goal is accessible in Source.   If this is the case, How
+%	specifies the reason why the predicate  is accessible. Note that
 %	this predicate does not deal with built-in or global predicates,
-%	just locally defined and imported ones.  How is one of:
+%	just locally defined and imported ones.  How   is  one of of the
+%	terms below. Location is one of Line (an integer) or File:Line
+%	if the definition comes from an included (using :-
+%	include(File)) directive.
 %
-%	  * dynamic(Line)
-%	  * thread_local(Line)
-%	  * multifile(Line)
-%	  * public(Line)
-%	  * local(Line)
-%	  * foreign(Line)
-%	  * constraint(Line)
+%	  * dynamic(Location)
+%	  * thread_local(Location)
+%	  * multifile(Location)
+%	  * public(Location)
+%	  * local(Location)
+%	  * foreign(Location)
+%	  * constraint(Location)
 %	  * imported(From)
 
 xref_defined(Source, Called, How) :-
@@ -409,13 +441,32 @@ xref_defined_class(Source, Class, file(File)) :-
 	prolog_canonical_source(Source, Src),
 	defined_class(Class, _, _, Src, file(File)).
 
-collect(Src, In) :-
+:- thread_local
+	current_cond/1,
+	source_line/1.
+
+current_source_line(Line) :-
+	source_line(Var), !,
+	Line = Var.
+
+%%	collect(+Source, +File, +Stream)
+%
+%	Process data from Source. If File  \== Source, we are processing
+%	an included file. Stream is the stream   from  shich we read the
+%	program.
+
+collect(Src, File, In) :-
+	(   Src == File
+	->  SrcSpec = Line
+	;   SrcSpec = (File:Line)
+	),
 	repeat,
-	    catch(prolog_read_source_term(In, _Term, Expanded,
+	    catch(prolog_read_source_term(In, Term, Expanded,
 					  [ process_comment(true),
 					    term_position(TermPos)
 					  ]),
 		  E, report_syntax_error(E)),
+	    update_condition(Term),
 	    (	is_list(Expanded)
 	    ->	member(T, Expanded)
 	    ;	T = Expanded
@@ -423,8 +474,10 @@ collect(Src, In) :-
 	    (   T == end_of_file
 	    ->  !
 	    ;   stream_position_data(line_count, TermPos, Line),
-		flag(xref_src_line, _, Line),
-		catch(process(T, Src), E, print_message(error, E)),
+		setup_call_cleanup(
+		    asserta(source_line(SrcSpec), Ref),
+		    catch(process(T, Src), E, print_message(error, E)),
+		    erase(Ref)),
 		fail
 	    ).
 
@@ -436,6 +489,44 @@ report_syntax_error(E) :-
 	fail.
 
 
+%%	update_condition(+Term) is det.
+%
+%	Update the condition under which the current code is compiled.
+
+update_condition((:-Directive)) :- !,
+	update_cond(Directive).
+update_condition(_).
+
+update_cond(if(Cond)) :- !,
+	asserta(current_cond(Cond)).
+update_cond(else) :-
+	retract(current_cond(C0)), !,
+	assert(current_cond(\+C0)).
+update_cond(elif(Cond)) :-
+	retract(current_cond(C0)), !,
+	assert(current_cond((\+C0,Cond))).
+update_cond(endif) :-
+	retract(current_cond(_)), !.
+update_cond(_).
+
+%%	current_condition(-Condition) is det.
+%
+%	Condition is the current compilation condition as defined by the
+%	:- if/1 directive and friends.
+
+current_condition(Condition) :-
+	\+ current_cond(_), !,
+	Condition = true.
+current_condition(Condition) :-
+	findall(C, current_cond(C), List),
+	list_to_conj(List, Condition).
+
+list_to_conj([], true).
+list_to_conj([C], C) :- !.
+list_to_conj([H|T], (H,C)) :-
+	list_to_conj(T, C).
+
+
 		 /*******************************
 		 *	     PROCESS		*
 		 *******************************/
@@ -490,18 +581,20 @@ process_directive(load_files(Files, _Options), Src) :-
 process_directive(include(Files), Src) :-
 	process_include(Files, Src).
 process_directive(dynamic(Dynamic), Src) :-
-	assert_dynamic(Src, Dynamic).
+	process_predicates(assert_dynamic, Dynamic, Src).
 process_directive(thread_local(Dynamic), Src) :-
-	assert_thread_local(Src, Dynamic).
+	process_predicates(assert_thread_local, Dynamic, Src).
 process_directive(multifile(Dynamic), Src) :-
-	assert_multifile(Src, Dynamic).
+	process_predicates(assert_multifile, Dynamic, Src).
 process_directive(public(Public), Src) :-
-	assert_public(Src, Public).
+	process_predicates(assert_public, Public, Src).
+process_directive(export(Export), Src) :-
+	process_predicates(assert_export, Export, Src).
 process_directive(module(Module, Export), Src) :-
 	assert_module(Src, Module),
-	assert_export(Src, Export).
-process_directive(system_mode(on), _Src) :- !,
-	style_check(+dollar).
+	assert_module_export(Src, Export).
+process_directive('$set_source_module'(_, system), Src) :-
+	assert_module(Src, system).	% hack for handling boot/init.pl
 process_directive(pce_begin_class_definition(Name, Meta, Super, Doc), Src) :-
 	assert_defined_class(Src, Name, Meta, Super, Doc).
 process_directive(pce_autoload(Name, From), Src) :-
@@ -516,8 +609,6 @@ process_directive(encoding(Enc), _) :-
 	->  catch(set_stream(Stream, encoding(Enc)), _, true)
 	;   true			% can this happen?
 	).
-process_directive(system_module, _) :-
-	style_check(+dollar).
 process_directive(set_prolog_flag(character_escapes, Esc), _) :-
 	set_prolog_flag(character_escapes, Esc).
 process_directive(pce_expansion:push_compile_operators, _) :-
@@ -529,21 +620,21 @@ process_directive(meta_predicate(Meta), _) :-
 	process_meta_predicate(Meta).
 process_directive(arithmetic_function(FSpec), Src) :-
 	arith_callable(FSpec, Goal), !,
-	flag(xref_src_line, Line, Line),
+	current_source_line(Line),
 	assert_called(Src, '<directive>'(Line), Goal).
 process_directive(format_predicate(_, Goal), Src) :- !,
-	flag(xref_src_line, Line, Line),
+	current_source_line(Line),
 	assert_called(Src, '<directive>'(Line), Goal).
 process_directive(if(Cond), Src) :- !,
-	flag(xref_src_line, Line, Line),
+	current_source_line(Line),
 	assert_called(Src, '<directive>'(Line), Cond).
 process_directive(elif(Cond), Src) :- !,
-	flag(xref_src_line, Line, Line),
+	current_source_line(Line),
 	assert_called(Src, '<directive>'(Line), Cond).
 process_directive(else, _) :- !.
 process_directive(endif, _) :- !.
 process_directive(Goal, Src) :-
-	flag(xref_src_line, Line, Line),
+	current_source_line(Line),
 	process_body(Goal, '<directive>'(Line), Src).
 
 %%	process_meta_predicate(+Decl)
@@ -562,7 +653,7 @@ process_meta_head(Decl) :-
 	functor(Head, Name, Arity),
 	meta_args(1, Arity, Decl, Head, Meta),
 	(   (   prolog:meta_goal(Head, _)
-	    ;   called_by(Head, _)
+	    ;   prolog:called_by(Head, _)
 	    ;   meta_goal(Head, _)
 	    )
 	->  true
@@ -571,11 +662,17 @@ process_meta_head(Decl) :-
 
 meta_args(I, Arity, _, _, []) :-
 	I > Arity, !.
-meta_args(I, Arity, Decl, Head, [H|T]) :- 		% 0
+meta_args(I, Arity, Decl, Head, [H|T]) :-		% 0
 	arg(I, Decl, 0), !,
 	arg(I, Head, H),
 	I2 is I + 1,
 	meta_args(I2, Arity, Decl, Head, T).
+meta_args(I, Arity, Decl, Head, [H|T]) :-		% ^
+	arg(I, Decl, ^), !,
+	arg(I, Head, EH),
+	setof_goal(EH, H),
+	I2 is I + 1,
+	meta_args(I2, Arity, Decl, Head, T).
 meta_args(I, Arity, Decl, Head, [H+A|T]) :-		% I --> H+I
 	arg(I, Decl, A),
 	integer(A), A > 0, !,
@@ -591,9 +688,9 @@ meta_args(I, Arity, Decl, Head, Meta) :-
 	      *             BODY	      *
 	      ********************************/
 
-xref_meta((A, B), 		[A, B]).
-xref_meta((A; B), 		[A, B]).
-xref_meta((A| B), 		[A, B]).
+xref_meta((A, B),		[A, B]).
+xref_meta((A; B),		[A, B]).
+xref_meta((A| B),		[A, B]).
 xref_meta((A -> B),		[A, B]).
 xref_meta((A *-> B),		[A, B]).
 xref_meta(findall(_V,G,_L),	[G]).
@@ -711,6 +808,7 @@ hook(attribute_goals(_,_,_)).
 hook(goal_expansion(_,_)).
 hook(term_expansion(_,_)).
 hook(resource(_,_,_)).
+hook('$pred_option'(_,_,_,_)).
 
 hook(emacs_prolog_colours:goal_classification(_,_)).
 hook(emacs_prolog_colours:term_colours(_,_)).
@@ -725,15 +823,20 @@ hook(pce_principal:pce_lazy_send_method(_,_,_)).
 hook(pce_principal:pce_uses_template(_,_)).
 hook(prolog:locate_clauses(_,_)).
 hook(prolog:message(_,_,_)).
+hook(prolog:error_message(_,_,_)).
+hook(prolog:message_location(_,_,_)).
 hook(prolog:message_context(_,_,_)).
+hook(prolog:message_line_element(_,_)).
 hook(prolog:debug_control_hook(_)).
 hook(prolog:help_hook(_)).
 hook(prolog:show_profile_hook(_,_)).
 hook(prolog:general_exception(_,_)).
+hook(prolog:predicate_summary(_,_)).
 hook(prolog_edit:load).
 hook(prolog_edit:locate(_,_,_)).
 hook(shlib:unload_all_foreign_libraries).
 hook(system:'$foreign_registered'(_, _)).
+hook(predicate_options:option_decl(_,_,_)).
 hook(user:exception(_,_,_)).
 hook(user:file_search_path(_,_)).
 hook(user:library_directory(_)).
@@ -759,34 +862,52 @@ arith_callable(Name/Arity, Goal) :-
 	PredArity is Arity + 1,
 	functor(Goal, Name, PredArity).
 
-
-%%	process_body(+Body, +Origin, +Src)
+%%	process_body(+Body, +Origin, +Src) is det.
 %
-%	Process a callable body (body of a clause or directive). Origin
-%	describes the origin of the call.
+%	Process a callable body (body of  a clause or directive). Origin
+%	describes the origin of the call. Partial evaluation may lead to
+%	non-determinism, which is why we backtrack over process_goal/3.
 
-process_body(Var, _, _) :-
+process_body(Body, Origin, Src) :-
+	forall(process_goal(Body, Origin, Src),
+	       true).
+
+process_goal(Var, _, _) :-
 	var(Var), !.
-process_body(Goal, Origin, Src) :-
-	called_by(Goal, Called), !,
+process_goal(Goal, Origin, Src) :-
+	Goal = (A;B), !,
+	setof(Goal,
+	      (   process_goal(A, Origin, Src)
+	      ;   process_goal(B, Origin, Src)
+	      ),
+	      Alts0),
+	variants(Alts0, Alts),
+	member(Goal, Alts).
+process_goal(Goal, Origin, Src) :-
+	prolog:called_by(Goal, Called), !,
 	must_be(list, Called),
 	assert_called(Src, Origin, Goal),
 	process_called_list(Called, Origin, Src).
-process_body(Goal, Origin, Src) :-
+process_goal(Goal, Origin, Src) :-
 	process_xpce_goal(Goal, Origin, Src), !.
-process_body(load_foreign_library(File), _Origin, Src) :-
+process_goal(load_foreign_library(File), _Origin, Src) :-
 	process_foreign(File, Src).
-process_body(load_foreign_library(File, _Init), _Origin, Src) :-
+process_goal(load_foreign_library(File, _Init), _Origin, Src) :-
 	process_foreign(File, Src).
-process_body(Goal, Origin, Src) :-
+process_goal(use_foreign_library(File), _Origin, Src) :-
+	process_foreign(File, Src).
+process_goal(use_foreign_library(File, _Init), _Origin, Src) :-
+	process_foreign(File, Src).
+process_goal(Goal, Origin, Src) :-
 	xref_meta(Goal, Metas), !,
 	assert_called(Src, Origin, Goal),
 	process_called_list(Metas, Origin, Src).
-process_body(Goal, Origin, Src) :-
+process_goal(Goal, Origin, Src) :-
 	asserting_goal(Goal, Rule), !,
 	assert_called(Src, Origin, Goal),
 	process_assert(Rule, Origin, Src).
-process_body(Goal, Origin, Src) :-
+process_goal(Goal, Origin, Src) :-
+	partial_evaluate(Goal),
 	assert_called(Src, Origin, Goal).
 
 process_called_list([], _, _).
@@ -796,11 +917,11 @@ process_called_list([H|T], Origin, Src) :-
 
 process_meta(A+N, Origin, Src) :- !,
 	(   extend(A, N, AX)
-	->  process_body(AX, Origin, Src)
+	->  process_goal(AX, Origin, Src)
 	;   true
 	).
 process_meta(G, Origin, Src) :-
-	process_body(G, Origin, Src).
+	process_goal(G, Origin, Src).
 
 extend(Var, _, _) :-
 	var(Var), !, fail.
@@ -826,6 +947,40 @@ process_assert((_:-Body), Origin, Src) :- !,
 	process_body(Body, Origin, Src).
 process_assert(_, _, _).
 
+%%	variants(+SortedList, -Variants) is det.
+
+variants([], []).
+variants([H|T], List) :-
+	variants(T, H, List).
+
+variants([], H, [H]).
+variants([H|T], V, List) :-
+	(   H =@= V
+	->  variants(T, V, List)
+	;   List = [V|List2],
+	    variants(T, H, List2)
+	).
+
+
+%%	partial_evaluate(Goal) is det.
+%
+%	Perform partial evaluation on Goal to trap cases such as below.
+%
+%	  ==
+%		T = hello(X),
+%		findall(T, T, List),
+%	  ==
+%
+%	@tbd	Make this user extensible? What about non-deterministic
+%		bindings?
+
+partial_evaluate(Goal) :-
+	eval(Goal), !.
+partial_evaluate(_).
+
+eval(X = Y) :-
+	unify_with_occurs_check(X, Y).
+
 
 		 /*******************************
 		 *	    XPCE STUFF		*
@@ -941,7 +1096,7 @@ process_pce_import(Name/Arity, Src, Path, Reexport) :-
 	integer(Arity), !,
 	functor(Term, Name, Arity),
 	(   \+ system_predicate(Term),
-	    \+ Term = pce_error(_) 	% hack!?
+	    \+ Term = pce_error(_)	% hack!?
 	->  assert_import(Src, [Name/Arity], _, Path, Reexport)
 	;   true
 	).
@@ -995,18 +1150,26 @@ public_list(Path, Module, Meta, Export, Public) :-
 	public_list(Path, Module, Meta, [], Export, [], Public, []).
 
 public_list(Path, Module, Meta, MT, Export, Rest, Public, PT) :-
-	setup_call_cleanup((prolog_open_source(Path, In),
-			    set_xref(Old)),
-			   phrase(read_directives(In), Directives),
-			   (set_prolog_flag(xref, Old),
-			    prolog_close_source(In))),
+	setup_call_cleanup(
+	    ( prolog_open_source(Path, In),
+	      set_xref(Old)
+	    ),
+	    phrase(read_directives(In), Directives),
+	    ( set_prolog_flag(xref, Old),
+	      prolog_close_source(In)
+	    )),
 	public_list(Directives, Path, Module, Meta, MT, Export, Rest, Public, PT).
 
 
 read_directives(In) -->
-	{ prolog_read_source_term(In, Term, Expanded, [process_comment(true)]),
-	  nonvar(Term),
-	  Term = (:-_)
+	{  repeat,
+	     catch(prolog_read_source_term(In, Term, Expanded,
+					   [ process_comment(true),
+					     syntax_errors(error)
+					   ]),
+		   E, report_syntax_error(E))
+	-> nonvar(Term),
+	   Term = (:-_)
 	}, !,
 	terms(Expanded),
 	read_directives(In).
@@ -1118,47 +1281,55 @@ process_include([H|T], Src) :- !,
 	process_include(T, Src).
 process_include(File, Src) :-
 	callable(File), !,
-	(   xref_source_file(File, Path, Src)
+	(   xref_input(ParentSrc, _),
+	    xref_source_file(File, Path, ParentSrc)
 	->  assert(uses_file(File, Src, Path)),
-	    (	catch(read_src_to_terms(Path, Terms), _, fail)
-	    ->	process_terms(Terms, Src)
-	    ;	true
-	    )
+	    setup_call_cleanup(
+		open_include_file(Path, In, Refs),
+		collect(Src, Path, In),
+		close_include(In, Refs))
 	;   assert(uses_file(File, Src, '<not_found>'))
 	).
 process_include(_, _).
 
-process_terms([], _).
-process_terms([H|T], Src) :-
-	process(H, Src),
-	process_terms(T, Src).
+%%	open_include_file(+Path, -In)
+%
+%	Opens an :- include(File) referenced file.   Note that we cannot
+%	use prolog_open_source/2 because we   should  _not_ safe/restore
+%	the lexical context.
 
-read_src_to_terms(Path, Terms) :-
-	prolog_open_source(Path, Fd),
-	call_cleanup(read_clauses(Fd, Terms),
-		     prolog_close_source(Fd)).
+open_include_file(Path, In, Ref) :-
+	xref_input(_, Parent),
+	stream_property(Parent, encoding(Enc)),
+	include_encoding(Enc, Options),
+	open(Path, read, In, Options),
+	(   peek_char(In, #)		% Deal with #! script
+	->  skip(In, 10)
+	;   true
+	),
+	asserta(xref_input(Path, In), Ref).
 
-read_clauses(In, Terms) :-
-	read_clause(In, C0),
-	read_clauses(C0, In, Terms).
+include_encoding(wchar_t, []) :- !.
+include_encoding(Enc, [encoding(Enc)]).
 
-read_clauses(end_of_file, _, []) :- !.
-read_clauses(Term, In, [Term|T]) :-
-	read_clause(In, C),
-	read_clauses(C, In, T).
 
+close_include(In, Refs) :-
+	maplist(erase, Refs),
+	close(In).
 
 %%	process_foreign(+Spec, +Src)
 %
 %	Process a load_foreign_library/1 call.
 
 process_foreign(Spec, Src) :-
-	current_foreign_library(Spec, Defined),
+	ground(Spec),
+	current_foreign_library(Spec, Defined), !,
 	(   xmodule(Module, Src)
 	->  true
 	;   Module = user
 	),
 	process_foreign_defined(Defined, Module, Src).
+process_foreign(_, _).
 
 process_foreign_defined([], _, _).
 process_foreign_defined([H|T], M, Src) :-
@@ -1237,7 +1408,7 @@ assert_constraint(Src, Head) :-
 assert_constraint(Src, Head) :-
 	functor(Head, Name, Arity),
 	functor(Term, Name, Arity),
-	flag(xref_src_line, Line, Line),
+	current_source_line(Line),
 	assert(constraint(Term, Src, Line)).
 
 
@@ -1245,7 +1416,7 @@ assert_constraint(Src, Head) :-
 		*       PHASE 1 ASSERTIONS	*
 		********************************/
 
-%%	assert_called(+Src, +From, +Head)
+%%	assert_called(+Src, +From, +Head) is det.
 %
 %	Assert the fact that Head is called by From in Src. We do not
 %	assert called system predicates.
@@ -1260,28 +1431,35 @@ assert_called(_, _, Goal) :-
 assert_called(Src, Origin, M:G) :- !,
 	(   atom(M),
 	    callable(G)
-	->  (   xmodule(M, Src)
+	->  current_condition(Cond),
+	    (   xmodule(M, Src)		% explicit call to own module
 	    ->  assert_called(Src, Origin, G)
-	    ;   called(M:G, Src, Origin)
+	    ;   called(M:G, Src, Origin, Cond) % already registered
 	    ->  true
-	    ;	system_predicate(G)
+	    ;	system_predicate(G)	% not interesting (now)
 	    ->	true
 	    ;   generalise(Origin, OTerm),
-		generalise(G, GTerm),
-		assert(called(M:GTerm, Src, OTerm))
+		generalise(G, GTerm)
+	    ->	assert(called(M:GTerm, Src, OTerm, Cond))
+	    ;	true
 	    )
 	;   true                        % call to variable module
 	).
-assert_called(_, _, Goal) :-
-	system_predicate(Goal), !.
+assert_called(Src, _, Goal) :-
+	system_predicate(Goal),
+	\+ xmodule(system, Src), !.
 assert_called(Src, Origin, Goal) :-
-	called(Goal, Src, Origin), !.
-assert_called(Src, Origin, Goal) :-
-	generalise(Origin, OTerm),
-	generalise(Goal, Term),
-	assert(called(Term, Src, OTerm)).
+	current_condition(Cond),
+	(   called(Goal, Src, Origin, Cond)
+	->  true
+	;   generalise(Origin, OTerm),
+	    generalise(Goal, Term)
+	->  assert(called(Term, Src, OTerm, Cond))
+	;   true
+	).
 
-%%	hide_called(:Callable)
+
+%%	hide_called(:Callable) is semidet.
 %
 %	Goals that should not turn up as being called. Hack. Eventually
 %	we should deal with that using an XPCE plugin.
@@ -1295,14 +1473,14 @@ assert_defined(Src, Goal) :-
 	defined(Goal, Src, _), !.
 assert_defined(Src, Goal) :-
 	generalise(Goal, Term),
-	flag(xref_src_line, Line, Line),
+	current_source_line(Line),
 	assert(defined(Term, Src, Line)).
 
 assert_foreign(Src, Goal) :-
 	foreign(Goal, Src, _), !.
 assert_foreign(Src, Goal) :-
 	generalise(Goal, Term),
-	flag(xref_src_line, Line, Line),
+	current_source_line(Line),
 	assert(foreign(Term, Src, Line)).
 
 %%	assert_import(+Src, +Import, +ExportList, +From, +Reexport) is det.
@@ -1330,7 +1508,7 @@ assert_import(Src, Import as Name, Export, From, Reexport) :- !,
 	(   in_export_list(Term0, Export)
 	->  assert(imported(Term, Src, From)),
 	    assert_reexport(Reexport, Src, Term)
-	;   flag(xref_src_line, Line, Line),
+	;   current_source_line(Line),
 	    assert_called(Src, '<directive>'(Line), Term0)
 	).
 assert_import(Src, Import, Export, From, Reexport) :-
@@ -1338,7 +1516,7 @@ assert_import(Src, Import, Export, From, Reexport) :-
 	(   in_export_list(Term, Export)
 	->  assert(imported(Term, Src, From)),
 	    assert_reexport(Reexport, Src, Term)
-	;   flag(xref_src_line, Line, Line),
+	;   current_source_line(Line),
 	    assert_called(Src, '<directive>'(Line), Term)
 	).
 assert_import(Src, op(P,T,N), _, _, _) :-
@@ -1347,8 +1525,8 @@ assert_import(Src, op(P,T,N), _, _, _) :-
 in_export_list(_Head, Export) :-
 	var(Export), !.
 in_export_list(Head, Export) :-
-	member(Export, Export),
-	pi_to_head(Export, Head).
+	member(PI, Export),
+	pi_to_head(PI, Head).
 
 assert_reexport(false, _, _) :- !.
 assert_reexport(true, Src, Term) :-
@@ -1391,64 +1569,78 @@ assert_module(Src, Module) :-
 	'$set_source_module'(_, Module),
 	assert(xmodule(Module, Src)).
 
-assert_export(_, []) :- !.
-assert_export(Src, [H|T]) :- !,
-	assert_export(Src, H),
-	assert_export(Src, T).
-assert_export(Src, PI) :-
+assert_module_export(_, []) :- !.
+assert_module_export(Src, [H|T]) :- !,
+	assert_module_export(Src, H),
+	assert_module_export(Src, T).
+assert_module_export(Src, PI) :-
 	pi_to_head(PI, Term), !,
 	assert(exported(Term, Src)).
-assert_export(Src, op(P, A, N)) :-
+assert_module_export(Src, op(P, A, N)) :-
 	xref_push_op(Src, P, A, N).
 
-assert_dynamic(Src, (A, B)) :- !,
-	assert_dynamic(Src, A),
-	assert_dynamic(Src, B).
-assert_dynamic(_, _M:_Name/_Arity) :- !. % not local
-assert_dynamic(Src, PI) :-
+%%	process_predicates(:Closure, +Predicates, +Src)
+%
+%	Process areguments of dynamic,  etc.,   using  call(Closure, PI,
+%	Src).  Handles  both  lists  of    specifications  and  (PI,...)
+%	specifications.
+
+process_predicates(Closure, Preds, Src) :-
+	is_list(Preds), !,
+	process_predicate_list(Preds, Closure, Src).
+process_predicates(Closure, Preds, Src) :-
+	process_predicate_comma(Preds, Closure, Src).
+
+process_predicate_list([], _, _).
+process_predicate_list([H|T], Closure, Src) :-
+	(   nonvar(H)
+	->  call(Closure, H, Src)
+	;   true
+	),
+	process_predicate_list(T, Closure, Src).
+
+process_predicate_comma(Var, _, _) :-
+	var(Var), !.
+process_predicate_comma(M:(A,B), Closure, Src) :- !,
+	process_predicate_comma(M:A, Closure, Src),
+	process_predicate_comma(M:B, Closure, Src).
+process_predicate_comma((A,B), Closure, Src) :- !,
+	process_predicate_comma(A, Closure, Src),
+	process_predicate_comma(B, Closure, Src).
+process_predicate_comma(A, Closure, Src) :-
+	call(Closure, A, Src).
+
+
+assert_dynamic(_M:_Name/_Arity, _Src) :- !.   % not local
+assert_dynamic(PI, Src) :-
 	pi_to_head(PI, Term),
 	(   thread_local(Term, Src, _)	% dynamic after thread_local has
 	->  true			% no effect
-	;   flag(xref_src_line, Line, Line),
+	;   current_source_line(Line),
 	    assert(dynamic(Term, Src, Line))
 	).
 
-assert_thread_local(Src, (A, B)) :- !,
-	assert_thread_local(Src, A),
-	assert_thread_local(Src, B).
-assert_thread_local(_, _M:_Name/_Arity) :- !. % not local
-assert_thread_local(Src, PI) :-
+assert_thread_local(_M:_Name/_Arity, _Src) :- !. % not local
+assert_thread_local(PI, Src) :-
 	pi_to_head(PI, Term),
-	flag(xref_src_line, Line, Line),
+	current_source_line(Line),
 	assert(thread_local(Term, Src, Line)).
 
-assert_multifile(_, Var) :-
-	var(Var), !, fail.
-assert_multifile(Src, (A,B)) :- !,
-	assert_multifile(Src, A),
-	assert_multifile(Src, B).
-assert_multifile(Src, M:(A,B)) :- !,
-	assert_multifile(Src, M:A),
-	assert_multifile(Src, M:B).
-assert_multifile(Src, PI) :-
+assert_multifile(PI, Src) :-			% :- multifile(Spec)
 	pi_to_head(PI, Term),
-	flag(xref_src_line, Line, Line),
+	current_source_line(Line),
 	assert(multifile(Term, Src, Line)).
 
-assert_public(_, Var) :-
-	var(Var), !, fail.
-assert_public(Src, (A,B)) :- !,
-	assert_public(Src, A),
-	assert_public(Src, B).
-assert_public(Src, M:(A,B)) :- !,
-	assert_public(Src, M:A),
-	assert_public(Src, M:B).
-assert_public(Src, PI) :-
+assert_public(PI, Src) :-			% :- public(Spec)
 	pi_to_head(PI, Term),
-	flag(xref_src_line, Line, Line),
+	current_source_line(Line),
 	assert_called(Src, '<public>'(Line), Term),
 	assert(public(Term, Src, Line)).
 
+assert_export(PI, Src) :-			% :- export(Spec)
+	pi_to_head(PI, Term), !,
+	assert(exported(Term, Src)).
+
 %%	pi_to_head(+PI, -Head) is semidet.
 %
 %	Translate Name/Arity or Name//Arity to a callable term. Fails if
@@ -1472,9 +1664,9 @@ assert_used_class(Src, Name) :-
 
 assert_defined_class(Src, Name, _Meta, _Super, _) :-
 	defined_class(Name, _, _, Src, _), !.
-assert_defined_class(_, _, _, -, _) :- !. 		% :- pce_extend_class
+assert_defined_class(_, _, _, -, _) :- !.		% :- pce_extend_class
 assert_defined_class(Src, Name, Meta, Super, Summary) :-
-	flag(xref_src_line, Line, Line),
+	current_source_line(Line),
 	(   Summary == @(default)
 	->  Atom = ''
 	;   is_list(Summary)
@@ -1571,9 +1763,9 @@ do_xref_source_file(Spec, File, Options) :-
 	prolog:xref_source_file(Spec, File, Options), !.
 do_xref_source_file(Spec, File, Options) :-
 	option(file_type(Type), Options, prolog),
-	absolute_file_name(Spec,
+	absolute_file_name(Spec, File,
 			   [ file_type(Type),
 			     access(read),
 			     file_errors(fail)
-			   ], File), !.
+			   ]), !.
 
