@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        J.Wielemaker@uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2009, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -30,83 +31,102 @@
 */
 
 :- module(qsave,
-	  [ qsave_program/1
-	  , qsave_program/2
+	  [ qsave_program/1,			% +File
+	    qsave_program/2			% +File, +Options
 	  ]).
 :- use_module(library(lists)).
+:- use_module(library(option)).
+:- use_module(library(error)).
 
-:- module_transparent
-	qsave_program/1,
-	qsave_program/2.
+/** <module> Save current program as a state or executable
 
-:- system_mode(on).
+This library provides qsave_program/1  and   qsave_program/2,  which are
+also used by the commandline sequence below.
+
+  ==
+  swipl -o exe -c file.pl ...
+  ==
+*/
+
+:- meta_predicate
+	qsave_program(+, :).
+
+:- predicate_options(qsave_program/2, 2,
+		     [ local(integer),
+		       global(integer),
+		       trail(integer),
+		       goal(callable),
+		       toplevel(callable),
+		       init_file(atom),
+		       class(oneof([runtime,kernel,development])),
+		       autoload(boolean),
+		       map(atom),
+		       op(oneof([save,standard])),
+		       stand_alone(boolean),
+		       foreign(oneof([save,no_save])),
+		       emulator(atom)
+		     ]).
+
+:- set_prolog_flag(generate_debug_info, false).
 
 :- dynamic verbose/1.
+:- volatile verbose/1.			% contains a stream-handle
 
 %%	qsave_program(+File) is det.
-%%	qsave_program(+File, +Options) is det.
+%%	qsave_program(+File, :Options) is det.
 %
 %	Make a saved state in file `File'.
 
 qsave_program(File) :-
 	qsave_program(File, []).
 
-qsave_program(FileSpec, Options0) :-
-	check_options(Options0),
-	strip_module(FileSpec, Module, FileBase),
+qsave_program(FileBase, Options0) :-
+	meta_options(is_meta, Options0, Options),
+	check_options(Options),
 	exe_file(FileBase, File),
-	option(Options0, autoload/true,	    Autoload,  Options1),
-	option(Options1, map/[],	    Map,       Options2),
-	option(Options2, goal/[],	    GoalTerm,  Options3),
-	option(Options3, op/save,	    SaveOps,   Options4),
-	option(Options4, class/runtime,	    SaveClass, Options5),
-	option(Options5, init_file/DefInit, InitFile,  Options6),
+	option(class(SaveClass),    Options, runtime),
+	option(init_file(InitFile), Options, DefInit),
 	default_init_file(SaveClass, DefInit),
-	(   GoalTerm == []
-	->  Options = Options6,
-	    flag('$banner_goal', BannerGoal, BannerGoal),
-	    define_predicate(user:BannerGoal)
-	;   term_to_atom(Module:GoalTerm, GoalAtom),
-	    term_to_atom(GT, GoalAtom),
-	    define_predicate(user:GT),
-	    Options = [goal=GoalAtom|Options6]
-	),
-	(   Autoload == true
-	->  save_autoload
-	;   true
-	),
-	open_map(Map),
-	set_prolog_flag(saved_program, true),
-	set_prolog_flag(saved_program_class, SaveClass),
+	save_autoload(Options),
+	open_map(Options),
+	create_prolog_flag(saved_program, true, []),
+	create_prolog_flag(saved_program_class, SaveClass, []),
 	(   exists_file(File)
 	->  delete_file(File)
 	;   true
 	),
 	'$rc_open_archive'(File, RC),
 	make_header(RC, SaveClass, Options),
-	save_options(RC, [ class(SaveClass),
-			   init_file(InitFile)
-			 | Options
-			 ]),
+	save_options(RC, SaveClass,
+		     [ init_file(InitFile)
+		     | Options
+		     ]),
 	save_resources(RC, SaveClass),
 	'$rc_open'(RC, '$state', '$prolog', write, StateFd),
 	'$open_wic'(StateFd),
-	system_mode(on),		% generate system modules too
-	save_modules(SaveClass),
-	save_records,
-	save_flags,
-	save_imports,
-	save_prolog_flags,
-	save_operators(SaveOps),
-	save_format_predicates,
-	save_functions,
-	system_mode(off),
+	setup_call_cleanup(
+	    ( current_prolog_flag(access_level, OldLevel),
+	      set_prolog_flag(access_level, system) % generate system modules
+	    ),
+	    ( save_modules(SaveClass),
+	      save_records,
+	      save_flags,
+	      save_imports,
+	      save_prolog_flags,
+	      save_operators(Options),
+	      save_format_predicates
+	    ),
+	    set_prolog_flag(access_level, OldLevel)),
 	'$close_wic',
 	close(StateFd),
+	save_foreign_libraries(RC, Options),
 	'$rc_close_archive'(RC),
 	'$mark_executable'(File),
 	close_map.
 
+is_meta(goal).
+is_meta(toplevel).
+
 exe_file(Base, Exe) :-
 	current_prolog_flag(windows, true),
 	file_name_extension(_, '', Base), !,
@@ -115,7 +135,7 @@ exe_file(Exe, Exe).
 
 default_init_file(runtime, none) :- !.
 default_init_file(_,       InitFile) :-
-	'$option'(init_file, InitFile, InitFile).
+	'$option'(init_file, InitFile).
 
 
 		 /*******************************
@@ -123,8 +143,7 @@ default_init_file(_,       InitFile) :-
 		 *******************************/
 
 make_header(RC, _, Options) :-
-	option(Options, emulator/(-), OptVal, _),
-	OptVal \== -, !,
+	option(emulator(OptVal), Options), !,
 	absolute_file_name(OptVal, [access(read)], Emulator),
 	'$rc_append_file'(RC, '$header', '$rc', none, Emulator).
 make_header(RC, _, Options) :-
@@ -132,8 +151,7 @@ make_header(RC, _, Options) :-
 	->  DefStandAlone = true
 	;   DefStandAlone = false
 	),
-	option(Options, stand_alone/DefStandAlone, OptVal, _),
-	OptVal == true, !,
+	option(stand_alone(true), Options, DefStandAlone), !,
 	current_prolog_flag(executable, Executable),
 	'$rc_append_file'(RC, '$header', '$rc', none, Executable).
 make_header(RC, SaveClass, _Options) :-
@@ -158,7 +176,6 @@ make_header(_, _, _).
 min_stack(local,    32).
 min_stack(global,   16).
 min_stack(trail,    16).
-min_stack(argument, 16).
 
 convert_option(Stack, Val, NewVal) :-	% stack-sizes are in K-bytes
 	min_stack(Stack, Min), !,
@@ -166,28 +183,38 @@ convert_option(Stack, Val, NewVal) :-	% stack-sizes are in K-bytes
 	->  NewVal = Val
 	;   NewVal is max(Min, Val*1024)
 	).
-convert_option(_, Val, Val).
+convert_option(goal, Callable, Atom) :- !,
+	term_to_atom(Callable, Atom).
+convert_option(toplevel, Callable, Atom) :- !,
+	term_to_atom(Callable, Atom).
+convert_option(_, Value, Value).
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Save the options in the '$options' resource.   The home directory is saved
-for development saves, so it keeps refering to the development home.
+doption(Name) :- min_stack(Name, _).
+doption(goal).
+doption(toplevel).
+doption(init_file).
+doption(system_init_file).
+doption(class).
+doption(home).
 
-The script-file (-s script) is not saved at all. I think this is fine to
-avoid a save-script loading itself.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+%%	save_options(+ArchiveHandle, +SaveClass, +Options)
+%
+%	Save the options in the '$options'  resource. The home directory
+%	is saved for development saves,  so   it  keeps  refering to the
+%	development home.
+%
+%	The script-file (-s script) is not saved at all. I think this is
+%	fine to avoid a save-script loading itself.
 
-save_options(RC, Options) :-
+save_options(RC, SaveClass, Options) :-
 	'$rc_open'(RC, '$options', '$prolog', write, Fd),
-	(   '$option'(OptionName, OptionVal0, _),
-	        (   OptionName == home	% save home if not runtime
-		->  \+ memberchk(class(runtime), Options)
-		;   true
-		),
-	        \+ OptionName == script_file,
-	        option(Options, OptionName/_, OptionVal1, _),
-	        (   var(OptionVal1)	% used the default
-		->  OptionVal = OptionVal0
-		;   convert_option(OptionName, OptionVal1, OptionVal)
+	(   doption(OptionName),
+	    '$option'(OptionName, OptionVal0),
+	        save_option_value(SaveClass, OptionName, OptionVal0, OptionVal1),
+	        OptTerm =.. [OptionName,OptionVal2],
+	        (   option(OptTerm, Options)
+		->  convert_option(OptionName, OptionVal2, OptionVal)
+		;   OptionVal = OptionVal1
 		),
 	        format(Fd, '~w=~w~n', [OptionName, OptionVal]),
 	    fail
@@ -195,6 +222,13 @@ save_options(RC, Options) :-
 	),
 	close(Fd).
 
+%%	save_option_value(+SaveClass, +OptionName, +OptionValue, -FinalValue)
+
+save_option_value(Class,   class, _,     Class) :- !.
+save_option_value(runtime, home,  _,     _) :- !, fail.
+save_option_value(_,       _,     Value, Value).
+
+
 		 /*******************************
 		 *	     RESOURCES		*
 		 *******************************/
@@ -249,13 +283,16 @@ reserved_resource('$state',	'$prolog').
 reserved_resource('$options',	'$prolog').
 
 copy_resource(FromRC, ToRC, Name, Class) :-
-	'$rc_open'(FromRC, Name, Class, read, FdIn),
-	'$rc_open'(ToRC,	 Name, Class, write, FdOut),
-	feedback('~t~8|~w~t~24|~w~t~40|~w~n',
-		 [Name, Class, '<Copied from running state>']),
-	copy_stream_data(FdIn, FdOut),
-	close(FdOut),
-	close(FdIn).
+	setup_call_cleanup(
+	    '$rc_open'(FromRC, Name, Class, read,  FdIn),
+	    setup_call_cleanup(
+		'$rc_open'(ToRC,   Name, Class, write, FdOut),
+		( feedback('~t~8|~w~t~24|~w~t~40|~w~n',
+			   [Name, Class, '<Copied from running state>']),
+		  copy_stream_data(FdIn, FdOut)
+		),
+		close(FdOut)),
+	    close(FdIn)).
 
 
 		 /*******************************
@@ -283,8 +320,24 @@ define_predicate(Head) :-
 		 *	      AUTOLOAD		*
 		 *******************************/
 
-save_autoload :-
-	autoload.
+define_init_goal(Options) :-
+	option(goal(Goal), Options), !,
+	define_predicate(Goal).
+define_init_goal(_) :-
+	flag('$banner_goal', BannerGoal, BannerGoal),
+	define_predicate(user:BannerGoal).
+
+define_toplevel_goal(Options) :-
+	option(toplevel(Goal), Options), !,
+	define_predicate(Goal).
+define_toplevel_goal(_).
+
+save_autoload(Options) :-
+	define_init_goal(Options),
+	define_toplevel_goal(Options),
+	option(autoload(true),  Options, true), !,
+	autoload(Options).
+
 
 		 /*******************************
 		 *	       MODULES		*
@@ -298,37 +351,49 @@ save_module(M, SaveClass) :-
 	'$qlf_start_module'(M),
 	feedback('~n~nMODULE ~w~n', [M]),
 	save_unknown(M),
-	(   P = (M:H),
+	(   P = (M:_H),
 	    current_predicate(_, P),
 	    \+ predicate_property(P, imported_from(_)),
-	    \+ predicate_property(P, foreign),
-	    functor(H, F, A),
-	    feedback('~nsaving ~w/~d ', [F, A]),
-	    (	H = resource(_,_,_),
-		SaveClass \== development
-	    ->	save_attribute(P, (dynamic)),
-		(   M == user
-		->  save_attribute(P, (multifile))
-		),
-		feedback('(Skipped clauses)', []),
-		fail
-	    ;	true
-	    ),
-	    save_attributes(P),
-	    \+ predicate_property(P, (volatile)),
-	    nth_clause(P, _, Ref),
-	    feedback('.', []),
-	    '$qlf_assert_clause'(Ref, SaveClass),
+	    save_predicate(P, SaveClass),
 	    fail
 	;   '$qlf_end_part',
 	    feedback('~n', [])
 	).
 
-pred_attrib(indexed(Term), Head, index(M:Term)) :- !,
-	    strip_module(Head, M, _).
+save_predicate(P, _SaveClass) :-
+	predicate_property(P, foreign), !,
+	P = (M:H),
+	functor(H, Name, Arity),
+	feedback('~npre-defining foreign ~w/~d ', [Name, Arity]),
+	'$add_directive_wic'('$predefine_foreign'(M:Name/Arity)).
+save_predicate(P, SaveClass) :-
+	P = (M:H),
+	functor(H, F, A),
+	feedback('~nsaving ~w/~d ', [F, A]),
+	(   H = resource(_,_,_),
+	    SaveClass \== development
+	->  save_attribute(P, (dynamic)),
+	    (   M == user
+	    ->  save_attribute(P, (multifile))
+	    ),
+	    feedback('(Skipped clauses)', []),
+	    fail
+	;   true
+	),
+	save_attributes(P),
+	\+ predicate_property(P, (volatile)),
+	(   nth_clause(P, _, Ref),
+	    feedback('.', []),
+	    '$qlf_assert_clause'(Ref, SaveClass),
+	    fail
+	;   true
+	).
+
+
 pred_attrib(meta_predicate(Term), Head, meta_predicate(M:Term)) :- !,
 	    strip_module(Head, M, _).
-pred_attrib(Attrib, Head, '$set_predicate_attribute'(M:Name/Arity, AttName, Val)) :-
+pred_attrib(Attrib, Head,
+	    '$set_predicate_attribute'(M:Name/Arity, AttName, Val)) :-
 	attrib_name(Attrib, AttName, Val),
 	strip_module(Head, M, Term),
 	functor(Term, Name, Arity).
@@ -337,18 +402,22 @@ attrib_name(dynamic,	   dynamic,	  1).
 attrib_name(volatile,	   volatile,	  1).
 attrib_name(thread_local,  thread_local,  1).
 attrib_name(multifile,	   multifile,	  1).
+attrib_name(public,	   public,	  1).
 attrib_name(transparent,   transparent,	  1).
 attrib_name(discontiguous, discontiguous, 1).
 attrib_name(notrace,	   trace,	  0).
 attrib_name(show_childs,   hide_childs,	  0).
+attrib_name(built_in,      system,	  1).
+attrib_name(nodebug,       hide_childs,	  1).
 
 
 save_attribute(P, Attribute) :-
 	pred_attrib(Attribute, P, D),
-	(   Attribute = indexed(Term)
-	->  \+(( arg(1, Term, 1),
-	         functor(Term, _, Arity),
-		 forall(between(2, Arity, N), arg(N, Term, 0))))
+	(   Attribute == built_in	% no need if there are clauses
+	->  (   predicate_property(P, number_of_clauses(0))
+	    ->	true
+	    ;	predicate_property(P, volatile)
+	    )
 	;   true
 	),
 	'$add_directive_wic'(D),
@@ -404,16 +473,13 @@ save_flags :-
 		 *	     IMPORTS		*
 		 *******************************/
 
-default_import(system, _, _) :- !, fail.
-default_import(To, Head, _) :-
-	'$get_predicate_attribute'(To:Head, (dynamic), 1), !,
-	fail.
-default_import(user, Head, _) :- !,
-	'$default_predicate'(user:Head, system:Head).
-default_import(To, Head, _From) :-
-	'$default_predicate'(To:Head, user:Head).
-default_import(To, Head, _From) :-
-	'$default_predicate'(To:Head, system:Head).
+%%	save_imports
+%
+%	Save  import  relations.  An  import  relation  is  saved  if  a
+%	predicate is imported from a module that is not a default module
+%	for the destination module. If  the   predicate  is  dynamic, we
+%	always define the explicit import relation to make clear that an
+%	assert must assert on the imported predicate.
 
 save_imports :-
 	feedback('~nIMPORTS~n~n', []),
@@ -426,6 +492,13 @@ save_imports :-
 	;   true
 	).
 
+default_import(To, Head, From) :-
+	'$get_predicate_attribute'(To:Head, (dynamic), 1),
+	predicate_property(From:Head, exported), !,
+	fail.
+default_import(Into, _, From) :-
+	default_module(Into, From).
+
 %%	restore_import(+TargetModule, +SourceModule, +PI) is det.
 %
 %	Restore import relation. This notably   deals  with imports from
@@ -439,40 +512,51 @@ restore_import(To, From, PI) :-
 	To:import(From:PI).
 
 		 /*******************************
-		 *	      FEATURES		*
+		 *	   PROLOG FLAGS		*
 		 *******************************/
 
 save_prolog_flags :-
 	feedback('~nPROLOG FLAGS~n~n', []),
-	'$current_prolog_flag'(Feature, Value, global, write, _Type),
-	\+ no_save_flag(Feature),
-	feedback('~t~8|~w: ~w~n', [Feature, Value]),
-	'$add_directive_wic'(qsave:restore_prolog_flag(Feature, Value)),
+	'$current_prolog_flag'(Flag, Value, _Scope, write, Type),
+	\+ no_save_flag(Flag),
+	feedback('~t~8|~w: ~w (type ~q)~n', [Flag, Value, Type]),
+	'$add_directive_wic'(qsave:restore_prolog_flag(Flag, Value, Type)),
 	fail.
 save_prolog_flags.
 
 no_save_flag(argv).
+no_save_flag(access_level).
+no_save_flag(tty_control).
+no_save_flag(readline).
 no_save_flag(associated_file).
 no_save_flag(hwnd).			% should be read-only, but comes
 					% from user-code
 
-%	Deal with possibly protected flags (debug_on_error and
+%%	restore_prolog_flag(+Name, +Value, +Type)
+%
+%	Deal  with  possibly   protected    flags   (debug_on_error  and
 %	report_error are protected flags for the runtime kernel).
 
-restore_prolog_flag(Flag, Value) :-
+restore_prolog_flag(Flag, Value, _Type) :-
+	current_prolog_flag(Flag, Value), !.
+restore_prolog_flag(Flag, Value, _Type) :-
+	current_prolog_flag(Flag, _), !,
 	catch(set_prolog_flag(Flag, Value), _, true).
+restore_prolog_flag(Flag, Value, Type) :-
+	create_prolog_flag(Flag, Value, [type(Type)]).
 
 
 		 /*******************************
 		 *	     OPERATORS		*
 		 *******************************/
 
-%%	save_operators(+Save) is det.
+%%	save_operators(+Options) is det.
 %
 %	Save operators for all modules.   Operators for =system= are
 %	not saved because these are read-only anyway.
 
-save_operators(save) :- !,
+save_operators(Options) :- !,
+	option(op(save), Options, save),
 	feedback('~nOPERATORS~n', []),
 	forall(current_module(M), save_module_operators(M)),
 	feedback('~n', []).
@@ -506,32 +590,62 @@ qualify_head(T, user:T).
 
 
 		 /*******************************
-		 *	     FUNCTIONS		*
+		 *	 FOREIGN LIBRARIES	*
 		 *******************************/
 
-save_functions :-
-	feedback('~nFUNCTIONS~n', []),
-	'$prolog_arithmetic_function'(M:Head, Index),
-	functor(Head, Name, Arity),
-	PredArity is Arity + 1,
-	functor(PredHead, Name, PredArity),
-	D = '$arithmetic_function'(M:PredHead, Index),
-	feedback('~n~t~8|~w ', [D]),
-	'$add_directive_wic'(D),
-	fail.
-save_functions.
+%%	save_foreign_libraries(+Archive, +Options) is det.
+%
+%	Save current foreign libraries into the archive.
 
+save_foreign_libraries(RC, Options) :-
+	option(foreign(save), Options), !,
+	feedback('~nFOREIGN LIBRARIES~n', []),
+	forall(current_foreign_library(FileSpec, _Predicates),
+	       ( find_foreign_library(FileSpec, File),
+		 term_to_atom(FileSpec, Name),
+		 '$rc_append_file'(RC, Name, shared, none, File)
+	       )).
+save_foreign_libraries(_, _).
+
+%%	find_foreign_library(+FileSpec, -File) is det.
+%
+%	Find the shared object specified by   FileSpec.  If posible, the
+%	shared object is stripped to reduce   its size. This is achieved
+%	by calling strip -o <tmp> <shared-object>. Note that the file is
+%	a Prolog tmp file and will be deleted on halt.
+%
+%	@bug	Should perform OS search on failure
+
+find_foreign_library(FileSpec, SharedObject) :-
+	absolute_file_name(FileSpec,
+			   [ file_type(executable),
+			     file_errors(fail)
+			   ], File), !,
+	(   absolute_file_name(path(strip), Strip,
+			       [ access(execute),
+				 file_errors(fail)
+			       ]),
+	    tmp_file(shared, Stripped),
+	    format(atom(Cmd), '"~w" -o "~w" "~w"',
+		   [ Strip, Stripped, File ]),
+	    shell(Cmd)
+	->  SharedObject = Stripped,
+	    Delete = true
+	;   SharedObject = File,
+	    Delete = false
+	).
 
 
 		 /*******************************
 		 *	       UTIL		*
 		 *******************************/
 
-open_map([]) :- !,
-	retractall(verbose(_)).
-open_map(File) :-
-	open(File, write, Fd),
+open_map(Options) :-
+	option(map(Map), Options), !,
+	open(Map, write, Fd),
 	asserta(verbose(Fd)).
+open_map(_) :-
+	retractall(verbose(_)).
 
 close_map :-
 	retract(verbose(Fd)),
@@ -544,27 +658,21 @@ feedback(Fmt, Args) :-
 feedback(_, _).
 
 
-option(List, Name/_Default, Value, Rest) :- % goal = Goal
-	select(Name=Value, List, Rest), !.
-option(List, Name/_Default, Value, Rest) :- % goal(Goal)
-	Term =.. [Name, Value],
-	select(Term, List, Rest), !.
-option(List, _Name/Default, Default, List).
-
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Option checking and exception generation.  This should be in a library!
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 option_type(Name,	 integer) :- min_stack(Name, _MinValue).
-option_type(class,	 atom([runtime,kernel,development])).
-option_type(autoload,	 bool).
+option_type(class,	 oneof([runtime,kernel,development])).
+option_type(autoload,	 boolean).
 option_type(map,	 atom).
-option_type(op,		 atom([save, standard])).
-option_type(stand_alone, bool).
-option_type(goal, 	 callable).
-option_type(toplevel, 	 callable).
-option_type(init_file, 	 atom).
-option_type(emulator, 	 ground).
+option_type(op,		 oneof([save, standard])).
+option_type(stand_alone, boolean).
+option_type(foreign,	 oneof([save, no_save])).
+option_type(goal,	 callable).
+option_type(toplevel,	 callable).
+option_type(init_file,	 atom).
+option_type(emulator,	 ground).
 
 check_options([]) :- !.
 check_options([Var|_]) :-
@@ -572,7 +680,7 @@ check_options([Var|_]) :-
 	throw(error(domain_error(save_options, Var), _)).
 check_options([Name=Value|T]) :- !,
 	(   option_type(Name, Type)
-	->  (   check_type(Type, Value)
+	->  (   must_be(Type, Value)
 	    ->  check_options(T)
 	    ;	throw(error(domain_error(Type, Value), _))
 	    )
@@ -586,21 +694,6 @@ check_options([Var|_]) :-
 check_options(Opt) :-
 	throw(error(domain_error(list, Opt), _)).
 
-check_type(integer, V) :-
-	integer(V).
-check_type(atom(List), V) :-
-	atom(V),
-	memberchk(V, List), !.
-check_type(atom, V) :-
-	atom(V).
-check_type(callable, V) :-
-	atom(V).
-check_type(callable, V) :-
-	compound(V).
-check_type(ground, V) :-
-	ground(V).
-check_type(bool, true).
-check_type(bool, false).
 
 		 /*******************************
 		 *	      MESSAGES		*
