@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <errno.h>
@@ -57,6 +57,7 @@ pl_error(const char *pred, int arity, const char *msg, int id, ...)
 
       switch(err)
       { case ENOMEM:
+	case EAGAIN:			/* fork(); might be other resource */
 	  rc = PL_unify_term(formal,
 			     CompoundArg("resource_error", 1),
 			       AtomArg("no_memory"));
@@ -85,10 +86,12 @@ pl_error(const char *pred, int arity, const char *msg, int id, ...)
       break;
     }
     case ERR_ARGTYPE:
-    { int argn	      = va_arg(args, int);
+    { int argn        = va_arg(args, int);	/* argument position (unused) */
       term_t actual   = va_arg(args, term_t);
       atom_t expected = PL_new_atom(va_arg(args, const char*));
 
+      (void)argn;				/* avoid unused warning */
+
       if ( PL_is_variable(actual) && expected != PL_new_atom("variable") )
 	rc = PL_unify_atom_chars(formal, "instantiation_error");
       else
