@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        J.Wielemaker@uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2009, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <SWI-Stream.h>
@@ -70,15 +69,21 @@ typedef struct
   size_t	data_size;		/* byte-size of data */
   size_t	size;			/* size in characters */
   IOSTREAM     *stream;			/* Stream hanging onto it */
-  atom_t 	atom;			/* Created from atom */
+  atom_t	atom;			/* Created from atom */
 } memfile;
 
 
 static int
 unify_memfile(term_t handle, memfile *f)
-{ return PL_unify_term(handle,
-		       PL_FUNCTOR, FUNCTOR_memory_file1,
-		         PL_POINTER, f);
+{ if ( PL_unify_term(handle,
+		     PL_FUNCTOR, FUNCTOR_memory_file1,
+		       PL_POINTER, f) )
+    return TRUE;
+
+  if ( !PL_is_variable(handle) )
+    return PL_uninstantiation_error(handle);
+
+  return FALSE;					/* (resource) error */
 }
 
 
@@ -188,7 +193,7 @@ static struct encname
 };
 
 
-IOENC
+static IOENC
 atom_to_encoding(atom_t a)
 { struct encname *en;
 
@@ -293,7 +298,12 @@ open_memory_file4(term_t handle, term_t mode, term_t stream, term_t options)
   fd->encoding = encoding;
   m->stream = fd;
 
-  return PL_unify_stream(stream, fd);
+  if ( PL_unify_stream(stream, fd) )
+  { return TRUE;
+  } else
+  { Sclose(fd);
+    return FALSE;
+  }
 }
 
 
