@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2013, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #define O_DEBUG 1
@@ -112,6 +111,17 @@ leave the details to this function.
 #undef HAVE_H_ERRNO
 #endif
 
+#if defined(__MINGW32__)
+#define __try
+#define __except(_) if (0)
+#define __finally
+#endif
+
+#if defined(__MINGW32__)
+#define WINVER 0x0501
+#include <ws2tcpip.h>
+#endif
+
 #include "nonblockio.h"
 
 #include <SWI-Stream.h>
@@ -154,30 +164,41 @@ leave the details to this function.
 static CRITICAL_SECTION mutex;
 static CRITICAL_SECTION mutex_free;
 
-#define LOCK() EnterCriticalSection(&mutex)
-#define UNLOCK() LeaveCriticalSection(&mutex)
-#define LOCK_FREE() EnterCriticalSection(&mutex_free)
-#define UNLOCK_FREE() LeaveCriticalSection(&mutex_free)
-#define INITLOCK() (InitializeCriticalSection(&mutex), \
-		    InitializeCriticalSection(&mutex_free))
-#else
+#define LOCK()			EnterCriticalSection(&mutex)
+#define UNLOCK()		LeaveCriticalSection(&mutex)
+#define LOCK_FREE()		EnterCriticalSection(&mutex_free)
+#define UNLOCK_FREE()		LeaveCriticalSection(&mutex_free)
+#define INITLOCK()		( InitializeCriticalSection(&mutex), \
+				  InitializeCriticalSection(&mutex_free))
+#define LOCK_SOCKET(s)		EnterCriticalSection(&s->socket_mutex)
+#define UNLOCK_SOCKET(s)	LeaveCriticalSection(&s->socket_mutex)
+#define INIT_SOCKET_LOCK(s)	InitializeCriticalSection(&s->socket_mutex)
+#define FREE_SOCKET_LOCK(s)	DeleteCriticalSection(&s->socket_mutex)
+#else /*__WINDOWS__*/
 #include <pthread.h>
 
 static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
 
-#define LOCK() pthread_mutex_lock(&mutex)
-#define UNLOCK() pthread_mutex_unlock(&mutex)
-#define LOCK_FREE()
-#define UNLOCK_FREE()
-#define INITLOCK()
-#endif
-#else
-#define LOCK()
-#define UNLOCK()
-#define LOCK_FREE()
-#define UNLOCK_FREE()
-#define INITLOCK()
+#define LOCK()			pthread_mutex_lock(&mutex)
+#define UNLOCK()		pthread_mutex_unlock(&mutex)
+#define LOCK_FREE()		(void)0
+#define UNLOCK_FREE()		(void)0
+#define INITLOCK()		(void)0
+#define FREE_SOCKET_LOCK(s)	(void)0
+#endif /*__WINDOWS__*/
+#else /* _REENTRANT */
+#define LOCK()			(void)0
+#define UNLOCK()		(void)0
+#define LOCK_FREE()		(void)0
+#define UNLOCK_FREE()		(void)0
+#define INITLOCK()		(void)0
+#define FREE_SOCKET_LOCK(s)	(void)0
+#if __WINDOWS__
+#define LOCK_SOCKET(s)		(void)0
+#define UNLOCK_SOCKET(s)	(void)0
+#define INIT_SOCKET_LOCK(s)	(void)0
 #endif
+#endif /* _REENTRANT */
 
 #define set(s, f)   ((s)->flags |= (f))
 #define clear(s, f) ((s)->flags &= ~(f))
@@ -206,6 +227,7 @@ typedef struct _plsocket
   DWORD             close_timeout;      /* Time to reap socket if FD_CLOSE is not received */
   int		    done;		/* request completed */
   int		    w32_flags;		/* or of received FD_* */
+  CRITICAL_SECTION  socket_mutex;       /* synchronization mutex */
   union
   { struct
     { struct sockaddr_in addr;		/* accepted address */
@@ -254,6 +276,7 @@ static int socketIsPendingClose(plsocket *s);
 static const char *WinSockError(unsigned long eno);
 #endif
 
+#ifndef __WINDOWS__
 static int
 need_retry(int error)
 { if ( error == EINTR || error == EAGAIN || error == EWOULDBLOCK )
@@ -261,6 +284,7 @@ need_retry(int error)
 
   return FALSE;
 }
+#endif
 
 #ifdef O_DEBUG
 static int debugging;
@@ -317,20 +341,32 @@ request_name(nbio_request request)
   }
 }
 
+static char *
+sepstrcatskip(char *buf, char *dest, char *src)
+{ if ( dest != buf )
+  { *dest++ = '|';
+    *dest = '\0';
+  }
+  strcat(dest, src);
+  dest += strlen(dest);
+
+  return dest;
+}
+
 static char *
 event_name(int ev)
 { char buf[256];
   char *o = buf;
 
   o[0] = '\0';
-  if ( (ev & FD_READ) )    strcat(o, "|FD_READ");
-  if ( (ev & FD_WRITE) )   strcat(o, "|FD_WRITE");
-  if ( (ev & FD_ACCEPT) )  strcat(o, "|FD_ACCEPT");
-  if ( (ev & FD_CONNECT) ) strcat(o, "|FD_CONNECT");
-  if ( (ev & FD_CLOSE) )   strcat(o, "|FD_CLOSE");
-  if ( (ev & FD_OOB) )	   strcat(o, "|FD_OOB");
+  if ( (ev & FD_READ) )    o=sepstrcatskip(buf, o, "FD_READ");
+  if ( (ev & FD_WRITE) )   o=sepstrcatskip(buf, o, "FD_WRITE");
+  if ( (ev & FD_ACCEPT) )  o=sepstrcatskip(buf, o, "FD_ACCEPT");
+  if ( (ev & FD_CONNECT) ) o=sepstrcatskip(buf, o, "FD_CONNECT");
+  if ( (ev & FD_CLOSE) )   o=sepstrcatskip(buf, o, "FD_CLOSE");
+  if ( (ev & FD_OOB) )	   o=sepstrcatskip(buf, o, "FD_OOB");
   if ( (ev & ~(FD_READ|FD_WRITE|FD_ACCEPT|FD_CONNECT|FD_CLOSE)) )
-    strcat(o, "|FD_???");
+    sepstrcatskip(buf, o, "FD_???");
 
   return strdup(buf);
 }
@@ -352,8 +388,12 @@ nbio_fcntl(nbio_sock_t socket, int op, int arg)
       switch(arg)
       { case O_NONBLOCK:
 	{ int rval;
+#if defined(__MINGW32__)
+	  u_long non_block;
+#else
+          /* FIXME: is this really `int' for MSC? */
 	  int non_block;
-
+#endif
 	  non_block = 1;
 	  rval = ioctlsocket(s->socket, FIONBIO, &non_block);
 	  if ( rval )
@@ -386,8 +426,10 @@ static int
 doneRequest(plsocket *s)
 { SOCKET sock;
 
+  LOCK_SOCKET(s);
   s->done = TRUE;
   s->request = REQ_NONE;
+  UNLOCK_SOCKET(s);
 
   /* If we have FD_CLOSE, then we know we can release the socket to the OS */
   if ( (s->w32_flags & FD_CLOSE) && (sock=s->socket) >= 0 )
@@ -396,7 +438,7 @@ doneRequest(plsocket *s)
   }
 
   if ( s->thread )
-  { DEBUG(2, Sdprintf("doneRequest(): posting %d\n", s->thread));
+  { DEBUG(2, Sdprintf("doneRequest(%d): posting %d\n", s->id, s->thread));
     PostThreadMessage(s->thread, WM_DONE, 0, (WPARAM)s);
   }
 
@@ -430,6 +472,8 @@ waitRequest(plsocket *s)
     } else if ( GetMessage(&msg, NULL, 0, 0) )
     { TranslateMessage(&msg);
       DispatchMessage(&msg);
+      if ( PL_exception(0) )
+	return FALSE;			/* DispatchMessage handled a signal */
     } else
     { ExitThread(0);			/* WM_QUIT received */
       return FALSE;			/* NOTREACHED */
@@ -445,15 +489,17 @@ nbio_wait(nbio_sock_t socket, nbio_request request)
   if ( !(s=nbio_to_plsocket(socket)) )
     return -1;
 
+  LOCK_SOCKET(s);
   s->flags  |= PLSOCK_WAITING;
   s->done    = FALSE;
   s->error   = 0;
   s->thread  = GetCurrentThreadId();
   s->request = request;
+  UNLOCK_SOCKET(s);
 
   SendMessage(State()->hwnd, WM_REQUEST, 1, (LPARAM)&s);
 
-  DEBUG(2, Sdprintf("[%d] (%ld): Waiting ...",
+  DEBUG(2, Sdprintf("[%d] (%ld): Waiting ... ",
 		    PL_thread_self(), s->thread));
 
   for(;;)
@@ -511,11 +557,13 @@ nbio_select(int n,
       { plsocket *s = nbio_to_plsocket(i);
 
 	if ( s )
-	{ s->flags  |= PLSOCK_WAITING;
+	{ LOCK_SOCKET(s);
+	  s->flags  |= PLSOCK_WAITING;
 	  s->done    = FALSE;
 	  s->error   = 0;
 	  s->thread  = GetCurrentThreadId();
 	  s->request = (s->flags & PLSOCK_LISTEN) ? REQ_ACCEPT : REQ_READ;
+	  UNLOCK_SOCKET(s);
 	  sockets[i] = s;
 	} else
 	{ DEBUG(2, Sdprintf("nbio_select(): no socket for %d\n", i));
@@ -596,14 +644,16 @@ placeRequest(plsocket *s, nbio_request request)
 { if ( s->magic != PLSOCK_MAGIC )
     Sdprintf("placeRequest: %p has bad magic\n", s);
 
+  LOCK_SOCKET(s);
   s->error   = 0;
   s->done    = FALSE;
   s->thread  = GetCurrentThreadId();
   s->request = request;
   clear(s, PLSOCK_WAITING);
+  UNLOCK_SOCKET(s);
 
   SendMessage(State()->hwnd, WM_REQUEST, 1, (LPARAM)&s);
-  DEBUG(2, Sdprintf("%d (%ld): Placed %s request for %d\n",
+  DEBUG(2, Sdprintf("[%d] (%ld): Placed %s request for %d\n",
 		    PL_thread_self(), s->thread,
 		    request_name(request), (int)s->socket));
 
@@ -671,7 +721,7 @@ doRequest(plsocket *s)
 			    (int)s->socket, WinSockError(s->error)));
 
 	  if ( s->error != WSAEWOULDBLOCK )
-	  { s->rdata.accept.slave = (SOCKET)-1;
+	  { s->rdata.accept.slave = (nbio_sock_t)-1;
 	    doneRequest(s);
 	  }
 	} else
@@ -785,7 +835,14 @@ doRequest(plsocket *s)
 	{ s->rdata.write.bytes = (int)s->rdata.write.written;
 	  doneRequest(s);
 	}
+      } else if ( s->w32_flags & FD_CLOSE )
+      { if ( s->rdata.write.written > 0 )
+	  s->rdata.write.bytes = (int)s->rdata.write.written;
+	else
+	  s->rdata.write.bytes = -1;
+	doneRequest(s);
       }
+      break;
     case REQ_SENDTO:
       if ( s->w32_flags & FD_WRITE )
       { int n;
@@ -819,6 +876,7 @@ doRequest(plsocket *s)
 	s->rdata.sendto.bytes = n;
 	doneRequest(s);
       }
+      break;
   }
 
   return TRUE;
@@ -880,7 +938,7 @@ socket_wnd_proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
       { DEBUG(1,
 	      { char *nm = event_name(evt);
 		Sdprintf("Got event %s (err=%s) on closed socket %d=%d\n",
-			 nm, WinSockError(err), s->id, sock);
+			 nm, err ? WinSockError(err) : "none", s->id, sock);
 		free(nm);
 	      })
       }
@@ -906,6 +964,9 @@ socket_wnd_proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 	switch(s->request)
 	{ case REQ_CONNECT:
 	    break;
+          case REQ_NONE:
+            /* FIXME: is this OK? */
+            break;
 	  case REQ_ACCEPT:
 	    s->rdata.accept.slave = SOCKET_ERROR;
 	    break;
@@ -947,7 +1008,7 @@ HiddenFrameClass()
   if ( !name )
   { char buf[50];
 
-    sprintf(buf, "PlSocketWin%d", (int)hinstance);
+    sprintf(buf, "PlSocketWin%d", (int)(uintptr_t)hinstance);
     name = strdup(buf);
 
     wndClass.style		= 0;
@@ -968,17 +1029,6 @@ HiddenFrameClass()
 }
 
 
-static void
-destroyHiddenWindow(int rval, void *closure)
-{ local_state *s = State();
-
-  if ( s->hwnd )
-  { DestroyWindow(s->hwnd);
-    s->hwnd = 0;
-  }
-}
-
-
 static HWND
 SocketHiddenWindow()
 { local_state *s = State();
@@ -1000,7 +1050,7 @@ SocketHiddenWindow()
 
 DWORD WINAPI
 socket_thread(LPVOID arg)
-{ DWORD parent = (DWORD)arg;
+{ DWORD parent = (DWORD)(uintptr_t)arg;
 
   SocketHiddenWindow();
   PostThreadMessage(parent, WM_READY, (WPARAM)0, (LPARAM)0);
@@ -1025,7 +1075,7 @@ startSocketThread()
 
   CreateThread(NULL,			/* security */
 	       2048,			/* stack */
-	       socket_thread, (LPVOID)me,	/* proc+arg */
+	       socket_thread, (LPVOID)(uintptr_t)me,	/* proc+arg */
 	       0,			/* flags */
 	       &State()->tid);
 
@@ -1246,8 +1296,8 @@ allocSocket(SOCKET socket)
   DWORD now = GetTickCount();
 
   if ( (p=lookupOSSocket(socket)) )
-  { DEBUG(1, Sdprintf("WinSock %d already registered on %d\n",
-		      (int)socket, p->id));
+  { DEBUG(1, Sdprintf("WinSock %d already registered on %d; tmo=%ld, now=%ld\n",
+		      (int)socket, p->id, (long)p->close_timeout, (long)now));
     p->socket = (SOCKET)-1;
   }
 #endif
@@ -1278,6 +1328,7 @@ allocSocket(SOCKET socket)
     { if ( p->close_timeout && p->close_timeout < now )
       { SOCKET sock;
 
+	p->close_timeout = 0;
 	if ( (sock=p->socket) )		/* is this ever the case? */
 	{ int rval;
 
@@ -1302,16 +1353,18 @@ allocSocket(SOCKET socket)
 #ifdef __WINDOWS__
   p->w32_flags = 0;
   p->request   = REQ_NONE;
+  INIT_SOCKET_LOCK(p);
 #endif
   p->input = p->output = (IOSTREAM*)NULL;
 
-  DEBUG(2, Sdprintf("[%d]: allocSocket(%d): bound to %p\n",
-		    PL_thread_self(), socket, p));
+  DEBUG(2, Sdprintf("[%d]: allocSocket(%d): bound to %d\n",
+		    PL_thread_self(), socket, p->id));
 
   return p;
 }
 
 
+#ifndef __WINDOWS__
 static int
 freeSocket(plsocket *s)
 { int rval;
@@ -1333,6 +1386,7 @@ freeSocket(plsocket *s)
   sock = s->socket;
   socket = s->id;
   s->magic = 0;
+  FREE_SOCKET_LOCK(s);
   PL_free(s);
   UNLOCK_FREE();
 
@@ -1350,6 +1404,7 @@ freeSocket(plsocket *s)
 
   return rval;
 }
+#endif
 
 
 		 /*******************************
@@ -1495,10 +1550,14 @@ int
 nbio_error(int code, nbio_error_map mapid)
 { const char *msg;
   term_t except = PL_new_term_ref();
+#ifdef HAVE_H_ERRNO
   error_codes *map;
+#endif
 
   if ( code == EPLEXCEPTION )
     return FALSE;
+
+#ifdef HAVE_H_ERRNO
   switch( mapid )
   { case TCP_HERRNO:
       map = h_errno_codes;
@@ -1506,6 +1565,8 @@ nbio_error(int code, nbio_error_map mapid)
     default:
       map = NULL;
   }
+#endif
+
   {
 #ifdef __WINDOWS__
   msg = WinSockError(code);
@@ -1563,15 +1624,12 @@ nbio_last_error(nbio_sock_t socket)
 
 NBIO_EXPORT(int)
 nbio_init(const char *module)
-{ INITLOCK();				/* is this ok? */
-
-  LOCK();
-  if ( initialised )
-  { UNLOCK();
-    return TRUE;
-  }
+{ if ( initialised )			/* called from install handlers, which */
+    return TRUE;			/* are serialized by the compiler mutex */
   initialised = TRUE;
 
+  INITLOCK();
+
   FUNCTOR_module2  = PL_new_functor(PL_new_atom(":"), 2);
   FUNCTOR_ip4	   = PL_new_functor(PL_new_atom("ip"), 4);
   FUNCTOR_ip1	   = PL_new_functor(PL_new_atom("ip"), 1);
@@ -1592,14 +1650,11 @@ nbio_init(const char *module)
 #endif
 
   if ( WSAStartup(MAKEWORD(2,0), &WSAData) )
-  { UNLOCK();
     return PL_warning("nbio_init() - WSAStartup failed.");
-  }
   startSocketThread();
 }
 #endif /*__WINDOWS__*/
 
-  UNLOCK();
   return TRUE;
 }
 
@@ -1610,6 +1665,7 @@ nbio_cleanup(void)
   {
 #ifdef __WINDOWS__
     WSACleanup();
+    SendMessage(State()->hwnd, WM_QUIT, 0, 0);
 #endif
   }
 
@@ -1618,7 +1674,7 @@ nbio_cleanup(void)
 
 #ifdef __WINDOWS__
 
-/* socketIsPendingClose() is called if a Windows socked is closed, but
+/* socketIsPendingClose() is called if a Windows socket is closed, but
    we did not yet see an FD_CLOSE message for it.  We will reuse this
    socket using lookupTimedOutSocket() if we have not seen the FD_CLOSE
    after 30 seconds.  This is a random time.  Are we guaranteed to get
@@ -1629,7 +1685,8 @@ static int
 socketIsPendingClose(plsocket *s)
 { s->close_timeout = GetTickCount() + 30 * 1000;
   shutdown(s->socket, SD_BOTH);
-  DEBUG(3, Sdprintf("Setting timeout for FD_CLOSE on socket %d...\n", s->id));
+  DEBUG(2, Sdprintf("Setting close_timeout on socket %d... to %ld\n",
+		    s->id, (long)s->close_timeout));
 
   return 0;
 }
@@ -1982,7 +2039,8 @@ nbio_connect(nbio_sock_t socket,
     { s->rdata.connect.addrlen = addrlen;
       memcpy(&s->rdata.connect.addr, serv_addr, addrlen);
       placeRequest(s, REQ_CONNECT);
-      waitRequest(s);
+      if ( !waitRequest(s) )
+	return -1;
     }
 
     if ( s->error )
