@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #ifdef HAVE_CONFIG_H
@@ -63,7 +63,6 @@ static atom_t ATOM_infinite;
 static functor_t FUNCTOR_error2;
 static functor_t FUNCTOR_type_error2;
 static functor_t FUNCTOR_domain_error2;
-static functor_t FUNCTOR_resource_error1;
 static functor_t FUNCTOR_process_error2;
 static functor_t FUNCTOR_system_error2;
 static functor_t FUNCTOR_pipe1;
@@ -84,7 +83,11 @@ static functor_t FUNCTOR_eq2;		/* =/2 */
 #include <stdio.h>
 #include <fcntl.h>
 #include <io.h>
+
+#if !defined(__MINGW32__)
 typedef DWORD  pid_t;
+#endif
+
 typedef wchar_t echar;			/* environment character */
 
 #ifndef CREATE_BREAKAWAY_FROM_JOB
@@ -141,22 +144,6 @@ domain_error(term_t actual, const char *expected)
 }
 
 
-static int
-resource_error(const char *resource)
-{ term_t ex;
-
-  if ( (ex=PL_new_term_ref()) &&
-       PL_unify_term(ex,
-		     PL_FUNCTOR, FUNCTOR_error2,
-		       PL_FUNCTOR, FUNCTOR_resource_error1,
-		         PL_CHARS, resource,
-		       PL_VARIABLE) )
-    return PL_raise_exception(ex);
-
-  return FALSE;
-}
-
-
 		 /*******************************
 		 *	       ADMIN		*
 		 *******************************/
@@ -295,10 +282,11 @@ parse_environment(term_t t, p_options *info)
   term_t head = PL_new_term_ref();
   term_t tmp  = PL_new_term_ref();
   ecbuf *eb   = &info->envbuf;
-  int count = 0, c = 0;
+  int count = 0;
 #ifndef __WINDOWS__
   echar *q;
   char **ep;
+  int c = 0;
 #endif
 
   assert(eb->size == 0);
@@ -446,9 +434,13 @@ get_exe(term_t exe, p_options *info)
   if ( !PL_get_chars(arg, &info->exe, CVT_ATOM|CVT_EXCEPTION|BUF_MALLOC|REP_FN) )
     return FALSE;
 
-  info->argv = PL_malloc((arity+2)*sizeof(char*));
+  if ( !(info->argv = PL_malloc((arity+2)*sizeof(char*))) )
+    return PL_resource_error("memory");
   memset(info->argv, 0, (arity+2)*sizeof(char*));
-  info->argv[0] = strdup(info->exe);
+  if ( !(info->argv[0] = PL_malloc(strlen(info->exe)+1)) )
+    return PL_resource_error("memory");
+  strcpy(info->argv[0], info->exe);
+
   { int i;
 
     for(i=1; i<=arity; i++)
@@ -621,7 +613,7 @@ open_process_pipe(process_context *pc, int which, int fd)
 
   pc->open_mask |= (1<<which);
 #ifdef __WINDOWS__
-  pc->pipes[which] = _open_osfhandle((long)fd, _O_BINARY);
+  pc->pipes[which] = _open_osfhandle((intptr_t)fd, _O_BINARY);
 #else
   pc->pipes[which] = fd;
 #endif
@@ -724,7 +716,7 @@ win_command_line(term_t t, int arity, const wchar_t *exe, wchar_t **cline)
     cmdlen = av[0].len+(av[0].quote?2:0)+1;
 
     for( i=1; i<=arity; i++)
-    { PL_get_arg(i, t, arg);
+    { _PL_get_arg(i, t, arg);
 
       if ( !PL_get_wchars(arg, &av[i].len, &av[i].text,
 			  CVT_ATOMIC|CVT_EXCEPTION|BUF_MALLOC) )
@@ -826,9 +818,8 @@ find_process_from_pid(DWORD pid, const char *pred)
   if ( pred )
   { term_t ex = PL_new_term_ref();
 
-    PL_put_integer(ex, pid);
-    pl_error(NULL, 2, NULL, ERR_EXISTENCE,
-	     "process", ex);
+    if ( PL_put_integer(ex, pid) )
+      PL_existence_error("process", ex);
   }
 
   return (HANDLE)0;
@@ -929,17 +920,17 @@ win_wait_success(atom_t exe, HANDLE process)
     return FALSE;
 
   if ( rc != 0 )
-  { term_t code = PL_new_term_ref();
-    term_t ex = PL_new_term_ref();
+  { term_t ex = PL_new_term_ref();
 
-    PL_unify_term(ex,
-		  PL_FUNCTOR, FUNCTOR_error2,
-		    PL_FUNCTOR, FUNCTOR_process_error2,
-		      PL_ATOM, exe,
-		      PL_FUNCTOR, FUNCTOR_exit1,
-		        PL_LONG, rc,
-		    PL_VARIABLE);
-    return PL_raise_exception(ex);
+    if ( PL_unify_term(ex,
+		       PL_FUNCTOR, FUNCTOR_error2,
+		         PL_FUNCTOR, FUNCTOR_process_error2,
+		           PL_ATOM, exe,
+		           PL_FUNCTOR, FUNCTOR_exit1,
+		             PL_LONG, rc,
+		         PL_VARIABLE) )
+      return PL_raise_exception(ex);
+    return FALSE;
   }
 
   return TRUE;
@@ -959,8 +950,14 @@ create_pipes(p_options *info)
   { p_stream *s = &info->streams[i];
 
     if ( s->term )
-    { if ( !CreatePipe(&s->fd[0], &s->fd[1], &sa, 1<<13) )
-      { return win_error("CreatePipe");
+    { if ( i == 2 && info->streams[1].term &&
+	   PL_compare(info->streams[1].term, info->streams[2].term) == 0 )
+      { s->fd[0] = info->streams[1].fd[0];
+	s->fd[1] = info->streams[1].fd[1];
+      } else
+      { if ( !CreatePipe(&s->fd[0], &s->fd[1], &sa, 1<<13) )
+	{ return win_error("CreatePipe");
+	}
       }
     }
   }
@@ -971,7 +968,7 @@ create_pipes(p_options *info)
 
 static IOSTREAM *
 Sopen_handle(HANDLE h, const char *mode)
-{ return Sfdopen(_open_osfhandle((long)h, _O_BINARY), mode);
+{ return Sfdopen(_open_osfhandle((intptr_t)h, _O_BINARY), mode);
 }
 
 
@@ -1028,6 +1025,8 @@ do_create_process(p_options *info)
 
   if ( info->detached )
     flags |= CREATE_BREAKAWAY_FROM_JOB;
+  if ( info->envbuf.buffer )
+    flags |= CREATE_UNICODE_ENVIRONMENT;
 
   memset(&si, 0, sizeof(si));
   si.cb = sizeof(si);
@@ -1156,9 +1155,15 @@ create_pipes(p_options *info)
   { p_stream *s = &info->streams[i];
 
     if ( s->term )
-    { if ( pipe(s->fd) )
-      { assert(errno = EMFILE);
-	return resource_error("open_files");
+    { if ( i == 2 && info->streams[1].term &&
+	   PL_compare(info->streams[1].term, info->streams[2].term) == 0 )
+      { s->fd[0] = info->streams[1].fd[0];
+	s->fd[1] = info->streams[1].fd[1];
+      } else
+      { if ( pipe(s->fd) )
+	{ assert(errno = EMFILE);
+	  return PL_resource_error("open_files");
+	}
       }
     }
   }
@@ -1503,7 +1508,7 @@ process_wait(term_t pid, term_t code, term_t options)
 
 static foreign_t
 process_kill(term_t pid, term_t signal)
-{ int p;
+{ pid_t p;
 
   if ( !get_pid(pid, &p) )
     return FALSE;
@@ -1573,7 +1578,6 @@ install_process()
   MKFUNCTOR(domain_error, 2);
   MKFUNCTOR(process_error, 2);
   MKFUNCTOR(system_error, 2);
-  MKFUNCTOR(resource_error, 1);
   MKFUNCTOR(exit, 1);
   MKFUNCTOR(killed, 1);
 
