@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #ifdef HAVE_CONFIG_H
@@ -41,6 +41,9 @@
 #ifdef HAVE_ALLOCA_H
 #include <alloca.h>
 #endif
+#ifdef HAVE_CRT_EXTERNS_H 
+#include <crt_externs.h>
+#endif
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Unix process management.
@@ -239,7 +242,12 @@ out:
 
 static foreign_t
 pl_environ(term_t l)
-{ extern char **environ;
+{
+#ifdef HAVE__NSGETENVIRON
+  char **environ = _NSGetEnviron();
+#else
+  extern char **environ;
+#endif
   char **e;
   term_t t = PL_copy_term_ref(l);
   term_t t2 = PL_new_term_ref();
@@ -260,8 +268,8 @@ pl_environ(term_t l)
       name[len] = '\0';
       PL_put_atom_chars(nt, name);
       PL_put_atom_chars(vt, s+1);
-      PL_cons_functor(nt, FUNCTOR_equal2, nt, vt);
-      if ( !PL_unify_list(t, t2, t) ||
+      if ( !PL_cons_functor(nt, FUNCTOR_equal2, nt, vt) ||
+	   !PL_unify_list(t, t2, t) ||
 	   !PL_unify(t2, nt) )
 	return FALSE;
     }
@@ -288,7 +296,7 @@ static ssize_t
 write_null(void *handle, char *buf, size_t count)
 { if ( error_fd )
   { if ( error_fd >= 0 )
-      write(error_fd, buf, count);
+      return write(error_fd, buf, count);
   } else if ( error_file )
   { error_fd = open(PL_atom_chars(error_file), O_WRONLY|O_CREAT|O_TRUNC, 0644);
     return write_null(handle, buf, count);
