@@ -1,11 +1,9 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2009, VU University Amsterdam
+    Copyright (C): 2009-2012, VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,7 +17,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #ifdef HAVE_CONFIG_H
@@ -1233,8 +1231,8 @@ ranges_in_charbuf(charbuf *cb, uri_component_ranges *ranges)
 
 
 typedef struct
-{ atom_t     	       atom;
-  pl_wchar_t 	      *text;
+{ atom_t	       atom;
+  pl_wchar_t	      *text;
   uri_component_ranges ranges;
 } base_cache;
 
@@ -1263,7 +1261,7 @@ myBase()
 
   if ( (base=pthread_getspecific(base_key)) )
     return base;
-  base = PL_malloc(sizeof(*base));
+  base = PL_malloc_uncollectable(sizeof(*base)); /* Due to bug in Boehm-GC */
   memset(base, 0, sizeof(*base));
 
   pthread_setspecific(base_key, base);
@@ -1605,5 +1603,8 @@ install_uri()
 
 install_t
 uninstall_uri()
-{ pthread_key_delete(base_key);
+{
+#ifdef _REENTRANT
+  pthread_key_delete(base_key);
+#endif
 }
