@@ -1,12 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2008-2010, University of Amsterdam
-			      Vu University Amsterdam
+    Copyright (C): 2008-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -20,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -58,6 +56,10 @@ two additional modules that acts as plugins:
     Loading this library causes http_open/3 to support the =POST= method
     in addition to =GET= and =HEAD=.
 
+    * library(http/http_ssl_plugin)
+    Loading this library causes http_open/3 HTTPS connections.  Relevant
+    options for SLL certificate handling are handled to ssl_context/3.
+
 Here is a simple example to fetch a web-page:
 
   ==
@@ -100,6 +102,23 @@ resource. See also parse_time/2.
 :- meta_predicate
 	http_open(+,-,:).
 
+:- predicate_options(http_open/3, 3,
+		     [ authorization(compound),
+		       final_url(-atom),
+		       header(+atom, -atom),
+		       method(oneof([get,head,post])),
+		       size(-integer),
+		       status_code(-integer),
+		       timeout(number),
+		       post(any),		    % library(http/http_header)
+		       proxy(atom, integer),
+		       proxy_authorization(compound),
+		       request_header(any),
+		       user_agent(atom),
+		       pem_password_hook(callable), % if SSL is loaded
+		       cert_verify_hook(callable)   % if SSL is loaded
+		     ]).
+
 %%	user_agent(-Agent) is det.
 %
 %	Default value for =|User-Agent|=,  can   be  overruled using the
@@ -186,10 +205,25 @@ user_agent('SWI-Prolog <http://www.swi-prolog.org>').
 %	@param	URL is either an atom (url) or a list of _parts_.
 %		If this list is provided, it may contain the fields
 %		=scheme=, =user=, =password=, =host=, =port=, =path= and
-%		=search= (where the argument of the latter is a
-%		Name(Value) list).  Only =host= is mandatory.
+%		=search= (where the argument of the latter is a list of
+%		Name(Value) or Name=Value).  Only =host= is mandatory.
+%		The following example below opens the URL
+%		=|http://www.example.com/my/path?q=Hello%20World&lang=en|=.
+%		Note that values must *not* be quoted because the
+%		library inserts the required quites.
+%
+%		==
+%		http_open([ host('www.example.com'),
+%			    path('/my/path'),
+%			    search([ q='Hello world',
+%			             lang=en
+%				   ])
+%			  ])
+%		==
 %
 %	@error existence_error(url, Id)
+%	@see ssl_context/3 for SSL related options if
+%	library(http/http_ssl_plugin) is loaded.
 
 http_open(URL, Stream, QOptions) :-
 	meta_options(is_meta, QOptions, Options),
@@ -214,7 +248,7 @@ http_open_parts(Parts, Stream, Options0) :-
 	open_socket(Host:ProxyPort, In, Out, Options),
 	parts_scheme(Parts, Scheme),
 	default_port(Scheme, DefPort),
-	option(port(Port), Parts, DefPort),
+	url_part(port(Port), Parts, DefPort),
 	host_and_port(Host, DefPort, Port, HostPort),
 	send_rec_header(Out, In, Stream, HostPort, RequestURI, Parts, Options),
 	return_final_url(Options).
@@ -222,7 +256,7 @@ http_open_parts(Parts, Stream, Options0) :-
 	memberchk(host(Host), Parts),
 	parts_scheme(Parts, Scheme),
 	default_port(Scheme, DefPort),
-	option(port(Port), Parts, DefPort),
+	url_part(port(Port), Parts, DefPort),
 	parts_request_uri(Parts, RequestURI),
 	Options = [visited(Parts)|Options0],
 	open_socket(Host:Port, SocketIn, SocketOut, Options),
@@ -269,9 +303,9 @@ guarded_send_rec_header(Out, In, Stream, Host, RequestURI, Parts, Options) :-
 	method(Options, MNAME),
 	http_version(Version),
 	format(Out,
-	       '~w ~w HTTP/~w\r\n\
-	       Host: ~w\r\n\
-	       User-Agent: ~w\r\n\
+	       '~w ~w HTTP/~w\r\n\c
+	       Host: ~w\r\n\c
+	       User-Agent: ~w\r\n\c
 	       Connection: close\r\n',
 	       [MNAME, RequestURI, Version, Host, Agent]),
 	x_headers(Options, Out),
@@ -378,7 +412,7 @@ do_open(Code, _, Lines, Options0, Parts, In, Stream) :-
 	redirect_options(Options0, Options),
 	http_open(RedirectedParts, Stream, Options).
 					% Accepted codes
-do_open(Code, _, Lines, Options, Parts, In0, In) :- !,
+do_open(Code, _, Lines, Options, Parts, In0, In) :-
 	(   option(status_code(Code), Options)
 	->  true
 	;   Code == 200
@@ -481,7 +515,7 @@ open_socket(Address, In, Out, Options) :-
 	      )),
 	debug(http(open), '\tok ~p --> ~p', [In, Out]),
 	set_stream(In, record_position(false)),
-	(   memberchk(Options, timeout(Timeout))
+	(   memberchk(timeout(Timeout), Options)
 	->  set_stream(In, timeout(Timeout))
 	;   true
 	).
@@ -546,9 +580,10 @@ transfer_encoding(Encoding) -->
 	field("transfer-encoding"),
 	rest(Encoding).
 
-%%	read_header(+In:stream, -Code:int, -Comment:atom, -Lines:list)
+%%	read_header(+In:stream, -Code:int, -Comment:atom, -Lines:list) is det.
 %
-%	Read the HTTP reply-header.
+%	Read the HTTP reply-header. If the replied header is invalid, it
+%	simulates a 500 error with the comment =|Invalid reply header|=.
 %
 %	@param Code	Numeric HTTP reply-code
 %	@param Comment	Comment of reply-code as atom
@@ -557,8 +592,15 @@ transfer_encoding(Encoding) -->
 read_header(In, Code, Comment, Lines) :-
 	read_line_to_codes(In, Line),
 	phrase(first_line(Code, Comment), Line),
+	debug(http(open), '~w ~w', [Code, Comment]),
 	read_line_to_codes(In, Line2),
-	rest_header(Line2, In, Lines).
+	rest_header(Line2, In, Lines), !,
+	(   debugging(http(open))
+	->  forall(member(HL, Lines),
+		   debug(http(open), '~s', [HL]))
+	;   true
+	).
+read_header(_, 500, 'Invalid reply header', []).
 
 rest_header("", _, []) :- !.		% blank line: end of header
 rest_header(L0, In, [L0|L]) :-
@@ -684,7 +726,7 @@ check_authorization(basic(User, Password)) :-
 	must_be(atom, User),
 	must_be(atom, Password).
 
-%%	authorization(+URL, -Authorization) is semdet.
+%%	authorization(+URL, -Authorization) is semidet.
 %
 %	True if Authorization must be supplied for URL.
 %
@@ -707,8 +749,8 @@ authorization(URL, Authorization) :-
 add_authorization(_, Options, Options) :-
 	option(authorization(_), Options), !.
 add_authorization(Parts, Options0, Options) :-
-	option(user(User), Parts),
-	option(password(Passwd), Parts),
+	url_part(user(User), Parts),
+	url_part(password(Passwd), Parts),
 	Options = [authorization(basic(User,Passwd))|Options0].
 add_authorization(Parts, Options0, Options) :-
 	stored_authorization(_, _) ->	% quick test to avoid work
@@ -785,25 +827,25 @@ uri_request_uri(Components) -->
 %%	parts_authority(+Parts, -Authority) is semidet.
 
 parts_scheme(Parts, Scheme) :-
-	option(scheme(Scheme), Parts), !.
+	url_part(scheme(Scheme), Parts), !.
 parts_scheme(Parts, Scheme) :-		% compatibility with library(url)
-	option(protocol(Scheme), Parts), !.
+	url_part(protocol(Scheme), Parts), !.
 parts_scheme(_, http).
 
 parts_authority(Parts, Auth) :-
-	option(authority(Auth), Parts), !.
+	url_part(authority(Auth), Parts), !.
 parts_authority(Parts, Auth) :-
-	option(host(Host), Parts, _),
-	option(port(Port), Parts, _),
-	option(user(User), Parts, _),
-	option(password(Password), Parts, _),
+	url_part(host(Host), Parts, _),
+	url_part(port(Port), Parts, _),
+	url_part(user(User), Parts, _),
+	url_part(password(Password), Parts, _),
 	uri_authority_components(Auth,
 				 uri_authority(User, Password, Host, Port)).
 
 parts_request_uri(Parts, RequestURI) :-
 	memberchk(request_uri(RequestURI), Parts), !.
 parts_request_uri(Parts, RequestURI) :-
-	option(path(Path), Parts, /),
+	url_part(path(Path), Parts, /),
 	ignore(parts_search(Parts, Search)),
 	uri_data(path, Data, Path),
 	uri_data(search, Data, Search),
@@ -827,6 +869,20 @@ parts_uri(Parts, URI) :-
 	uri_data(authority, Data, Auth),
 	uri_components(URI, Data).
 
+url_part(Part, Parts) :-
+	Part =.. [Name,Value],
+	Gen =.. [Name,RawValue],
+	memberchk(Gen, Parts), !,
+	Value = RawValue.
+
+url_part(Part, Parts, Default) :-
+	Part =.. [Name,Value],
+	Gen =.. [Name,RawValue],
+	(   memberchk(Gen, Parts)
+	->  Value = RawValue
+	;   Value = Default
+	).
+
 
 		 /*******************************
 		 *	      COOKIES		*
@@ -859,12 +915,12 @@ update_cookies(Lines, Parts, Options) :-
 %
 %	    ==
 %	    :- multifile
-%	    	http:open_options/2.
+%		http:open_options/2.
 %
 %	    http:open_options(Parts, Options) :-
-%	    	memberchk(host(Host), Parts),
-%	    	Host \== localhost,
-%	    	Options = [proxy('proxy.local', 3128)].
+%		memberchk(host(Host), Parts),
+%		Host \== localhost,
+%		Options = [proxy('proxy.local', 3128)].
 %	    ==
 
 %%	http:write_cookies(+Out, +Parts, +Options) is semidet.
