@@ -39,9 +39,10 @@ public class TestJUnit extends TestCase {
 	}
 	protected void setUp() {
 		// initialization code
-		// Prolog.set_default_init_args(new String[] { "libpl.dll", "-f", "none", "-g", "set_prolog_flag(debug_on_error,false)", "-q" });
-		Prolog.set_default_init_args(new String[] { "libpl.dll", "-f", "none", "-g", "true", "-q" });
+		// Prolog.set_default_init_args(new String[] { "swipl.dll", "-f", "none", "-g", "set_prolog_flag(debug_on_error,false)", "-q" });
+		Prolog.set_default_init_args(new String[] { "swipl.dll", "-x", "../../src/swipl.prc", "-f", "none", "-g", "true", "-q", "--home=../.." });
 		assertTrue((new Query("consult(test_jpl)")).hasSolution());
+		assertTrue((new Query("use_module(library(jpl))")).hasSolution());
 	}
 	protected void tearDown() {
 		// cleanup code
