@@ -1,6 +1,4 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@uva.nl
@@ -19,7 +17,7 @@
 
     You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -40,7 +38,7 @@
 :- use_module(library(http/http_parameters)).
 :- use_module(library(http/html_write)).
 :- use_module(library(http/mimetype)).
-:- use_module(library(http/dcg_basics)).
+:- use_module(library(dcg/basics)).
 :- use_module(library(http/http_dispatch)).
 :- use_module(library(http/http_hook)).
 :- use_module(library(http/http_path)).
@@ -62,6 +60,7 @@
 :- use_module(pldoc(doc_wiki)).
 :- use_module(pldoc(doc_util)).
 :- use_module(pldoc(doc_access)).
+:- use_module(pldoc(doc_pack)).
 
 /** <module> Documentation server
 
@@ -130,13 +129,10 @@ doc_server(Port, Options) :-
 	merge_options(ServerOptions,
 		      [ port(Port),
 			timeout(60),
-			keep_alive_timeout(1),
-			local(4000),	% small stacks for now
-			global(4000),
-			trail(4000)
+			keep_alive_timeout(1)
 		      ], HTTPOptions),
 	http_server(http_dispatch, HTTPOptions),
-	assert(doc_server_port(Port)),
+	assertz(doc_server_port(Port)),
 	print_message(informational, pldoc(server_started(Port))).
 
 %%	doc_current_server(-Port) is det.
@@ -166,7 +162,9 @@ doc_current_server(Port) :-
 doc_browser :-
 	doc_browser([]).
 doc_browser(Spec) :-
-	doc_current_server(Port),
+	catch(doc_current_server(Port),
+	      error(existence_error(http_server, pldoc), _),
+	      doc_server(Port)),
 	browser_url(Spec, Request),
 	format(string(URL), 'http://localhost:~w~w', [Port, Request]),
 	www_open_url(URL).
@@ -216,7 +214,7 @@ prepare_editor.
 		[prefix, authentication(pldoc(read))]).
 :- http_handler(pldoc('index.html'), pldoc_index,   []).
 :- http_handler(pldoc(file),	   pldoc_file,	   []).
-:- http_handler(pldoc(directory),  pldoc_dir,	   []).
+:- http_handler(pldoc(place),	   go_place,	   []).
 :- http_handler(pldoc(edit),	   pldoc_edit,
 		[authentication(pldoc(edit))]).
 :- http_handler(pldoc(doc),	   pldoc_doc,	   [prefix]).
@@ -267,10 +265,21 @@ pldoc_index(_Request) :-
 			title('SWI-Prolog documentation'),
 			[ \doc_links('', []),
 			   h1('SWI-Prolog documentation'),
-			  \man_overview([])
+			  \man_overview([]),
+			  \paperback
 			]).
 
 
+paperback -->
+	{ expand_url_path(swipl_book(.), HREF)
+	},
+	html([ h2('The manual as a book'),
+	       p([ 'A paperback version of the manual is ',
+		   a(href(HREF), 'available'), '.'
+		 ])
+	     ]).
+
+
 %%	pldoc_file(+Request)
 %
 %	Hander for /file?file=File, providing documentation for File.
@@ -341,15 +350,20 @@ pldoc_edit(_Request) :-
 	throw(http_reply(forbidden(Location))).
 
 
-%%	pldoc_dir(+Request)
+%%	go_place(+Request)
 %
-%	Handler for /directory?dir=Dir, providing an index for
-%	Dir.  Mapped to /doc/Dir/index.html.
+%	HTTP handler to handle the places menu.
 
-pldoc_dir( Request) :-
+go_place(Request) :-
 	http_parameters(Request,
-			[ dir(Dir0, [])
+			[ place(Place, [])
 			]),
+	places(Place).
+
+places(':packs:') :- !,
+	http_link_to_id(pldoc_pack, [], HREF),
+	throw(http_reply(moved(HREF))).
+places(Dir0) :-
 	expand_alias(Dir0, Dir),
 	(   allowed_directory(Dir)
 	->  format(string(IndexFile), '~w/index.html', [Dir]),
@@ -464,6 +478,17 @@ documentation(File, Request) :-
 	edit_options(Request, Options),
 	doc_for_wiki_file(WikiFile, Options).
 documentation(Path, Request) :-
+	pl_file(Path, File),
+	(   allowed_file(File)
+	->  true
+	;   throw(http_reply(forbidden(File)))
+	),
+	doc_reply_file(File, Request).
+
+:- public
+	doc_reply_file/2.
+
+doc_reply_file(File, Request) :-
 	http_parameters(Request,
 			[ public_only(Public),
 			  reload(Reload),
@@ -471,11 +496,10 @@ documentation(Path, Request) :-
 			],
 			[ attribute_declarations(param)
 			]),
-	pl_file(Path, File),
-	(   allowed_file(File)
-	->  true
-	;   throw(http_reply(forbidden(File)))
-	),
+        (   exists_file(File)
+        ->  true
+        ;   throw(http_reply(not_found(File)))
+        ),
 	(   Reload == true,
 	    source_file(File)
 	->  load_files(File, [if(changed), imports([])])
@@ -499,6 +523,8 @@ documentation(Path, Request) :-
 	).
 
 
+:- public src_skin/4.			% called through source_to_html/3.
+
 src_skin(Request, _Show, header, Out) :- !,
 	memberchk(request_uri(ReqURI), Request), !,
 	replace_parameters(ReqURI, [show(raw)], RawLink),
@@ -536,10 +562,10 @@ edit_options(_, []).
 
 pl_file(File, PlFile) :-
 	file_name_extension(Base, html, File), !,
-	absolute_file_name(Base,
+	absolute_file_name(Base, PlFile,
 			   [ file_type(prolog),
 			     access(read)
-			   ], PlFile).
+			   ]).
 pl_file(File, File).
 
 %%	wiki_file(+File, -TxtFile) is semidet.
@@ -577,24 +603,57 @@ clean_path(Path, Path).
 %
 %	    * predicate=PI
 %	    providing documentation from the manual on the predicate PI.
+%	    * function=PI
+%	    providing documentation from the manual on the function PI.
 %	    * 'CAPI'=F
 %	    providing documentation from the manual on the C-function F.
 
 pldoc_man(Request) :-
 	http_parameters(Request,
 			[ predicate(PI, [optional(true)]),
+			  function(Fun, [optional(true)]),
 			  'CAPI'(F,     [optional(true)])
 			]),
-	format(string(Title), 'Manual -- ~w', [PI]),
 	(   ground(PI)
-	->  Obj = PI
+	->  split_pi(PI, Obj)
+	;   ground(Fun)
+	->  atomic_list_concat([Name,ArityAtom], /, Fun),
+	    atom_number(ArityAtom, Arity),
+	    Obj = f(Name/Arity)
 	;   ground(F)
 	->  Obj = c(F)
 	),
+	man_title(Obj, Title),
 	reply_html_page(pldoc(man),
 			title(Title),
 			\man_page(Obj, [])).
 
+man_title(f(Obj), Title) :-
+	format(atom(Title), 'SWI-Prolog -- function ~w', [Obj]).
+man_title(c(Obj), Title) :-
+	format(atom(Title), 'SWI-Prolog -- API-function ~w', [Obj]).
+man_title(Obj, Title) :-
+	format(atom(Title), 'SWI-Prolog -- ~w', [Obj]).
+
+split_pi(Atom, Module:PI) :-
+	atomic_list_concat([Module, PIAtom], :, Atom), !,
+	split_pi2(PIAtom, PI).
+split_pi(Atom, PI) :-
+	split_pi2(Atom, PI).
+
+split_pi2(Atom, Name//Arity) :-
+	sub_atom(Atom, B, _, A, //),
+	sub_atom(Atom, _, A, 0, ArityA),
+	atom_number(ArityA, Arity), !,
+	sub_atom(Atom, 0, B, _, Name).
+split_pi2(Atom, Name/Arity) :-
+	sub_atom(Atom, B, _, A, /),
+	sub_atom(Atom, _, A, 0, ArityA),
+	atom_number(ArityA, Arity), !,
+	sub_atom(Atom, 0, B, _, Name).
+
+
+
 %%	pldoc_object(+Request)
 %
 %	Handler for /doc_for?object=Term, Provide  documentation for the
