@@ -5,7 +5,8 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2004-2009, University of Amsterdam
+    Copyright (C): 2004-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -36,7 +37,7 @@
 	  ]).
 :- use_module(library(assoc)).
 :- use_module(library(option)).
-:- use_module(library('semweb/rdf_db')).
+:- use_module(library(semweb/rdf_db)).
 :- use_module(library(debug)).
 :- use_module(library(uri)).
 :- use_module(library(record)).
@@ -46,6 +47,31 @@
 :- meta_predicate
 	rdf_process_turtle(+,2,+).
 
+:- predicate_options(rdf_load_turtle/3, 3,
+		     [pass_to(rdf_read_turtle/3, 3)]).
+:- predicate_options(rdf_process_turtle/3, 3,
+		     [ anon_prefix(atom),
+		       base_uri(atom),
+		       base_used(-atom),
+		       db(atom),
+		       error_count(-integer),
+		       namespaces(-list),
+		       on_error(oneof([warning,error])),
+		       prefixes(-list),
+		       resources(oneof([uri,iri]))
+		     ]).
+:- predicate_options(rdf_read_turtle/3, 3,
+		     [ anon_prefix(atom),
+		       base_uri(atom),
+		       base_used(-atom),
+		       db(atom),
+		       error_count(-integer),
+		       namespaces(-list),
+		       on_error(oneof([warning,error])),
+		       prefixes(-list),
+		       resources(oneof([uri,iri]))
+		     ]).
+
 /** <module> Turtle: Terse RDF Triple Language
 
 This module implements the Turtle  language   for  representing  the RDF
@@ -122,6 +148,10 @@ one of the extensions =|.ttl|=, =|.n3|= or =|.nt|=.
 %		* error_count(-Count)
 %		If on_error(warning) is active, this option cane be
 %		used to retrieve the number of generated errors.
+%
+%	@param	Input is one of stream(Stream), atom(Atom), a =http=,
+%		=https= or =file= url or a filename specification as
+%		accepted by absolute_file_name/3.
 
 rdf_read_turtle(In, Triples, Options) :-
 	open_input(In, Stream, Close),
@@ -233,7 +263,7 @@ step_error(State, Error) :-
 
 open_input(stream(Stream), Stream, true) :- !,
 	stream_property(Stream, encoding(Old)),
-	(   Old == utf8
+	(   unicode_encoding(Old)
 	->  Close = true
 	;   set_stream(Stream, encoding(utf8)),
 	    Close = set_stream(Stream, encoding(Old))
@@ -241,10 +271,18 @@ open_input(stream(Stream), Stream, true) :- !,
 open_input(Stream, Stream, Close) :-
 	is_stream(Stream), !,
 	open_input(stream(Stream), Stream, Close).
+open_input(atom(Atom), Stream, close(Stream)) :- !,
+	atom_to_memory_file(Atom, MF),
+	open_memory_file(MF, read, Stream, [free_on_close(true)]).
 open_input(URL, Stream, close(Stream)) :-
-	sub_atom(URL, 0, _, _, 'http://'), !,
+	(   sub_atom(URL, 0, _, _, 'http://')
+	;   sub_atom(URL, 0, _, _, 'https://')
+	), !,
 	http_open(URL, Stream, []),
 	set_stream(Stream, encoding(utf8)).
+open_input(URL, Stream, close(Stream)) :-
+	uri_file_name(URL, Path), !,
+	open(Path, read, Stream, [encoding(utf8)]).
 open_input(File, Stream, close(Stream)) :-
 	absolute_file_name(File, Path,
 			   [ access(read),
@@ -252,6 +290,11 @@ open_input(File, Stream, close(Stream)) :-
 			   ]),
 	open(Path, read, Stream, [encoding(utf8)]).
 
+unicode_encoding(utf8).
+unicode_encoding(wchar_t).
+unicode_encoding(unicode_be).
+unicode_encoding(unicode_le).
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 The parser is a two-stage processor. The  first reads the raw file input
 and generates a list of tokens, stripping   comments and white space. It
@@ -265,7 +308,10 @@ init_state(In, Stream, Options, State) :-
 	->  true
 	;   In = stream(_)
 	->  BaseURI = []
-	;   uri_is_global(In)
+	;   In = atom(_)
+	->  BaseURI = []
+	;   uri_is_global(In),
+	    \+ is_absolute_file_name(In)	% Avoid C:Path in Windows
 	->  uri_normalized(In, BaseURI)
 	;   uri_file_name(BaseURI, In)
 	),
@@ -324,7 +370,9 @@ triples(State, []) -->
 triples(State, []) -->
 	[ '@', name(prefix), ':' ], !,
 	iri(State, URI),
-	{ set_base_uri_of_ttl_state(URI, State)
+	{ ttl_state_prefix_map(State, Map0),
+	  put_assoc('', Map0, URI, Map),
+	  set_prefix_map_of_ttl_state(Map, State)
 	}.
 triples(State, []) -->
 	[ '@', name(base) ], !,
@@ -430,28 +478,46 @@ resource(State, IRI) -->
 	iri(State, IRI), !.
 resource(State, IRI) -->
 	[ :(Name) ], !,
-	{ ttl_state_base_uri(State, Base),
-	  atom_concat(Base, Name, URI),
-	  uri_iri(State, URI, IRI)
-	}.
+	{ prefix_iri(:(Name), State, IRI) }.
 resource(State, IRI) -->
 	[ name(Prefix), : ], !,
-	{ ttl_state_prefix_map(State, Map),
-	  get_assoc(Prefix, Map, IRI)
-	}.
+	{ prefix_iri(prefix(Prefix), State, IRI) }.
 resource(State, IRI) -->
 	[ Prefix:Name ], !,
-	{ ttl_state_prefix_map(State, Map),
-	  (   get_assoc(Prefix, Map, Base)
-	  ->  atom_concat(Base, Name, URI),
-	      uri_iri(State, URI, IRI)
-	  ;   throw(error(existence_error(prefix, Prefix), _))
-	  )
-	}.
-resource(State, BaseIRI) -->
+	{ prefix_iri(Prefix:Name, State, IRI) }.
+resource(State, IRI) -->
 	[ : ], !,
-	{ ttl_state_base_uri(State, BaseIRI)
-	}.
+	{ prefix_iri(:, State, IRI) }.
+
+%%	prefix_iri(+PrefixSpec, +State, -IRI).
+
+prefix_iri(:(Name), State, IRI) :-		% :<prefix>
+	prefix('', Base, State),
+	atom_concat(Base, Name, URI),
+	uri_iri(State, URI, IRI).
+prefix_iri(prefix(Prefix), State, IRI) :-	% <prefix>:
+	prefix(Prefix, IRI, State).
+prefix_iri(Prefix:Name, State, IRI) :-		% <prefix>:<local>
+	prefix(Prefix, Base, State),
+	atom_concat(Base, Name, URI),
+	uri_iri(State, URI, IRI).
+prefix_iri(:, State, IRI) :-			% :
+	prefix('', URI, State),
+	uri_iri(State, URI, IRI).
+
+prefix(Prefix, URI, State) :-
+	ttl_state_prefix_map(State, Map),
+	get_assoc(Prefix, Map, URI), !.
+prefix(Prefix, URI, _) :-
+	predefined(Prefix), !,
+	rdf_current_ns(Prefix, URI).
+prefix(Prefix, _, _) :-
+	throw(error(existence_error(prefix, Prefix), _)).
+
+predefined(rdf).
+predefined(rdfs).
+predefined(owl).
+predefined(xsd).
 
 uri_iri(State, URI, IRI) :-
 	(   ttl_state_resources(State, uri)
@@ -526,21 +592,37 @@ anonid(State, _NodeId, Node) :-
 	anonid(State, Node).
 anonid(_State, NodeId, node(NodeId)).
 
-mk_object(type(Prefix:Name, Value), State, literal(type(Type, Value))) :- !,
-	  ttl_state_prefix_map(State, Map),
-	  get_assoc(Prefix, Map, Base),
-	  atom_concat(Base, Name, Type).
-mk_object(type(relative_uri(Rel), Value), State, literal(type(Type, Value))) :- !,
-	  ttl_state_base_uri(State, Base),
-	  (   Rel == ''			% must be in global_url?
-	  ->  Type = Base
-	  ;   uri_normalized_iri(Rel, Base, Type)
-	  ).
-mk_object(type(:(Name), Value), State, literal(type(Type, Value))) :- !,
-	  ttl_state_base_uri(State, Base),
-	  atom_concat(Base, Name, Type).
+mk_object(type(TypeSpec, Value0), State, literal(type(TypeIRI, Value))) :- !,
+	type_iri(TypeSpec, State, TypeIRI),
+	convert_literal(TypeIRI, Value0, Value).
 mk_object(Value, _State, literal(Value)).
 
+type_iri(relative_uri(Rel), State, TypeIRI) :- !,
+	ttl_state_base_uri(State, Base),
+	(   Rel == ''			% must be in global_url?
+	->  TypeIRI = Base
+	;   uri_normalized_iri(Rel, Base, TypeIRI)
+	).
+type_iri(Spec, State, TypeIRI) :-
+	prefix_iri(Spec, State, TypeIRI).
+
+
+%%	convert_literal(+Type, +Text, -Value) is det.
+%
+%	Convert  rdf:XMLLiteral  values  into  the   XML  DOM.  This  is
+%	consistent with the XML based version. Unclear whether this is a
+%	good idea.
+
+:- rdf_meta
+	convert_literal(r,+,-).
+
+convert_literal(rdf:'XMLLiteral', Text, DOM) :-
+	atom_to_memory_file(Text, MF),
+	open_memory_file(MF, read, In, [free_on_close(true)]),
+	load_structure(stream(In), DOM, [dialect(xml)]),
+	close(In).
+convert_literal(_, Value, Value).
+
 syntax_rule(State, Error) -->
 	error_tokens(7, Tokens),
 	{ ttl_state_input(State, Stream),
@@ -678,8 +760,7 @@ turtle_tokens(C0, In, List) :-
 	).
 
 turtle_token(-1, _, -1, end_of_file) :- !.
-turtle_token(0'., In, C, '.') :- !,
-	get_code(In, C).
+turtle_token(0'., _, end, '.') :- !.	% Turtle does not demand a space here!
 turtle_token(0'#, In, C, Token) :- !,
 	get_code(In, C1),
 	skip_line(C1, In, C2),
@@ -750,7 +831,7 @@ turtle_number(C0, In, CN, Value) :-
 	turtle_number_nn(C0, In, CN, Value).
 
 turtle_number_nn(C, In, CN, numeric(Type, Codes)) :-
-	turtle_integer_codes(C, In, CN0, Codes, T0), 	% [0-9]+
+	turtle_integer_codes(C, In, CN0, Codes, T0),	% [0-9]+
 	(   CN0 == 0'.
 	->  T0 = [CN0|T1],
 	    get_code(In, C1),
@@ -869,23 +950,29 @@ turtle_ws(0x9).
 turtle_ws(0xA).
 turtle_ws(0xD).
 turtle_ws(0x20).
+turtle_ws(0xFEFF).			% Unicode BOM marker and zero-width space
 
 syntax_error(Stream, Line, Which) :-
 	syntax_error_term(Stream, Line, Which, Error),
 	throw(Error).
 
 syntax_error_term(Stream, -1, Which, Error) :- !,
-	stream_property(Stream, file_name(File)),
+	stream_file_name(Stream, File),
 	line_count(Stream, LineNo),
 	line_position(Stream, LinePos),
 	character_count(Stream, CharIndex),
 	Error = error(syntax_error(Which),
 		      file(File, LineNo, LinePos, CharIndex)).
 syntax_error_term(Stream, LineNo, Which, Error) :-
-	stream_property(Stream, file_name(File)),
+	stream_file_name(Stream, File),
 	Error = error(syntax_error(Which),
 		      file(File, LineNo, -1, -1)).
 
+stream_file_name(Stream, File) :-
+	stream_property(Stream, file_name(File)), !.
+stream_file_name(_, '<no file>').
+
+
 
 		 /*******************************
 		 *	    RDF-DB HOOK		*
@@ -898,7 +985,8 @@ syntax_error_term(Stream, LineNo, Which, Error) :-
 rdf_db:rdf_load_stream(turtle, Stream, _Module:Options) :-
 	rdf_db:graph(Options, Id),
 	rdf_transaction(rdf_process_turtle(Stream, assert_triples, Options),
-			parse(Id)).
+			parse(Id)),
+	rdf_set_graph(Id, modified(false)).
 
 assert_triples([], _).
 assert_triples([rdf(S,P,O)|T], Location) :-
