@@ -1,9 +1,10 @@
 /*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        J.Wielemaker@uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2009, University of Amsterdam
+    Copyright (C): 2009-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -17,7 +18,7 @@
 
     You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -44,6 +45,27 @@
 :- use_module(library(debug)).
 :- use_module(library(sgml_write)).
 
+:- predicate_options(rdf_save_turtle/2, 2,
+		     [ graph(atom),
+		       base(atom),
+		       encoding(oneof([utf8])),
+		       indent(nonneg),
+		       tab_distance(nonneg),
+		       silent(boolean),
+		       subject_white_lines(nonneg),
+		       align_prefixes(boolean),
+		       user_prefixes(boolean),
+		       only_known_prefixes(boolean),
+		       comment(boolean),
+		       group(boolean),
+		       single_line_bnodes(boolean),
+		       canonize_numbers(boolean),
+		       canonical(boolean),
+		       expand(any)
+		     ]).
+:- predicate_options(rdf_save_canonical_turtle/2, 2,
+		     [ pass_to(rdf_save_turtle/2, 2)
+		     ]).
 
 /** <module> Turtle - Terse RDF Triple Language writer
 
@@ -164,9 +186,9 @@ has the following properties:
 %
 %	    ==
 %	    triple_in(RDF, S,P,O,_G) :-
-%	    	member(rdf(S,P,O), RDF).
+%		member(rdf(S,P,O), RDF).
 %
-%	    	...,
+%		...,
 %	        rdf_save_turtle(Out, [ expand(triple_in(RDF)) ]),
 %	    ==
 %
@@ -237,10 +259,15 @@ rdf_save_canonical_turtle(Spec, M:Options) :-
 %	@param	Cleanup is a goal that must be used to revert the side
 %		effects of open_output/4.
 
-open_output(stream(Out), Encoding, Out,
-	    set_stream(Out, encoding(Old))) :- !,
+open_output(stream(Out), Encoding, Out, Cleanup) :- !,
 	stream_property(Out, encoding(Old)),
-	set_stream(Out, encoding(Encoding)).
+	(   (   Old == Encoding
+	    ;	Old == wchar_t		% Internal encoding
+	    )
+	->  Cleanup = true
+	;   set_stream(Out, encoding(Encoding)),
+	    Cleanup = set_stream(Out, encoding(Old))
+	).
 open_output(Stream, Encoding, Out, Cleanup) :-
 	\+ atom(Stream),
 	is_stream(Stream), !,
@@ -289,6 +316,8 @@ init_prefix_map(State0, State) :-
 %	Test whether we want  to  include   the  proposed  prefix in the
 %	@prefix declaration.
 
+:- public turtle_prefix/4.		% called through rdf_graph_prefixes/3.
+
 turtle_prefix(true, _, Prefix, _) :- !,
 	rdf_current_ns(_, Prefix), !.
 turtle_prefix(_, _, Prefix, URI) :-
@@ -394,7 +423,7 @@ namespace_parts(URL, _) :-
 parts(List) -->	sep2, parts2(List).
 
 parts2([H|T]) -->
-	string(Codes), 	{Codes \== []},
+	string(Codes),	{Codes \== []},
 	sep, !,
 	{atom_codes(H, Codes)},
 	parts2(T).
@@ -961,12 +990,14 @@ subject(State, Subject) :-
 	(   atom(Graph)
 	->  rdf_subject(Subject),
 	    (   rdf(Subject, _, _, Graph)
-		->  true
+	    ->  true
 	    )
 	;   rdf_subject(Subject)
 	).
 
 
+:- public lookup/4.			% called from expand hook.
+
 lookup(S,P,O,G) :-
 	(   var(G)
 	->  rdf(S,P,O)
@@ -1109,11 +1140,8 @@ tw_literal(literal(type(Type, Value)), State, Out) :- !,
 	tw_typed_literal(Type, Value, State, Out).
 tw_literal(literal(lang(Lang, Value)), State, Out) :- !,
 	tw_quoted_string(Value, State, Out),
-	downcase_atom(Lang, TurtleLang), 	% Turtle lang = [a-z]+('-'[a-z0-9]+)*
+	downcase_atom(Lang, TurtleLang),	% Turtle lang = [a-z]+('-'[a-z0-9]+)*
 	format(Out, '@~w', [TurtleLang]).
-tw_literal(literal(Value), State, Out) :-
-	atom(Value), !,
-	tw_quoted_string(Value, State, Out).
 tw_literal(literal(Value), State, Out) :-
 	atom(Value), !,
 	tw_quoted_string(Value, State, Out).
