@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #ifndef DTD_H_INCLUDED
@@ -466,7 +466,7 @@ dtd *		new_dtd(const ichar *doctype);
 int		set_dialect_dtd(dtd *dtd, dtd_dialect dialect);
 int		set_option_dtd(dtd *dtd, dtd_option option, int set);
 
-void		putchar_dtd_parser(dtd_parser *p, int chr);
+int		putchar_dtd_parser(dtd_parser *p, int chr);
 int		begin_document_dtd_parser(dtd_parser *p);
 int		end_document_dtd_parser(dtd_parser *p);
 void		reset_document_dtd_parser(dtd_parser *p);
