@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2009, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -59,6 +58,13 @@
 	    xml_quote_cdata/2,		% +In, -Quoted
 	    xml_name/1,			% +In
 	    xml_name/2,			% +In, +Encoding
+
+	    xml_basechar/1,		% +Code
+	    xml_ideographic/1,		% +Code
+	    xml_combining_char/1,	% +Code
+	    xml_digit/1,		% +Code
+	    xml_extender/1,		% +Code
+
 	    iri_xml_namespace/2,	% +IRI, -Namespace
 	    iri_xml_namespace/3,	% +IRI, -Namespace, -LocalName
 	    xml_is_dom/1		% +Term
@@ -66,6 +72,59 @@
 :- use_module(library(lists)).
 :- use_module(library(option)).
 
+:- meta_predicate
+	load_structure(+, -, :).
+
+:- predicate_options(load_structure/3, 3,
+		     [ charpos(integer),
+		       defaults(boolean),
+		       dialect(oneof([sgml,xml,xmlns])),
+		       doctype(atom),
+		       dtd(any),
+		       encoding(oneof(['iso-8859-1', 'utf-8'])),
+		       entity(atom,atom),
+		       file(atom),
+		       line(integer),
+		       number(oneof([token,integer])),
+		       qualify_attributes(boolean),
+		       shorttag(boolean),
+		       space(oneof([sgml,preserve,defailt,remove])),
+		       xmlns(atom),
+		       xmlns(atom,atom),
+		       pass_to(sgml_parse/2, 2)
+		     ]).
+:- predicate_options(load_dtd/3, 3,
+		     [ dialect(oneof([sgml,xml,xmlns])),
+		       pass_to(open/4, 4)
+		     ]).
+:- predicate_options(sgml_parse/2, 2,
+		     [ call(oneof([begin,end,cdata,pi,decl,error,xmlns,urlns]),
+			    callable),
+		       content_length(integer),
+		       document(-any),
+		       max_errors(integer),
+		       parse(oneof([file,element,content,declaration,input])),
+		       source(any),
+		       syntax_errors(oneof([quiet,print,style])),
+		       xml_no_ns(oneof([error,quiet]))
+		     ]).
+:- predicate_options(new_sgml_parser/2, 2,
+		     [ dtd(any)
+		     ]).
+
+
+/** <module> SGML, XML and HTML parser
+
+This library allows you to parse SGML, XML   and HTML data into a Prolog
+data  structure.  The  high-level  interface  defined  in  library(sgml)
+provides  access  at  the  file-level,  while  the  low-level  interface
+(load_structure/3) defined in  the  foreign   module  works  with Prolog
+streams. Please examine the source as a  starting point for dealing with
+data from other sources  than  files,   such  as  SWI-Prolog  resources,
+network-sockets, character strings, etc. The   first example below loads
+an HTML file.
+*/
+
 :- multifile user:file_search_path/2.
 :- dynamic   user:file_search_path/2.
 
@@ -116,6 +175,16 @@ diagnosed to mess with the entity resolution by Fabien Todescato.
 
 dtd_alias(html, 'HTML4').
 
+%%	dtd(+Type, -DTD) is det.
+%
+%	DTD is a DTD object created from  the file dtd(Type). Loaded DTD
+%	objects are cached. Note that  DTD   objects  may  not be shared
+%	between threads. Therefore, dtd/2  maintains   the  pool  of DTD
+%	objects  using  a  thread_local  predicate.    DTD  objects  are
+%	destroyed if a thread terminates.
+%
+%	@error existence_error(source_sink, dtd(Type))
+
 dtd(Type, DTD) :-
 	current_dtd(Type, DTD), !.
 dtd(Type, DTD) :-
@@ -134,23 +203,28 @@ dtd(Type, DTD) :-
 
 %%	load_dtd(+DTD, +DtdFile, +Options)
 %
-%	Load file into a DTD.  Defined options are:
+%	Load DtdFile into a DTD.  Defined options are:
 %
 %		* dialect(+Dialect)
 %		Dialect to use (xml, xmlns, sgml)
 %
 %		* encoding(+Encoding)
 %		Encoding of DTD file
+%
+%	@param	DTD is a fresh DTD object, normally created using
+%		new_dtd/1.
 
 load_dtd(DTD, DtdFile) :-
 	load_dtd(DTD, DtdFile, []).
 load_dtd(DTD, DtdFile, Options) :-
 	split_dtd_options(Options, DTDOptions, FileOptions),
-	open_dtd(DTD, DTDOptions, DtdOut),
-	open(DtdFile, read, DtdIn, FileOptions),
-	copy_stream_data(DtdIn, DtdOut),
-	close(DtdIn),
-	close(DtdOut).
+	setup_call_cleanup(
+	    open_dtd(DTD, DTDOptions, DtdOut),
+	    setup_call_cleanup(
+		open(DtdFile, read, DtdIn, FileOptions),
+		copy_stream_data(DtdIn, DtdOut),
+		close(DtdIn)),
+	    close(DtdOut)).
 
 split_dtd_options([], [], []).
 split_dtd_options([H|T], [H|TD], S) :-
@@ -257,8 +331,17 @@ set_parser_options(Parser, Options, RestOptions) :-
 set_parser_options(_, Options, Options).
 
 
-:- meta_predicate
-	load_structure(+, -, :).
+%%	load_structure(+Source, -ListOfContent, :Options) is det.
+%
+%	Parse   Source   and   return   the   resulting   structure   in
+%	ListOfContent.  Source  is  either   a    term   of  the  format
+%	stream(StreamHandle) or a  file-name.  Options   is  a  list  of
+%	options controlling the conversion process.
+%
+%	A proper XML document contains only   a  single toplevel element
+%	whose name matches the document type.   Nevertheless,  a list is
+%	returned for consistency with  the   representation  of  element
+%	content.
 
 load_structure(stream(In), Term, M:Options) :- !,
 	(   select_option(offset(Offset), Options, Options1)
@@ -289,9 +372,10 @@ load_structure(Stream, Term, Options) :-
 	is_stream(Stream), !,
 	load_structure(stream(Stream), Term, Options).
 load_structure(File, Term, M:Options) :-
-	open(File, read, In, [type(binary)]),
-	load_structure(stream(In), Term, M:[file(File)|Options]),
-	close(In).
+	setup_call_cleanup(
+	    open(File, read, In, [type(binary)]),
+	    load_structure(stream(In), Term, M:[file(File)|Options]),
+	    close(In)).
 
 parse(Parser, M:Options, Document, In) :-
 	set_parser_options(Parser, Options, Options1),
@@ -352,8 +436,8 @@ load_html_file(File, Term) :-
 		 *	      ENCODING		*
 		 *******************************/
 
-%	xml_quote_attribute(+In, -Quoted)
-%	xml_quote_cdata(+In, -Quoted)
+%%	xml_quote_attribute(+In, -Quoted) is det.
+%%	xml_quote_cdata(+In, -Quoted) is det.
 %
 %	Backward  compatibility  for  versions  that  allow  to  specify
 %	encoding. All characters that cannot fit the encoding are mapped
@@ -366,32 +450,64 @@ xml_quote_attribute(In, Quoted) :-
 xml_quote_cdata(In, Quoted) :-
 	xml_quote_cdata(In, Quoted, ascii).
 
+%%	xml_name(+Atom) is semidet.
+%
+%	True if Atom is a valid XML name.
+
 xml_name(In) :-
 	xml_name(In, ascii).
 
 
+		 /*******************************
+		 *    XML CHARACTER CLASSES	*
+		 *******************************/
+
+%%	xml_basechar(+CodeOrChar) is semidet.
+%%	xml_ideographic(+CodeOrChar) is semidet.
+%%	xml_combining_char(+CodeOrChar) is semidet.
+%%	xml_digit(+CodeOrChar) is semidet.
+%%	xml_extender(+CodeOrChar) is semidet.
+%
+%	XML  character  classification   predicates.    Each   of  these
+%	predicates accept both a character   (one-character  atom) and a
+%	code (integer).
+%
+%	@see http://www.w3.org/TR/2006/REC-xml-20060816
+
+
 		 /*******************************
 		 *	   TYPE CHECKING	*
 		 *******************************/
 
-%	xml_is_dome(@Term)
+%%	xml_is_dom(@Term) is semidet.
 %
 %	True  if  term  statisfies   the    structure   as  returned  by
 %	load_structure/3 and friends.
 
 xml_is_dom(0) :- !, fail.		% catch variables
-xml_is_dom([]) :- !.
-xml_is_dom([H|T]) :- !,
-	xml_is_dom(H),
-	xml_is_dom(T).
-xml_is_dom(element(Name, Attributes, Content)) :- !,
+xml_is_dom(List) :-
+	is_list(List), !,
+	xml_is_content_list(List).
+xml_is_dom(Term) :-
+	xml_is_element(Term).
+
+xml_is_content_list([]).
+xml_is_content_list([H|T]) :-
+	xml_is_content(H),
+	xml_is_content_list(T).
+
+xml_is_content(0) :- !, fail.
+xml_is_content(pi(Pi)) :- !,
+	atom(Pi).
+xml_is_content(CDATA) :-
+	atom(CDATA), !.
+xml_is_content(Term) :-
+	xml_is_element(Term).
+
+xml_is_element(element(Name, Attributes, Content)) :-
 	dom_name(Name),
 	dom_attributes(Attributes),
-	xml_is_dom(Content).
-xml_is_dom(pi(Pi)) :- !,
-	atom(Pi).
-xml_is_dom(CDATA) :-
-	atom(CDATA).
+	xml_is_content_list(Content).
 
 dom_name(NS:Local) :-
 	atom(NS),
@@ -432,9 +548,26 @@ prolog:message(sgml(Parser, File, Line, Message)) -->
 	prolog:called_by/2.
 
 prolog:called_by(sgml_parse(_, Options), Called) :-
-	is_list(Options),
-	findall(G+3,
-		(   member(call(_, G), Options),
-		    callable(G)
-		),
-		Called).
+	findall(Meta, meta_call_term(_, Meta, Options), Called).
+
+meta_call_term(T, G+N, Options) :-
+	T = call(Event, G),
+	pmember(T, Options),
+	call_params(Event, Term),
+	functor(Term, _, N).
+
+pmember(X, List) :-			% member for partial lists
+	nonvar(List),
+	List = [H|T],
+	(   X = H
+	;   pmember(X, T)
+	).
+
+call_params(begin, begin(tag,attributes,parser)).
+call_params(end,   end(tag,parser)).
+call_params(cdata, cdata(cdata,parser)).
+call_params(pi,	   pi(cdata,parser)).
+call_params(decl,  decl(cdata,parser)).
+call_params(error, error(severity,message,parser)).
+call_params(xmlns, xmlns(namespace,url,parser)).
+call_params(urlns, urlns(url,url,parser)).
