@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 
@@ -31,13 +31,6 @@
 
 #ifdef _REENTRANT
 #include <pthread.h>
-
-static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
-#define LOCK() pthread_mutex_lock(&mutex)
-#define UNLOCK() pthread_mutex_unlock(&mutex)
-#else
-#define LOCK()
-#define UNLOCK()
 #endif
 
 static atom_t ATOM_server;
@@ -56,11 +49,7 @@ static atom_t ATOM_close_parent;
 
 static functor_t FUNCTOR_ssl1;
 static functor_t FUNCTOR_error2;
-static functor_t FUNCTOR_type_error2;
-static functor_t FUNCTOR_domain_error2;
-static functor_t FUNCTOR_resource_error1;
 static functor_t FUNCTOR_ssl_error1;
-static functor_t FUNCTOR_existence_error1;
 static functor_t FUNCTOR_permission_error3;
 static functor_t FUNCTOR_ip4;
 static functor_t FUNCTOR_version1;
@@ -73,62 +62,17 @@ static functor_t FUNCTOR_public_key5;
 static functor_t FUNCTOR_private_key8;
 static functor_t FUNCTOR_key1;
 static functor_t FUNCTOR_hash1;
+static functor_t FUNCTOR_next_update1;
 static functor_t FUNCTOR_signature1;
 static functor_t FUNCTOR_equals2;
 static functor_t FUNCTOR_crl1;
 static functor_t FUNCTOR_revocations1;
 static functor_t FUNCTOR_revoked2;
-
-
-
-static int
-type_error(term_t val, const char *type)
-{ term_t ex;
-
-  if ( (ex=PL_new_term_ref()) &&
-       PL_unify_term(ex,
-		     PL_FUNCTOR, FUNCTOR_error2,
-		       PL_FUNCTOR, FUNCTOR_type_error2,
-		         PL_CHARS, type,
-		         PL_TERM, val,
-		       PL_VARIABLE) )
-    return PL_raise_exception(ex);
-
-  return FALSE;
-}
-
-
-static int
-domain_error(term_t val, const char *type)
-{ term_t ex;
-
-  if ( (ex=PL_new_term_ref()) &&
-       PL_unify_term(ex,
-		     PL_FUNCTOR, FUNCTOR_error2,
-		       PL_FUNCTOR, FUNCTOR_domain_error2,
-		         PL_TERM, val,
-		         PL_CHARS, type,
-		       PL_VARIABLE) )
-    return PL_raise_exception(ex);
-
-  return FALSE;
-}
-
-
-static int
-resource_error(const char *resource)
-{ term_t ex;
-
-  if ( (ex=PL_new_term_ref()) &&
-       PL_unify_term(ex,
-		     PL_FUNCTOR, FUNCTOR_error2,
-		       PL_FUNCTOR, FUNCTOR_resource_error1,
-		         PL_CHARS, resource,
-		       PL_VARIABLE) )
-    return PL_raise_exception(ex);
-
-  return FALSE;
-}
+static functor_t FUNCTOR_session_key1;
+static functor_t FUNCTOR_master_key1;
+static functor_t FUNCTOR_session_id1;
+static functor_t FUNCTOR_client_random1;
+static functor_t FUNCTOR_server_random1;
 
 static int
 ssl_error(const char *id)
@@ -162,23 +106,6 @@ permission_error(const char *action, const char *type, term_t obj)
   return FALSE;
 }
 
-/* not used now
-static int
-existence_error(term_t resource)
-{ term_t ex;
-
-  if ( (ex=PL_new_term_ref()) &&
-       PL_unify_term(ex,
-		     PL_FUNCTOR, FUNCTOR_error2,
-		       PL_FUNCTOR, FUNCTOR_existence_error1,
-		         PL_TERM, resource,
-		       PL_VARIABLE) )
-    return PL_raise_exception(ex);
-
-  return FALSE;
-}
-*/
-
 static int i2d_X509_CRL_INFO_wrapper(void* i, unsigned char** d)
 {
    return i2d_X509_CRL_INFO(i, d);
@@ -189,22 +116,13 @@ static int i2d_X509_CINF_wrapper(void* i, unsigned char** d)
    return i2d_X509_CINF(i, d);
 }
 
-static int
-get_atom_ex(term_t t, atom_t *a)
-{ if ( !PL_get_atom(t, a) )
-    return type_error(t, "atom");
-
-  return TRUE;
-}
-
-
 static int
 get_char_arg(int a, term_t t, char **s)
 { term_t t2 = PL_new_term_ref();
 
   _PL_get_arg(a, t, t2);
-  if ( !PL_get_atom_chars(t2, s) )
-    return type_error(t2, "atom");
+  if ( !PL_get_chars(t2, s, CVT_ATOM|CVT_EXCEPTION) )
+    return FALSE;
 
   return TRUE;
 }
@@ -215,8 +133,8 @@ get_int_arg(int a, term_t t, int *i)
 { term_t t2 = PL_new_term_ref();
 
   _PL_get_arg(a, t, t2);
-  if ( !PL_get_integer(t2, i) )
-    return type_error(t2, "integer");
+  if ( !PL_get_integer_ex(t2, i) )
+    return FALSE;
 
   return TRUE;
 }
@@ -227,8 +145,8 @@ get_bool_arg(int a, term_t t, int *i)
 { term_t t2 = PL_new_term_ref();
 
   _PL_get_arg(a, t, t2);
-  if ( !PL_get_bool(t2, i) )
-    return type_error(t2, "boolean");
+  if ( !PL_get_bool_ex(t2, i) )
+    return FALSE;
 
   return TRUE;
 }
@@ -240,7 +158,7 @@ get_file_arg(int a, term_t t, char **f)
 
   _PL_get_arg(a, t, t2);
   if ( !PL_get_file_name(t2, f, PL_FILE_EXIST) )
-    return type_error(t2, "file");	/* TBD: check errors */
+    return FALSE;
 
   return TRUE;
 }
@@ -253,7 +171,7 @@ get_predicate_arg(int a, module_t m, term_t t, int arity, predicate_t *pred)
 
   _PL_get_arg(a, t, t2);
   PL_strip_module(t2, &m, t2);
-  if ( !get_atom_ex(t2, &name) )
+  if ( !PL_get_atom_ex(t2, &name) )
     return FALSE;
 
   *pred = PL_pred(PL_new_functor(name, arity), m);
@@ -272,11 +190,11 @@ recover_public_key(term_t public_t, RSA** rsa)
 
    if(!(PL_get_arg(1, public_t, n_t) &&
         PL_get_arg(2, public_t, e_t)))
-      return type_error(public_t, "public_key");
+      return PL_type_error("public_key", public_t);
 
    if (!(PL_get_atom_chars(n_t, &n) &&
          PL_get_atom_chars(e_t, &e)))
-      return type_error(public_t, "public_key");
+      return PL_type_error("public_key", public_t);
    *rsa = RSA_new();
    BN_hex2bn(&((*rsa)->n), n);
    BN_hex2bn(&((*rsa)->e), e);
@@ -305,7 +223,7 @@ recover_private_key(term_t private_t, RSA** rsa)
         PL_get_arg(6, private_t, dmp1_t) &&
         PL_get_arg(7, private_t, dmq1_t) &&
         PL_get_arg(8, private_t, iqmp_t)))
-      return type_error(private_t, "private_key");
+      return PL_type_error("private_key", private_t);
    ssl_deb(1, "Dismantling key");
    if (!(PL_get_atom_chars(n_t, &n) &&
          PL_get_atom_chars(e_t, &e) &&
@@ -315,7 +233,7 @@ recover_private_key(term_t private_t, RSA** rsa)
          PL_get_atom_chars(dmp1_t, &dmp1) &&
          PL_get_atom_chars(dmq1_t, &dmq1) &&
          PL_get_atom_chars(iqmp_t, &iqmp)))
-      return type_error(private_t, "private_key");
+      return PL_type_error("private_key", private_t);
    ssl_deb(1, "Assembling RSA");
    *rsa = RSA_new();
 
@@ -541,16 +459,28 @@ unify_asn1_time(term_t term, ASN1_TIME *time)
   time_tm.tm_wday = 0;
   time_tm.tm_yday = 0;
   time_tm.tm_isdst = 0;  /* No DST adjustment requested, though mktime might do it anyway */
-  result = mktime(&time_tm);
+
+#ifdef HAVE_TIMEGM
+  result = timegm(&time_tm);
   if ((time_t)-1 != result)
-  { if (time_tm.tm_isdst != 0)
-        result -= 3600; /* if mktime has adjusted the time for DST, adjust it back 1 hour*/
-     result += lSecondsFromUTC; /* Add in the UTC offset, or should we return UTC time? */
+  { result += lSecondsFromUTC;
+  } else
+  { ssl_deb(2, "timegm() failed");
+    return FALSE;
+  }
+#else
+  result = mktime(&time_tm);
+  /* mktime assumes that the time_tm contains information for localtime. Convert back to UTC: */
+  if ((time_t)-1 != result)
+  { result += lSecondsFromUTC; /* Add in the UTC offset of the original value */
+    result -= timezone; /* Adjust for localtime */
   } else
   { ssl_deb(2, "mktime() failed");
     return FALSE;
   }
-  return PL_unify_integer(term, result);
+#endif
+
+  return PL_unify_int64(term, result);
 }
 
 static int
@@ -573,7 +503,7 @@ unify_hash(term_t hash, ASN1_OBJECT* algorithm, int (*i2d)(void*, unsigned char*
   digestible_length=i2d(data,NULL);
   digest_buffer = PL_malloc(digestible_length);
   if (digest_buffer == NULL)
-     return resource_error("memory");
+    return PL_resource_error("memory");
 
   /* i2d_X509_CINF will change the value of p. We need to pass in a copy */
   p = digest_buffer;
@@ -631,6 +561,8 @@ unify_crl(term_t term, X509_CRL* crl)
   term_t issuer = PL_new_term_ref();
   term_t revocations = PL_new_term_ref();
   term_t list = PL_copy_term_ref(revocations);
+  term_t next_update = PL_new_term_ref();
+
   int result = 1;
   long n;
   unsigned char* p;
@@ -639,28 +571,31 @@ unify_crl(term_t term, X509_CRL* crl)
 
   mem = BIO_new(BIO_s_mem());
   if (mem == NULL)
-     return resource_error("memory");
+    return PL_resource_error("memory");
 
   i2a_ASN1_INTEGER(mem, crl->signature);
   if (!(unify_name(issuer, X509_CRL_get_issuer(crl)) &&
         unify_hash(hash, crl->sig_alg->algorithm, i2d_X509_CRL_INFO_wrapper, crl->crl) &&
+        unify_asn1_time(next_update, X509_CRL_get_nextUpdate(crl)) &&
         PL_unify_term(term,
-                      PL_LIST, 4,
+                      PL_LIST, 5,
                       PL_FUNCTOR, FUNCTOR_issuername1,
                       PL_TERM, issuer,
                       PL_FUNCTOR, FUNCTOR_signature1,
                       PL_NCHARS, (size_t)crl->signature->length, crl->signature->data,
                       PL_FUNCTOR, FUNCTOR_hash1,
                       PL_TERM, hash,
+                      PL_FUNCTOR, FUNCTOR_next_update1,
+                      PL_TERM, next_update,
                       PL_FUNCTOR, FUNCTOR_revocations1,
                       PL_TERM, revocations)))
   {
      return FALSE;
   }
-
   for (i = 0; i < sk_X509_REVOKED_num(info->revoked); i++)
   {
      X509_REVOKED* revoked = sk_X509_REVOKED_value(info->revoked, i);
+     (void)BIO_reset(mem);
      i2a_ASN1_INTEGER(mem, revoked->serialNumber);
      result &= (((n = BIO_get_mem_data(mem, &p)) > 0) &&
                 PL_unify_list(list, item, list) &&
@@ -674,7 +609,7 @@ unify_crl(term_t term, X509_CRL* crl)
      {
         BIO_free(mem);
         // The only reason I can imagine this would fail is out of memory
-        return resource_error("memory");
+        return PL_resource_error("memory");
      }
   }
   BIO_free(mem);
@@ -849,7 +784,7 @@ pl_load_public_key(term_t source, term_t key_t)
   int c;
 
   if ( !PL_get_stream_handle(source, &stream) )
-     return type_error(source, "stream");
+    return FALSE;
   bio = BIO_new(&bio_read_functions);
   BIO_set_ex_data(bio, 0, stream);
 
@@ -895,10 +830,11 @@ pl_load_private_key(term_t source, term_t password, term_t key_t)
   char* password_chars;
   int c;
 
-  if (!PL_get_atom_chars(password, &password_chars))
-    return type_error(password, "atom");
+  if ( !PL_get_chars(password, &password_chars,
+		     CVT_ATOM|CVT_STRING|CVT_LIST|CVT_EXCEPTION) )
+    return FALSE;
   if ( !PL_get_stream_handle(source, &stream) )
-     return type_error(source, "stream");
+    return FALSE;
   bio = BIO_new(&bio_read_functions);
   BIO_set_ex_data(bio, 0, stream);
 
@@ -934,7 +870,7 @@ pl_load_crl(term_t source, term_t list)
   int c;
 
   if ( !PL_get_stream_handle(source, &stream) )
-     return type_error(source, "stream");
+    return FALSE;
 
   bio = BIO_new(&bio_read_functions);
   BIO_set_ex_data(bio, 0, stream);
@@ -965,7 +901,7 @@ pl_load_certificate(term_t source, term_t cert)
   int c = 0;
 
   if ( !PL_get_stream_handle(source, &stream) )
-     return type_error(source, "stream");
+    return FALSE;
   bio = BIO_new(&bio_read_functions);
   BIO_set_ex_data(bio, 0, stream);
   /* Determine format */
@@ -1003,14 +939,12 @@ get_conf(term_t config, PL_SSL **conf)
   void *ptr;
   PL_SSL *ssl;
 
-  if ( !PL_is_functor(config, FUNCTOR_ssl1) )
-    return type_error(config, "ssl_config");
-  _PL_get_arg(1, config, a);
-  if ( !PL_get_pointer(a, &ptr) )
-    return type_error(config, "ssl_config");
-  ssl = ptr;
-  if ( ssl->magic != SSL_CONFIG_MAGIC )
-    return type_error(config, "ssl_config");
+  if ( !PL_is_functor(config, FUNCTOR_ssl1) ||
+       !PL_get_arg(1, config, a) ||
+       !PL_get_pointer(a, &ptr)	||
+       !(ssl = ptr) ||
+       ssl->magic != SSL_CONFIG_MAGIC )
+    return PL_type_error("ssl_config", config);
 
   *conf = ssl;
 
@@ -1036,7 +970,7 @@ pl_pem_passwd_hook(PL_SSL *config, char *buf, int size)
    */
 
   unify_conf(av+0, config);
-  if ( PL_call_predicate(NULL, PL_Q_NORMAL, pred, av) )
+  if ( PL_call_predicate(NULL, PL_Q_PASS_EXCEPTION, pred, av) )
   { if ( PL_get_nchars(av+1, &len, &passwd, CVT_ALL) )
     { if ( len >= (unsigned int)size )
 	PL_warning("pem_passwd too long");
@@ -1077,7 +1011,7 @@ pl_cert_verify_hook(PL_SSL *config,
   val = ( unify_certificate(av+1, cert) &&
           unify_certificates(av+2, av+3, stack) &&
 	  PL_unify_atom_chars(av+4, error) &&
-	  PL_call_predicate(NULL, PL_Q_NORMAL, pred, av) );
+	  PL_call_predicate(NULL, PL_Q_PASS_EXCEPTION, pred, av) );
 
   /* free any items still on stack, since X509_STORE_CTX_get1_chain returns a copy */
   sk_X509_pop_free(stack, X509_free);
@@ -1087,35 +1021,6 @@ pl_cert_verify_hook(PL_SSL *config,
 }
 
 
-		 /*******************************
-		 *	       INIT		*
-		 *******************************/
-
-static BOOL initialised  = FALSE;
-
-static int
-threads_init()
-{ LOCK();
-  if ( initialised )
-  { UNLOCK();
-    return TRUE;
-  }
-  initialised = TRUE;
-
-#ifdef _REENTRANT
-  if ( !ssl_thread_setup() )
-  { term_t o = PL_new_term_ref();
-
-    PL_put_atom_chars(o, "ssl");
-    return permission_error("setup_threads", "library", o);
-  }
-#endif
-
-  UNLOCK();
-  return TRUE;
-}
-
-
 static foreign_t
 pl_ssl_context(term_t role, term_t config, term_t options)
 { atom_t a;
@@ -1128,27 +1033,23 @@ pl_ssl_context(term_t role, term_t config, term_t options)
   PL_strip_module(options, &module, options);
   tail = PL_copy_term_ref(options);
 
-  if ( !get_atom_ex(role, &a) )
+  if ( !PL_get_atom_ex(role, &a) )
     return FALSE;
   if ( a == ATOM_server )
     r = PL_SSL_SERVER;
   else if ( a == ATOM_client )
     r = PL_SSL_CLIENT;
   else
-    return domain_error(a, "ssl_role");
-
- if ( !threads_init() )
-    return FALSE;
-
+    return PL_domain_error("ssl_role", a);
 
   if ( !(conf = ssl_init(r)) )
-    return resource_error("memory");
+    return PL_resource_error("memory");
   while( PL_get_list(tail, head, tail) )
   { atom_t name;
     int arity;
 
     if ( !PL_get_name_arity(head, &name, &arity) )
-      return type_error(head, "ssl_option");
+      return PL_type_error("ssl_option", head);
 
     if ( name == ATOM_password && arity == 1 )
     { char *s;
@@ -1221,18 +1122,21 @@ pl_ssl_context(term_t role, term_t config, term_t options)
 
       ssl_set_cb_cert_verify(conf, pl_cert_verify_hook, (void *)hook);
     } else if ( name == ATOM_close_parent && arity == 1 )
-    { char* s;
+    { int val;
 
-      if ( !get_char_arg(1, head, &s) )
+      if ( !get_bool_arg(1, head, &val) )
 	return FALSE;
 
-      ssl_set_close_parent(conf, strcmp(s, "true") == 0);
+      ssl_set_close_parent(conf, val);
     } else
       continue;
   }
 
-  if ( !PL_get_nil(tail) )
-    return type_error(tail, "list");
+  if ( !PL_get_nil_ex(tail) )
+    return FALSE;
+
+  if ( ssl_config(conf) < 0 )
+    return FALSE;
 
   return unify_conf(config, conf);
 }
@@ -1252,20 +1156,40 @@ pl_ssl_close(PL_SSL_INSTANCE *instance)
 static int
 pl_ssl_control(PL_SSL_INSTANCE *instance, int action, void *data)
 { switch(action)
-  { case SIO_GETFILENO:
-    { if (instance->sread != NULL)
-    {  SOCKET fd = Sfileno(instance->sread);
-       SOCKET *fdp = data;
-       *fdp = fd;
-       return 0;
-    } else if (instance->swrite != NULL)
-    {  SOCKET fd = Sfileno(instance->swrite);
-       SOCKET *fdp = data;
-       *fdp = fd;
-       return 0;
-    }
-    }
-    return -1;
+  {
+#ifdef __WINDOWS__
+    case SIO_GETFILENO:
+      return -1;
+    case SIO_GETWINSOCK:
+      { if (instance->sread != NULL)
+        { (*instance->sread->functions->control)(instance->sread->handle,
+                                                 SIO_GETWINSOCK,
+                                                 data);
+          return 0;
+        } else if (instance->swrite != NULL)
+        { (*instance->swrite->functions->control)(instance->swrite->handle,
+                                                  SIO_GETWINSOCK,
+                                                  data);
+          return 0;
+        }
+      }
+      return -1;
+#else
+    case SIO_GETFILENO:
+      { if (instance->sread != NULL)
+        {  SOCKET fd = Sfileno(instance->sread);
+           SOCKET *fdp = data;
+           *fdp = fd;
+           return 0;
+        } else if (instance->swrite != NULL)
+        {  SOCKET fd = Sfileno(instance->swrite);
+           SOCKET *fdp = data;
+           *fdp = fd;
+           return 0;
+        }
+      }
+      return -1;
+#endif
     case SIO_SETENCODING:
     case SIO_FLUSHOUTPUT:
       return 0;
@@ -1322,19 +1246,19 @@ pl_ssl_negotiate(term_t config, term_t org_in, term_t org_out, term_t in, term_t
   if ( !get_conf(config, &conf) )
     return FALSE;
   if ( !PL_get_stream_handle(org_in, &sorg_in) )
-     return FALSE;
+    return FALSE;
   if ( !PL_get_stream_handle(org_out, &sorg_out) )
-     return FALSE;
+    return FALSE;
   if ( !(instance = ssl_ssl_bio(conf, sorg_in, sorg_out)) )
-  {  PL_release_stream(sorg_in);
-     PL_release_stream(sorg_out);
-     return FALSE;			/* TBD: error */
+  { PL_release_stream(sorg_in);
+    PL_release_stream(sorg_out);
+    return ssl_error("ssl_verify");
   }
 
   if ( !(i=Snew(instance, SIO_INPUT|SIO_RECORDPOS|SIO_FBUF, &ssl_funcs)) )
-  {  PL_release_stream(sorg_in);
-     PL_release_stream(sorg_out);
-    return FALSE;
+  { PL_release_stream(sorg_in);
+    PL_release_stream(sorg_out);
+    return PL_resource_error("memory");
   }
   instance->close_needed++;
   if ( !PL_unify_stream(in, i) )
@@ -1346,8 +1270,8 @@ pl_ssl_negotiate(term_t config, term_t org_in, term_t org_out, term_t in, term_t
   Sset_filter(sorg_in, i);
   PL_release_stream(sorg_in);
   if ( !(o=Snew(instance, SIO_OUTPUT|SIO_RECORDPOS|SIO_FBUF, &ssl_funcs)) )
-  {  PL_release_stream(sorg_out);
-    return FALSE;
+  { PL_release_stream(sorg_out);
+    return PL_resource_error("memory");
   }
   instance->close_needed++;
   if ( !PL_unify_stream(out, o) )
@@ -1370,10 +1294,11 @@ foreign_t pl_rsa_private_decrypt(term_t private_t, term_t cipher_t, term_t plain
   RSA* key;
   int retval;
 
-  if(!PL_get_atom_nchars(cipher_t, &cipher_length, (char**)&cipher))
-     return type_error(cipher_t, "atom");
-  if (!recover_private_key(private_t, &key))
-     return FALSE;
+  if( !PL_get_nchars(cipher_t, &cipher_length, (char**)&cipher,
+		     CVT_ATOM|CVT_STRING|CVT_LIST|CVT_EXCEPTION) )
+    return FALSE;
+  if ( !recover_private_key(private_t, &key) )
+    return FALSE;
 
   outsize = RSA_size(key);
   ssl_deb(1, "Output size is going to be %d", outsize);
@@ -1404,10 +1329,11 @@ foreign_t pl_rsa_public_decrypt(term_t public_t, term_t cipher_t, term_t plain_t
   RSA* key;
   int retval;
 
-  if(!PL_get_atom_nchars(cipher_t, &cipher_length, (char**)&cipher))
-     return type_error(cipher_t, "atom");
+  if ( !PL_get_nchars(cipher_t, &cipher_length, (char**)&cipher,
+		      CVT_ATOM|CVT_STRING|CVT_LIST|CVT_EXCEPTION) )
+    return FALSE;
   if (!recover_public_key(public_t, &key))
-     return FALSE;
+    return FALSE;
 
   outsize = RSA_size(key);
   ssl_deb(1, "Output size is going to be %d", outsize);
@@ -1440,8 +1366,9 @@ foreign_t pl_rsa_public_encrypt(term_t public_t, term_t plain_t, term_t cipher_t
 
   ssl_deb(1, "Generating terms");
   ssl_deb(1, "Collecting plaintext");
-  if(!PL_get_atom_nchars(plain_t, &plain_length, (char**)&plain))
-     return type_error(plain_t, "atom");
+  if ( !PL_get_nchars(plain_t, &plain_length, (char**)&plain,
+		      CVT_ATOM|CVT_STRING|CVT_LIST|CVT_EXCEPTION) )
+     return FALSE;
   if (!recover_public_key(public_t, &key))
      return FALSE;
 
@@ -1474,16 +1401,17 @@ foreign_t pl_rsa_private_encrypt(term_t private_t, term_t plain_t, term_t cipher
   int outsize;
   RSA* key;
   int retval;
-  if(!PL_get_atom_nchars(plain_t, &plain_length, (char**)&plain))
-     return type_error(plain_t, "atom");
+  if ( !PL_get_nchars(plain_t, &plain_length, (char**)&plain,
+		      CVT_ATOM|CVT_STRING|CVT_LIST|CVT_EXCEPTION))
+    return FALSE;
   if (!recover_private_key(private_t, &key))
-     return FALSE;
+    return FALSE;
 
   outsize = RSA_size(key);
   ssl_deb(1, "Output size is going to be %d", outsize);
   cipher = PL_malloc(outsize);
   ssl_deb(1, "Allocated %d bytes for ciphertext", outsize);
-  if ((outsize = RSA_public_encrypt((int)plain_length, plain, cipher, key, RSA_PKCS1_PADDING)) <= 0)
+  if ((outsize = RSA_private_encrypt((int)plain_length, plain, cipher, key, RSA_PKCS1_PADDING)) <= 0)
   { ssl_deb(1, "Failure to encrypt!");
     PL_free(plain);
     RSA_free(key);
@@ -1505,8 +1433,8 @@ static foreign_t
 pl_ssl_debug(term_t level)
 { int l;
 
-  if ( !PL_get_integer(level, &l) )
-    return type_error(level, "integer");
+  if ( !PL_get_integer_ex(level, &l) )
+    return FALSE;
 
   ssl_set_debug(l);
 
@@ -1514,6 +1442,80 @@ pl_ssl_debug(term_t level)
 }
 
 
+static foreign_t
+pl_ssl_session(term_t stream_t, term_t session_t)
+{ IOSTREAM* stream;
+  PL_SSL_INSTANCE* instance;
+  SSL* ssl;
+  SSL_SESSION* session;
+  term_t list_t = PL_copy_term_ref(session_t);
+  term_t node_t = PL_new_term_ref();
+
+  if ( !PL_get_stream_handle(stream_t, &stream) )
+     return FALSE;
+  if ( stream->functions != &ssl_funcs )
+  { PL_release_stream(stream);
+    return PL_domain_error("ssl_stream", stream_t);
+  }
+
+  instance = stream->handle;
+  PL_release_stream(stream);
+
+  if ( !(ssl = instance->ssl) ||
+       !(session = SSL_get_session(ssl)) )
+    return PL_existence_error("ssl_session", stream_t);
+
+  if ( !PL_unify_list_ex(list_t, node_t, list_t) )
+    return FALSE;
+  if ( !PL_unify_term(node_t,
+		      PL_FUNCTOR, FUNCTOR_version1,
+		        PL_INTEGER, (int)session->ssl_version))
+     return FALSE;
+
+  if ( !PL_unify_list_ex(list_t, node_t, list_t) )
+    return FALSE;
+  if ( !PL_unify_term(node_t,
+		      PL_FUNCTOR, FUNCTOR_session_key1,
+		        PL_NCHARS, session->key_arg_length, session->key_arg))
+    return FALSE;
+
+  if ( !PL_unify_list_ex(list_t, node_t, list_t))
+    return FALSE;
+  if ( !PL_unify_term(node_t,
+		      PL_FUNCTOR, FUNCTOR_master_key1,
+		        PL_NCHARS, session->master_key_length,
+		                   session->master_key))
+     return FALSE;
+
+  if ( !PL_unify_list_ex(list_t, node_t, list_t) )
+    return FALSE;
+  if ( !PL_unify_term(node_t,
+                     PL_FUNCTOR, FUNCTOR_session_id1,
+		       PL_NCHARS, session->session_id_length,
+		                  session->session_id) )
+    return FALSE;
+  if ( ssl->s3 != NULL ) /* If the connection is SSLv2?! */
+  { if ( !PL_unify_list_ex(list_t, node_t, list_t) )
+      return FALSE;
+    if ( !PL_unify_term(node_t,
+			PL_FUNCTOR, FUNCTOR_client_random1,
+			  PL_NCHARS, SSL3_RANDOM_SIZE, ssl->s3->client_random) )
+      return FALSE;
+
+    if ( !PL_unify_list_ex(list_t, node_t, list_t) )
+      return FALSE;
+    if ( !PL_unify_term(node_t,
+			PL_FUNCTOR, FUNCTOR_server_random1,
+			  PL_NCHARS, SSL3_RANDOM_SIZE, ssl->s3->server_random) )
+      return FALSE;
+  }
+
+  return PL_unify_nil_ex(list_t);
+}
+
+
+
+
 
 		 /*******************************
 		 *	     INSTALL		*
@@ -1538,11 +1540,7 @@ install_ssl4pl()
 
   FUNCTOR_ssl1            = PL_new_functor(PL_new_atom("$ssl"), 1);
   FUNCTOR_error2          = PL_new_functor(PL_new_atom("error"), 2);
-  FUNCTOR_domain_error2   = PL_new_functor(PL_new_atom("domain_error"), 2);
-  FUNCTOR_type_error2     = PL_new_functor(PL_new_atom("type_error"), 2);
-  FUNCTOR_resource_error1 = PL_new_functor(PL_new_atom("resource_error"), 1);
   FUNCTOR_ssl_error1      = PL_new_functor(PL_new_atom("ssl_error"), 1);
-  FUNCTOR_existence_error1 =PL_new_functor(PL_new_atom("existence_error"), 1);
   FUNCTOR_permission_error3=PL_new_functor(PL_new_atom("permission_error"), 3);
   FUNCTOR_ip4		  = PL_new_functor(PL_new_atom("ip"), 4);
   FUNCTOR_version1	  = PL_new_functor(PL_new_atom("version"), 1);
@@ -1555,11 +1553,17 @@ install_ssl4pl()
   FUNCTOR_public_key5     = PL_new_functor(PL_new_atom("public_key"), 5);
   FUNCTOR_private_key8    = PL_new_functor(PL_new_atom("private_key"), 8);
   FUNCTOR_hash1	          = PL_new_functor(PL_new_atom("hash"), 1);
+  FUNCTOR_next_update1    = PL_new_functor(PL_new_atom("next_update"), 1);
   FUNCTOR_signature1      = PL_new_functor(PL_new_atom("signature"), 1);
   FUNCTOR_equals2         = PL_new_functor(PL_new_atom("="), 2);
   FUNCTOR_crl1            = PL_new_functor(PL_new_atom("crl"), 1);
   FUNCTOR_revoked2        = PL_new_functor(PL_new_atom("revoked"), 2);
   FUNCTOR_revocations1    = PL_new_functor(PL_new_atom("revocations"), 1);
+  FUNCTOR_session_key1    = PL_new_functor(PL_new_atom("session_key"), 1);
+  FUNCTOR_master_key1     = PL_new_functor(PL_new_atom("master_key"), 1);
+  FUNCTOR_session_id1     = PL_new_functor(PL_new_atom("session_id"), 1);
+  FUNCTOR_client_random1  = PL_new_functor(PL_new_atom("client_random"), 1);
+  FUNCTOR_server_random1  = PL_new_functor(PL_new_atom("server_random"), 1);
 
   PL_register_foreign("_ssl_context",	3, pl_ssl_context,    0);
   PL_register_foreign("ssl_exit",	1, pl_ssl_exit,	      0);
@@ -1567,6 +1571,7 @@ install_ssl4pl()
   PL_register_foreign("ssl_get_socket",	2, pl_ssl_get_socket, 0);
   PL_register_foreign("ssl_negotiate",	5, pl_ssl_negotiate,  0);
   PL_register_foreign("ssl_debug",	1, pl_ssl_debug,      0);
+  PL_register_foreign("ssl_session",    2, pl_ssl_session,    0);
   PL_register_foreign("load_crl",       2, pl_load_crl,      0);
   PL_register_foreign("load_certificate",2,pl_load_certificate,      0);
   PL_register_foreign("load_private_key",3,pl_load_private_key,      0);
