@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <string.h>
@@ -123,7 +123,7 @@ static SSL_PL_STATUS
 ssl_inspect_status(SSL *ssl, int ssl_ret)
 {   int code;
 
-    if (ssl_ret >= 0) {
+    if (ssl_ret > 0) {
         return SSL_PL_OK;
     }
 
@@ -249,7 +249,7 @@ ssl_new(void)
         new->pl_ssl_role                = PL_SSL_NONE;
 
 	new->sock			= -1;
-        new->closeparent	       	= 0;
+        new->closeparent		= 0;
 
         new->pl_ssl_peer_cert           = NULL;
         new->pl_ssl_ctx                 = NULL;
@@ -269,7 +269,7 @@ ssl_new(void)
         new->pl_ssl_cb_cert_verify_data = NULL;
         new->pl_ssl_cb_pem_passwd       = NULL;
         new->pl_ssl_cb_pem_passwd_data  = NULL;
-	new->magic 		        = SSL_CONFIG_MAGIC;
+	new->magic		        = SSL_CONFIG_MAGIC;
     }
     ssl_deb(1, "Allocated config structure\n");
 
@@ -537,14 +537,15 @@ ssl_cb_cert_verify(int preverify_ok, X509_STORE_CTX *ctx)
             char  issuer [256];
             int   depth;
 
-            depth = X509_STORE_CTX_get_error_depth (ctx);
-            X509_NAME_oneline( X509_get_subject_name(cert)
-                             , subject, sizeof(subject));
-            X509_NAME_oneline( X509_get_issuer_name (cert)
-                             , issuer, sizeof(issuer));
-            ssl_deb(1,  "error:%s\n", error);
+            depth = X509_STORE_CTX_get_error_depth(ctx);
+            X509_NAME_oneline(X509_get_subject_name(cert),
+			      subject, sizeof(subject));
+            X509_NAME_oneline(X509_get_issuer_name (cert),
+			      issuer, sizeof(issuer));
+            ssl_deb(1,   "depth:%d\n", depth);
+            ssl_deb(1,   "error:%s\n", error);
             ssl_deb(1, "subject:%s\n", subject);
-            ssl_deb(1, "issuer:%s\n", issuer);
+            ssl_deb(1,  "issuer:%s\n", issuer);
         }
         ssl_deb(1, "Certificate preverified not ok\n");
     } else {
@@ -668,7 +669,7 @@ ssl_exit(PL_SSL *config)
 
       if (config->pl_ssl_ctx)
       { ssl_deb(1, "Calling SSL_CTX_free()\n");
-	SSL_CTX_free(config->pl_ssl_ctx); 	/* doesn't call free hook? */
+	SSL_CTX_free(config->pl_ssl_ctx);	/* doesn't call free hook? */
       }
       else
       { ssl_deb(1, "config without CTX encountered\n");
@@ -705,28 +706,9 @@ ssl_init(PL_SSL_ROLE role)
  * Create SSL context.
  */
 {
-    PL_SSL      * config     = NULL;
-    SSL_METHOD  * ssl_method = NULL;
-    SSL_CTX     * ssl_ctx    = NULL;
-
-#if 0
-    switch (role) {
-        case PL_SSL_SERVER:
-            ssl_method = SSLv23_server_method();
-            break;
-        case PL_SSL_NONE:
-        case PL_SSL_CLIENT:
-            /*
-             * Either v2 or v3, we choose v3 to make the shutdown more
-             * controlled and clean (see util.c)
-             */
-            ssl_method = SSLv3_client_method();
-            break;
-    }
-#else
-    ssl_method = SSLv23_method();
-#endif
-    ssl_ctx = SSL_CTX_new(ssl_method);
+    PL_SSL           * config    = NULL;
+    const SSL_METHOD *ssl_method = SSLv23_method();
+    SSL_CTX          *ssl_ctx    = SSL_CTX_new(ssl_method);
 
     if (!ssl_ctx) {
         ERR_print_errors_pl();
@@ -794,7 +776,7 @@ ssl_debug(PL_SSL *config)
     return 0;
 }
 
-static int
+int
 ssl_config(PL_SSL *config)
 /*
  * Initialize various SSL layer parameters using the supplied
@@ -903,34 +885,51 @@ ssl_lib_init(void)
     }
 #endif
 
+#ifdef _REENTRANT
+    ssl_thread_setup();
+#endif
+
     return 0;
 }
 
-int
-ssl_lib_exit(void)
-/*
- * One-time library exit calls
- */
-{
-#ifdef __SWI_PROLOG__
-    nbio_cleanup();
-#endif
-    return 0;
-}
 
 /*
  * BIO routines for SSL over streams
  */
 
 /*
- * Read function
+ * Read function.  To allow for setting a timeout on the SSL stream, we
+ * use the timeout of this stream if we do not have a timeout ourselves.
+ *
+ * Note that if the underlying stream received a timeout, we lift this
+ * error to the ssl stream and clear the error on the underlying stream.
+ * This way, the normal timeout-reset in pl-stream.c correctly resets
+ * a possible timeout.  See also test_ssl.c.  Patch and analysis by
+ * Keri Harris.
  */
 
 int bio_read(BIO* bio, char* buf, int len)
 {
-   IOSTREAM* stream;
-   stream  = BIO_get_ex_data(bio, 0);
-   return (int)Sfread(buf, sizeof(char), len, stream);
+   IOSTREAM *stream = BIO_get_ex_data(bio, 0);
+   IOSTREAM *ssl_stream = stream->upstream;
+   int rc;
+
+   if ( ssl_stream &&
+	stream->timeout < 0 &&
+	ssl_stream->timeout > 0 )
+   { stream->timeout = ssl_stream->timeout;
+     rc = (int)Sfread(buf, sizeof(char), len, stream);
+     stream->timeout = -1;
+   } else
+   { rc = (int)Sfread(buf, sizeof(char), len, stream);
+   }
+
+   if ( ssl_stream && (stream->flags & SIO_TIMEOUT) )
+   { ssl_stream->flags |= (SIO_FERR|SIO_TIMEOUT);
+     Sclearerr(stream);
+   }
+
+   return rc;
 }
 
 /*
@@ -961,12 +960,28 @@ int bio_gets(BIO* bio, char* buf, int len)
 
 int bio_write(BIO* bio, const char* buf, int len)
 {
-   IOSTREAM* stream;
+   IOSTREAM* stream = BIO_get_ex_data(bio, 0);
+   IOSTREAM* ssl_stream = stream->upstream;
    int r;
-   stream  = BIO_get_ex_data(bio, 0);
-   r = (int)Sfwrite(buf, sizeof(char), len, stream);
-   /* OpenSSL expects there to be no buffering when it writes. Flush here */
-   Sflush(stream);
+
+   if ( ssl_stream &&
+	stream->timeout < 0 &&
+	ssl_stream->timeout > 0 )
+   { stream->timeout = ssl_stream->timeout;
+     r = (int)Sfwrite(buf, sizeof(char), len, stream);
+     /* OpenSSL expects there to be no buffering when it writes. Flush here */
+     Sflush(stream);
+     stream->timeout = -1;
+   } else
+   { r = (int)Sfwrite(buf, sizeof(char), len, stream);
+     Sflush(stream);
+   }
+
+   if ( ssl_stream && (stream->flags & SIO_TIMEOUT) )
+   { ssl_stream->flags |= (SIO_FERR|SIO_TIMEOUT);
+     Sclearerr(stream);
+   }
+
    return r;
 }
 
@@ -1057,12 +1072,6 @@ ssl_ssl_bio(PL_SSL *config, IOSTREAM* sread, IOSTREAM* swrite)
         ssl_deb(1, "ssl instance malloc failed\n");
         return NULL;
     }
-    /*
-     * Configure SSL behaviour with install configuration parameters
-     */
-    if (ssl_config(config) < 0) {
-        return NULL;
-    }
 
     /*
      * Create BIOs
@@ -1130,7 +1139,7 @@ ssl_ssl_bio(PL_SSL *config, IOSTREAM* sread, IOSTREAM* swrite)
     return NULL;
 }
 
-int
+ssize_t
 ssl_read(PL_SSL_INSTANCE *instance, char *buf, int size)
 /*
  * Perform read on SSL session
@@ -1142,6 +1151,8 @@ ssl_read(PL_SSL_INSTANCE *instance, char *buf, int size)
 
     do {
         int rbytes = SSL_read(ssl, buf, size);
+        if (rbytes == 0) /* EOF - error, but we handle in prolog */
+          return 0;
         switch(ssl_inspect_status(ssl, rbytes)) {
             case SSL_PL_OK:
                 /* success */
@@ -1156,7 +1167,7 @@ ssl_read(PL_SSL_INSTANCE *instance, char *buf, int size)
     } while (1);
 }
 
-int
+ssize_t
 ssl_write(PL_SSL_INSTANCE *instance, const char *buf, int size)
 /*
  * Perform write on SSL session
@@ -1168,6 +1179,8 @@ ssl_write(PL_SSL_INSTANCE *instance, const char *buf, int size)
 
     do {
         int wbytes = SSL_write(ssl, buf, size);
+        if (wbytes == 0) /* EOF - error, but we handle in prolog */
+          return 0;
         switch(ssl_inspect_status(ssl, wbytes)) {
             case SSL_PL_OK:
                 /* success */
@@ -1198,6 +1211,8 @@ code is based on mttest.c distributed with the OpenSSL library.
 
 static pthread_mutex_t *lock_cs;
 static long *lock_count;
+static void (*old_locking_callback)(int, int, const char*, int) = NULL;
+static unsigned long (*old_id_callback)(void) = NULL;
 
 static void
 pthreads_locking_callback(int mode, int type, const char *file, int line)
@@ -1242,6 +1257,8 @@ ssl_thread_setup(void)
     pthread_mutex_init(&(lock_cs[i]), NULL);
   }
 
+  old_id_callback = CRYPTO_get_id_callback();
+  old_locking_callback = CRYPTO_get_locking_callback();
 #ifndef __WINDOWS__
   CRYPTO_set_id_callback(pthreads_thread_id);
 #endif
@@ -1260,6 +1277,30 @@ ssl_thread_init(void)
 #endif /*_REENTRANT*/
 
 
+int
+ssl_lib_exit(void)
+/*
+ * One-time library exit calls
+ */
+{
+#ifdef __SWI_PROLOG__
+    nbio_cleanup();
+#endif
+
+/*
+ * If the module is being unloaded, we should remove callbacks pointing to
+ * our address space
+ */
+#ifdef _REENTRANT
+#ifndef __WINDOWS__
+  CRYPTO_set_id_callback(old_id_callback);
+#endif
+  CRYPTO_set_locking_callback(old_locking_callback);
+#endif
+    return 0;
+}
+
+
 /***********************************************************************
  * Warning, error and debug reporting
  ***********************************************************************/
