@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of XPCE --- The SWI-Prolog GUI toolkit
+/*  Part of XPCE --- The SWI-Prolog GUI toolkit
 
     Author:        Jan Wielemaker and Anjo Anjewierden
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi.psy.uva.nl/projects/xpce/
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -35,7 +34,8 @@
 :- use_module(library(pce_report)).
 :- use_module(library(toolbar)).
 :- use_module(library(helpidx)).
-:- use_module(library('trace/util')).
+:- use_module(library(trace/util)).
+:- use_module(library(prolog_source)).
 :- use_module(browse_xref).
 :- use_module(library(persistent_frame)).
 :- require([ '$qlf_info'/5
@@ -129,8 +129,8 @@ initialise(FB, Root:directory) :->
 	asserta(prolog_overview_window(FB)).
 
 source_pattern(Pat) :-
-	findall(E, (prolog_file_type(E, prolog),
-		    \+ prolog_file_type(E, qlf)), Exts),
+	findall(E, (user:prolog_file_type(E, prolog),
+		    \+ user:prolog_file_type(E, qlf)), Exts),
 	(   Exts = [Ext]
 	->  format(atom(Pat), '.*\\.~w$', [Ext])
 	;   atomic_list_concat(Exts, '|', P1),
@@ -208,8 +208,16 @@ goto(FB, File:'file|node', Line:int) :->
 popup(FB, Id:any, Popup:popup) :<-
 	"Return popup from current node"::
 	get(FB, node, Id, Node),
-	send(Node, has_get_method, popup),
-	get(Node, popup, Popup).
+	(   send(Node, has_get_method, popup),
+	    get(Node, popup, Popup)
+	->  true
+	;   send(Node, instance_of, toc_directory)
+	->  new(Popup, popup(options)),
+	    send(Popup, append,
+		 menu_item(expand_all,
+			   message(Node, expand_all)))
+	).
+
 
 :- pce_group(edit).
 
@@ -260,10 +268,7 @@ event(FB, Ev:event) :->
 initialise(TF, File:file) :->
 	get(File, name, FileName),
 	canonical_source_file(FileName, Path),
-	(   source_file(Path)
-	->  Img = 'plloadedfile.xpm'
-	;   Img = 'plfile.xpm'
-	),
+	file_image(Path, Img),
 	file_base_name(FileName, Base),
 	send_super(TF, initialise, Base, Path, Img),
 	send(TF, name, Base).
@@ -275,6 +280,20 @@ loaded(TF) :->
 	get(TF, identifier, Path),
 	source_file(Path).
 
+included(TF) :->
+	get(TF, identifier, Path),
+	included_file(Path).
+
+file_image(Path, 'plloadedfile.xpm') :-
+	source_file(Path), !.
+file_image(Path, 'plincludedfile.xpm') :-
+	included_file(Path), !.
+file_image(_, 'plfile.xpm').
+
+included_file(Path) :-
+	source_file_property(Path, included_in(_,_)).
+
+
 module(TF, Module:name) :<-
 	"Return module defined in this file"::
 	get(TF, identifier, Path),
@@ -284,13 +303,11 @@ module(TF, Module:name) :<-
 	).
 
 module_of_path(Path, Module) :-
-	open(Path, read, Fd),
-	(   peek_char(Fd, '#')		% skip !# line
-	->  skip(Fd, 10)
-	;   true
-	),
-	read(Fd, Term),
-	close(Fd),
+	catch(setup_call_cleanup(
+		  prolog_open_source(Path, Stream),
+		  prolog_read_source_term(Stream, Term, _, []),
+		  prolog_close_source(Stream)), _,
+	      fail),
 	Term = (:- module(Module, _Public)).
 
 hidden_entity(module(_)).
@@ -318,6 +335,9 @@ expand(TF) :->
 	;   true
 	).
 
+expand_all(_TF) :->
+	true.
+
 split_head(M:Head, Name, Arity, M) :- !,
 	callable(Head),
 	functor(Head, Name, Arity).
@@ -361,6 +381,8 @@ identify(TF) :->
 	get(TF, identifier, Path),
 	(   send(TF, loaded)
 	->  send(TF, report, status, 'Loaded file %s', Path)
+	;   send(TF, included)
+	->  send(TF, report, status, 'Included file %s', Path)
 	;   send(TF, report, status, 'File %s', Path)
 	).
 
@@ -932,26 +954,35 @@ image(predicate,	dcg,		'grammar.xpm').
 :- multifile
 	user:message_hook/3.
 
-image_of_load_state(start,	'loading.xpm').
-image_of_load_state(true,	'plloadedfile.xpm').
-image_of_load_state(false,	'loadfailed.xpm').
+image_of_load_state(start, _,	    'loading.xpm').
+image_of_load_state(true,  load,    'plloadedfile.xpm').
+image_of_load_state(true,  include, 'plincludedfile.xpm').
+image_of_load_state(false, _,	    'loadfailed.xpm').
 
 user:message_hook(load_file(What), _Kind, _Lines) :-
+	loading(What, load).
+user:message_hook(include_file(What), _Kind, _Lines) :-
+	loading(What, include).
+
+loading(What, How) :-
 	load_info(What, File, Stage),
 	prolog_overview_window(Win),
 	(   file_name_extension(_, qlf, File)
-	->  debug('Looking for ~q~n', [File]),
+	->  debug(gtrace(qlf), 'Looking for ~q', [File]),
 	    '$qlf_info'(File, V, V, _WordSize, Sources),
-	    debug('Contains ~q~n', [Sources]),
+	    debug(gtrace(qlf), 'Contains ~q', [Sources]),
 	    member(TheFile, Sources)
 	;   TheFile = File
 	),
 	get(Win, file_node, TheFile, Node),
-	image_of_load_state(Stage, Img),
-	send(Node, image, Img),
-	send(Node, flush),
+	image_of_load_state(Stage, How, Img),
+	in_pce_thread(update_image(Node, Img)),
 	fail.
 
+update_image(Node, Img) :-
+	send(Node, image, Img),
+	send(Node, flush).
+
 load_info(start(_Level, file(_, Path)),
 	  Path, start).
 load_info(failed(Spec),
@@ -963,4 +994,6 @@ load_info(failed(Spec),
 			   Path).
 load_info(done(_Level, file(_, Path), _, _, _, _),
 	  Path, true).
+load_info(done(_Level, file(_, Path)),
+	  Path, true).
 
