@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of XPCE --- The SWI-Prolog GUI toolkit
+/*  Part of XPCE --- The SWI-Prolog GUI toolkit
 
     Author:        Jan Wielemaker and Anjo Anjewierden
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog/projects/xpce/
-    Copyright (C): 1985-2006, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -39,7 +38,8 @@
 	    get_tracer/2,		% :Goal, -Result
 	    in_debug_thread/2,		% +ObjOrThread, :Goal
 	    thread_debug_queue/2,	% +Thread, -Queue
-	    prolog_frame_attribute/4	% +GUI, +Frame, +Attr, -Value
+	    prolog_frame_attribute/4,	% +GUI, +Frame, +Attr, -Value
+	    prolog_choice_attribute/4	% +GUI, +Choice, +Attr, -Value
 	  ]).
 :- use_module(library(pce)).
 :- use_module(library(lists)).
@@ -62,6 +62,11 @@
 	user:prolog_event_hook/1,
 	user:message_hook/3.
 
+:- meta_predicate
+	in_debug_thread(+, 0),
+	send_pce(0),
+	send_pce_async(0).
+
 register_directories :-
 	(   member(SpyBase, ['icons/nospy', library('trace/icons/nospy')]),
 	    absolute_file_name(SpyBase,
@@ -107,22 +112,18 @@ prolog_tracer(Thread, Ref, Create) :-
 	(   gui(Thread, Level, Ref)
 	->  true
 	;   Create == true
-	->  debug('New GUI for thread ~p, break level ~p~n', [Thread, Level]),
+	->  debug(gtrace(gui),
+		  'New GUI for thread ~p, break level ~p', [Thread, Level]),
 	    send_pce(send(new(Ref, prolog_debugger(Level, Thread)), open))
 	).
 
 %%	break_level(-Level)
 %
 %	Current break-level.
-%
-%	@bug	Breaks only work in the main thread. We need to change I/O
-%		handling and where the break-level is stored to fix
-%		this.
 
 break_level(Level) :-
-	thread_self(main), !,
-	flag('$break_level', Level, Level).
-break_level(1).
+	current_prolog_flag(break_level, Level), !.
+break_level(-1).				% non-interactive thread.
 
 
 %%	send_tracer(+ThreadOrGUI, +Term) is semidet.
@@ -169,10 +170,6 @@ get_tracer(Thread, Term, Result) :-
 		 *    THREAD SYNCHRONISATION	*
 		 *******************************/
 
-:- meta_predicate
-	in_debug_thread(+, :),
-	send_pce(:).
-
 %%	thread_debug_queue(+Thread, -Queue) is det.
 %
 %	Queue is the debugging queue for Thread.  We do not use the main
@@ -182,6 +179,10 @@ get_tracer(Thread, Term, Result) :-
 	thread_debug_queue_store/2.
 
 thread_debug_queue(Thread, Queue) :-
+	with_mutex(debug_msg_queue,
+		   thread_debug_queue_locked(Thread, Queue)).
+
+thread_debug_queue_locked(Thread, Queue) :-
 	(   thread_debug_queue_store(Thread, Q)
 	->  Queue = Q
 	;   message_queue_create(Q),
@@ -197,6 +198,22 @@ user:prolog_event_hook(thread_finished(TID)) :-
 	       message_queue_destroy(Queue)),
 	fail.				% allow other hooks
 
+msg_id(Id) :-
+	with_mutex(debug_msg_id,
+		   msg_id_locked(Id)).
+
+:- dynamic
+	trace_msg_id/1.
+
+msg_id_locked(Id) :-
+	(   retract(trace_msg_id(Id0))
+	->  NId is Id0+1
+	;   NId = 1
+	),
+	assert(trace_msg_id(NId)),
+	Id = NId.
+
+
 %%	send_pce(:Goal)
 %
 %	Run Goal in XPCE thread. Wait  for completion. In the meanwhile,
@@ -209,24 +226,30 @@ send_pce(Goal) :-
 send_pce(Goal) :-
 	thread_self(Self),
 	term_variables(Goal, GVars),
-	in_pce_thread(run_pce(Goal, GVars, Self)),
+	msg_id(Id),
+	in_pce_thread(run_pce(Goal, GVars, Self, Id)),
 	thread_debug_queue(Self, Queue),
 	repeat,
-	thread_get_message(Queue, '$trace'(Result)),
-	debug(' ---> send_pce: result = ~p~n', [Result]),
-	(   Result = error(E)
-	->  throw(E)
-	;   Result = call(CallBack, CBVars, Caller)
-	->  run_pce(CallBack, CBVars, Caller),
-	    fail
-	;   Result = true(BGVars)
-	->  !, BGVars = GVars
-	;   assertion(Result == false), !,
+	thread_get_message(Queue, '$trace'(Result, Id2)),
+	debug(gtrace(thread),
+	      ' ---> ~w: send_pce: result = ~p', [Id2, Result]),
+	(   Result = call(CallBack, CBVars, Caller)
+	->  run_pce(CallBack, CBVars, Caller, Id2),
 	    fail
+	;   assertion(Id == Id2),
+	    (	Result = true(BGVars)
+	    ->  !, BGVars = GVars
+	    ;   Result == false
+	    ->	fail
+	    ;	Result = error(E)
+	    ->  throw(E)
+	    ;	assertion(false)
+	    )
 	).
 
-run_pce(Goal, Vars, Caller) :-
-	debug('Running ~p for thread ~p~n', [Goal, Caller]),
+run_pce(Goal, Vars, Caller, Id) :-
+	debug(gtrace(thread), '~w: running ~p for thread ~p',
+	      [Id, Goal, Caller]),
 	(   catch(Goal, Error, true)
 	->  (   var(Error)
 	    ->	Result = true(Vars)
@@ -234,9 +257,9 @@ run_pce(Goal, Vars, Caller) :-
 	    )
 	;   Result = false
 	),
-	debug('Ok, returning ~p~n', [Result]),
+	debug(gtrace(thread), '~w: ok, returning ~p', [Id, Result]),
 	thread_debug_queue(Caller, Queue),
-	thread_send_message(Queue, '$trace'(Result)).
+	thread_send_message(Queue, '$trace'(Result, Id)).
 
 %%	in_debug_thread(+Thread, :Goal) is semidet.
 %%	in_debug_thread(+Object, :Goal) is semidet.
@@ -255,13 +278,15 @@ in_debug_thread(Thread, Goal) :-
 	Goal, !.
 in_debug_thread(Thread, Goal) :-
 	thread_self(Self),
-	debug('Call [Thread ~p] ~p~n', [Thread, Goal]),
+	msg_id(Id),
+	debug(gtrace(thread), 'Call [Thread ~p] ~p', [Thread, Goal]),
 	term_variables(Goal, GVars),
 	thread_debug_queue(Thread, Queue),
-	thread_send_message(Queue, '$trace'(call(Goal, GVars, Self))),
+	thread_send_message(Queue, '$trace'(call(Goal, GVars, Self), Id)),
 	thread_debug_queue(Self, MyQueue),
-	thread_get_message(MyQueue, '$trace'(Result)),
-	debug(' ---> in_debug_thread: result = ~p~n', [Result]),
+	thread_get_message(MyQueue, '$trace'(Result, Id)),
+	debug(gtrace(thread),
+	      ' ---> in_debug_thread: result = ~p', [Result]),
 	(   Result = error(E)
 	->  throw(E)
 	;   Result = true(BGVars)
@@ -269,6 +294,18 @@ in_debug_thread(Thread, Goal) :-
 	).
 
 
+%%	send_pce_async(:Goal) is det.
+%
+%	Send to the debug thread asynchronously.
+
+send_pce_async(Goal) :-
+	thread_self(Me),
+	pce_thread(Me), !,
+	Goal.
+send_pce_async(Goal) :-
+	in_pce_thread(Goal).
+
+
 %%	prolog_frame_attribute(+GUI, +Frame, +Attribute, -Value) is det.
 %%	prolog_frame_attribute(+Thread, +Frame, +Attribute, -Value) is det.
 %
@@ -279,6 +316,11 @@ prolog_frame_attribute(Thread, Frame, Attribute, Value) :-
 	in_debug_thread(Thread,
 			prolog_frame_attribute(Frame, Attribute, Value)).
 
+prolog_choice_attribute(GUI, Choice, Attribute, Value) :-
+	in_debug_thread(GUI,
+			prolog_choice_attribute(Choice, Attribute, Value)).
+
+
 
 		 /*******************************
 		 *     DEBUGGER APPLICATION	*
@@ -306,10 +348,11 @@ initialise(App) :->
 variable(source,	any,		both, "Source view").
 variable(break_level,	int,		get,  "Break-level I'm associated to").
 variable(thread,	'int|name*',	get,  "Associated thread").
-variable(current_frame, int*,   	both, "The most recent frame").
-variable(current_break,	tuple*,		both, "tuple(ClauseRef, PC)").
+variable(trap_frame,    int*,		get,  "Last trapped frame").
+variable(trap_port,     name*,		get,  "Last trapped port").
+variable(current_frame, int*,		both, "The most recent frame").
 variable(quitted,	bool := @off,   both, "Asked to quit").
-variable(mode,		name := created,both, "Current mode").
+variable(mode,		name := created,get,  "Current mode").
 
 running_in_pce_thread :-
 	pce_thread(Pce), thread_self(Pce).
@@ -412,11 +455,16 @@ fill_menu_bar(F) :->
 			      message(F, exceptions),
 			      end_group := @on),
 		    menu_item(toggle_edit_mode,
-			      message(F, edit))
+			      message(F, edit),
+			      end_group := @on),
+		    menu_item(copy_goal,
+			      message(F, copy_goal))
 		  ]),
 	send_list(View, append,
 		  [ menu_item(threads,
-			      message(F, show_threads))
+			      message(F, show_threads)),
+		    new(PT, menu_item(portray_code_lists,
+				      message(F, portray_text)))
 		  ]),
 	send_list(Comp, append,
 		  [ menu_item(make,
@@ -433,7 +481,10 @@ fill_menu_bar(F) :->
 			      message(@prolog, prolog_help)),
 		    menu_item('XPCE manual',
 			      message(@prolog, manpce))
-		  ]).
+		  ]),
+	send(View, show_current, @on),
+	send(View, multiple_selection, @on),
+	send(PT, condition, message(F, update_portray_text, PT)).
 
 settings(_F) :->
 	"Edit the preferences"::
@@ -444,7 +495,7 @@ about(_) :->
 	"Display aout message"::
 	version(Version),
 	send(@display, inform,
-	     'SWI-Prolog debugger version %s\n\
+	     'SWI-Prolog debugger version %s\n\c
 	      By Jan Wielemaker',
 	     Version).
 
@@ -452,9 +503,18 @@ help(_) :->
         "Show window with help-text"::
         send(@helper, give_help, pltracer, main).
 
-show_frame(GUI, Frame:int, PC:'int|name') :->
+show_frame(GUI, Frame:int, PC:prolog) :->
 	"Show the variables of this frame"::
-	prolog_show_frame(Frame, [gui(GUI), pc(PC), source, bindings]).
+	(   get(GUI, trap_frame, Frame)		% the initial trapped port
+	->  get(GUI, trap_port, Style)
+	;   PC = choice(_)			% A choice-point
+	->  Style = choice
+	;   Style = frame			% Somewhere up the stack
+	),
+	prolog_show_frame(Frame, [ gui(GUI), pc(PC),
+				   source, bindings,
+				   style(Style)
+				 ]).
 
 show_stack(GUI, CallFrames:prolog, ChoiceFrames:prolog) :->
 	"Show the stack and choicepoints"::
@@ -466,6 +526,44 @@ show_threads(_GUI) :->
 	"Open Thread monitor"::
 	prolog_ide(thread_monitor).
 
+portray_text(GUI) :->
+	"Toggle portray of text"::
+	portraying_text(Old),
+	negate(Old, New),
+	portray_text(New),
+	send(GUI, refresh_bindings),
+	send(GUI, report, status, 'Portray code-list as text: %s', New).
+
+negate(true, false).
+negate(false, true).
+
+update_portray_text(_GUI, MI:menu_item) :->
+	"Update selected of portray text item"::
+	portraying_text(Bool),
+	send(MI, selected, Bool).
+
+portraying_text(Bool) :-
+	current_predicate(portray_text:do_portray_text/1), !,
+	portray_text:do_portray_text(Bool).
+portraying_text(false).
+
+trapped_location(GUI, StartFrame:int, Frame:int, Port:name) :->
+	"The last trapped location"::
+	send(GUI, slot, trap_frame, Frame),
+	send(GUI, slot, trap_port, Port),
+	send(GUI, current_frame, StartFrame).
+
+refresh_bindings(GUI) :->
+	"Refresch the binding view after changing parameters"::
+	(   get(GUI, member, bindings, Bindings),
+	    get(Bindings, prolog_frame, Frame),
+	    Frame \== @nil
+	->  prolog_show_frame(Frame, [ gui(GUI),
+				       bindings
+				     ])
+	;   true
+	).
+
 
 		 /*******************************
 		 *	      EVENT		*
@@ -516,10 +614,11 @@ return(Frame, Result:any) :->
 		->  send(Frame, destroy)
 		;   true
 		),
-		debug(' ---> frame for thread = ~p: result = ~p~n',
+		debug(gtrace(thread),
+		      ' ---> frame for thread = ~p: result = ~p',
 		      [Thread, Result]),
 		thread_debug_queue(Thread, Queue),
-		thread_send_message(Queue, '$trace'(action(Result)))
+		thread_send_message(Queue, '$trace'(action(Result), action))
 	    )
 	;   get(Frame, quitted, @on)
 	->  send(Frame, destroy)
@@ -610,8 +709,6 @@ nostop_or_spy(F) :->
 	"Clear spy-point"::
 	(   send(F?source, delete_selected_stop)
 	->  true
-	;   get(F, current_break, tuple(ClauseRef, PC))
-	->  '$break_at'(ClauseRef, PC, false)
 	;   (   get(F, current_frame, Frame)
 	    ;   get(F, selected_frame, Frame)
 	    ),
@@ -651,10 +748,50 @@ details(F) :->
 	get(F, member, bindings, Bindings),
 	send(Bindings, details).
 
+nodebug(F) :->
+	"User hit n(odebug)"::
+	(   setting(auto_close, true)
+	->  send(F, quitted, @on)
+	;   true
+	),
+	send(F, return, nodebug).
+
+abort(F) :->
+	"User hit a(bort)"::
+	(   setting(auto_close, true)
+	->  send(F, quitted, @on)
+	;   true
+	),
+	send(F, return, abort).
+
+interrupt(F) :->
+	"User hit t (interrupt, trace)"::
+	get(F, thread, Thread),
+	thread_signal(Thread, trace).
+
 query(_F) :->
 	"Enter and run a query"::
 	prolog_ide(open_query_window).
 
+copy_goal(F) :->
+	"Copy the current goal into the copy-buffer"::
+	get(F, selected_frame, Frame),
+	(   Frame \== @nil
+	->  true
+	;   send(F, report, warning, 'No current frame'),
+	    fail
+	),
+	prolog_frame_attribute(F, Frame, goal, Goal),
+	prolog_frame_attribute(F, Frame, predicate_indicator, PI),
+	(   numbervars(Goal, 0, _, [attvar(skip)]),
+	    format(string(Text), '~q', [Goal]),
+	    send(@display, copy, Text),
+	    fail
+	;   true
+	),
+	format(atom(PIA), '~q', [PI]),
+	send(F, report, inform, 'Copied goal (%s) to clipboard', PIA).
+
 :- pce_group(delegate).
 
 file(F, File:'name|emacs_buffer*') :->
@@ -713,6 +850,17 @@ query_finished(F, Message:char_array) :->
 	send(F, mode, query_finished),
 	send(F, report, status, Message).
 
+mode(F, Mode:name) :->
+	"Switch modes"::
+	send(F, slot, mode, Mode),
+	(   get(F, member, buttons, D)
+	->  (   Mode == wait_user
+	    ->  send(D, running, @off)
+	    ;   send(D, running, @on)
+	    )
+	;   true
+	).
+
 :- pce_end_class(prolog_debugger).
 
 		 /*******************************
@@ -734,9 +882,10 @@ button(finish,	       "f",   'finish.xpm',	     'Finish selected goal').
 button(gap,	       -,     -,		     -).
 button(retry,	       "r",   'retry.xpm',	     'Retry selected goal').
 button(gap,	       -,     -,		     -).
-button(nodebug,	       "n",   'nodebug.xpm',	     'Continue without debugging').
+button(+nodebug,       "n",   'nodebug.xpm',	     'Continue without debugging').
+button(+abort,	       "a",   'abort.xpm',	     'Abort to the Prolog toplevel').
+button(+interrupt,     "t",   'interrupt.xpm',	     'Interrupt (trace)').
 button(+query,	       "b",   'break.xpm',	     'Enter a query').
-button(abort,	       "a",   'abort.xpm',	     'Abort to the Prolog toplevel').
 button(fail,	       "F",   'fail.xpm',	     'Force query to fail').
 button(gap,	       -,     -,		     -).
 button(+up,	       "u",   'up.xpm',		     'Select parent frame').
@@ -773,11 +922,12 @@ initialise(D) :->
 	    (	Action == gap
 	    ->	send(TB, append, gap)
 	    ;   tag_balloon(Balloon0, Keys, Balloon),
-		make_message(Action, D, Message),
+		make_message(Action, Name, D, Message),
 	        send(TB, append,
 		     new(B, tool_button(Message,
 					image(Image),
-					Balloon))),
+					Balloon,
+				        name := Name))),
 	        chain_list(KL, Keys),
 	        send(B, attribute, keys, KL)
 	    ),
@@ -785,8 +935,8 @@ initialise(D) :->
 	;   true
 	).
 
-make_message(+Action, D, message(D?frame, Action)) :- !.
-make_message(Action,  D, message(D, return, Action)).
+make_message(+Action, Action, D, message(D?frame, Action)) :- !.
+make_message(Action,  Action, D, message(D, return, Action)).
 
 typed(D, Id:event_id, Delegate:[bool]) :->
 	"Handle typing"::
@@ -810,6 +960,27 @@ button(D, Name:name, Button:button) :<-
 	get(D, member, tool_bar, TB),
 	get(TB, member, Name, Button).
 
+running(D, Running:bool) :->
+	"Make some buttons (in)active"::
+	get(Running, negate, NotRunning),
+	forall(running_button(Name),
+	       (   get(D, button, Name, Button)
+	       ->  send(Button, active, NotRunning)
+	       ;   format('No button ~w~n', [Name])
+	       )),
+	(   get(D, button, interrupt, Interrupt)
+	->  send(Interrupt, active, Running)
+	;   true
+	).
+
+running_button(into).
+running_button(creep).
+running_button(skip).
+running_button(retry).
+running_button(finish).
+running_button(nodebug).
+running_button(abort).
+
 :- pce_end_class(prolog_button_dialog).
 
 
@@ -828,13 +999,15 @@ it.
 :- pce_begin_class(prolog_bindings_view, view,
 		   "Overview of bindings of the current frame").
 
-class_variable(font,	font, 	normal, "Font for bindings").
+class_variable(font,	font,	normal, "Font for bindings").
 class_variable(size,    size,   size(40,11), "Initial size").
 
 variable(prolog_frame, int*, both, "Frame who's variables we are showing").
 
 :- pce_global(@prolog_binding_recogniser,
 	      make_prolog_binding_recogniser).
+:- pce_global(@prolog_binding_popup,
+	      make_prolog_binding_popup).
 
 make_prolog_binding_recogniser(G) :-
 	new(View, @event?window),
@@ -843,8 +1016,18 @@ make_prolog_binding_recogniser(G) :-
 			      message(View, on_click, Index))),
 	new(C2, click_gesture(left, '', double,
 			      message(View, details))),
-	new(G, handler_group(C1, C2)).
+	new(C3, popup_gesture(@prolog_binding_popup)),
+	send(@prolog_binding_popup, update_message,
+	     message(View, on_click, Index)),
+	new(G, handler_group(C1, C2, C3)).
 
+make_prolog_binding_popup(P) :-
+	new(P, popup),
+	send_list(P, append,
+		  [ menu_item(details, message(@arg1?window, details)),
+		    menu_item(copy,    message(@arg1?window, details,
+					       @default, copy))
+		  ]).
 
 initialise(B) :->
 	send_super(B, initialise),
@@ -867,7 +1050,7 @@ clear(B, Content:[bool]) :->
 	;   send_super(B, clear)
 	).
 
-details(B, Fragment:[prolog_frame_var_fragment]) :->
+details(B, Fragment:[prolog_frame_var_fragment], Action:[{view,copy}]) :->
 	"View details of the binding"::
 	get(B, prolog_frame, Frame),
 	(   Frame \== @nil
@@ -884,22 +1067,33 @@ details(B, Fragment:[prolog_frame_var_fragment]) :->
 	    )
 	;   Frag = Fragment
 	),
-	get(Frag, var_name, VarName),
-	get(Frag, value, Value),
-	prolog_frame_attribute(B, Frame, level, Level),
-	prolog_frame_attribute(B, Frame, goal, Goal),
-	predicate_name(Goal, PredName),
-	(   integer(VarName)
-	->  VarType = 'Argument'
-	;   VarType = 'Variable'
-	),
-	format(string(Label), '~w ~w of frame at level ~d running ~w',
-		[ VarType, VarName, Level, PredName ]),
-	view_term(Value,
-		  [ comment(Label),
-		    source_object(Frag),
-		    expose(true)
-		  ]).
+	(   get(Frag, var_name, VarName)
+	->  get(Frag, value, Value),
+	    prolog_frame_attribute(B, Frame, level, Level),
+	    prolog_frame_attribute(B, Frame, goal, Goal),
+	    predicate_name(Goal, PredName),
+	    (   integer(VarName)
+	    ->  VarType = 'Argument'
+	    ;   VarType = 'Variable'
+	    ),
+	    format(string(Label), '~w ~w of frame at level ~d running ~w',
+		   [ VarType, VarName, Level, PredName ]),
+	    debug(gtrace(bindings), 'Action ~w on ~w', [Action, Value]),
+	    (   Action == copy
+	    ->  (   numbervars(Value, 0, _, [attvar(skip)]),
+		    format(string(Text), '~q', [Value]),
+		    send(@display, copy, Text),
+		    fail
+		;   send(B, report, status, Label)
+		)
+	    ;   view_term(Value,
+			  [ comment(Label),
+			    source_object(Frag),
+			    expose(true)
+			  ])
+	    )
+	;   send(B, report, warning, 'Not a variable value')
+	).
 
 on_click(B, Index:int) :->
 	"Select fragment clicked"::
@@ -910,6 +1104,10 @@ on_click(B, Index:int) :->
 	;   send(B, selected_fragment, @nil)
 	).
 
+% Bindings is a list of Vars = Value,   where Vars is a list of variable
+% identifiers that take the form Name:ArgN,   were  Name is the variable
+% name (atom) and ArgN is the location in the frame.
+
 bindings(B, Bindings:prolog) :->
 	"Display complete list of bindings"::
 	(   term_attvars(Bindings, [])
@@ -917,13 +1115,16 @@ bindings(B, Bindings:prolog) :->
 	    Constraints = []
 	;   copy_term(Bindings, Plain, Constraints)
 	),
+	bind_vars(Plain),
+	cycles(Plain, Template, Cycles, Plain),
 	send(B, background, white),
 	pce_open(B, write, Fd),
-	(   bind_vars(Plain),
-	    forall(member(Vars=Value, Plain),
+	(   forall(member(Vars=Value, Template),
 		   send(B, append_binding, Vars, value(Value), Fd)),
 	    forall(member(C, Constraints),
-		   send(B, append_constraint, C, Fd)),
+		   send(B, append_extra, C, Fd, constraint)),
+	    forall(member(C, Cycles),
+		   send(B, append_extra, C, Fd, cycle)),
 	    fail
 	;   true
 	),
@@ -939,13 +1140,45 @@ bind_vars([Vars=Value|T]) :-
 	),
 	bind_vars(T).
 
+cycles(Term, Template, Cycles, _) :-
+	acyclic_term(Term), !,
+	Template = Term,
+	Cycles = [].
+cycles(Term, Template, Cycles, Bindings) :-
+	'$factorize_term'(Term, Template, Factors),
+	bind_non_cycles(Factors, Cycles),
+	name_cycle_vars(Cycles, 1, Bindings).
+
+bind_non_cycles([], []).
+bind_non_cycles([V=Term|T], L) :-
+	unify_with_occurs_check(V, Term), !,
+	bind_non_cycles(T, L).
+bind_non_cycles([H|T0], [H|T]) :-
+	bind_non_cycles(T0, T).
+
+
+name_cycle_vars([], _, _).
+name_cycle_vars([H|T], I, Bindings) :-
+	H = (Var=_Value),
+	(   member(Vars=VarsValue, Bindings),
+	    VarsValue == Var,
+	    Vars = [Name:_|_]
+	->  I2 = I
+	;   atom_concat('_S', I, Name),
+	    I2 is I + 1
+	),
+	Var = '$VAR'(Name),
+	name_cycle_vars(T, I2, Bindings).
+
 
 append_binding(B, Names0:prolog, ValueTerm:prolog, Fd:prolog) :->
 	"Add a binding to the browser"::
 	ValueTerm = value(Value0),	% protect :=, ?, etc.
-	(   Value0 = '$VAR'(_), Names0 = [_],
-	    setting(show_unbound, false)
-	->  true
+	(   Value0 = '$VAR'(Name), Names0 = [Name:_]
+	->  (   setting(show_unbound, false)
+	    ->  true
+	    ;	format(Fd, '~w\t= _~n', [Name])
+	    )
 	;   (   Value0 = '$VAR'(_), Names0 = [_,_|_]
 	    ->	append(Names, [VarN:_], Names0),
 		Value = '$VAR'(VarN)
@@ -972,12 +1205,12 @@ write_varnames(Fd, [N:_|T]) :-
 	format(Fd, '~w = ', N),
 	write_varnames(Fd, T).
 
-append_constraint(B, Constraint:prolog, Fd:prolog) :->
+append_extra(B, Constraint:prolog, Fd:prolog, Comment:name) :->
 	"Display current constraints"::
 	get(B, text_buffer, TB),
 	current_prolog_flag(toplevel_print_options, Options),
 	get(TB, size, S0),
-	format(Fd, '(constraint)\t~W~n', [Constraint, Options]),
+	format(Fd, '(~w)\t~W~n', [Comment, Constraint, Options]),
 	flush_output(Fd),
 	get(TB, size, S1),
 	new(_, prolog_frame_constraint_fragment(TB, S0, S1)).
@@ -1018,6 +1251,10 @@ initialise(F, TB:text_buffer, From:int, To:int) :->
 	Len is To-From,
 	send_super(F, initialise, TB, From, Len, constraint).
 
+var_name(_F, _Name:name) :<-
+	"Cannot show details"::
+	fail.
+
 :- pce_end_class(prolog_frame_constraint_fragment).
 
 
@@ -1027,28 +1264,28 @@ initialise(F, TB:text_buffer, From:int, To:int) :->
 		 *******************************/
 
 %user:prolog_event_hook(Term) :-
-%	debug('prolog_event_hook(~w).~n', [Term]),
+%	debug('prolog_event_hook(~w).', [Term]),
 %	fail.
 user:prolog_event_hook(frame_finished(Frame)) :-
 	thread_self(Thread),
-	gui(Thread, _, Gui),
-	send(Gui, frame_finished, Frame),
+	gui(Thread, _, Gui),		% has a gui
+	send_pce_async(send(Gui, frame_finished(Frame))),
 	fail.
 user:prolog_event_hook(exit_break(Level)) :-
 	thread_self(Thread),
 	gui(Thread, Level, Gui),
-	send(Gui, destroy),
+	send_pce_async(send(Gui, destroy)),
 	fail.
 user:prolog_event_hook(finished_query(_Qid, YesNo)) :-
 	thread_self(Thread),		% only main?
 	break_level(Level),
 	gui(Thread, Level, Ref),
-	send(Ref, query_finished, YesNo),
+	send_pce_async(send(Ref, query_finished(YesNo))),
 	fail.
 user:prolog_event_hook(thread_finished(Thread)) :-
 	gui(Thread, _, Gui),
 	thread_property(Thread, status(Status)),
-	in_pce_thread(send(Gui, thread_finished, Status)),
+	send_pce_async(send(Gui, thread_finished(Status))),
 	fail.
 
 user:message_hook('$aborted', _, _Lines) :-
@@ -1062,16 +1299,17 @@ aborted :-
 	thread_self(Thread),
 	gui(Thread, Level, Gui),
 	(   Level \== 0
-	->  send(Gui, destroy)
-	;   send(Gui, aborted)
-	).
+	->  Message = destroy
+	;   Message = aborted
+	),
+	send_pce_async(send(Gui, Message)).
 
 query_finished(YesNo) :-
 	finished(YesNo, Message),
 	thread_self(Thread),
 	break_level(Level),
 	gui(Thread, Level, Gui),
-	send(Gui, query_finished, Message).
+	send_pce_async(send(Gui, query_finished(Message))).
 
 finished(no, 'Query failed').
 finished(yes, 'Query succeeded').
