@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <h/kernel.h>
@@ -442,7 +442,7 @@ static getdecl get_font[] =
 /* Resources */
 
 static classvardecl rc_font[] =
-{ RC(NAME_scale, "real", "1.4",
+{ RC(NAME_scale, "real", UXWIN("1.0", "1.4"),
      "Multiplication factor for all fonts")
 };
 
