@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #ifndef _XPCE_INTERFACE_H_INCLUDED
@@ -31,7 +31,7 @@
 		********************************/
 
 #ifndef PCE_VERSION
-#define PCE_VERSION "6.6.64, January 2009"
+#define PCE_VERSION "6.6.66, July 2009"
 #endif
 
 #ifndef OS_VERSION
@@ -49,7 +49,7 @@
 
 #ifndef __pce_export
 #if defined(PCE_INCLUDED) && (defined(WIN32) || defined(__CYWIN32__))
-#define __pce_export _declspec(dllexport)
+#define __pce_export __declspec(dllexport)
 #else
 #define __pce_export extern
 #endif /*PCE_INCLUDED*/
@@ -146,8 +146,8 @@ __pce_export PceITFSymbol getITFSymbolName(PceName);
 
 typedef union
 { char		character;		/* Value is a C char */
-  long		integer;		/* Value of PCE int or reference */
-  double 	real;			/* Value of PCE real */
+  intptr_t	integer;		/* Value of PCE int or reference */
+  double	real;			/* Value of PCE real */
   char *	string;			/* Value is a C char * */
   void *	pointer;		/* Anonymous pointer */
   PceITFSymbol	itf_symbol;		/* Interface symbol */
@@ -157,7 +157,7 @@ typedef union
 #define PCE_INTEGER	1
 #define PCE_NAME	2
 #define PCE_REFERENCE	3
-#define PCE_ASSOC	4 
+#define PCE_ASSOC	4
 #define PCE_REAL	5
 #define PCE_HOSTDATA	6
 
@@ -296,6 +296,9 @@ __pce_export void	pceMTLock(int lock);
 __pce_export void	pceMTUnlock(int lock);
 __pce_export int	pceMTTryLock(int lock);
 __pce_export int	pceMTdetach(void);
+#ifdef __WINDOWS__
+__pce_export unsigned	setPceThread(unsigned id);
+#endif
 
 		/********************************
 		*          PCE CALLING C	*
