@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <h/kernel.h>
@@ -66,7 +66,7 @@ static const encname enc_names[] =
 { { NAME_octet,		ENC_OCTET },
   { NAME_ascii,		ENC_ASCII },
   { NAME_iso_latin_1,   ENC_ISO_LATIN_1 },
-  { NAME_text,   	ENC_ANSI },
+  { NAME_text,		ENC_ANSI },
   { NAME_utf8,		ENC_UTF8 },
   { NAME_unicodeBe,	ENC_UNICODE_BE },
   { NAME_unicodeLe,	ENC_UNICODE_LE },
@@ -116,7 +116,7 @@ getContentsSourceSink(SourceSink ss, Int from, Int len)
   IOSTREAM *fd = Sopen_object(ss, mode);
 
   if ( fd )
-  { long size = Ssize(fd);
+  { int64_t size = Ssize(fd);
     string s;
     status ok;
 
