@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <h/kernel.h>
@@ -300,10 +300,22 @@ value hexadecimal in case of error.
 We have structure exception-handling of   Windows, Unix with traditional
 signal(), Unix with sigaction (preserves more  context) and systems with
 and without SIGBUS ...
+
+This is hard on MinGW. There is a  library libSEH that is supposed to do
+the job, but this is basically just to   make XPCE deal with errors more
+elegantly, so for now we just run the code unprotected.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #ifdef __WINDOWS__
 
+#ifdef __MINGW32__
+
+char *
+pcePP(Any obj)
+{ return do_pp(obj);
+}
+
+#else
 #include <excpt.h>
 
 char *
@@ -330,7 +342,7 @@ pcePP(Any obj)
 
   return s;
 }
-
+#endif /*__MINGW32__*/
 
 #else /*__WINDOWS__*/
 
