@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <h/kernel.h>
@@ -75,9 +75,9 @@ isqrt(long a)
 
 int
 distance(int x1, int y1, int x2, int y2)
-{ int dx = x1-x2, dy = y1-y2;
+{ double dx = (double)(x1-x2), dy = (double)(y1-y2);
 
-  return isqrt(dx*dx + dy*dy);
+  return rfloat(sqrt(dx*dx + dy*dy));
 }
 
 
@@ -349,6 +349,8 @@ characterName(Any chr)
       } else
       { size_t l;
 
+	if ( ctrl )
+	  wcscat(buf, L"\\C-");
         buf[l=wcslen(buf)] = c;
 	buf[l+1] = EOS;
       }
@@ -390,7 +392,7 @@ writef_arguments(char *fm, va_list args, int *argc, Any *argv)
 	}
 	continue;
       default:
-      	fm++;
+	fm++;
 	continue;
     }
   }
@@ -535,7 +537,7 @@ swritefv(int (*out)(void*, wint_t), void *closure,
 	      char buf[64];
 
 	      if ( argc <= 0 )
-	      	a = 0;
+		a = 0;
 	      else
 	      { if ( (i = checkType(argv[0], TypeInt, NIL)) )
 		  a = valInt(i);
@@ -554,7 +556,7 @@ swritefv(int (*out)(void*, wint_t), void *closure,
 	      if ( arg == NOT_SET )
 		sprintf(buf, fmtbuf, a);
 	      else
-	      	sprintf(buf, fmtbuf, arg, a);
+		sprintf(buf, fmtbuf, arg, a);
 
 	      PutString(buf);
 
@@ -570,7 +572,7 @@ swritefv(int (*out)(void*, wint_t), void *closure,
 	      char buf[64];
 
 	      if ( argc <= 0 )
-	      	a = 0.0;
+		a = 0.0;
 	      else
 	      { if ( (f = checkType(argv[0], TypeReal, NIL)) )
 		  a = valReal(f);
@@ -588,7 +590,7 @@ swritefv(int (*out)(void*, wint_t), void *closure,
 	      if ( arg == NOT_SET )
 		sprintf(buf, fmtbuf, a);
 	      else
-	      	sprintf(buf, fmtbuf, arg, a);
+		sprintf(buf, fmtbuf, arg, a);
 
 	      PutString(buf);
 
@@ -775,7 +777,7 @@ scanstr(char *str, char *fmt, Any *r)
 	  length = 0;
 
 	switch(*s)			/* conversion char */
-    	{ case 'u':
+	{ case 'u':
 	    if ( !supress )
 	    { types[argn] = length|T_INT|S_UNSIGNED;
 	      ptrs[argn]  = (void *)alloca(sizeof(long));
@@ -833,7 +835,7 @@ scanstr(char *str, char *fmt, Any *r)
 	    continue;
 	}
       default:
-      	s++;
+	s++;
 	continue;
     }
   }
@@ -854,7 +856,7 @@ scanstr(char *str, char *fmt, Any *r)
 			              ptrs[3], ptrs[4]); break;
     case 6:	ar = sscanf(str, fmt, ptrs[0], ptrs[1], ptrs[2],
 			              ptrs[3], ptrs[4], ptrs[5]);
-    				      break;
+				      break;
     case 7:	ar = sscanf(str, fmt, ptrs[0], ptrs[1], ptrs[2],
 			              ptrs[3], ptrs[4], ptrs[5],
 			              ptrs[6]); break;
@@ -868,75 +870,75 @@ scanstr(char *str, char *fmt, Any *r)
     case 10:	ar = sscanf(str, fmt, ptrs[0], ptrs[1], ptrs[2],
 			              ptrs[3], ptrs[4], ptrs[5],
 			              ptrs[6], ptrs[7], ptrs[8],
-			    	      ptrs[9]);
+				      ptrs[9]);
 				      break;
     case 11:	ar = sscanf(str, fmt, ptrs[0], ptrs[1], ptrs[2],
 			              ptrs[3], ptrs[4], ptrs[5],
 			              ptrs[6], ptrs[7], ptrs[8],
-			    	      ptrs[9], ptrs[10]);
+				      ptrs[9], ptrs[10]);
 				      break;
     case 12:	ar = sscanf(str, fmt, ptrs[0], ptrs[1], ptrs[2],
 			              ptrs[3], ptrs[4], ptrs[5],
 			              ptrs[6], ptrs[7], ptrs[8],
-			    	      ptrs[9], ptrs[10], ptrs[11]);
+				      ptrs[9], ptrs[10], ptrs[11]);
 				      break;
     case 13:	ar = sscanf(str, fmt, ptrs[0], ptrs[1], ptrs[2],
 			              ptrs[3], ptrs[4], ptrs[5],
 			              ptrs[6], ptrs[7], ptrs[8],
-			    	      ptrs[9], ptrs[10], ptrs[11],
-			    	      ptrs[12]);
+				      ptrs[9], ptrs[10], ptrs[11],
+				      ptrs[12]);
 				      break;
     case 14:	ar = sscanf(str, fmt, ptrs[0], ptrs[1], ptrs[2],
 			              ptrs[3], ptrs[4], ptrs[5],
 			              ptrs[6], ptrs[7], ptrs[8],
-			    	      ptrs[9], ptrs[10], ptrs[11],
-			    	      ptrs[12], ptrs[13]);
+				      ptrs[9], ptrs[10], ptrs[11],
+				      ptrs[12], ptrs[13]);
 				      break;
     case 15:	ar = sscanf(str, fmt, ptrs[0], ptrs[1], ptrs[2],
 			              ptrs[3], ptrs[4], ptrs[5],
 			              ptrs[6], ptrs[7], ptrs[8],
-			    	      ptrs[9], ptrs[10], ptrs[11],
-			    	      ptrs[12], ptrs[13], ptrs[14]);
+				      ptrs[9], ptrs[10], ptrs[11],
+				      ptrs[12], ptrs[13], ptrs[14]);
 				      break;
     case 16:	ar = sscanf(str, fmt, ptrs[0], ptrs[1], ptrs[2],
 			              ptrs[3], ptrs[4], ptrs[5],
 			              ptrs[6], ptrs[7], ptrs[8],
-			    	      ptrs[9], ptrs[10], ptrs[11],
-			    	      ptrs[12], ptrs[13], ptrs[14],
-			    	      ptrs[15]);
+				      ptrs[9], ptrs[10], ptrs[11],
+				      ptrs[12], ptrs[13], ptrs[14],
+				      ptrs[15]);
 				      break;
     case 17:	ar = sscanf(str, fmt, ptrs[0], ptrs[1], ptrs[2],
 			              ptrs[3], ptrs[4], ptrs[5],
 			              ptrs[6], ptrs[7], ptrs[8],
-			    	      ptrs[9], ptrs[10], ptrs[11],
-			    	      ptrs[12], ptrs[13], ptrs[14],
-			    	      ptrs[15], ptrs[16]);
+				      ptrs[9], ptrs[10], ptrs[11],
+				      ptrs[12], ptrs[13], ptrs[14],
+				      ptrs[15], ptrs[16]);
 				      break;
     case 18:	ar = sscanf(str, fmt, ptrs[0], ptrs[1], ptrs[2],
 			              ptrs[3], ptrs[4], ptrs[5],
 			              ptrs[6], ptrs[7], ptrs[8],
-			    	      ptrs[9], ptrs[10], ptrs[11],
-			    	      ptrs[12], ptrs[13], ptrs[14],
-			    	      ptrs[15], ptrs[16], ptrs[17]);
+				      ptrs[9], ptrs[10], ptrs[11],
+				      ptrs[12], ptrs[13], ptrs[14],
+				      ptrs[15], ptrs[16], ptrs[17]);
 				      break;
     case 19:	ar = sscanf(str, fmt, ptrs[0], ptrs[1], ptrs[2],
 			              ptrs[3], ptrs[4], ptrs[5],
 			              ptrs[6], ptrs[7], ptrs[8],
-			    	      ptrs[9], ptrs[10], ptrs[11],
-			    	      ptrs[12], ptrs[13], ptrs[14],
-			    	      ptrs[15], ptrs[16], ptrs[17],
-			    	      ptrs[18]);
+				      ptrs[9], ptrs[10], ptrs[11],
+				      ptrs[12], ptrs[13], ptrs[14],
+				      ptrs[15], ptrs[16], ptrs[17],
+				      ptrs[18]);
 				      break;
     case 20:	ar = sscanf(str, fmt, ptrs[0], ptrs[1], ptrs[2],
 			              ptrs[3], ptrs[4], ptrs[5],
 			              ptrs[6], ptrs[7], ptrs[8],
-			    	      ptrs[9], ptrs[10], ptrs[11],
-			    	      ptrs[12], ptrs[13], ptrs[14],
-			    	      ptrs[15], ptrs[16], ptrs[17],
+				      ptrs[9], ptrs[10], ptrs[11],
+				      ptrs[12], ptrs[13], ptrs[14],
+				      ptrs[15], ptrs[16], ptrs[17],
 			              ptrs[18], ptrs[19]);
 				      break;
     default:	errorPce(NIL, NAME_tooManyArguments);
-    		fail;
+		fail;
   }
 #endif /*HAVE_VSSCANF*/
 
@@ -1106,18 +1108,36 @@ mclock()
 		*            SLEEP		*
 		********************************/
 
-#if defined(__WATCOMC__)
-#include <dos.h>
+#if !defined(MSLEEP_DONE) && defined(HAVE_NANOSLEEP)
+#define MSLEEP_DONE 1
+#include <errno.h>
 
 void
 msleep(int time)
-{ if ( time > 0 )
-    delay(time);
+{ struct timespec req;
+  int rc;
+
+  if ( time < 0 )
+    return;
+
+  DEBUG(NAME_flash,
+	Cprintf("nanosleep() %d milliseconds ...\n", time));
+
+  req.tv_sec = time/1000;
+  req.tv_nsec = (time%1000)*1000000;
+
+  do
+  { rc = nanosleep(&req, &req);
+  } while ( rc == -1 && errno == EINTR );
+
+  DEBUG(NAME_flash, Cprintf("ok\n"));
 }
 
-#else  /* !defined(__WATCOMC__) */
+#endif
+
+#if !defined(MSLEEP_DONE) && defined(HAVE_SELECT)
+#define MSLEEP_DONE 1
 
-#ifdef HAVE_SELECT
 #ifdef HAVE_SYS_SELECT_H
 #include <sys/select.h>
 #endif
@@ -1129,7 +1149,7 @@ msleep(int time)
   timeout.tv_usec = (time%1000)*1000;
 
   DEBUG(NAME_flash,
-	Cprintf("waiting %d milliseconds ...", time));
+	Cprintf("waiting %d milliseconds ...\n", time));
 
 #ifdef SELLIST
   { SELLIST(1,1) readbits = {0, 0};
@@ -1139,13 +1159,15 @@ msleep(int time)
     select(0, &readbits, &writebits, &exceptbits, &timeout);
   }
 #else
-  select(32, NULL, NULL, NULL, &timeout);
+  select(0, NULL, NULL, NULL, &timeout);
 #endif
 
   DEBUG(NAME_flash, Cprintf("ok\n"));
 }
 
-#else /*HAVE_SELECT*/
+#endif
+
+#ifndef MSLEEP_DONE
 
 extern void ws_msleep(int time);
 
@@ -1154,8 +1176,7 @@ msleep(int time)
 { ws_msleep(time);
 }
 
-#endif /*HAVE_SELECT*/
-#endif /*__WATCOMC__*/
+#endif
 
 		 /*******************************
 		 *	 ASSERT() SUPPORT	*
