@@ -19,15 +19,23 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
-#undef  O_IMGLIB  			/* Generic IMGLIB support */
-#undef  O_GIFREAD 			/* The GIF library */
+#undef  O_IMGLIB			/* Generic IMGLIB support */
+#undef  O_GIFREAD			/* The GIF library */
 #define O_XPM	  1			/* use Xpm library */
 #define O_GIF	  1			/* use our GIF loader */
 #define O_GIFWRITE 1			/* allow writing GIF files */
 
+#ifndef WINVER
+#define WINVER 0x0500
+#endif
+
+#ifndef _WIN32_IE
+#define _WIN32_IE 0x0501
+#endif
+
 #define RedrawWindow WinRedrawWindow
 #include <windows.h>
 #ifdef USE_WINSOCK2
@@ -58,6 +66,7 @@
 typedef struct win_mf *WinMF;		/* win_metafile pointer type */
 
 extern HINSTANCE ThePceHInstance;	/* HINSTANCE from pceDLLEntry() */
+extern DWORD	 ThePceThread;
 
 typedef char	cwidth;			/* width of a character */
 
