@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #define UNICODE 1
@@ -208,7 +208,7 @@ do_frame_wnd_proc(FrameObj fr,
 	    send(fr, NAME_resize, EAV);
 	    SetWindowText(hwnd, nameToWC(fr->label, NULL));
 	    assign(fr, status, wParam == SIZE_MAXIMIZED ? NAME_fullScreen
-		   					: NAME_window);
+							: NAME_window);
 	    for_cell(cell, fr->members)
 	      DisplayedGraphical(cell->value, ON);
 	  }
@@ -656,6 +656,7 @@ ws_create_frame(FrameObj fr)
   HWND owner = NULL;
   DWORD style = WS_CLIPCHILDREN;
   DWORD exstyle = 0;
+  DWORD tid;
   RECT rect;
   int w, h;
 
@@ -731,6 +732,13 @@ ws_create_frame(FrameObj fr)
   if ( !ref )
     return errorPce(fr, NAME_xOpen, fr->display);
 
+  if ( ThePceThread && (tid=GetCurrentThreadId()) != ThePceThread )
+  { int rc = AttachThreadInput(tid, ThePceThread, TRUE);
+
+    DEBUG(NAME_thread, Cprintf("%s: AttachThreadInput(%d, %d) --> %d\n",
+			       pp(fr), (int)tid, (int)ThePceThread, rc));
+  }
+
   setHwndFrame(fr, ref);
   assocObjectToHWND(ref, fr);
   current_frame = NULL;
