@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include "include.h"
@@ -31,12 +31,14 @@ by msimage.c, which in  turn  implements   the  OS  specific  version of
 gra/image.c implementing class image.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#ifdef __CYGWIN__
+#if defined(__CYGWIN__) || defined(__MINGW32__)
 #define XMD_H
 #endif
 
 #ifdef HAVE_LIBJPEG
-/*#define boolean jpeg_boolean*/
+#ifdef __RPCNDR_H__
+#define HAVE_BOOLEAN		/* prevent jmorecfg.h from redefining it */
+#endif
 #undef GLOBAL				/* conflict */
 #include <jpeglib.h>
 #include <jerror.h>
