@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include "include.h"
@@ -114,7 +114,7 @@ do_window_wnd_proc(HWND hwnd, UINT message, UINT wParam, LONG lParam)
       if ( DragQueryPoint(hdrop, &pt)  &&
 	   (nfiles = DragQueryFile(hdrop, (UINT)-1, NULL, 0)) >= 0 )
       { Chain files;
-      	Point pos;
+	Point pos;
 	wchar_t buf[MAXPATHLEN];
 	AnswerMark mark;
 	int i;
@@ -266,12 +266,8 @@ do_window_wnd_proc(HWND hwnd, UINT message, UINT wParam, LONG lParam)
 
 	if ( oproc )			/* refining alien window */
 					/* see winHandleWindow() below */
-	{
-#if (_MSC_VER < 1400)
-	  SetWindowLong(hwnd, GWL_WNDPROC, (LONG) oproc);
-#else
-	  SetWindowLongPtr(hwnd, GWLP_WNDPROC, (LONG_PTR) oproc);
-#endif
+	{ SetWindowLongPtr(hwnd, GWLP_WNDPROC, (LONG_PTR) oproc);
+
 	  return CallWindowProc(oproc, hwnd, message, wParam, lParam);
 	}
 
@@ -436,6 +432,7 @@ ws_create_window(PceWindow sw, PceWindow parent)
 { HWND hwnd;
   HWND parent_handle;
   DWORD style = WS_CHILD|WS_CLIPCHILDREN|WS_CLIPSIBLINGS|WS_VISIBLE;
+  DWORD tid;
 
   DEBUG(NAME_window, Cprintf("ws_create_window(%s %s)\n", pp(sw), pp(parent)));
 
@@ -459,6 +456,13 @@ ws_create_window(PceWindow sw, PceWindow parent)
   if ( !hwnd )
     return errorPce(sw, NAME_createFailed);
 
+  if ( ThePceThread && (tid=GetCurrentThreadId()) != ThePceThread )
+  { int rc = AttachThreadInput(tid, ThePceThread, TRUE);
+
+    DEBUG(NAME_thread, Cprintf("AttachThreadInput(%d, %d) --> %d\n",
+			       (int)tid, (int)ThePceThread, rc));
+  }
+
   DEBUG(NAME_window, Cprintf("Windows hwnd = %ld\n", (long) hwnd));
 
   setHwndWindow(sw, hwnd);
@@ -849,13 +853,8 @@ winHandleWindow(PceWindow sw, Int handle)
   setHwndWindow(sw, hwnd);
   assocObjectToHWND(hwnd, sw);
   w = sw->ws_ref;
-#if (_MSC_VER < 1400)
-  w->saved_window_procedure = (WNDPROC)GetWindowLong(hwnd, GWL_WNDPROC);
-  SetWindowLong(hwnd, GWL_WNDPROC, (LONG) window_wnd_proc);
-#else
   w->saved_window_procedure = (WNDPROC)GetWindowLongPtr(hwnd, GWLP_WNDPROC);
   SetWindowLongPtr(hwnd, GWLP_WNDPROC, (LONG_PTR) window_wnd_proc);
-#endif
   GetWindowRect(hwnd, &rect);
   ServiceMode(is_service_window(sw),
 	      { Area a = sw->area;
