@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <h/kernel.h>
@@ -35,10 +35,9 @@
 
 #ifdef HAVE_SYS_PARAM_H
 #include <sys/param.h>
-#else
-#ifndef MAXPATHLEN
-#define MAXPATHLEN 256
 #endif
+#ifndef MAXPATHLEN
+#define MAXPATHLEN 1024
 #endif
 
 #if HAVE_SYS_ACCESS_H			/* AIX 3.2.5 */
@@ -76,6 +75,7 @@ initialiseFile(FileObj f, Name name, Name encoding)
   if ( isDefault(name) )
   {
 #ifdef HAVE_MKSTEMP
+#define TMP_FILE_DONE
     char namebuf[100];
     int fileno;
     char *s;
@@ -95,8 +95,27 @@ initialiseFile(FileObj f, Name name, Name encoding)
 
     name = CtoName(namebuf);
     assign(f, status, NAME_tmpWrite);
-#else					/* use unsafe tmpnam */
-#ifdef HAVE_TMPNAM
+#endif
+
+#if !defined(TMP_FILE_DONE) && defined(HAVE_TEMPNAM)
+#define TMP_FILE_DONE			/* Prefer this on __WINDOWS__ */
+#ifdef __WINDOWS__
+    char *s = tempnam("c:\\tmp", "xpce");
+#else
+    char *s = tempnam("/tmp", "xpce");
+#endif
+
+    if ( s )
+    { name = CtoName(s);
+      /*Cprintf("tempnam() returns %s\n", s);*/
+      free(s);
+    } else
+    { return errorPce(f, NAME_noTempFile, getOsErrorPce(PCE));
+    }
+#endif
+
+#if !defined(TMP_FILE_DONE) && defined(HAVE_TMPNAM)
+#define TMP_FILE_DONE				/* use unsafe tmpnam */
     char namebuf[L_tmpnam];
     char *s = tmpnam(namebuf);
 
@@ -106,22 +125,11 @@ initialiseFile(FileObj f, Name name, Name encoding)
     } else
     { return errorPce(f, NAME_noTempFile, getOsErrorPce(PCE));
     }
-#else
-#ifdef HAVE_TEMPNAM			/* Prefer this on __WINDOWS__ */
-    char *s = tempnam("c:\\tmp", "xpce");
+#endif
 
-    if ( s )
-    { name = CtoName(s);
-      /*Cprintf("tempnam() returns %s\n", s);*/
-      free(s);
-    } else
-    { return errorPce(f, NAME_noTempFile, getOsErrorPce(PCE));
-    }
-#else
+#if !defined(TMP_FILE_DONE)
     Cprintf("No temporary files on this platform");
     fail;
-#endif
-#endif
 #endif
   }
 
@@ -980,7 +988,7 @@ getReadFile(FileObj f, Int n)
     str_alloc(&s->data);
 
     if ( (m = Sfread(s->data.s_textA, 1, size, f->fd)) != size )
-    { deleteString(s, toInt(m), DEFAULT); 			/* TBD: error? */
+    { deleteString(s, toInt(m), DEFAULT);			/* TBD: error? */
     }
   } else
   { tmp_string tmp;
