@@ -19,10 +19,10 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
-#ifdef __linux__
+#if defined(__linux__) || defined(__GLIBC__) || defined(__GNU__)
 #define _XOPEN_SOURCE			/* GNU glibc grantpt() prototypes */
 #endif
 #include <h/kernel.h>
@@ -67,6 +67,9 @@ reduced to the facility to terminate the inferior process.
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
+#ifdef HAVE_CRT_EXTERNS_H
+#include <crt_externs.h>
+#endif
 
 #ifdef HAVE_TERMIOS_H
 #include <termios.h>
@@ -85,9 +88,12 @@ reduced to the facility to terminate the inferior process.
 #ifdef __WINDOWS__
 #define environ _environ	/* declared in STDLIB.H */
 #else
+#ifdef HAVE__NSGETENVIRON
+#define environ (*_NSGetEnviron())
+#else
 extern char **environ;		/* Unix version */
 #endif
-
+#endif
 					/* this fixes RS6000/AIX problems */
 #if !defined(TCGETS) && defined(HAVE_TERMIO_H)
 #include <termio.h>
@@ -628,7 +634,7 @@ openProcess(Process p, CharArray cmd, int argc, CharArray *argv)
       int pid;
 
       if ( pipe(wrfd) )
-      	return errorPce(p, NAME_noPipe, OsError());
+	return errorPce(p, NAME_noPipe, OsError());
       if ( pipe(rdfd) )
       { close(wrfd[0]);
 	close(wrfd[1]);
@@ -1046,7 +1052,9 @@ copyTty(Process p, char *pty, int fd)
     buf.c_cc[VKILL]    = Control('U');
     buf.c_cc[VEOF]     = Control('D');
     buf.c_cc[VEOL]     = Control('@');
+#ifdef VEOL2
     buf.c_cc[VEOL2]    = Control('@');
+#endif
 #ifdef VSWTCH
     buf.c_cc[VSWTCH]   = Control('@');
 #endif
