@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <h/kernel.h>
@@ -647,7 +647,7 @@ collect_selection_display(Widget w, XtPointer xtp,
 	return;
       }
 
-      outA = bufA = malloc(l);
+      outA = bufA = pceMalloc(l);
       while(in<end)
       { in = utf8_get_uchar(in, &chr);
 
