@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 :- module(test_files, [test_files/0]).
@@ -41,7 +41,7 @@ test_files :-
 % this test verifies that atoms associated with temporary files
 % are properly deleted.
 
-test(tmp_cleanup, Atoms0 == Atoms1) :-
+test(tmp_cleanup, Atoms1 == Atoms0) :-
 	tmp_atoms(Atoms0),
 	(   between(1, 10, _),
 	    tmp_file(magic_sjwefrbas, Tmp),
@@ -53,6 +53,7 @@ test(tmp_cleanup, Atoms0 == Atoms1) :-
 	).
 
 tmp_atoms(List) :-
+	flush_unregistering,
 	agc,
 	findall(X, tmp_atom(X), Xs),
 	sort(Xs, List).
@@ -61,6 +62,17 @@ tmp_atom(X) :-
 	current_atom(X),
 	sub_atom(X, _, _, _, magic_sjwefrbas).
 
+:- dynamic a/1.
+
+%%	flush_unregistering
+%
+%	Register/unregister    an    arbitrary    atom      to     flush
+%	LD->atoms.unregistering.
+
+flush_unregistering :-
+	assertz(a(a)),
+	retractall(a(_)).
+
 %%	agc/0
 %
 %	If   other   threads   are   active,   it   is   possible   that
@@ -75,12 +87,16 @@ agc :-
 	repeat,
 	    garbage_collect_atoms,
 	    statistics(agc, AGC1),
- 	(   AGC1 > AGC0
+	(   AGC1 > AGC0
 	->  !
 	;   sleep(0.01),
 	    fail
 	).
 
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+% more tests
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+
 magic_pred_123.
 
 test(directory_files, error(existence_error(_,_))) :-
@@ -91,6 +107,10 @@ test(directory_files, true) :-
 	file_base_name(File, Plain),
 	directory_files(Dir, Files),
 	memberchk(Plain, Files).
-
+test(max_path_len, error(representation_error(max_path_length))) :-
+	length(L, 10000),
+	maplist(=('a'), L),
+	atom_chars(F, L),
+	absolute_file_name(F, _, [access(read)]).
 
 :- end_tests(files).
