@@ -1,11 +1,9 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2011, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
 			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
@@ -20,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 /*#define O_DEBUG 1*/
@@ -33,10 +31,10 @@ Implementation issues
 
 There are two parts in the atom   administration. One is a dynamic array
 (called buffer) atom_array, which is there to   find  the atoms back. An
-atom as it appears is a intptr_t   of the form (n<<LMASK_BITS)|TAG_ATOM. The
-atom  structure  is  located  by  getting  the  n-th  pointer  from  the
-atom_array dynamic array.  See atomValue() for translating the intptr_t into
-the address of the structure.
+atom   as   it   appears   is   of   type   word   and   of   the   form
+(n<<LMASK_BITS)|TAG_ATOM. The atom structure is   located by getting the
+n-th pointer from the atom_array  dynamic   array.  See  atomValue() for
+translating the word into the address of the structure.
 
 Next, there is a hash-table, which is a normal `open' hash-table mapping
 char * to the atom structure. This   thing is dynamically rehashed. This
@@ -45,112 +43,96 @@ table is used by lookupAtom() below.
 Atom garbage collection
 -----------------------
 
-There is no such thing, but below is an outline of what in entails.
-
 There are various categories of atoms:
 
-	# Built-in atoms
+	* Built-in atoms
 	These are used directly in the C-source of the system and cannot
 	be removed. These are the atoms upto a certain number. This
 	number is sizeof(atoms)/sizeof(char *).
 
-	# Foreign referenced atoms
+	* Foreign referenced atoms
 	These are references hold in foreign code by means of
 	PL_new_atom() or other calls returning an atom. The system has
-	no way to determine the lifetime of them. Most probably the best
-	approach is to offer a locking/unlocking flag to deal this type
-	of atoms. The lock/unlock may be nested. The proposed functions
-	are:
+	no way to determine the lifetime of them.  Foreign code must
+	keep track of references to atoms using these two functions:
 
-		PL_register_atom(atom_t atom)
-		PL_unregister_atom(atom_t atom)
+	      - PL_register_atom(atom_t atom)
+	      - PL_unregister_atom(atom_t atom)
 
-	# References from the Prolog stacks
-	Reference counting is unacceptable here, which implies a pass
-	similar to the normal garbage-collector is required to deal with
-	them. It might be worthwhile to include the atom
-	garbage-collection in the normal garbage collector.
+	* References from the Prolog stacks
+	Reference counting is unacceptable here, which implies we must
+	mark atoms that are accessible from the stacks.  This is done
+	by markAtomsOnStacks().
 
-	# References from other structures
+	* References from other structures
 	Various of the structures contain or may contain atom
 	references.  There are two options: lock/unlock them using
 	PL_register_atom() on creation/destruction of the structure
-	or enumerate them and flag the atoms.  The choice depends a
-	bit on the structure.  FunctorDef for example is not garbage
-	collected itself, so simply locking it makes sence.
+	or enumerate them and flag the atoms.  Currently, we use
+	registration everywhere, except for temporary structures
+	used by findall/3 and message queues, which are marked
+	by markAtomsThreads().
 
-	# References from compiled code and records
-	Again both aproaches are feasible.  Reference counting is
-	easy, except that the destruction of clauses and records
-	will be more slowly as the code needs to be analysed for
-	atom-references.
+	* References from compiled code and records
+	This uses PL_register_atom(), except for the cases mentioned
+	above.
 
 Reclaiming
 ----------
 
-To reclaim an atom, it needs to be   deleted from the hash-table, a NULL
-pointer should be set in the dynamic array and the structure needs to be
-disposed using unalloc(). Basically, this is not a hard job.
-
-The dynamic array will get holes   and  registerAtom() should first spot
-for a hole (probably using a globally   defined index that points to the
-first location that might be a hole   to avoid repetive scanning for the
-array while there is no place anyway).   It cannot be shrunk, unless all
-atoms above a certain index are gone. There is simply no way to find all
-atom references (that why we need the PL_register_atom()).
-
-In an advanced form, the hash-table could be shrunk, but it is debatable
-whether this is worth the trouble. So, alltogether the system will waist
-an average 1.5 machine word per reclaimed  atom, which will be reused as
-the atom-space grows again.
-
-
-Status
-------
-
-Some prelimary parts of atom garbage   collection  have been implemented
-and flagged using #ifdef O_ATOMGC  ...   #endif.  The foreign code hooks
-have been defined as well.
+To reclaim an atom, it is deleted   from  the hash-table, a NULL pointer
+should be set in the dynamic array and the structure is disposed.
 
+The dynamic array gets holes and  we   remember  the  first free hole to
+avoid too much search. Alternatively, we could  turn the holes into some
+form of linked list, for example by   encoding an integer that expresses
+the location of the next hole. We   cannot  shrink the array, unless all
+atoms above a certain index are gone.
 
 Atom GC and multi-threading
 ---------------------------
 
-This is a hard problem. I think the   best  solution is to add something
-like PL_thread_signal_async(int tid, void (*f)(void)) and call this from
-the invoking thread on all other threads.   These  thread will then scan
-their stacks and mark any references from their. Next they can carry on,
-as intptr_t as the invoking thread keeps   the  atom mutex locked during the
-whole atom garbage collection process. This   implies  the thread cannot
-create any atoms as intptr_t as the collection is going on.
+This is a hard problem. Atom-GC cannot   run  while some thread performs
+normal GC because the pointer relocation makes it extremely hard to find
+referenced atoms. Otherwise, ask all  threads   to  mark their reachable
+atoms and run collectAtoms() to reclaim the unreferenced atoms.
 
-We do have to define some mechanism to   know  all threads are done with
-their marking.
+On Unix, we signal all threads. Upon receiving the signal, they mark all
+accessible atoms and continue.  On  Windows,   we  have  no asynchronous
+signals, so we silence the threads one-by-one   and do the marking. This
+is realised by forThreadLocalData().  Note  that   this  means  only one
+thread is collecting  in  Windows.  This   could  be  enhanced  by using
+multiple threads during the collection phase.
 
-Don't know yet about Windows.  They   can't  do anything asynchronously.
-Maybe they have ways to ensure  all   other  threads  are sleeping for a
-while, so we can control the whole  process from the invoking thread. If
-this is the case we could also do this in Unix:
+Note that threads can mark their atoms and continue execution because:
 
-	thread_kill(<thread>, SIG_STOP);
-	<mark from thread>;
-	thread_kill(<thread>, SIG_CONT);
-
-Might be wise  to  design  the  marking   routine  suitable  to  take  a
-PL_local_data term as argument, so it can be called from any thread.
-
-All this will only work if we can call the atom garbage synchronously!!!
-
-Measures to allow for asynchronous atom GC
-------------------------------------------
-
-	* lookupAtom() returns a referenced atom
-	If not, it can be collected right-away!  Actually this might be
-	a good idea anyway to avoid foreign-code that caches atoms from
-	having to be updated.
+  - If a marked atom is no longer needed it is merely not reclaimed this
+    time (but might be in the next collection).
+  - If a new atom is referenced from the stack it is either a
+    - builtin atom (no problem)
+    - an atom from a structure using reference counting (is referenced
+      by this structure, so no problem, unless the reference count drops
+      to zero in PL_unregister_atom().  See PL_unregister_atom() for
+      handling the no-locking case.
+    - It is created.  This case blocks on L_ATOM being locked from
+      lookupBlob().
+  - Finally, message queues and bags as used by findall/3 complicate
+    the issue.  An atom sent to these structures subsequently may
+    become inaccessible from the stack (the normal case for findall/3,
+    which backtracks).  If the atom is copied back from the structure
+    to the stack ('$collect_findall_bag'/3 or thread_get_message/1,2),
+    the atom can no longer be marked from the structure but is added
+    to the stacks without locking.   We resolve this issue as follows:
+    - Run marking from queues/bags after marking the stack.  This
+      ensures that atoms added to the these structures get marked,
+      also if the atom is no longer on the stack.
+    - If an atom is copied to the stack from such a structure while
+      AGC is running, we are ok, because this is merely the same issue
+      as atoms living on the stack.  TBD: redesign the structures such
+      that they can safely be walked.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static void	rehashAtoms();
+static void	rehashAtoms(void);
 
 #define atom_buckets GD->atoms.buckets
 #define atomTable    GD->atoms.table
@@ -165,6 +147,7 @@ static void	rehashAtoms();
 #undef LD
 #define LD LOCAL_LD
 
+
 		 /*******************************
 		 *	      TYPES		*
 		 *******************************/
@@ -228,10 +211,10 @@ PL_find_blob_type(const char *name)
 
 int
 PL_unregister_blob_type(PL_blob_t *type)
-{ unsigned int i;
+{ size_t index;
+  int i, last=FALSE;
   PL_blob_t **t;
   int discarded = 0;
-  Atom *ap;
 
   PL_LOCK(L_MISC);
   for(t = &GD->atoms.types; *t; t = &(*t)->next)
@@ -245,13 +228,19 @@ PL_unregister_blob_type(PL_blob_t *type)
   PL_register_blob_type(&unregistered_blob_atom);
 
   LOCK();
+  for(index=1, i=0; !last; i++)
+  { size_t upto = (size_t)2<<i;
+    Atom *b = GD->atoms.array.blocks[i];
 
-  ap = GD->atoms.array;
-  for(i=0; i < GD->atoms.count; i++, ap++ )
-  { Atom atom;
+    if ( upto >= GD->atoms.highest )
+    { upto = GD->atoms.highest;
+      last = TRUE;
+    }
 
-    if ( (atom = *ap) )
-    { if ( atom->type == type )
+    for(; index<upto; index++)
+    { Atom atom = b[index];
+
+      if ( atom && atom->type == type )
       { atom->type = &unregistered_blob_atom;
 
 	atom->name = "<discarded blob>";
@@ -280,6 +269,37 @@ static const ccharp atoms[] = {
 };
 #undef ATOM
 
+/* Note that we use PL_malloc_uncollectable() here because the pointer in
+   our block is not the real memory pointer.  Probably it is better to
+   have two pointers; one to the allocated memory and one with the
+   necessary offset.
+*/
+
+static void
+putAtomArray(size_t where, Atom a)
+{ int idx = MSB(where);
+
+  assert(where >= 0);
+
+  if ( !GD->atoms.array.blocks[idx] )
+  { PL_LOCK(L_MISC);
+    if ( !GD->atoms.array.blocks[idx] )
+    { size_t bs = (size_t)1<<idx;
+      Atom *newblock;
+
+      if ( !(newblock=PL_malloc_uncollectable(bs*sizeof(Atom))) )
+	outOfCore();
+
+      memset(newblock, 0, bs*sizeof(Atom));
+      GD->atoms.array.blocks[idx] = newblock-bs;
+    }
+    PL_UNLOCK(L_MISC);
+  }
+
+  GD->atoms.array.blocks[idx][where] = a;
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 It might be wise to  provide  for   an  option  that does not reallocate
 atoms. In that case accessing a GC'ed   atom  causes a crash rather then
@@ -288,43 +308,50 @@ another atom.
 
 static void
 registerAtom(Atom a)
-{ size_t n = GD->atoms.count;
+{ size_t index;
 #ifdef O_ATOMGC				/* try to find a hole! */
-  Atom *ap = GD->atoms.array;
-  Atom *ep = ap+n;
-  Atom *p;
+  int i;
+  int last = FALSE;
 
-  for(p = &ap[GD->atoms.no_hole_before]; p < ep; p++)
-  { if ( *p == NULL )
-    { n = p - ap;
-      *p = a;
-      a->atom = (n<<LMASK_BITS)|TAG_ATOM;
-      if ( indexAtom(a->atom) != (uintptr_t)n )
-      {	/* TBD: user-level exception */
-	fatalError("Too many (%d) atoms", n);
+  for(index=GD->atoms.no_hole_before, i=MSB(index); !last; i++)
+  { size_t upto = (size_t)2<<i;
+    Atom *b = GD->atoms.array.blocks[i];
+
+    if ( upto >= GD->atoms.highest )
+    { upto = GD->atoms.highest;
+      last = TRUE;
+    }
+
+    for(; index<upto; index++)
+    { if ( b[index] == NULL )
+      { a->atom = (index<<LMASK_BITS)|TAG_ATOM;
+	b[index] = a;
+	GD->atoms.no_hole_before = index+1;
+
+	return;
       }
-      GD->atoms.no_hole_before = n+1;
-
-      return;
     }
   }
-  GD->atoms.no_hole_before = n+1;
+  GD->atoms.no_hole_before = index+1;
+#else
+  index = GD->atoms.highest;
 #endif /*O_ATOMGC*/
 
-  a->atom = (n<<LMASK_BITS)|TAG_ATOM;
-  if ( n >= GD->atoms.array_allocated )
-  { size_t newcount = GD->atoms.array_allocated * 2;
-    size_t newsize  = newcount*sizeof(Atom);
-    Atom *np = PL_malloc(newsize);
+  a->atom = (index<<LMASK_BITS)|TAG_ATOM;
+  if ( indexAtom(a->atom) != index )	/* TBD: user-level exception */
+    fatalError("Too many (%d) atoms", index);
+  putAtomArray(index, a);
+  GD->atoms.highest = index+1;
+}
 
-    memcpy(np, ap, newsize/2);
-    GD->atoms.array = np;
-    GD->atoms.array_allocated = newcount;
-    PL_free(ap);
-    ap = np;
+
+static size_t
+paddingBlob(PL_blob_t *type)
+{ if ( true(type, PL_BLOB_TEXT) )
+  { return true(type, PL_BLOB_WCHAR) ? sizeof(pl_wchar_t) : sizeof(char);
+  } else
+  { return 0;
   }
-  ap[n++] = a;
-  GD->atoms.count = n;
 }
 
 
@@ -347,9 +374,7 @@ treat the signal as bogus if agc has already been performed.
 
 word
 lookupBlob(const char *s, size_t length, PL_blob_t *type, int *new)
-{ GET_LD
-  unsigned int v0, v;
-  uintptr_t oldheap;
+{ unsigned int v0, v;
   Atom a;
 
   if ( !type->registered )		/* avoid deadlock */
@@ -358,20 +383,26 @@ lookupBlob(const char *s, size_t length, PL_blob_t *type, int *new)
 
   LOCK();
   v  = v0 & (atom_buckets-1);
-  DEBUG(0, lookups++);
+  DEBUG(MSG_HASH_STAT, lookups++);
 
   if ( true(type, PL_BLOB_UNIQUE) )
   { if ( false(type, PL_BLOB_NOCOPY) )
     { for(a = atomTable[v]; a; a = a->next)
-      { DEBUG(0, cmps++);
+      { DEBUG(MSG_HASH_STAT, cmps++);
 	if ( length == a->length &&
 	     type == a->type &&
 	     memcmp(s, a->name, length) == 0 )
 	{
 #ifdef O_ATOMGC
 	  if ( indexAtom(a->atom) >= GD->atoms.builtin )
-	  { if ( a->references++ == 0 )
+	  {
+#ifdef ATOMIC_REFERENCES
+	    if ( ATOMIC_INC(&a->references) == 1 )
+	      ATOMIC_DEC(&GD->atoms.unregistered);
+#else
+	    if ( ++a->references == 1 )
 	      GD->atoms.unregistered--;
+#endif
 	  }
 #endif
           UNLOCK();
@@ -381,15 +412,20 @@ lookupBlob(const char *s, size_t length, PL_blob_t *type, int *new)
       }
     } else
     { for(a = atomTable[v]; a; a = a->next)
-      { DEBUG(0, cmps++);
+      { DEBUG(MSG_HASH_STAT, cmps++);
 
 	if ( length == a->length &&
 	     type == a->type &&
 	     s == a->name )
 	{
 #ifdef O_ATOMGC
+#ifdef ATOMIC_REFERENCES
+	  if ( ATOMIC_INC(&a->references) == 1 )
+	    ATOMIC_DEC(&GD->atoms.unregistered);
+#else
 	  if ( a->references++ == 0 )
 	    GD->atoms.unregistered--;
+#endif
 #endif
           UNLOCK();
 	  *new = FALSE;
@@ -399,14 +435,22 @@ lookupBlob(const char *s, size_t length, PL_blob_t *type, int *new)
     }
   }
 
-  oldheap = GD->statistics.heap;
-  a = allocHeap(sizeof(struct atom));
+  a = allocHeapOrHalt(sizeof(struct atom));
   a->length = length;
   a->type = type;
   if ( false(type, PL_BLOB_NOCOPY) )
-  { a->name = allocHeap(length+1);
-    memcpy(a->name, s, length);
-    a->name[length] = EOS;
+  { if ( true(type, PL_BLOB_TEXT) )
+    { size_t pad = paddingBlob(type);
+
+      a->name = PL_malloc_atomic(length+pad);
+      memcpy(a->name, s, length);
+      memset(a->name+length, 0, pad);
+      GD->statistics.atom_string_space += length+pad;
+    } else
+    { a->name = PL_malloc(length);
+      memcpy(a->name, s, length);
+      GD->statistics.atom_string_space += length;
+    }
   } else
   { a->name = (char *)s;
   }
@@ -436,8 +480,6 @@ lookupBlob(const char *s, size_t length, PL_blob_t *type, int *new)
   if ( atom_buckets * 2 < GD->statistics.atoms )
     rehashAtoms();
 
-  GD->statistics.atomspace += (GD->statistics.heap - oldheap);
-
   UNLOCK();
 
   *new = TRUE;
@@ -513,7 +555,7 @@ _PL_debug_atom_value(atom_t a)
     Sdprintf("*** No atom at index (#%d) ***", i);
     trap_gdb();
 
-    atom = allocHeap(sizeof(*atom));
+    atom = allocHeapOrHalt(sizeof(*atom));
     Ssprintf(buf, "***(#%d)***", i);
     atom->name = store_string(buf);
     atom->length = strlen(atom->name);
@@ -553,9 +595,9 @@ To be used after loading the program,   so we won't traverse the program
 atoms each pass.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-void
-lockAtoms()
-{ GD->atoms.builtin      = GD->atoms.count;
+static void
+lockAtoms(void)
+{ GD->atoms.builtin      = GD->atoms.highest;
   GD->atoms.unregistered = 0;
 }
 
@@ -568,12 +610,12 @@ markAtom(atom_t a)
 { size_t i = indexAtom(a);
   Atom ap;
 
-  if ( i >= GD->atoms.count )
+  if ( i >= GD->atoms.highest )
     return;				/* not an atom */
   if ( i < GD->atoms.builtin )
     return;				/* locked range */
 
-  ap = GD->atoms.array[i];
+  ap = fetchAtomArray(i);
 
   if ( ap )
   {
@@ -581,7 +623,11 @@ markAtom(atom_t a)
     if ( atomLogFd )
       Sfprintf(atomLogFd, "Marked `%s' at (#%d)\n", ap->name, i);
 #endif
+#ifdef ATOMIC_REFERENCES
+    ATOMIC_OR(&ap->references, ATOM_MARKED_REFERENCE);
+#else
     ap->references |= ATOM_MARKED_REFERENCE;
+#endif
   }
 }
 
@@ -594,7 +640,7 @@ know we trapped a bug.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static int
-destroyAtom(Atom *ap, uintptr_t mask ARG_LD)
+destroyAtom(Atom *ap, uintptr_t mask)
 { Atom a = *ap;
   Atom *ap2 = &atomTable[a->hash_value & mask];
 
@@ -607,7 +653,7 @@ destroyAtom(Atom *ap, uintptr_t mask ARG_LD)
   }
 
 #if 0
-  if ( strncmp(a->name, "xxxx", 4) == 0 ) 	/* (*) see above */
+  if ( strncmp(a->name, "xxxx", 4) == 0 )	/* (*) see above */
   { Sdprintf("Deleting %s\n", a->name);
     assert(0);
   }
@@ -631,7 +677,11 @@ destroyAtom(Atom *ap, uintptr_t mask ARG_LD)
 
   *ap = NULL;			/* delete from index array */
   if ( false(a->type, PL_BLOB_NOCOPY) )
-    freeHeap(a->name, a->length+1);
+  { size_t slen = a->length+paddingBlob(a->type);
+    GD->statistics.atom_string_space -= slen;
+    GD->statistics.atom_string_space_freed += slen;
+    PL_free(a->name);
+  }
   freeHeap(a, sizeof(*a));
 
   return TRUE;
@@ -640,53 +690,69 @@ destroyAtom(Atom *ap, uintptr_t mask ARG_LD)
 
 static size_t
 collectAtoms(void)
-{ GET_LD
-  Atom *ap0 = GD->atoms.array;
-  Atom *ap  = ap0 + GD->atoms.builtin;
-  Atom *ep  = ap0 + GD->atoms.count;
-  int hole_seen = FALSE;
-  uintptr_t mask = atom_buckets-1;
+{ int hole_seen = FALSE;
   size_t reclaimed = 0;
+  size_t unregistered = 0;
+  size_t index;
+  int i, last=FALSE;
 
-  ap--;
-  while(++ap < ep)
-  { Atom a = *ap;
+  for(index=GD->atoms.builtin, i=MSB(index); !last; i++)
+  { size_t upto = (size_t)2<<i;
+    Atom *b = GD->atoms.array.blocks[i];
 
-    if ( !a )
-    { if ( !hole_seen )
-      { hole_seen = TRUE;
-	GD->atoms.no_hole_before = ap-ap0;
-      }
-      continue;
+    if ( upto >= GD->atoms.highest )
+    { upto = GD->atoms.highest;
+      last = TRUE;
     }
 
-    if ( a->references == 0 )
-    { if ( destroyAtom(ap, mask PASS_LD) )
-      { reclaimed++;
-	if ( !hole_seen && *ap == NULL )
+    for(; index<upto; index++)
+    { Atom a = b[index];
+
+      if ( !a )
+      { if ( !hole_seen )
 	{ hole_seen = TRUE;
-	  GD->atoms.no_hole_before = ap-ap0;
+	  GD->atoms.no_hole_before = index;
 	}
+	continue;
+      }
+
+      if ( a->references == 0 )
+      { if ( destroyAtom(&b[index], atom_buckets-1) )
+	{ reclaimed++;
+	  if ( !hole_seen )
+	  { hole_seen = TRUE;
+	    GD->atoms.no_hole_before = index;
+	  }
+	}
+      } else
+      {
+#ifdef ATOMIC_REFERENCES
+	ATOMIC_AND(&a->references, ~ATOM_MARKED_REFERENCE);
+#else
+	a->references &= ~ATOM_MARKED_REFERENCE;
+#endif
+        if ( a->references == 0 )
+	  unregistered++;
       }
-    } else
-    { a->references &= ~ATOM_MARKED_REFERENCE;
     }
   }
 
+  GD->atoms.unregistered = GD->atoms.non_garbage = unregistered;
+
   return reclaimed;
 }
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 pl_garbage_collect_atoms() realised the atom   garbage  collector (AGC).
-This is a tricky beast that   needs  careful synchronisation with normal
-GC. These issues are described with enterGC() in pl-gc.c.
+
+Issues around the design of the atom  garbage collector are explained at
+the start of this file.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-word
-pl_garbage_collect_atoms()
+foreign_t
+pl_garbage_collect_atoms(void)
 { GET_LD
-  intptr_t oldheap, freed;
   int64_t oldcollected;
   int verbose;
   double t;
@@ -700,8 +766,8 @@ pl_garbage_collect_atoms()
   }
 
 #ifdef O_PLMT
-  if ( GD->gc.active ) 			/* GC in progress: delay */
-  { DEBUG(2, Sdprintf("GC active; delaying AGC\n"));
+  if ( GD->gc.active )			/* GC in progress: delay */
+  { DEBUG(MSG_AGC, Sdprintf("GC active; delaying AGC\n"));
     GD->gc.agc_waiting = TRUE;
     PL_UNLOCK(L_GC);
     succeed;
@@ -715,9 +781,7 @@ pl_garbage_collect_atoms()
 #ifdef O_DEBUG_ATOMGC
 /*
     Sdprintf("Starting ATOM-GC.  Stack:\n");
-    systemMode(TRUE);
-    backTrace(NULL, 5);
-    systemMode(FALSE);
+    PL_backtrace(5, 0);
 */
 #endif
     printMessage(ATOM_informational,
@@ -727,29 +791,27 @@ pl_garbage_collect_atoms()
 
   PL_LOCK(L_THREAD);
   PL_LOCK(L_AGC);
+  PL_LOCK(L_STOPTHEWORLD);
   LOCK();
+  GD->atoms.gc_active = TRUE;
   blockSignals(&set);
   t = CpuTime(CPU_USER);
   markAtomsOnStacks(LD);
 #ifdef O_PLMT
-  markAtomsThreads();
   forThreadLocalData(markAtomsOnStacks, 0);
+  markAtomsMessageQueues();
 #endif
   oldcollected = GD->atoms.collected;
-  oldheap = GD->statistics.heap;
   reclaimed = collectAtoms();
+  GD->atoms.gc_active = FALSE;
   GD->atoms.collected += reclaimed;
   GD->statistics.atoms -= reclaimed;
-  GD->atoms.unregistered -= reclaimed;
-  GD->atoms.non_garbage = GD->atoms.unregistered;
   t = CpuTime(CPU_USER) - t;
   GD->atoms.gc_time += t;
   GD->atoms.gc++;
-  freed = oldheap - GD->statistics.heap;
-  GD->statistics.atomspacefreed += freed;
-  GD->statistics.atomspace -= freed;
   unblockSignals(&set);
   UNLOCK();
+  PL_UNLOCK(L_STOPTHEWORLD);
   PL_UNLOCK(L_AGC);
   PL_UNLOCK(L_THREAD);
   gc_status.blocked--;
@@ -786,6 +848,29 @@ resetAtoms()
 {
 }
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+(un)register atoms. If  possible,  this   is  implemented  using  atomic
+operations. This should be safe because:
+
+    - When we register an atom, it must be referenced from somewhere
+      else.
+    - When we unregister an atom, it must have at least one reference.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static void
+register_atom(Atom p)
+{
+#ifdef ATOMIC_REFERENCES
+  if ( ATOMIC_INC(&p->references) == 1 )
+    ATOMIC_DEC(&GD->atoms.unregistered);
+#else
+  LOCK();
+  if ( p->references++ == 0 )
+    GD->atoms.unregistered--;
+  UNLOCK();
+#endif
+}
+
 
 void
 PL_register_atom(atom_t a)
@@ -794,17 +879,48 @@ PL_register_atom(atom_t a)
   size_t index = indexAtom(a);
 
   if ( index >= GD->atoms.builtin )
-  { Atom p;
+  { Atom p = fetchAtomArray(index);
 
-    LOCK();
-    p = GD->atoms.array[index];
-    if ( p->references++ == 0 )
-      GD->atoms.unregistered--;
-    UNLOCK();
+    register_atom(p);
   }
 #endif
 }
 
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Foreign code reduces the reference count. This is safe, unless we are in
+the following scenario:
+
+  - A threads has done its atom-marking during a GC and is continued.
+  - Now, it fetches an atom from foreign code and the foreign code calls
+    PL_unregister_atom() which drops the reference count to zero. We can
+    now get into the position where the atom is no longer accessible
+    from foreign code and has not be seen while marking atoms from the
+    stack.
+
+The locking version of this code  is   not  a  problem, as the reference
+count cannot be dropped as  long  as   AGC  is  running. In the unlocked
+version, we need  to  replace  1   by  ATOM_MARKED_REFERENCE  if  AGC is
+running. We can be a bit sloppy here:  if   we  do this while AGC is not
+running we merely prevent the atom to be  collected in the next AGC. The
+next AGC resets the flag  and  thus   the  atom  becomes a candidate for
+collection afterwards.  So, basically we must do something like this:
+
+  if ( agc_running )
+  { do
+    { unsigned int oldref = p->references;
+      unsigned int newref = oldref == 1 ? ATOM_MARKED_REFERENCE : oldref-1;
+    } while( !compare_and_swap(&p->references, oldref, newref) );
+  } else
+  { atomic_dec(&p->references);
+  }
+
+But, this fails because AGC might kick in between agc_running was tested
+FALSE the atomic decrement. This is  fixed   by  putting the atom we are
+unregistering  in  LD->atoms.unregistered  and  mark    this  atom  from
+markAtomsOnStacks().
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 void
 PL_unregister_atom(atom_t a)
 {
@@ -813,16 +929,39 @@ PL_unregister_atom(atom_t a)
 
   if ( index >= GD->atoms.builtin )
   { Atom p;
+    unsigned int refs;
 
+    p = fetchAtomArray(index);
+#ifdef ATOMIC_REFERENCES
+    if ( GD->atoms.gc_active )
+    { unsigned int oldref, newref;
+
+      do
+      { oldref = p->references;
+	newref = oldref == 1 ? ATOM_MARKED_REFERENCE : oldref-1;
+      } while( !COMPARE_AND_SWAP(&p->references, oldref, newref) );
+      refs = newref;
+
+      if ( newref == ATOM_MARKED_REFERENCE )
+	ATOMIC_INC(&GD->atoms.unregistered);
+    } else
+    { GET_LD
+
+      if ( LD )
+	LD->atoms.unregistering = a;
+      if ( (refs=ATOMIC_DEC(&p->references)) == 0 )
+	ATOMIC_INC(&GD->atoms.unregistered);
+    }
+#else
     LOCK();
-    p = GD->atoms.array[index];
-    if ( --p->references == 0 )
+    if ( (refs = --p->references) == 0 )
       GD->atoms.unregistered++;
-    if ( p->references == (unsigned)-1 )
+    UNLOCK();
+#endif
+    if ( refs == (unsigned int)-1 )
     { Sdprintf("OOPS: -1 references to '%s'\n", p->name);
       trap_gdb();
     }
-    UNLOCK();
   }
 #endif
 }
@@ -835,34 +974,43 @@ PL_unregister_atom(atom_t a)
 		 *******************************/
 
 static void
-rehashAtoms()
-{ GET_LD
-  Atom *oldtab   = atomTable;
+rehashAtoms(void)
+{ Atom *oldtab   = atomTable;
   int   oldbucks = atom_buckets;
-  size_t mx = GD->atoms.count;
   uintptr_t mask;
-  Atom *ap, *ep;
+  size_t index;
+  int i, last=FALSE;
 
-  startCritical;
   atom_buckets *= 2;
   mask = atom_buckets-1;
-  atomTable = allocHeap(atom_buckets * sizeof(Atom));
+  atomTable = allocHeapOrHalt(atom_buckets * sizeof(Atom));
   memset(atomTable, 0, atom_buckets * sizeof(Atom));
 
-  DEBUG(0, Sdprintf("rehashing atoms (%d --> %d)\n", oldbucks, atom_buckets));
+  DEBUG(MSG_HASH_STAT,
+	Sdprintf("rehashing atoms (%d --> %d)\n", oldbucks, atom_buckets));
 
-  for(ap = GD->atoms.array, ep = ap+mx;
-      ap < ep;
-      ap++)
-  { Atom a = *ap;
-    size_t v = a->hash_value & mask;
+  for(index=1, i=0; !last; i++)
+  { size_t upto = (size_t)2<<i;
+    Atom *b = GD->atoms.array.blocks[i];
 
-    a->next = atomTable[v];
-    atomTable[v] = a;
+    if ( upto >= GD->atoms.highest )
+    { upto = GD->atoms.highest;
+      last = TRUE;
+    }
+
+    for(; index<upto; index++)
+    { Atom a = b[index];
+
+      if ( a && true(a->type, PL_BLOB_UNIQUE) )
+      { size_t v = a->hash_value & mask;
+
+	a->next = atomTable[v];
+	atomTable[v] = a;
+      }
+    }
   }
 
   freeHeap(oldtab, oldbucks * sizeof(Atom));
-  endCritical;
 }
 
 
@@ -882,15 +1030,15 @@ pl_atom_hashstat(term_t idx, term_t n)
 
 
 static void
-registerBuiltinAtoms()
-{ GET_LD
-  int size = sizeof(atoms)/sizeof(char *) - 1;
-  Atom a = allocHeap(size * sizeof(struct atom));
+registerBuiltinAtoms(void)
+{ int size = sizeof(atoms)/sizeof(char *) - 1;
+  Atom a;
   const ccharp *s;
 
+  GD->atoms.builtin_array = PL_malloc(size * sizeof(struct atom));
   GD->statistics.atoms = size;
 
-  for(s = atoms; *s; s++, a++)
+  for(s = atoms, a = GD->atoms.builtin_array; *s; s++, a++)
   { size_t len = strlen(*s);
     unsigned int v0 = MurmurHashAligned2(*s, len, MURMUR_SEED);
     unsigned int v = v0 & (atom_buckets-1);
@@ -913,8 +1061,11 @@ registerBuiltinAtoms()
 
 #if O_DEBUG
 static void
-exitAtoms(int status, void *arg)
-{ Sdprintf("hashstat: %d lookupAtom() calls used %d strcmp() calls\n",
+exitAtoms(int status, void *context)
+{ (void)status;
+  (void)context;
+
+  Sdprintf("hashstat: %d lookupAtom() calls used %d strcmp() calls\n",
 	   lookups, cmps);
 }
 #endif
@@ -923,14 +1074,13 @@ exitAtoms(int status, void *arg)
 void
 initAtoms(void)
 { LOCK();
-  if ( !atomTable )
-  { GET_LD
-    atom_buckets = ATOMHASHSIZE;
-    atomTable = allocHeap(atom_buckets * sizeof(Atom));
-
+  if ( !atomTable )			/* Atom hash table */
+  { atom_buckets = ATOMHASHSIZE;
+    atomTable = allocHeapOrHalt(atom_buckets * sizeof(Atom));
     memset(atomTable, 0, atom_buckets * sizeof(Atom));
-    GD->atoms.array_allocated = 4096;
-    GD->atoms.array = PL_malloc(GD->atoms.array_allocated*sizeof(Atom));
+
+    GD->atoms.highest = 1;
+    GD->atoms.no_hole_before = 1;
     registerBuiltinAtoms();
 #ifdef O_ATOMGC
     GD->atoms.margin = 10000;
@@ -938,21 +1088,81 @@ initAtoms(void)
 #endif
     PL_register_blob_type(&text_atom);
 
-    DEBUG(0, PL_on_halt(exitAtoms, NULL));
+    DEBUG(MSG_HASH_STAT, PL_on_halt(exitAtoms, NULL));
   }
   UNLOCK();
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+cleanupAtoms() is called at shutdown. There are three possible scenarios
+these days: (1) do not cleanup at  all, (2) cleanup the main structures,
+leaving the rest to GC or (3) cleanup the whole thing.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 void
 cleanupAtoms(void)
-{ PL_free(GD->atoms.array);
+{ int i;
+  int builtin_count = sizeof(atoms)/sizeof(char *) - 1;
+  Atom builtin_start = GD->atoms.builtin_array;
+  Atom builtin_end   = builtin_start+builtin_count;
+  Atom *ap0;
+
+  for(i=0; (ap0=GD->atoms.array.blocks[i]); i++)
+  { size_t bs = (size_t)1<<i;
+    size_t upto = (size_t)2<<i;
+    Atom *ap, *ep;
+
+    ap0 += bs;
+    ap = ap0;
+    ep = ap+bs;
+    if ( upto > GD->atoms.highest )
+      ep -= upto-GD->atoms.highest;
+
+    for(; ap<ep; ap++)
+    { if ( *ap )
+      { Atom a = *ap;
+
+	if ( !(a>=builtin_start && a<builtin_end) )
+	{ if ( a->type->release )
+	    (*a->type->release)(a->atom);
+	  else if ( GD->atoms.gc_hook )
+	    (*GD->atoms.gc_hook)(a->atom);
+
+	  if ( false(a->type, PL_BLOB_NOCOPY) )
+	    PL_free(a->name);
+	  freeHeap(a, sizeof(*a));
+	}
+      }
+    }
+
+    GD->atoms.array.blocks[i] = NULL;
+    PL_free(ap0);
+  }
+
+  PL_free(builtin_start);
+
+  for(i=0; i<256; i++)			/* char-code -> char-atom map */
+  { atom_t *p;
+
+    if ( (p=GD->atoms.for_code[i]) )
+    { GD->atoms.for_code[i] = NULL;
+      PL_free(p);
+    }
+  }
+
+  if ( atomTable )
+  { freeHeap(atomTable, atom_buckets * sizeof(Atom));
+    atomTable = NULL;
+  }
 }
 
 
 static word
-current_blob(term_t a, term_t type, frg_code call, intptr_t i ARG_LD)
+current_blob(term_t a, term_t type, frg_code call, intptr_t state ARG_LD)
 { atom_t type_name = 0;
+  size_t index;
+  int i, last=0;
 
   switch( call )
   { case FRG_FIRST_CALL:
@@ -967,12 +1177,13 @@ current_blob(term_t a, term_t type, frg_code call, intptr_t i ARG_LD)
 	succeed;
       }
       if ( !PL_is_variable(a) )
-	return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_atom, a);
+	return FALSE;
 
-      i = 0;
+      index = 1;
       break;
     }
     case FRG_REDO:
+      index = state;
       break;
     case FRG_CUTTED:
     default:
@@ -986,29 +1197,37 @@ current_blob(term_t a, term_t type, frg_code call, intptr_t i ARG_LD)
   }
 
   PL_LOCK(L_AGC);
-  for( ; i < (intptr_t)GD->atoms.count; i++ )
-  { Atom atom;
+  for(i=MSB(index); !last; i++)
+  { size_t upto = (size_t)2<<i;
+    Atom *b = GD->atoms.array.blocks[i];
 
-    if ( (atom = GD->atoms.array[i]) )
-    { if ( atom->atom == ATOM_garbage_collected )
-	continue;
+    if ( upto >= GD->atoms.highest )
+    { upto = GD->atoms.highest;
+      last = TRUE;
+    }
 
-      if ( type )
-      { if ( type_name && type_name != atom->type->atom_name )
+    for(; index<upto; index++)
+    { Atom atom = b[index];
+
+      if ( atom &&
+	   atom->atom != ATOM_garbage_collected )
+      { if ( type )
+	{ if ( type_name && type_name != atom->type->atom_name )
+	    continue;
+
+	  PL_unify_atom(type, atom->type->atom_name);
+	} else if ( false(atom->type, PL_BLOB_TEXT) )
 	  continue;
 
-	PL_unify_atom(type, atom->type->atom_name);
-      } else if ( false(atom->type, PL_BLOB_TEXT) )
-	continue;
-
-      PL_unify_atom(a, atom->atom);
-      PL_UNLOCK(L_AGC);
-      ForeignRedoInt(i+1);
+	PL_unify_atom(a, atom->atom);
+	PL_UNLOCK(L_AGC);
+	ForeignRedoInt(index+1);
+      }
     }
   }
   PL_UNLOCK(L_AGC);
 
-  fail;
+  return FALSE;
 }
 
 
@@ -1068,15 +1287,15 @@ PRED_IMPL("$atom_references", 2, atom_references, 0)
 #define stringMatch(m)	((m)->name->name)
 
 typedef struct match
-{ Atom	name;
-  int	length;
+{ Atom		name;
+  size_t	length;
 } *Match;
 
 
 static bool
-allAlpha(const char *s)
-{ for( ; *s; s++)
-  { if ( !isAlpha(*s) )
+allAlpha(const char *s, size_t len)
+{ for( ; len-- > 0; s++)
+  { if ( !*s || !isAlpha(*s) )
       fail;
   }
   succeed;
@@ -1085,28 +1304,39 @@ allAlpha(const char *s)
 
 static int
 extendAtom(char *prefix, bool *unique, char *common)
-{ intptr_t i, mx = GD->atoms.count;
-  Atom a;
+{ size_t index;
+  int i, last=FALSE;
   bool first = TRUE;
-  int lp = (int) strlen(prefix);
-  Atom *ap = GD->atoms.array;
+  size_t lp = strlen(prefix);
 
   *unique = TRUE;
 
-  for(i=0; i<mx; i++, ap++)
-  { if ( (a=*ap) && strprefix(a->name, prefix) )
-    { if ( strlen(a->name) >= LINESIZ )
-	continue;
-      if ( first == TRUE )
-      { strcpy(common, a->name+lp);
-	first = FALSE;
-      } else
-      { char *s = common;
-	char *q = a->name+lp;
-	while( *s && *s == *q )
-	  s++, q++;
-	*s = EOS;
-	*unique = FALSE;
+  for(index=1, i=0; !last; i++)
+  { size_t upto = (size_t)2<<i;
+    Atom *b = GD->atoms.array.blocks[i];
+
+    if ( upto >= GD->atoms.highest )
+    { upto = GD->atoms.highest;
+      last = TRUE;
+    }
+
+    for(; index<upto; index++)
+    { Atom a = b[index];
+
+      if ( a && a->type == &text_atom &&
+	   strprefix(a->name, prefix) &&
+	   strlen(a->name) < LINESIZ )
+      { if ( first == TRUE )
+	{ strcpy(common, a->name+lp);
+	  first = FALSE;
+	} else
+	{ char *s = common;
+	  char *q = a->name+lp;
+	  while( *s && *s == *q )
+	    s++, q++;
+	  *s = EOS;
+	  *unique = FALSE;
+	}
       }
     }
   }
@@ -1148,29 +1378,37 @@ compareMatch(const void *m1, const void *m2)
 
 static bool
 extend_alternatives(char *prefix, struct match *altv, int *altn)
-{ intptr_t i, mx = GD->atoms.count;
-  Atom a, *ap;
-  char *as;
-  int l;
+{ size_t index;
+  int i, last=FALSE;
 
   *altn = 0;
-  ap = GD->atoms.array;
-  for(i=0; i<mx; i++, ap++)
-  { if ( !(a=*ap) )
-      continue;
+  for(index=1, i=0; !last; i++)
+  { size_t upto = (size_t)2<<i;
+    Atom *b = GD->atoms.array.blocks[i];
 
-    as = a->name;
-    if ( strprefix(as, prefix) &&
-	 allAlpha(as) &&
-	 (l = (int)strlen(as)) < ALT_SIZ )
-    { Match m = &altv[(*altn)++];
-      m->name = a;
-      m->length = l;
-      if ( *altn > ALT_MAX )
-	break;
+    if ( upto >= GD->atoms.highest )
+    { upto = GD->atoms.highest;
+      last = TRUE;
+    }
+
+    for(; index<upto; index++)
+    { Atom a = b[index];
+
+      if ( a && a->type == &text_atom &&
+	   strprefix(a->name, prefix) &&
+	   a->length < ALT_SIZ &&
+	   allAlpha(a->name, a->length) )
+      { Match m = &altv[(*altn)++];
+
+	m->name = a;
+	m->length = a->length;
+	if ( *altn > ALT_MAX )
+	  goto out;
+      }
     }
   }
 
+out:
   qsort(altv, *altn, sizeof(struct match), compareMatch);
 
   succeed;
@@ -1254,8 +1492,8 @@ alnum_text(PL_chars_t *txt)
 static int
 atom_generator(PL_chars_t *prefix, PL_chars_t *hit, int state)
 { GET_LD
-  intptr_t i, mx = GD->atoms.count;
-  Atom *ap;
+  size_t index;
+  int i, last=FALSE;
 
 #ifdef O_PLMT
   if ( !key )
@@ -1263,37 +1501,43 @@ atom_generator(PL_chars_t *prefix, PL_chars_t *hit, int state)
 #endif
 
   if ( !state )
-    i = 0;
-  else
+  { index = 1;
+  } else
   {
 #ifdef O_PLMT
-    i = (intptr_t)pthread_getspecific(key);
+    index = (size_t)pthread_getspecific(key);
 #else
-    i = LD->atoms.generator;
+    index = LD->atoms.generator;
 #endif
   }
 
-  for(ap=&GD->atoms.array[i]; i<mx; i++, ap++)
-  { Atom a;
+  for(i=MSB(index); !last; i++)
+  { size_t upto = (size_t)2<<i;
+    Atom *b = GD->atoms.array.blocks[i];
 
-    if ( !(a = *ap) )
-      continue;
+    if ( upto >= GD->atoms.highest )
+    { upto = GD->atoms.highest;
+      last = TRUE;
+    }
 
-    if ( is_signalled() )		/* Notably allow windows version */
-      PL_handle_signals();		/* to break out on ^C */
+    for(; index<upto; index++)
+    { Atom a = b[index];
 
-    if ( get_atom_ptr_text(a, hit) &&
+      if ( is_signalled() )		/* Notably allow windows version */
+	PL_handle_signals();		/* to break out on ^C */
+
+      if ( a && get_atom_ptr_text(a, hit) &&
 	 hit->length < ALT_SIZ &&
 	 PL_cmp_text(prefix, 0, hit, 0, prefix->length) == 0 &&
 	 alnum_text(hit) )
-    {
+      {
 #ifdef O_PLMT
-      pthread_setspecific(key, (void *)(i+1));
+        pthread_setspecific(key, (void *)(index+1));
 #else
-      LD->atoms.generator = i+1;
+        LD->atoms.generator = index+1;
 #endif
-
-      return TRUE;
+        return TRUE;
+      }
     }
   }
 
