@@ -1,11 +1,9 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2010, VU University Amsterdam
+    Copyright (C): 2010-2012, VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,7 +17,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include "pl-incl.h"
@@ -37,6 +35,7 @@ typedef struct recref
 static int
 write_clause_ref(IOSTREAM *s, atom_t aref, int flags)
 { clref *ref = PL_blob_data(aref, NULL, NULL);
+  (void)flags;
 
   Sfprintf(s, "<clause>(%p)", ref->clause);
   return TRUE;
@@ -57,9 +56,8 @@ release_clause(atom_t aref)
 
   clear(ref->clause, DBREF_CLAUSE);
   if ( true(ref->clause, DBREF_ERASED_CLAUSE) )
-  { GET_LD
-    unallocClause(ref->clause PASS_LD);
-  }
+    unallocClause(ref->clause);
+
   return TRUE;
 }
 
@@ -67,6 +65,7 @@ release_clause(atom_t aref)
 static int
 save_clause_ref(atom_t aref, IOSTREAM *fd)
 { clref *ref = PL_blob_data(aref, NULL, NULL);
+  (void)fd;
 
   return PL_warning("Cannot save reference to <clause>(%p)", ref->clause);
 }
@@ -74,7 +73,9 @@ save_clause_ref(atom_t aref, IOSTREAM *fd)
 
 static atom_t
 load_clause_ref(IOSTREAM *fd)
-{ return PL_new_atom("<saved-clause-ref>");
+{ (void)fd;
+
+  return PL_new_atom("<saved-clause-ref>");
 }
 
 
@@ -113,11 +114,9 @@ release_record(atom_t aref)
 { recref *ref = PL_blob_data(aref, NULL, NULL);
 
   if ( ref->record->record )
-  { clear(ref->record->record, R_DBREF);
-  } else
-  { GET_LD
-    unallocRecordRef(ref->record PASS_LD);
-  }
+    clear(ref->record->record, R_DBREF);
+  else
+    unallocRecordRef(ref->record);
 
   return TRUE;
 }
@@ -179,7 +178,7 @@ PL_get_dbref(term_t t, db_ref_type *type_ptr)
   if ( type == &clause_blob )
   { clref *ref = data;
 
-    if ( false(ref->clause, ERASED) )
+    if ( false(ref->clause, CL_ERASED) )
     { *type_ptr = DB_REF_CLAUSE;
       return ref->clause;
     }
@@ -199,6 +198,12 @@ PL_get_dbref(term_t t, db_ref_type *type_ptr)
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Returns FALSE: error
+         TRUE: existing clause
+           -1: erased clause
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 int
 PL_get_clref(term_t t, Clause *cl)
 { struct clref *ref;
@@ -208,10 +213,11 @@ PL_get_clref(term_t t, Clause *cl)
        type != &clause_blob )
     return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_db_reference, t);
 
-  if ( true(ref->clause, ERASED) )
-    return FALSE;
-
   *cl = ref->clause;
+
+  if ( true(ref->clause, CL_ERASED) )
+    return -1;
+
   return TRUE;
 }
 
