@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 /*#define O_DEBUG 1*/			/* include crash/0 */
@@ -53,7 +52,7 @@ Link all foreign language predicates.  The arguments to FRG are:
 Flags almost always is TRACE_ME.  Additional common flags:
 
 	P_TRANSPARENT		Predicate is module transparent
-	NONDETERMINISTIC	Predicate can be resatisfied
+	P_NONDET	Predicate can be resatisfied
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #define NOTRACE PL_FA_NOTRACE
@@ -69,7 +68,6 @@ static const PL_extension foreigns[] = {
 #if O_DEBUG
   FRG("crash",			0, pl_crash,			0),
 #endif
-  FRG("nl",			0, pl_nl,			ISO),
 #ifdef __WINDOWS__
   FRG("win_exec",		2, pl_win_exec,			0),
   FRG("win_module_file",	2, pl_win_module_file,		0),
@@ -83,14 +81,14 @@ static const PL_extension foreigns[] = {
   FRG("sub_atom",		5, pl_sub_atom,		 NDET|ISO),
   FRG("sleep",			1, pl_sleep,			0),
   FRG("break",			0, pl_break,			0),
-  FRG("$break",			1, pl_break1,			0),
   FRG("notrace",		1, pl_notrace1,		     META),
 
   FRG("write_canonical",	1, pl_write_canonical,	      ISO),
-  FRG("write_term",		2, pl_write_term,	      ISO),
-  FRG("write_term",		3, pl_write_term3,	      ISO),
+  FRG("write_term",		2, pl_write_term,	 META|ISO),
+  FRG("write_term",		3, pl_write_term3,	 META|ISO),
   FRG("write",			1, pl_write,		      ISO),
   FRG("writeq",			1, pl_writeq,		      ISO),
+  FRG("writeln",		1, pl_writeln,		        0),
   FRG("print",			1, pl_print,			0),
 
   FRG("read_term",		2, pl_read_term,	      ISO),
@@ -114,27 +112,21 @@ static const PL_extension foreigns[] = {
   FRG("$rc_save_archive",	2, pl_rc_save_archive,		0),
   FRG("$rc_append_file",	5, pl_rc_append_file,		0),
 
-  FRG("abolish",    		1, pl_abolish1,		     META|ISO),
-  FRG("abolish",    		2, pl_abolish,		     META),
-  FRG("nth_clause", 		3, pl_nth_clause,       NDET|META|CREF),
+  FRG("abolish",		1, pl_abolish1,		     META|ISO),
+  FRG("abolish",		2, pl_abolish,		     META),
+  FRG("nth_clause",		3, pl_nth_clause,       NDET|META|CREF),
   FRG("retractall",		1, pl_retractall,	 META|ISO),
 #ifdef O_MAINTENANCE
   FRG("$list_generations",	1, pl_list_generations,	     META),
   FRG("$check_procedure",	1, pl_check_procedure,	     META),
 #endif
 
-  FRG("$term_complexity",	3, pl_term_complexity,		0),
-  FRG("redefine_system_predicate", 1, pl_redefine_system_predicate,
-							     META),
-
   FRG("$c_current_predicate",	2, pl_current_predicate,  NDET|META),
   FRG("current_predicate",	1, pl_current_predicate1, NDET|META|ISO),
   FRG("$set_predicate_attribute", 3, pl_set_predicate_attribute,META),
   FRG("$get_predicate_attribute", 3, pl_get_predicate_attribute,META),
-  FRG("$get_clause_attribute",  3, pl_get_clause_attribute,	0),
   FRG("$require",		1, pl_require,		     META),
   FRG("source_file",		2, pl_source_file,      NDET|META),
-  FRG("$start_consult",		1, pl_start_consult,		0),
   FRG("$make_system_source_files",0,pl_make_system_source_files,0),
   FRG("$default_predicate",	2, pl_default_predicate,     META),
 
@@ -164,28 +156,14 @@ static const PL_extension foreigns[] = {
 #ifdef O_PROLOG_HOOK
   FRG("set_prolog_hook",	3, pl_set_prolog_hook,	        0),
 #endif
-  FRG("$current_module",	2, pl_current_module,	     NDET),
   FRG("$module",		2, pl_module,			0),
   FRG("$set_source_module",	2, pl_set_source_module,	0),
   FRG("context_module",		1, pl_context_module,	     META),
-  FRG("import",			1, pl_import,		     META),
-  FRG("index",			1, pl_index,		     META),
-  FRG("hash",			1, pl_hash,		     META),
-
-#if O_DDE
-  FRG("open_dde_conversation",	3, pl_open_dde_conversation,	0),
-  FRG("close_dde_conversation",	1, pl_close_dde_conversation,	0),
-  FRG("dde_request",		4, pl_dde_request,		0),
-  FRG("dde_execute",		3, pl_dde_execute,		0),
-  FRG("dde_poke",		4, pl_dde_poke,			0),
-  FRG("$dde_register_service",	2, pl_dde_register_service,	0),
-#endif /*O_DDE*/
 
 #if O_STRING
   FRG("sub_string",		5, pl_sub_string,	     NDET),
 #endif /* O_STRING */
 
-  FRG("$length",		2, pl_length,			0),
   FRG("format",			2, pl_format,		     META),
 #ifdef O_DEBUG
   FRG("$check_definition",	1, pl_check_definition,      META),
@@ -204,7 +182,6 @@ static const PL_extension foreigns[] = {
 #endif
   FRG("current_flag",		1, pl_current_flag,	     NDET),
 
-  FRG("nl",			1, pl_nl1,		      ISO),
   FRG("read",			2, pl_read2,		      ISO),
   FRG("write",			2, pl_write2,		      ISO),
   FRG("writeq",			2, pl_writeq2,		      ISO),
@@ -212,22 +189,13 @@ static const PL_extension foreigns[] = {
   FRG("write_canonical",	2, pl_write_canonical2,	      ISO),
   FRG("format",			3, pl_format3,		     META),
 
-  FRG("tty_get_capability",	3, pl_tty_get_capability,	0),
-  FRG("tty_goto",		2, pl_tty_goto,			0),
-  FRG("tty_put",		2, pl_tty_put,			0),
-  FRG("tty_size",		2, pl_tty_size,			0),
   FRG("format_predicate",	2, pl_format_predicate,	     META),
   FRG("current_format_predicate", 2, pl_current_format_predicate,
 						        META|NDET),
   FRG("get_time",		1, pl_get_time,			0),
-#if O_PROLOG_FUNCTIONS
-  FRG("current_arithmetic_function", 1, pl_current_arithmetic_function,
-							NDET|META),
-#endif
 
 #ifdef O_PLMT
   FRG("thread_create",		3, pl_thread_create,	 META|ISO),
-  FRG("thread_join",		2, pl_thread_join,	      ISO),
   FRG("thread_exit",		1, pl_thread_exit,		0),
   FRG("thread_signal",		2, pl_thread_signal,	 META|ISO),
   FRG("thread_at_exit",		1, pl_thread_at_exit,	     META),
@@ -356,18 +324,16 @@ registerBuiltins(const PL_extension *f)
 
     PL_unregister_atom(name);
     def = lookupProcedure(fdef, m)->definition;
-    set(def, FOREIGN|SYSTEM|HIDE_CHILDS|LOCKED);
+    set(def, P_FOREIGN|HIDE_CHILDS|P_LOCKED);
 
     if ( f->flags & PL_FA_NOTRACE )	     clear(def, TRACE_ME);
     if ( f->flags & PL_FA_TRANSPARENT )	     set(def, P_TRANSPARENT);
-    if ( f->flags & PL_FA_NONDETERMINISTIC ) set(def, NONDETERMINISTIC);
+    if ( f->flags & PL_FA_NONDETERMINISTIC ) set(def, P_NONDET);
     if ( f->flags & PL_FA_VARARGS )	     set(def, P_VARARG);
     if ( f->flags & PL_FA_CREF )	     set(def, P_FOREIGN_CREF);
     if ( f->flags & PL_FA_ISO )		     set(def, P_ISO);
 
-    def->definition.function = f->function;
-    def->indexPattern = 0;
-    def->indexCardinality = 0;
+    def->impl.function = f->function;
     createForeignSupervisor(def, f->function);
   }
 }
@@ -378,14 +344,18 @@ registerBuiltins(const PL_extension *f)
 #define REG_PLIST(id) \
 	registerBuiltins(PL_predicates_from_ ## id)
 
+DECL_PLIST(alloc);
 DECL_PLIST(atom);
 DECL_PLIST(arith);
 DECL_PLIST(bag);
 DECL_PLIST(comp);
 DECL_PLIST(flag);
+DECL_PLIST(index);
 DECL_PLIST(list);
 DECL_PLIST(module);
 DECL_PLIST(prims);
+DECL_PLIST(variant);
+DECL_PLIST(copyterm);
 DECL_PLIST(prologflag);
 DECL_PLIST(trace);
 DECL_PLIST(pro);
@@ -411,6 +381,9 @@ DECL_PLIST(system);
 DECL_PLIST(op);
 DECL_PLIST(rec);
 DECL_PLIST(termhash);
+DECL_PLIST(dde);
+DECL_PLIST(term);
+DECL_PLIST(debug);
 
 void
 initBuildIns(void)
@@ -418,14 +391,18 @@ initBuildIns(void)
   Module m = MODULE_system;
 
   registerBuiltins(foreigns);
+  REG_PLIST(alloc);
   REG_PLIST(atom);
   REG_PLIST(arith);
   REG_PLIST(bag);
   REG_PLIST(comp);
   REG_PLIST(flag);
+  REG_PLIST(index);
   REG_PLIST(list);
   REG_PLIST(module);
   REG_PLIST(prims);
+  REG_PLIST(variant);
+  REG_PLIST(copyterm);
   REG_PLIST(prologflag);
   REG_PLIST(trace);
   REG_PLIST(pro);
@@ -447,6 +424,7 @@ initBuildIns(void)
   REG_PLIST(system);
   REG_PLIST(op);
   REG_PLIST(rec);
+  REG_PLIST(term);
   REG_PLIST(termhash);
 #ifdef O_ATTVAR
   REG_PLIST(attvar);
@@ -456,7 +434,9 @@ initBuildIns(void)
 #endif
 #ifdef __WINDOWS__
   REG_PLIST(win);
+  REG_PLIST(dde);
 #endif
+  REG_PLIST(debug);
 
 #define LOOKUPPROC(name) \
 	GD->procedures.name = lookupProcedure(FUNCTOR_ ## name, m);
@@ -485,7 +465,32 @@ initBuildIns(void)
 	PL_predicate("prolog_exception_hook", 4, "user");
 					/* allow debugging in call/1 */
   clear(PROCEDURE_dcall1->definition, HIDE_CHILDS|TRACE_ME);
-  set(PROCEDURE_dcall1->definition, DYNAMIC|SYSTEM);
+  set(PROCEDURE_dcall1->definition, P_DYNAMIC|P_LOCKED);
+
+  PL_meta_predicate(PL_predicate("assert",           1, "system"), ":");
+  PL_meta_predicate(PL_predicate("asserta",          1, "system"), ":");
+  PL_meta_predicate(PL_predicate("assertz",          1, "system"), ":");
+  PL_meta_predicate(PL_predicate("assert",           2, "system"), ":-");
+  PL_meta_predicate(PL_predicate("asserta",          2, "system"), ":-");
+  PL_meta_predicate(PL_predicate("assertz",          2, "system"), ":-");
+  PL_meta_predicate(PL_predicate("retract",          1, "system"), ":");
+  PL_meta_predicate(PL_predicate("retractall",       1, "system"), ":");
+  PL_meta_predicate(PL_predicate("clause",           2, "system"), ":?");
+
+  PL_meta_predicate(PL_predicate("format",           2, "system"), "+:");
+  PL_meta_predicate(PL_predicate("format",           3, "system"), "++:");
+  PL_meta_predicate(PL_predicate("format_predicate", 2, "system"), "+0");
+
+  PL_meta_predicate(PL_predicate("notrace",          1, "system"), "0");
+  PL_meta_predicate(PL_predicate("with_mutex",       2, "system"), "+0");
+  PL_meta_predicate(PL_predicate("with_output_to",   2, "system"), "+0");
+#ifdef O_PLMT
+  PL_meta_predicate(PL_predicate("thread_create",    3, "system"), "0?+");
+  PL_meta_predicate(PL_predicate("thread_at_exit",   1, "system"), "0");
+  PL_meta_predicate(PL_predicate("thread_signal",    2, "system"), "+0");
+#endif
+  PL_meta_predicate(PL_predicate("prolog_frame_attribute", 3, "system"), "++:");
+  PL_meta_predicate(PL_predicate("compile_predicates", 1, "system"), ":");
 
   for( ecell = ext_head; ecell; ecell = ecell->next )
     bindExtensions(ecell->module, ecell->extensions);
