@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,10 +18,9 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
-/*#define O_SECURE 1*/
 /*#define O_DEBUG 1*/
 #include "pl-incl.h"
 #include "os/pl-ctype.h"
@@ -65,16 +63,16 @@ number of term-refs allocated. This  information   (stored  as  a tagged
 Prolog int) is used by the garbage collector to update the stack frames.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#if O_SECURE
-#define setHandle(h, w)		{ assert(*valTermRef(h) != QID_MAGIC); \
-				  (*valTermRef(h) = (w)); \
-				}
-#else
-#define setHandle(h, w)		(*valTermRef(h) = (w))
+#if O_DEBUG || defined(O_MAINTENANCE)
+#ifndef O_CHECK_TERM_REFS
+#define O_CHECK_TERM_REFS 1
 #endif
+#endif
+
+#define setHandle(h, w)		(*valTermRef(h) = (w))
 #define valHandleP(h)		valTermRef(h)
 
-static int	PL_unify_int64__LD(term_t t, int64_t i, int ex ARG_LD);
+static int	unify_int64_ex__LD(term_t t, int64_t i, int ex ARG_LD);
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Deduce the value to store a copy of the  contents of p. This is a *very*
@@ -101,23 +99,44 @@ linkVal__LD(Word p ARG_LD)
     w = *p;
   }
 
-  SECURE(assert(w != ATOM_garbage_collected));
+  DEBUG(CHK_SECURE, assert(w != ATOM_garbage_collected));
 
   return w;
 }
 
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+term_t pushWordAsTermRef(Word p)
+       popTermRef()
+
+These two functions are used to create a term-ref from a `Word'. This is
+typically needed for calling  PL_error().  In   many  cases  there is no
+foreign  environment  around,  which   makes    that   we   cannot  call
+PL_new_term_ref(). These functions use the   tmp-references, shared with
+PushPtr()/PopPtr() (see pl-incl.h).  Push and pop *must* match.
+
+Note that this protects creating a term-ref  if there is no environment.
+However, the function called still must   either not use term-references
+or must create an environment.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 term_t
-wordToTermRef(Word p)
-{ GET_LD
+pushWordAsTermRef__LD(Word p ARG_LD)
+{ int i = LD->tmp.top++;
+  term_t t = LD->tmp.h[i];
 
-  if ( p > (Word) lBase )
-    return p - (Word)lBase;
-  else
-  { term_t t = PL_new_term_ref();
+  assert(i<TMP_PTR_SIZE);
+  setHandle(t, linkVal(p));
 
-    setHandle(t, linkVal(p));
-    return t;
-  }
+  return t;
+}
+
+void
+popTermRef__LD(ARG1_LD)
+{ int i = --LD->tmp.top;
+
+  assert(i>=0);
+  setVar(*valTermRef(LD->tmp.h[i]));
 }
 
 
@@ -130,9 +149,10 @@ PL_new_term_refs__LD(int n ARG_LD)
 { Word t;
   term_t r;
   int i;
+  FliFrame fr;
 
   if ( addPointer(lTop, n*sizeof(word)) > (void*) lMax )
-  { int rc = ensureLocalSpace(sizeof(word), ALLOW_SHIFT);
+  { int rc = ensureLocalSpace(n*sizeof(word), ALLOW_SHIFT);
 
     if ( rc != TRUE )
     { raiseStackOverflow(rc);
@@ -146,7 +166,13 @@ PL_new_term_refs__LD(int n ARG_LD)
   for(i=0; i<n; i++)
     setVar(*t++);
   lTop = (LocalFrame)t;
-  fli_context->size += n;
+  fr = fli_context;
+  fr->size += n;
+#ifdef O_CHECK_TERM_REFS
+  { int s = (int)((Word) lTop - (Word)(fr+1));
+    assert(s == fr->size);
+  }
+#endif
 
   return r;
 }
@@ -160,15 +186,16 @@ new_term_ref(ARG1_LD)
 
   t = (Word)lTop;
   r = consTermRef(t);
-  SECURE(assert(*t != QID_MAGIC));
-  setVar(*t);
+  setVar(*t++);
 
-  lTop = (LocalFrame)(t+1);
+  lTop = (LocalFrame)t;
   fr = fli_context;
   fr->size++;
-  SECURE({ int s = (Word) lTop - (Word)(fr+1);
-	   assert(s == fr->size);
-	 });
+#ifdef O_CHECK_TERM_REFS
+  { int s = (int)((Word) lTop - (Word)(fr+1));
+    assert(s == fr->size);
+  }
+#endif
 
   return r;
 }
@@ -253,8 +280,8 @@ PL_reset_term_refs(term_t r)
 
   lTop = (LocalFrame) valTermRef(r);
   fr->size = (int)((Word) lTop - (Word)addPointer(fr, sizeof(struct fliFrame)));
-  SECURE(if ( fr->size < 0 || fr->size > 100 )
-	   Sdprintf("Suspect foreign frame size: %d\n", fr->size));
+  DEBUG(CHK_SECURE, if ( fr->size < 0 || fr->size > 100 )
+		      Sdprintf("Suspect foreign frame size: %d\n", fr->size));
 }
 
 
@@ -282,9 +309,10 @@ PL_copy_term_ref(term_t from)
   lTop = (LocalFrame)(t+1);
   fr = fli_context;
   fr->size++;
-  SECURE({ int s = (Word) lTop - (Word)(fr+1);
-	   assert(s == fr->size);
-	 });
+  DEBUG(CHK_SECURE,
+	{ int s = (Word) lTop - (Word)(fr+1);
+	  assert(s == fr->size);
+	});
 
   return r;
 }
@@ -493,11 +521,14 @@ compareUCSAtom(atom_t h1, atom_t h2)
 
   for( ; len-- > 0; s1++, s2++)
   { if ( *s1 != *s2 )
-      return *s1 - *s2;
+    { int d = *s1 - *s2;
+
+      return d<0 ? CMP_LESS : d>0 ? CMP_GREATER : CMP_EQUAL;
+    }
   }
 
-  return a1->length >  a2->length ? 1 :
-	 a1->length == a2->length ? 0 : -1;
+  return a1->length >  a2->length ? CMP_GREATER :
+	 a1->length == a2->length ? CMP_EQUAL : CMP_LESS;
 }
 
 
@@ -745,7 +776,7 @@ PL_cvt_i_address(term_t p, void *address)
 bool
 PL_cvt_o_int64(int64_t c, term_t p)
 { GET_LD
-  return PL_unify_int64__LD(p, c, TRUE PASS_LD);
+  return unify_int64_ex__LD(p, c, TRUE PASS_LD);
 }
 
 
@@ -1419,7 +1450,7 @@ PL_get_float(term_t t, double *f)
 }
 
 
-#ifdef __WINDOWS__
+#ifdef _MSC_VER
 #define ULL(x) x ## ui64
 #else
 #define ULL(x) x ## ULL
@@ -1964,7 +1995,7 @@ PL_unify_string_nchars(term_t t, size_t len, const char *s)
 
 
 		 /*******************************
-		 *             PUT-*  		*
+		 *             PUT-*		*
 		 *******************************/
 
 int
@@ -2002,6 +2033,15 @@ PL_put_atom(term_t t, atom_t a)
 #define PL_put_atom(t, a) PL_put_atom__LD(t, a PASS_LD)
 
 
+int
+PL_put_bool(term_t t, int val)
+{ GET_LD
+
+  PL_put_atom__LD(t, val ? ATOM_true : ATOM_false PASS_LD);
+  return TRUE;
+}
+
+
 int
 PL_put_atom_chars(term_t t, const char *s)
 { GET_LD
@@ -2572,7 +2612,7 @@ PL_unify_chars(term_t t, int flags, size_t len, const char *s)
 
 
 static int
-PL_unify_int64__LD(term_t t, int64_t i, int ex ARG_LD)
+unify_int64_ex__LD(term_t t, int64_t i, int ex ARG_LD)
 { word w = consInt(i);
   Word p = valHandleP(t);
 
@@ -2610,13 +2650,19 @@ PL_unify_int64__LD(term_t t, int64_t i, int ex ARG_LD)
 
 int
 PL_unify_int64_ex__LD(term_t t, int64_t i ARG_LD)
-{ return PL_unify_int64__LD(t, i, TRUE PASS_LD);
+{ return unify_int64_ex__LD(t, i, TRUE PASS_LD);
+}
+
+
+int
+PL_unify_int64__LD(term_t t, int64_t i ARG_LD)
+{ return unify_int64_ex__LD(t, i, FALSE PASS_LD);
 }
 
 
 int
 PL_unify_integer__LD(term_t t, intptr_t i ARG_LD)
-{ return PL_unify_int64__LD(t, i, FALSE PASS_LD);
+{ return unify_int64_ex__LD(t, i, FALSE PASS_LD);
 }
 
 
@@ -2624,24 +2670,24 @@ PL_unify_integer__LD(term_t t, intptr_t i ARG_LD)
 int
 PL_unify_integer(term_t t, intptr_t i)
 { GET_LD
-  return PL_unify_int64__LD(t, i, FALSE PASS_LD);
+  return unify_int64_ex__LD(t, i, FALSE PASS_LD);
 }
 #define PL_unify_integer(t, i)	PL_unify_integer__LD(t, i PASS_LD)
 
-
+#undef PL_unify_int64
 int
 PL_unify_int64(term_t t, int64_t i)
 { GET_LD
 
-  return PL_unify_int64__LD(t, i, FALSE PASS_LD);
+  return unify_int64_ex__LD(t, i, FALSE PASS_LD);
 }
-
+#define PL_unify_int64(t, i)	PL_unify_int64__LD(t, i PASS_LD)
 
 int
 PL_unify_pointer__LD(term_t t, void *ptr ARG_LD)
 { uint64_t i = pointerToInt(ptr);
 
-  return PL_unify_int64__LD(t, (int64_t)i, FALSE PASS_LD);
+  return unify_int64_ex__LD(t, (int64_t)i, FALSE PASS_LD);
 }
 
 
@@ -2816,7 +2862,8 @@ PL_unify_termv(term_t t, va_list args)
   int rval;
   int op;
 
-  t = PL_copy_term_ref(t);
+  if ( !(t = PL_copy_term_ref(t)) )
+    return FALSE;
   initBuffer(&buf);
 
 cont:
@@ -2901,7 +2948,7 @@ cont:
       rval = PL_unify_text(t, 0, &txt,
 			   op == PL_NUTF8_CHARS ? PL_ATOM :
 			   op == PL_NUTF8_CODES ? PL_CODE_LIST :
-			   			  PL_STRING);
+						  PL_STRING);
       PL_free_text(&txt);
 
       break;
@@ -2967,7 +3014,8 @@ cont:
 	goto failout;
 
       w.type  = w_term;
-      w.value.term.term  = PL_copy_term_ref(t);
+      if ( !(w.value.term.term  = PL_copy_term_ref(t)) )
+	return FALSE;
       w.value.term.arg   = 0;
       w.value.term.arity = arity;
       addBuffer(&buf, w, work);
@@ -2981,7 +3029,8 @@ cont:
     { work w;
 
       w.type = w_list;
-      w.value.list.tail = PL_copy_term_ref(t);
+      if ( !(w.value.list.tail = PL_copy_term_ref(t)) )
+	return FALSE;
       w.value.list.len  = va_arg(args, int);
 
       addBuffer(&buf, w, work);
@@ -3269,7 +3318,8 @@ PL_strip_module__LD(term_t raw, module_t *m, term_t plain ARG_LD)
   { if ( *m == NULL )
       *m = environment_frame ? contextModule(environment_frame)
 			     : MODULE_user;
-    setHandle(plain, needsRef(*p) ? makeRef(p) : *p);
+    if ( raw != plain )
+      setHandle(plain, needsRef(*p) ? makeRef(p) : *p);
   }
 
   succeed;
@@ -3283,6 +3333,35 @@ PL_strip_module(term_t raw, module_t *m, term_t plain)
 }
 #define PL_strip_module(q, m, t) PL_strip_module__LD(q, m, t PASS_LD)
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+PL_strip_module_ex() is similar to  PL_strip_module(),   but  returns an
+error if it encounters a term <m>:<t>, where <m> is not an atom.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+int
+PL_strip_module_ex__LD(term_t raw, module_t *m, term_t plain ARG_LD)
+{ Word p = valTermRef(raw);
+
+  deRef(p);
+  if ( hasFunctor(*p, FUNCTOR_colon2) )
+  { p = stripModule(p, m PASS_LD);
+    if ( hasFunctor(*p, FUNCTOR_colon2) )
+    { Word a1 = argTermP(*p, 0);
+      deRef(a1);
+      setHandle(plain, needsRef(*a1) ? makeRef(a1) : *a1);
+      return PL_type_error("module", plain);
+    }
+    setHandle(plain, linkVal(p));
+  } else
+  { if ( *m == NULL )
+      *m = environment_frame ? contextModule(environment_frame)
+			     : MODULE_user;
+    setHandle(plain, needsRef(*p) ? makeRef(p) : *p);
+  }
+
+  return TRUE;
+}
+
 module_t
 PL_context()
 { GET_LD
@@ -3300,6 +3379,21 @@ PL_new_module(atom_t name)
 { return lookupModule(name);
 }
 
+int
+PL_qualify(term_t raw, term_t qualified)
+{ GET_LD
+  Module m = NULL;
+  term_t mname;
+
+  if ( !(mname = PL_new_term_ref()) ||
+       !PL_strip_module(raw, &m, qualified) )
+    return FALSE;
+
+  setHandle(mname, m->name);
+
+  return PL_cons_functor(qualified, FUNCTOR_colon2, mname, qualified);
+}
+
 
 		 /*******************************
 		 *	    PREDICATES		*
@@ -3345,20 +3439,16 @@ _PL_predicate(const char *name, int arity, const char *module,
 
 int
 PL_predicate_info(predicate_t pred, atom_t *name, int *arity, module_t *m)
-{ if ( pred->type == PROCEDURE_TYPE )
-  { Definition def = pred->definition;
+{ Definition def = pred->definition;
 
-    if ( name )
-      *name  = def->functor->name;
-    if ( arity )
-      *arity = def->functor->arity;
-    if ( m )
-      *m     = def->module;
+  if ( name )
+    *name  = def->functor->name;
+  if ( arity )
+    *arity = def->functor->arity;
+  if ( m )
+    *m     = def->module;
 
-    succeed;
-  }
-
-  fail;
+  return TRUE;
 }
 
 
@@ -3427,6 +3517,9 @@ int
 PL_raise_exception(term_t exception)
 { GET_LD
 
+  if ( PL_is_variable(exception) )
+    fatalError("Cannot throw variable exception");
+
   LD->exception.processing = TRUE;
   if ( !PL_same_term(exception, exception_bin) ) /* re-throwing */
   { setVar(*valTermRef(exception_bin));
@@ -3506,7 +3599,7 @@ notify_registered_foreign(functor_t fd, Module m)
 }
 
 
-static bool
+static predicate_t
 bindForeign(Module m, const char *name, int arity, Func f, int flags)
 { GET_LD
   Procedure proc;
@@ -3517,39 +3610,38 @@ bindForeign(Module m, const char *name, int arity, Func f, int flags)
   aname = PL_new_atom(name);
 
   fdef = lookupFunctorDef(aname, arity);
-  proc = lookupProcedure(fdef, m);
+  if ( !(proc = lookupProcedureToDefine(fdef, m)) )
+  { warning("PL_register_foreign(): attempt to redefine "
+	    "a system predicate: %s:%s",
+	    PL_atom_chars(m->name), functorName(fdef));
+    return NULL;
+  }
   def = proc->definition;
-
-  if ( true(def, LOCKED) && !SYSTEM_MODE )
-  { warning("PL_register_foreign: attempt to redefine a system predicate: %s",
-	    procedureName(proc));
-    fail;
+  if ( def->module != m || def->impl.any )
+  { DEBUG(MSG_PROC, Sdprintf("Abolish %s from %s\n",
+			     procedureName(proc), PL_atom_chars(m->name)));
+    abolishProcedure(proc, m);
+    def = proc->definition;
   }
 
-  if ( def->definition.function )
-    warning("PL_register_foreign: redefined %s", procedureName(proc));
-  if ( false(def, FOREIGN) && def->definition.clauses != NULL )
-    abolishProcedure(proc, m);
+  if ( def->impl.any )
+    PL_linger(def->impl.any);
+  def->impl.function = f;
+  def->flags &= ~(P_DYNAMIC|P_THREAD_LOCAL|P_TRANSPARENT|P_NONDET|P_VARARG);
+  def->flags |= (P_FOREIGN|TRACE_ME);
 
-  def->definition.function = f;
-  def->indexPattern = 0;
-  def->indexCardinality = 0;
-  def->flags = FOREIGN|TRACE_ME;
-
-  if ( m == MODULE_system )
-    set(def, SYSTEM|HIDE_CHILDS);
-  else if ( SYSTEM_MODE )
-    set(def, SYSTEM|HIDE_CHILDS);
+  if ( m == MODULE_system || SYSTEM_MODE )
+    set(def, P_LOCKED|HIDE_CHILDS);
 
   if ( (flags & PL_FA_NOTRACE) )	  clear(def, TRACE_ME);
   if ( (flags & PL_FA_TRANSPARENT) )	  set(def, P_TRANSPARENT);
-  if ( (flags & PL_FA_NONDETERMINISTIC) ) set(def, NONDETERMINISTIC);
+  if ( (flags & PL_FA_NONDETERMINISTIC) ) set(def, P_NONDET);
   if ( (flags & PL_FA_VARARGS) )	  set(def, P_VARARG);
 
   createForeignSupervisor(def, f);
   notify_registered_foreign(fdef, m);
 
-  succeed;
+  return proc;
 }
 
 
@@ -3592,12 +3684,18 @@ PL_register_extensions(const PL_extension *e)
 }
 
 
-int
-PL_register_foreign_in_module(const char *module,
-			      const char *name, int arity, Func f, int flags)
+static int
+register_foreignv(const char *module,
+		  const char *name, int arity, Func f, int flags,
+		  va_list args)
 { if ( extensions_loaded )
   { Module m = resolveModule(module);
-    return bindForeign(m, name, arity, f, flags);
+    predicate_t p = bindForeign(m, name, arity, f, flags);
+
+    if ( p && (flags&PL_FA_META) )
+      PL_meta_predicate(p, va_arg(args, char*));
+
+    return (p != NULL);
   } else
   { PL_extension ext[2];
     ext->predicate_name = (char *)name;
@@ -3613,8 +3711,29 @@ PL_register_foreign_in_module(const char *module,
 
 
 int
-PL_register_foreign(const char *name, int arity, Func f, int flags)
-{ return PL_register_foreign_in_module(NULL, name, arity, f, flags);
+PL_register_foreign_in_module(const char *module,
+			      const char *name, int arity, Func f, int flags, ...)
+{ va_list args;
+  int rc;
+
+  va_start(args, flags);
+  rc = register_foreignv(module, name, arity, f, flags, args);
+  va_end(args);
+
+  return rc;
+}
+
+
+int
+PL_register_foreign(const char *name, int arity, Func f, int flags, ...)
+{ va_list args;
+  int rc;
+
+  va_start(args, flags);
+  rc = register_foreignv(NULL, name, arity, f, flags, args);
+  va_end(args);
+
+  return rc;
 }
 
 		    /* deprecated */
@@ -3641,7 +3760,13 @@ PL_toplevel(void)
 
 void
 PL_halt(int status)
-{ PL_cleanup(status);
+{ int reclaim_memory = FALSE;
+
+#if defined(GC_DEBUG) || defined(O_DEBUG)
+  reclaim_memory = TRUE;
+#endif
+
+  cleanupProlog(status, reclaim_memory);
 
   exit(status);
 }
@@ -3737,7 +3862,7 @@ struct abort_handle
 void
 PL_abort_hook(PL_abort_hook_t func)
 { GET_LD
-  AbortHandle h = (AbortHandle) allocHeap(sizeof(struct abort_handle));
+  AbortHandle h = (AbortHandle) allocHeapOrHalt(sizeof(struct abort_handle));
   h->next = NULL;
   h->function = func;
 
@@ -3914,9 +4039,12 @@ PL_dispatch_hook(PL_dispatch_hook_t hook)
 #include <sys/select.h>
 #endif
 
-#ifdef __WINDOWS__
-#include <winsock2.h>
-#endif
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Note that this is  used  to   integrate  X11  event-dispatching into the
+SWI-Prolog  toplevel.  Integration  of  event-handling   in  Windows  is
+achieved through the plterm DLL (see  win32/console). For this reason we
+do never want this code in Windows.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static int
 input_on_fd(int fd)
@@ -3981,9 +4109,7 @@ PL_recorded(record_t r, term_t t)
 
 void
 PL_erase(record_t r)
-{ GET_LD
-
-  freeRecord(r);
+{ freeRecord(r);
 }
 
 
@@ -3998,7 +4124,7 @@ PL_duplicate_record(record_t r)
 
 
 		 /*******************************
-		 *	    FEATURES		*
+		 *	   PROLOG FLAGS		*
 		 *******************************/
 
 int
@@ -4007,8 +4133,6 @@ PL_set_prolog_flag(const char *name, int type, ...)
   int rval = TRUE;
   int flags = (type & FF_MASK);
 
-  initPrologFlagTable();
-
   va_start(args, type);
   switch(type & ~FF_MASK)
   { case PL_BOOL:
@@ -4032,8 +4156,8 @@ PL_set_prolog_flag(const char *name, int type, ...)
     default:
       rval = FALSE;
   }
-
   va_end(args);
+
   return rval;
 }
 
@@ -4085,7 +4209,6 @@ PL_action(int action, ...)
 #ifdef O_DEBUGGER
     { GET_LD
       int a = va_arg(args, int);
-      int om;
 
       if ( gc_status.active )
       { Sfprintf(Serror,
@@ -4100,9 +4223,7 @@ PL_action(int action, ...)
 	rval = FALSE;
 	break;
       }
-      om = systemMode(TRUE);		/* Also show hidden frames */
-      backTrace(environment_frame, a);
-      systemMode(om);
+      PL_backtrace(a, 0);
     }
 #else
       warning("No Prolog backtrace in runtime version");
@@ -4120,7 +4241,7 @@ PL_action(int action, ...)
       break;
     }
     case PL_ACTION_ABORT:
-      rval = (int)abortProlog(ABORT_THROW);
+      rval = (int)abortProlog();
       break;
     case PL_ACTION_GUIAPP:
     { int guiapp = va_arg(args, int);
@@ -4184,13 +4305,12 @@ PL_action(int action, ...)
 static void
 init_c_args()
 { if ( c_argc == -1 )
-  { GET_LD
-    int i;
+  { int i;
     int opts = 1;
     int argc    = GD->cmdline.argc;
     char **argv = GD->cmdline.argv;
 
-    c_argv = allocHeap(argc * sizeof(char *));
+    c_argv = allocHeapOrHalt(argc * sizeof(char *));
     c_argv[0] = argv[0];
     c_argc = 1;
 
@@ -4308,7 +4428,7 @@ PL_license(const char *license, const char *module)
 
     PL_discard_foreign_frame(fid);
   } else
-  { struct license *l = allocHeap(sizeof(*l));
+  { struct license *l = allocHeapOrHalt(sizeof(*l));
 
     l->license_id = store_string(license);
     l->module_id  = store_string(module);
@@ -4320,8 +4440,7 @@ PL_license(const char *license, const char *module)
 
 void
 registerForeignLicenses(void)
-{ GET_LD
-  struct license *l, *next;
+{ struct license *l, *next;
 
   for(l=pre_registered; l; l=next)
   { next = l->next;
