@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2009, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #ifndef PL_GLOBAL_H_INCLUDED
@@ -65,9 +64,10 @@ typedef struct
 
 typedef struct
 { atom_t	file;			/* current source file */
-  int	  	line;			/* current line */
+  int		line;			/* current line */
   int		linepos;		/* position in the line */
   int64_t	character;		/* current character location */
+  int64_t	byte;			/* byte offset of location */
 } source_location;
 
 		 /*******************************
@@ -84,13 +84,14 @@ struct PL_global_data
   cleanup_status cleaning;		/* Inside PL_cleanup() */
   int		bootsession;		/* -b boot compilation */
   int		debug_level;		/* Maintenance debugging: 0..9 */
+  struct bit_vector *debug_topics;	/* debug topics enabled */
   void *	resourceDB;		/* program resource database */
 
 #ifdef HAVE_SIGNAL
   sig_handler sig_handlers[MAXSIGNAL];	/* How Prolog preceives signals */
 #endif
 #ifdef O_LOGICAL_UPDATE
-  uintptr_t generation;			/* generation of the database */
+  gen_t		generation;		/* generation of the database */
 #endif
 
   struct
@@ -103,25 +104,27 @@ struct PL_global_data
   } cmdline;
 
   struct
-  { char *	executable;		/* Running executable */
+  { char *	CWDdir;
+    size_t	CWDlen;
+    char *	executable;		/* Running executable */
 #ifdef __WINDOWS__
     char *	module;			/* argv[0] module passed */
 #endif
   } paths;
 
   struct
-  { size_t	heap;			/* heap in use */
-    size_t	atoms;			/* No. of atoms defined */
-    size_t	atomspace;		/* # bytes used to store atoms */
+  { size_t	atoms;			/* No. of atoms defined */
+    size_t	atom_string_space;	/* # bytes used to store atoms */
+    size_t	atom_string_space_freed;/* # bytes in freed atoms */
     size_t	stack_space;		/* # bytes on stacks */
-#ifdef O_ATOMGC
-    size_t	atomspacefreed;		/* Freed atom-space */
-#endif
     int		functors;		/* No. of functors defined */
     int		predicates;		/* No. of predicates defined */
     int		modules;		/* No. of modules in the system */
-    intptr_t	codes;			/* No. of byte codes generated */
+    size_t	clauses;		/* No. clauses */
+    size_t	codes;			/* No. of VM codes generated */
     double	start_time;		/* When Prolog was started */
+    double	user_cputime;		/* User CPU time (whole process) */
+    double	system_cputime;		/* Kernel CPU time (whole process) */
 #ifdef O_PLMT
     int		threads_created;	/* # threads created */
     int		threads_finished;	/* # finished threads */
@@ -145,16 +148,16 @@ struct PL_global_data
   } recorded_db;
 
   struct
-  { buffer	functions;		/* index --> function */
-    ArithFunction table[ARITHHASHSIZE];	/* functor --> function table */
+  { ArithF     *functions;		/* index --> function */
+    size_t	functions_allocated;	/* Size of above array */
   } arith;
 
   struct
-  { Atom *	array;			/* index --> atom */
-    size_t	count;			/* elements in array */
-    size_t	array_allocated;	/* allocated size of array */
+  { size_t	highest;		/* Highest atom index */
+    atom_array	array;
     unsigned int buckets;		/* # buckets in char * --> atom */
     Atom *	table;			/* hash-table */
+    Atom	builtin_array;		/* Builtin atoms */
     int		lookups;		/* # atom lookups */
     int		cmps;			/* # string compares for lookup */
 #ifdef O_ATOMGC
@@ -215,7 +218,8 @@ struct PL_global_data
   } prolog_flag;
 
   struct
-  { buffer	array;			/* index --> functor */
+  { size_t	highest;		/* Next index to handout */
+    functor_array array;		/* index --> functor */
     int		buckets;		/* # buckets in atom --> functor */
     FunctorDef* table;			/* hash-table */
   } functors;
@@ -234,7 +238,7 @@ struct PL_global_data
     int			halting;	/* process is shutting down */
     int			gui_app;	/* Win32: Application is a gui app */
     IOFUNCTIONS		iofunctions;	/* initial IO functions */
-    IOFUNCTIONS 	org_terminal;	/* IO+Prolog terminal functions */
+    IOFUNCTIONS		org_terminal;	/* IO+Prolog terminal functions */
     IOFUNCTIONS		rl_functions;	/* IO+Terminal+Readline functions */
   } os;
 
@@ -254,6 +258,7 @@ struct PL_global_data
     Procedure	prolog_trace_interception4;
     Procedure	portray;		/* portray/1 */
     Procedure   dcall1;			/* $call/1 */
+    Procedure   call3;			/* call/3*/
     Procedure	setup_call_catcher_cleanup4; /* setup_call_catcher_cleanup/4 */
     Procedure	undefinterc4;		/* $undefined_procedure/4 */
     Procedure   dthread_init0;		/* $thread_init/0 */
@@ -265,6 +270,7 @@ struct PL_global_data
 #ifdef O_CALL_RESIDUE
     Procedure	call_residue_vars2;	/* $attvar:call_residue_vars/2 */
 #endif
+    Procedure   comment_hook3;		/* prolog:comment_hook/3 */
 
     SourceFile  reloading;		/* source file we are re-loading */
     int		active_marked;		/* #prodedures marked active */
@@ -290,18 +296,14 @@ struct PL_global_data
   } terminal;
 #endif
 
-  struct alloc_pool alloc_pool;		/* Main allocation pool */
 #ifdef O_PLMT
-  FreeChunk	    left_over_pool;	/* Left-over from threads */
-
   struct
   { struct _at_exit_goal *exit_goals;	/* Global thread_at_exit/1 goals */
-    int		    	enabled;	/* threads are enabled */
+    int			enabled;	/* threads are enabled */
     Table		mutexTable;	/* Name --> mutex table */
     int			mutex_next_id;	/* next id for anonymous mutexes */
-    struct pl_mutex*	MUTEX_load;	/* The $load mutex */
 #ifdef __WINDOWS__
-    HINSTANCE	    	instance;	/* Win32 process instance */
+    HINSTANCE		instance;	/* Win32 process instance */
 #endif
     counting_mutex     *mutexes;	/* Registered mutexes */
     int			thread_max;	/* Maximum # threads */
@@ -331,7 +333,7 @@ struct PL_local_data
   uintptr_t	bases[STG_MASK+1];	/* area base addresses */
   int		alerted;		/* Special mode. See updateAlerted() */
   int		critical;		/* heap is being modified */
-  abort_type	aborted;		/* !ABORT_NONE: abort in Critical */
+  int		break_level;		/* current break level */
   Stack		outofstack;		/* thread is out of stack */
   int		trim_stack_requested;	/* perform a trim-stack */
 #ifdef O_PLMT
@@ -359,6 +361,12 @@ struct PL_local_data
     jmp_buf	context;		/* context of longjmp() */
   } pipe;
 
+  struct
+  { char       *getstr_buffer;		/* getString() buffer */
+    size_t	getstr_buffer_size;	/* size of getstr_buffer */
+    struct wic_state *current_state;	/* qlf-creation state */
+  } qlf;
+
   struct
   { atom_t	current;		/* current global prompt */
     atom_t	first;			/* how to prompt first line */
@@ -443,7 +451,8 @@ struct PL_local_data
   } modules;
 
   struct
-  { intptr_t 	generator;		/* See PL_atom_generator() */
+  { intptr_t	generator;		/* See PL_atom_generator() */
+    atom_t	unregistering;		/* See PL_unregister_atom() */
   } atoms;
 
   struct
@@ -452,10 +461,6 @@ struct PL_local_data
     int		filledVars;
   } comp;
 
-  struct
-  { int		active;
-  } read;
-
   struct
   { struct
     { Number	base;
@@ -478,12 +483,14 @@ struct PL_local_data
 #endif
 
   struct
-  { char *	_CWDdir;
-    size_t	_CWDlen;
+  {
 #ifdef __BEOS__
     status_t	dl_error;		/* dlopen() emulation in pl-beos.c */
 #endif
     int		rand_initialised;	/* have we initialised random? */
+#ifdef O_DDE
+    unsigned	dde_instance;		/* Actually DWORD */
+#endif
   } os;
 
   struct
@@ -491,17 +498,17 @@ struct PL_local_data
     pl_features_t mask;			/* Masked access to booleans */
     int		  write_attributes;	/* how to write attvars? */
     occurs_check_t occurs_check;	/* Unify and occurs check */
+    access_level_t access_level;	/* Current access level */
   } prolog_flag;
 
   struct
   { FindData	find;			/* /<ports> <goal> in tracer */
   } trace;
 
-#ifdef O_ATOMGC
   struct
   { struct findall_bag *bags;		/* Known bags */
+    struct findall_bag *default_bag;	/* Bag we keep around */
   } bags;
-#endif
 
   struct
   { AbortHandle	_abort_head;		/* PL_abort_hook() */
@@ -516,10 +523,16 @@ struct PL_local_data
 
   struct				/* Local IO stuff */
   { IOSTREAM *streams[6];		/* handles for standard streams */
+    st_check stream_type_check;		/* Check bin/text streams? */
+					/* do not copy from parent */
     struct input_context *input_stack;	/* maintain input stream info */
     struct output_context *output_stack; /* maintain output stream info */
   } IO;
 
+  struct
+  { fid_t	numbervars_frame;	/* Numbervars choice-point */
+  } var_names;
+
 #ifdef O_LIMIT_DEPTH
   struct
   { uintptr_t limit;
@@ -527,6 +540,24 @@ struct PL_local_data
   } depth_info;
 #endif
 
+  pl_shift_status_t shift_status;	/* Stack shifter status */
+  pl_debugstatus_t _debugstatus;	/* status of the debugger */
+  struct btrace *btrace_store;		/* C-backtraces */
+
+#ifdef O_PLMT
+  struct
+  { intptr_t   magic;			/* PL_THREAD_MAGIC (checking) */
+    struct _PL_thread_info_t *info;	/* info structure */
+    unsigned forall_flags;		/* forThreadLocalData() flags */
+					/* Communication */
+    message_queue messages;		/* Message queue */
+    struct _thread_sig   *sig_head;	/* Head of signal queue */
+    struct _thread_sig   *sig_tail;	/* Tail of signal queue */
+    struct _at_exit_goal *exit_goals;	/* thread_at_exit/1 goals */
+    DefinitionChain local_definitions;	/* P_THREAD_LOCAL predicates */
+  } thread;
+#endif
+
   struct
   { intptr_t _total_marked;		/* # marked global cells */
     intptr_t _trailcells_deleted;	/* # garbage trailcells */
@@ -542,40 +573,25 @@ struct PL_local_data
     intptr_t _choice_count;		/* choice-point count for debugging */
     int  *_start_map;			/* bitmap with legal global starts */
     sigset_t saved_sigmask;		/* Saved signal mask */
-#if defined(O_SECURE) || defined(SECURE_GC)
+    int64_t inferences;			/* #inferences at last GC */
+    pl_gc_status_t	status;		/* Garbage collection status */
+#ifdef O_CALL_RESIDUE
+    int			marked_attvars;	/* do not GC attvars */
+#endif
+    int active;				/* GC is running in this thread */
+					/* These must be at the end to be */
+					/* able to define O_DEBUG in only */
+					/* some modules */
+#if defined(O_DEBUG) || defined(SECURE_GC)
     intptr_t _trailtops_marked;		/* # marked trailtops */
     Word *_mark_base;			/* Array of marked cells addresses */
     Word *_mark_top;			/* Top of this array */
     Table _check_table;			/* relocation address table */
     Table _local_table;			/* marked local variables */
     int  _relocated_check;		/* Verify relocated addresses? */
-#endif
-    int64_t inferences;			/* #inferences at last GC */
-
-    pl_gc_status_t	status;		/* Garbage collection status */
-#ifdef O_CALL_RESIDUE
-    int			marked_attvars;	/* do not GC attvars */
+    unsigned int incr_seed;		/* Seed for random stack increments */
 #endif
   } gc;
-
-  pl_shift_status_t shift_status;	/* Stack shifter status */
-  pl_debugstatus_t _debugstatus;	/* status of the debugger */
-
-#ifdef O_PLMT
-  struct
-  { intptr_t   magic;			/* PL_THREAD_MAGIC (checking) */
-    struct _PL_thread_info_t *info;	/* info structure */
-    unsigned forall_flags;		/* forThreadLocalData() flags */
-					/* Communication */
-    message_queue messages;		/* Message queue */
-    struct _thread_sig   *sig_head;	/* Head of signal queue */
-    struct _thread_sig   *sig_tail;	/* Tail of signal queue */
-    struct _at_exit_goal *exit_goals;	/* thread_at_exit/1 goals */
-    DefinitionChain local_definitions;	/* P_THREAD_LOCAL predicates */
-  } thread;
-
-  struct alloc_pool alloc_pool;		/* Thread allocation pool */
-#endif
 };
 
 GLOBAL PL_global_data_t PL_global_data;
@@ -585,28 +601,19 @@ GLOBAL PL_local_data_t  PL_local_data;
 GLOBAL PL_local_data_t *PL_current_engine_ptr;
 #endif
 
-#if !defined(O_PLMT) && !defined(O_MULTIPLE_ENGINES)
-#define GET_LD
-#define ARG_LD
-#define ARG1_LD void
-#define PASS_LD
-#define PASS_LD1
-#define LOCAL_LD  (&PL_local_data)
-#define GLOBAL_LD (&PL_local_data)
-#define LD	  GLOBAL_LD
-#endif /*O_PLMT && O_MULTIPLE_ENGINES*/
 #define GD (&PL_global_data)
 #define CD (&PL_code_data)
 
 #define functor_array		(GD->functors.array)
 #define systemDefaults		(GD->defaults)
 
-#define environment_frame 	(LD->environment)
-#define fli_context	  	(LD->foreign_environment)
+#define environment_frame	(LD->environment)
+#define fli_context		(LD->foreign_environment)
 #define source_file_name	(LD->read_source.file)
 #define source_line_no		(LD->read_source.line)
 #define source_line_pos		(LD->read_source.linepos)
 #define source_char_no		(LD->read_source.character)
+#define source_byte_no		(LD->read_source.byte)
 #define exception_term		(LD->exception.term)
 #define exception_bin		(LD->exception.bin)
 #define exception_printed	(LD->exception.printed)
@@ -621,7 +628,7 @@ GLOBAL PL_local_data_t *PL_current_engine_ptr;
 #define Scurin			(LD->IO.streams[3])
 #define Scurout			(LD->IO.streams[4])
 #define Sprotocol		(LD->IO.streams[5])
-#define Sdin			Suser_input 		/* not used for now */
+#define Sdin			Suser_input		/* not used for now */
 #define Sdout			Suser_output
 
 #ifdef VMCODE_IS_ADDRESS
