@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        J.Wielemaker@uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2009, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include "pl-incl.h"
@@ -41,7 +40,15 @@ the  global  module  for  the  user  and imports from `system' all other
 modules import from `user' (and indirect from `system').
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static int addSuperModule_no_lock(Module m, Module s, int where);
+static int	addSuperModule_no_lock(Module m, Module s, int where);
+static void	unallocModule(Module m);
+
+static void
+unallocProcedureSymbol(Symbol s)
+{ DEBUG(MSG_CLEANUP,
+	Sdprintf("unallocProcedure(%s)\n", functorName((functor_t)s->name)));
+  unallocProcedure(s->value);
+}
 
 
 static Module
@@ -52,11 +59,9 @@ _lookupModule(atom_t name)
   if ((s = lookupHTable(GD->tables.modules, (void*)name)) != (Symbol) NULL)
     return (Module) s->value;
 
-  { GET_LD
-    m = allocHeap(sizeof(struct module));
-  }
+  m = allocHeapOrHalt(sizeof(struct module));
   m->name = name;
-  m->file = (SourceFile) NULL;
+  m->file = NULL;
   m->operators = NULL;
   m->level = 0;
 #ifdef O_PROLOG_HOOK
@@ -66,24 +71,28 @@ _lookupModule(atom_t name)
   m->mutex = allocSimpleMutex(PL_atom_chars(m->name));
 #endif
   clearFlags(m);
-  set(m, CHARESCAPE);
+  set(m, M_CHARESCAPE);
 
   if ( name == ATOM_user || name == ATOM_system )
     m->procedures = newHTable(PROCEDUREHASHSIZE);
   else
     m->procedures = newHTable(MODULEPROCEDUREHASHSIZE);
+  m->procedures->free_symbol = unallocProcedureSymbol;
 
   m->public = newHTable(PUBLICHASHSIZE);
-
   m->supers = NULL;
+  m->class  = ATOM_user;
+
   if ( name == ATOM_user )
   { super = MODULE_system;
   } else if ( name == ATOM_system )
-  { set(m, SYSTEM|UNKNOWN_ERROR);
+  { set(m, M_SYSTEM|UNKNOWN_ERROR);
     super = NULL;
+    m->class = ATOM_system;
   } else if ( stringAtom(name)[0] == '$' )
-  { set(m, SYSTEM);
+  { set(m, M_SYSTEM);
     super = MODULE_system;
+    m->class = ATOM_system;
   } else
   { super = MODULE_user;
   }
@@ -124,6 +133,12 @@ isCurrentModule(atom_t name)
 }
 
 
+static void
+unallocModuleSymbol(Symbol s)
+{ unallocModule(s->value);
+}
+
+
 void
 initModules(void)
 { LOCK();
@@ -136,6 +151,7 @@ initModules(void)
     initFunctors();
 
     GD->tables.modules = newHTable(MODULEHASHSIZE);
+    GD->tables.modules->free_symbol = unallocModuleSymbol;
     GD->modules.system = _lookupModule(ATOM_system);
     GD->modules.user   = _lookupModule(ATOM_user);
   }
@@ -143,6 +159,49 @@ initModules(void)
 }
 
 
+static void
+unallocList(ListCell c)
+{ ListCell n;
+
+  for(; c; c=n)
+  { n = c->next;
+
+    freeHeap(c, sizeof(*c));
+  }
+}
+
+static void
+unallocModule(Module m)
+{ if ( m->procedures ) destroyHTable(m->procedures);
+  if ( m->public )     destroyHTable(m->public);
+  if ( m->operators )  destroyHTable(m->operators);
+  if ( m->supers )     unallocList(m->supers);
+  if ( m->mutex )      freeSimpleMutex(m->mutex);
+
+  freeHeap(m, sizeof(*m));
+}
+
+
+static void
+emptyModule(Module m)
+{ DEBUG(MSG_CLEANUP, Sdprintf("emptyModule(%s)\n", PL_atom_chars(m->name)));
+  if ( m->procedures ) clearHTable(m->procedures);
+}
+
+
+void
+cleanupModules(void)
+{ Table t;
+
+  if ( (t=GD->tables.modules) )
+  { for_unlocked_table(t, s, emptyModule(s->value));
+
+    GD->tables.modules = NULL;
+    destroyHTable(t);
+  }
+}
+
+
 int
 isSuperModule(Module s, Module m)	/* s is a super-module of m */
 { ListCell c;
@@ -193,6 +252,7 @@ static int
 cannotSetSuperModule(Module m, Module s)
 { GET_LD
   term_t t = PL_new_term_ref();
+  (void)s;				/* would be nice to add to message */
 
   PL_put_atom(t, m->name);
 
@@ -224,8 +284,7 @@ reachableModule(Module here, Module end)
 
 static int
 addSuperModule_no_lock(Module m, Module s, int where)
-{ GET_LD
-  ListCell c;
+{ ListCell c;
 
   if ( reachableModule(s, m) )
     return cannotSetSuperModule(m, s);
@@ -235,7 +294,7 @@ addSuperModule_no_lock(Module m, Module s, int where)
       return TRUE;			/* already a super-module */
   }
 
-  c = allocHeap(sizeof(*c));
+  c = allocHeapOrHalt(sizeof(*c));
   c->value = s;
 
   if ( where == 'A' )
@@ -270,8 +329,7 @@ addSuperModule(Module m, Module s, int where)
 
 static int
 delSuperModule(Module m, Module s)
-{ GET_LD
-  ListCell *p;
+{ ListCell *p;
 
   for(p = &m->supers; *p; p = &(*p)->next)
   { ListCell c = *p;
@@ -290,9 +348,8 @@ delSuperModule(Module m, Module s)
 
 
 static void
-clearSupersModule(Module m)
-{ GET_LD
-  ListCell c = m->supers;
+clearSupersModule_no_lock(Module m)
+{ ListCell c = m->supers;
   ListCell next;
 
   m->supers = NULL;
@@ -304,6 +361,13 @@ clearSupersModule(Module m)
   m->level = 0;
 }
 
+void
+clearSupersModule(Module m)
+{ LOCK();
+  clearSupersModule_no_lock(m);
+  UNLOCK();
+}
+
 
 int
 setSuperModule(Module m, Module s)
@@ -318,28 +382,55 @@ setSuperModule(Module m, Module s)
       succeed;
     }
   }
-  clearSupersModule(m);
+  clearSupersModule_no_lock(m);
 
   return addSuperModule_no_lock(m, s, 'A');
 }
 
 
 static
-PRED_IMPL("set_base_module", 1, set_base_module, PL_FA_TRANSPARENT)
+PRED_IMPL("set_module", 1, set_module, PL_FA_TRANSPARENT)
 { PRED_LD
   Module m = MODULE_parse;
-  atom_t mname;
-  int rc;
+  term_t prop = PL_new_term_ref();
+  atom_t pname;
+  int arity;
 
-  PL_strip_module(A1, &m, A1);
-  if ( !PL_get_atom_ex(A1, &mname) )
-    fail;
+  PL_strip_module(A1, &m, prop);
+  if ( PL_get_name_arity(prop, &pname, &arity) && arity == 1 )
+  { term_t arg = PL_new_term_ref();
 
-  LOCK();
-  rc = setSuperModule(m, _lookupModule(mname));
-  UNLOCK();
+    _PL_get_arg(1, prop, arg);
 
-  return rc;
+    if ( pname == ATOM_base )
+    { atom_t mname;
+      int rc;
+
+      if ( !PL_get_atom_ex(arg, &mname) )
+	return FALSE;
+      LOCK();
+      rc = setSuperModule(m, _lookupModule(mname));
+      UNLOCK();
+      return rc;
+    } else if ( pname == ATOM_class )
+    { atom_t class;
+
+      if ( !PL_get_atom_ex(arg, &class) )
+	return FALSE;
+      if ( class == ATOM_user ||
+	   class == ATOM_system ||
+	   class == ATOM_library ||
+	   class == ATOM_test ||
+	   class == ATOM_development )
+      { m->class = class;
+	return TRUE;
+      } else
+	return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_module_class, arg);
+    } else
+    { return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_module_property, prop);
+    }
+  } else
+    return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_module_property, prop);
 }
 
 
@@ -374,7 +465,7 @@ getUnknownModule(Module m)
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 stripModule() takes an atom or term, possible embedded in the :/2 module
-term.  It will assing *module with the associated module and return  the
+term.  It will assign *module with the associated module and return  the
 remaining term.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
@@ -395,7 +486,7 @@ stripModule(Word term, Module *module ARG_LD)
 
   if ( ! *module )
     *module = (environment_frame ? contextModule(environment_frame)
-	       			 : MODULE_user);
+				 : MODULE_user);
 
   return term;
 }
@@ -530,96 +621,108 @@ get_existing_source_file(term_t file, SourceFile *sfp ARG_LD)
 }
 
 
-Module
-moduleFromFile(SourceFile sf)
-{ TableEnum e;
-  Symbol symb;
+/** '$current_module'(+Module, -File) is semidet.
+    '$current_module'(-ModuleOrList, +File) is semidet.
+    '$current_module'(-Module, -File) is nondet.
 
-  for(e = newTableEnum(GD->tables.modules),
-      symb = advanceTableEnum(e);
-      symb;
-      symb = advanceTableEnum(e))
-  { Module m = symb->value;
+Query module<->file association. This association  is N:1 in SWI-Prolog.
+Think e.g., of test-units that are mapped  to modules. When used in mode
+(-, +), this predicate unifies Module with  a non-empty list if the file
+is associated to multiple modules.
+*/
 
-    if ( m->file == sf )
-    { freeTableEnum(e);
-      return m;
-    }
-  }
-
-  freeTableEnum(e);
-  return NULL;
-}
-
-
-word
-pl_current_module(term_t module, term_t file, control_t h)
-{ GET_LD
-  TableEnum e = NULL;
+static
+PRED_IMPL("$current_module", 2, current_module, PL_FA_NONDETERMINISTIC)
+{ PRED_LD
+  TableEnum e;
   Symbol symb;
   atom_t name;
   SourceFile sf = NULL;
 
-  if ( ForeignControl(h) == FRG_CUTTED )
-  { e = ForeignContextPtr(h);
-    freeTableEnum(e);
-    succeed;
-  }
-				/* deterministic case: module --> file */
-  if ( PL_get_atom(module, &name) )
-  { Module m;
+  term_t module = A1;
+  term_t file   = A2;
 
-    if ( (m=isCurrentModule(name)) )
-    { atom_t f = (!m->file ? ATOM_nil : m->file->name);
-      return PL_unify_atom(file, f);
-    }
-
-    fail;
-  }
-
-  switch(ForeignControl(h))
+  switch(CTX_CNTRL)
   { case FRG_FIRST_CALL:
+				/* deterministic case: module --> file */
+      if ( PL_get_atom(module, &name) )
+      { Module m;
+
+	if ( (m=isCurrentModule(name)) )
+	{ atom_t f = (!m->file ? ATOM_nil : m->file->name);
+	  return PL_unify_atom(file, f);
+	}
+
+	return FALSE;
+      }
+
       if ( !get_existing_source_file(file, &sf PASS_LD) )
-	fail;				/* given, but non-existing file */
+	return FALSE;			/* given, but non-existing file */
+
+      if ( sf )
+      { if ( sf->modules )
+	{ int rc = FALSE;
+
+	  PL_LOCK(L_PREDICATE);
+	  if ( sf->modules->next )
+	  { term_t tail = PL_copy_term_ref(module);
+	    term_t head = PL_new_term_ref();
+	    ListCell c;
+
+	    for(c=sf->modules; c; c=c->next)
+	    { Module m = c->value;
+
+	      if ( !(PL_unify_list(tail, head, tail) &&
+		     PL_unify_atom(head, m->name)) )
+		goto out;
+	    }
+	    rc = PL_unify_nil(tail);
+	  } else
+	  { Module m = sf->modules->value;
+	    rc = PL_unify_atom(module, m->name);
+	  }
+
+	out:
+	  PL_UNLOCK(L_PREDICATE);
+	  return rc;
+	}
+	return FALSE;			/* source-file has no modules */
+      }
+
       e = newTableEnum(GD->tables.modules);
       break;
     case FRG_REDO:
-      e = ForeignContextPtr(h);
+      e = CTX_PTR;
       get_existing_source_file(file, &sf PASS_LD);
       break;
     case FRG_CUTTED:
+      e = CTX_PTR;
       freeTableEnum(e);
       succeed;
+    default:
+      assert(0);
+      return FALSE;
   }
 
+					/* mode (-,-) */
+
   while( (symb = advanceTableEnum(e)) )
   { Module m = symb->value;
+    atom_t f = ( !m->file ? ATOM_nil : m->file->name);
 
-    if ( stringAtom(m->name)[0] == '$' &&
+    if ( m->class == ATOM_system && m->name != ATOM_system &&
 	 !SYSTEM_MODE && PL_is_variable(module) )
       continue;
 
-    { fid_t cid = PL_open_foreign_frame();
-      atom_t f = ( !m->file ? ATOM_nil : m->file->name);
+    if ( PL_unify_atom(module, m->name) &&
+	 PL_unify_atom(file, f) )
+      ForeignRedoPtr(e);
 
-      if ( (!sf || (m->file == sf)) &&
-	   PL_unify_atom(module, m->name) &&
-	   PL_unify_atom(file, f) )
-      { PL_close_foreign_frame(cid);
-
-	if ( sf && sf->module_count == 1 )
-	{ freeTableEnum(e);
-	  succeed;
-	}
-	ForeignRedoPtr(e);
-      }
-
-      PL_discard_foreign_frame(cid);
-    }
+    break;				/* must be an error */
   }
 
   freeTableEnum(e);
-  fail;
+  return FALSE;
 }
 
 
@@ -716,7 +819,7 @@ find_modules_with_def(Module m, functor_t fdef,
   }
 
   if ( (proc = isCurrentProcedure(fdef, m)) &&
-       proc->definition->definition.clauses )
+       proc->definition->impl.any )
   { if ( !(PL_unify_list(t, h, t) &&
 	   PL_unify_atom(h, m->name)) )
       fail;
@@ -770,7 +873,7 @@ in it are abolished.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 int
-declareModule(atom_t name, atom_t super,
+declareModule(atom_t name, atom_t class, atom_t super,
 	      SourceFile sf, int line,
 	      int allow_newfile)
 { GET_LD
@@ -779,6 +882,8 @@ declareModule(atom_t name, atom_t super,
 
   LOCK();
   module = _lookupModule(name);
+  if ( class )
+    module->class = class;
 
   if ( !allow_newfile && module->file && module->file != sf)
   { term_t obj;
@@ -795,7 +900,7 @@ declareModule(atom_t name, atom_t super,
 
   if ( module->file != sf )
   { module->file = sf;
-    sf->module_count++;		/* current determinism in $current_module/2 */
+    addModuleSourceFile(sf, module);
   }
   module->line_no = line;
   LD->modules.source = module;
@@ -804,7 +909,7 @@ declareModule(atom_t name, atom_t super,
 	    { Procedure proc = s->value;
 	      Definition def = proc->definition;
 	      if ( /*def->module == module &&*/
-		   !true(def, DYNAMIC|MULTIFILE|FOREIGN) )
+		   !true(def, P_DYNAMIC|P_MULTIFILE|P_FOREIGN) )
 	      { if ( def->module == module &&
 		     hasClausesDefinition(def) )
 		{ if ( !rdef )
@@ -840,7 +945,7 @@ declareModule(atom_t name, atom_t super,
 }
 
 
-/** '$declare_module'(+Module, +Super, +File, +Line, +Redefine) is det.
+/** '$declare_module'(+Module, +Class, +Super, +File, +Line, +Redefine) is det.
 
 Start a new (source-)module
 
@@ -852,19 +957,21 @@ Start a new (source-)module
 */
 
 static
-PRED_IMPL("$declare_module", 5, declare_module, 0)
+PRED_IMPL("$declare_module", 6, declare_module, 0)
 { PRED_LD
   SourceFile sf;
-  atom_t mname, sname, fname;
+  atom_t mname, cname, sname, fname;
   int line_no, rdef;
 
   term_t module   = A1;
-  term_t super    = A2;
-  term_t file     = A3;
-  term_t line     = A4;
-  term_t redefine = A5;
+  term_t class    = A2;
+  term_t super    = A3;
+  term_t file     = A4;
+  term_t line     = A5;
+  term_t redefine = A6;
 
   if ( !PL_get_atom_ex(module, &mname) ||
+       !PL_get_atom_ex(class, &cname) ||
        !PL_get_atom_ex(super, &sname) ||
        !PL_get_atom_ex(file, &fname) ||
        !PL_get_integer_ex(line, &line_no) ||
@@ -872,7 +979,7 @@ PRED_IMPL("$declare_module", 5, declare_module, 0)
     fail;
 
   sf = lookupSourceFile(fname, TRUE);
-  return declareModule(mname, sname, sf, line_no, rdef);
+  return declareModule(mname, cname, sname, sf, line_no, rdef);
 }
 
 
@@ -912,7 +1019,10 @@ PRED_IMPL("$module_property", 2, module_property, 0)
 		    ATOM_module_property, A2);
 
   if ( PL_is_functor(A2, FUNCTOR_line_count1) )
-  { return PL_unify_integer(a, m->line_no);
+  { if ( m->line_no > 0 )
+      return PL_unify_integer(a, m->line_no);
+    else
+      fail;
   } else if ( PL_is_functor(A2, FUNCTOR_file1) )
   { if ( m->file )
       return PL_unify_atom(a, m->file->name);
@@ -920,6 +1030,8 @@ PRED_IMPL("$module_property", 2, module_property, 0)
       fail;
   } else if ( PL_is_functor(A2, FUNCTOR_exports1) )
   { return unify_export_list(a, m PASS_LD);
+  } else if ( PL_is_functor(A2, FUNCTOR_class1) )
+  { return PL_unify_atom(a, m->class);
   } else
     return PL_error(NULL, 0, NULL, ERR_DOMAIN,
 		    ATOM_module_property, A2);
@@ -1070,16 +1182,35 @@ fixExport(Definition old, Definition new)
 }
 
 
-word
-pl_import(term_t pred)
-{ GET_LD
-  Module source = NULL;
+int
+atomToImportStrength(atom_t a)
+{ if ( a == ATOM_weak )
+    return PROC_WEAK;
+  else if ( a == ATOM_strong )
+    return 0;
+  else
+    return -1;				/* domain error */
+}
+
+
+static int
+import(term_t pred, term_t strength ARG_LD)
+{ Module source = NULL;
   Module destination = contextModule(environment_frame);
   functor_t fd;
   Procedure proc, old;
+  int pflags = 0;
 
   if ( !get_functor(pred, &fd, &source, 0, GF_PROCEDURE) )
-    fail;
+    return FALSE;
+  if ( strength )
+  { atom_t a;
+
+    if ( !PL_get_atom_ex(strength, &a) )
+      return FALSE;
+    if ( (pflags=atomToImportStrength(a)) < 0 )
+      return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_import_type, strength);
+  }
 
   proc = lookupProcedure(fd, source);
 
@@ -1095,24 +1226,40 @@ pl_import(term_t pred)
     { Definition odef = old->definition;
 
       old->definition = proc->definition;
-      if ( true(odef, P_SHARED) )
+      shareDefinition(proc->definition);
+      if ( odef->shared > 1 )
 	fixExport(odef, proc->definition);
-      set(proc->definition, P_SHARED);
+      shareDefinition(odef);
+      GC_LINGER(odef);
+      set(old, pflags);
 
       succeed;
     }
 
     if ( old->definition->module == destination )
-      return warning("Cannot import %s into module %s: name clash",
-		     procedureName(proc),
-		     stringAtom(destination->name) );
+    { if ( (pflags & PROC_WEAK) )
+      { if ( truePrologFlag(PLFLAG_WARN_OVERRIDE_IMPLICIT_IMPORT) )
+	{ term_t pi = PL_new_term_ref();
 
-    if ( old->definition->module != source )
-    { warning("Cannot import %s into module %s: already imported from %s",
-	      procedureName(proc),
-	      stringAtom(destination->name),
-	      stringAtom(old->definition->module->name) );
-      fail;
+	  if ( !PL_unify_predicate(pi, proc, GP_NAMEARITY) )
+	    return FALSE;
+
+	  printMessage(ATOM_warning,
+		       PL_FUNCTOR_CHARS, "ignored_weak_import", 2,
+		         PL_ATOM, destination->name,
+		         PL_TERM, pi);
+	}
+
+	return TRUE;
+      } else
+	return PL_error("import", 1, "name clash", ERR_IMPORT_PROC,
+			proc, destination->name, 0);
+    }
+
+    if ( old->definition->module != source )	/* already imported */
+    { return PL_error("import", 1, NULL, ERR_IMPORT_PROC,
+		      proc, destination->name,
+		      old->definition->module->name);
     }
 
     sysError("Unknown problem importing %s into module %s",
@@ -1122,15 +1269,21 @@ pl_import(term_t pred)
   }
 
   if ( !isPublicModule(source, proc) )
-  { warning("import/1: %s is not declared public (still imported)",
-	    procedureName(proc));
+  { term_t pi = PL_new_term_ref();
+
+    if ( !PL_unify_predicate(pi, proc, GP_NAMEARITY) )
+      return FALSE;
+    printMessage(ATOM_warning,
+		 PL_FUNCTOR_CHARS, "import_private", 2,
+		   PL_ATOM, destination->name,
+		   PL_TERM, pi);
   }
 
-  { Procedure nproc = (Procedure)  allocHeap(sizeof(struct procedure));
+  { Procedure nproc = (Procedure)  allocHeapOrHalt(sizeof(struct procedure));
 
-    nproc->type = PROCEDURE_TYPE;
+    nproc->flags = pflags;
     nproc->definition = proc->definition;
-    set(proc->definition, P_SHARED);
+    shareDefinition(proc->definition);
 
     LOCKMODULE(destination);
     addHTable(destination->procedures,
@@ -1141,6 +1294,20 @@ pl_import(term_t pred)
   succeed;
 }
 
+static
+PRED_IMPL("import", 1, import, PL_FA_TRANSPARENT)
+{ PRED_LD
+
+  return import(A1, 0 PASS_LD);
+}
+
+static
+PRED_IMPL("$import", 2, import, PL_FA_TRANSPARENT)
+{ PRED_LD
+
+  return import(A1, A2 PASS_LD);
+}
+
 		 /*******************************
 		 *      PUBLISH PREDICATES	*
 		 *******************************/
@@ -1149,12 +1316,15 @@ BeginPredDefs(module)
   PRED_DEF("import_module", 2, import_module,
 	   PL_FA_NONDETERMINISTIC)
   PRED_DEF("$def_modules", 2, def_modules, PL_FA_TRANSPARENT)
-  PRED_DEF("$declare_module", 5, declare_module, 0)
+  PRED_DEF("$declare_module", 6, declare_module, 0)
   PRED_DEF("add_import_module", 3, add_import_module, 0)
   PRED_DEF("delete_import_module", 2, delete_import_module, 0)
-  PRED_DEF("set_base_module", 1, set_base_module, PL_FA_TRANSPARENT)
+  PRED_DEF("set_module", 1, set_module, PL_FA_TRANSPARENT)
+  PRED_DEF("$current_module", 2, current_module, PL_FA_NONDETERMINISTIC)
   PRED_DEF("$module_property", 2, module_property, 0)
   PRED_DEF("strip_module", 3, strip_module, PL_FA_TRANSPARENT)
+  PRED_DEF("import", 1, import, PL_FA_TRANSPARENT)
+  PRED_DEF("$import", 2, import, PL_FA_TRANSPARENT)
   PRED_DEF("export", 1, export, PL_FA_TRANSPARENT)
   PRED_DEF("$undefined_export", 2, undefined_export, 0)
 EndPredDefs
