@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include "pl-incl.h"
@@ -80,9 +80,8 @@ typedef struct _opdef			/* predefined and enumerated */
 
 static void
 copyOperatorSymbol(Symbol s)
-{ GET_LD
-  operator *op = s->value;
-  operator *o2 = allocHeap(sizeof(*o2));
+{ operator *op = s->value;
+  operator *o2 = allocHeapOrHalt(sizeof(*o2));
 
   *o2 = *op;
 }
@@ -90,17 +89,16 @@ copyOperatorSymbol(Symbol s)
 
 static void
 freeOperatorSymbol(Symbol s)
-{ GET_LD
-  operator *op = s->value;
+{ operator *op = s->value;
 
   PL_unregister_atom((atom_t) s->name);
   freeHeap(op, sizeof(*op));
 }
 
 
-Table
-newOperatorTable()
-{ Table t = newHTable(8);
+static Table
+newOperatorTable(int size)
+{ Table t = newHTable(size);
 
   t->copy_symbol = copyOperatorSymbol;
   t->free_symbol = freeOperatorSymbol;
@@ -118,9 +116,10 @@ Define operator in some table.  See   discussion  above for the relevant
 tables.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static void
-defOperator(Module m, atom_t name, int type, int priority)
-{ Symbol s;
+static int
+defOperator(Module m, atom_t name, int type, int priority, int force)
+{ GET_LD
+  Symbol s;
   operator *op;
   int t = (type & OP_MASK);		/* OP_PREFIX, ... */
 
@@ -132,19 +131,31 @@ defOperator(Module m, atom_t name, int type, int priority)
 
   assert(t>=OP_PREFIX && t<=OP_POSTFIX);
 
+  if ( !force && !SYSTEM_MODE )
+  { if ( name == ATOM_comma || name == ATOM_nil || name == ATOM_curl ||
+	 (name == ATOM_bar && ((t&OP_MASK) != OP_INFIX ||
+			       (priority < 1001 && priority != 0))) )
+    { GET_LD
+      atom_t action = (name == ATOM_comma ? ATOM_modify : ATOM_create);
+      term_t t = PL_new_term_ref();
+
+      PL_put_atom(t, name);
+      return PL_error(NULL, 0, NULL, ERR_PERMISSION,
+		      action, ATOM_operator, t);
+    }
+  }
+
   LOCK();
   if ( !m->operators )
-    m->operators = newOperatorTable();
+    m->operators = newOperatorTable(8);
 
   if ( (s = lookupHTable(m->operators, (void *)name)) )
   { op = s->value;
   } else if ( priority < 0 )
   { UNLOCK();				/* already inherited: do not change */
-    return;
+    return TRUE;
   } else
-  { GET_LD
-
-    op = allocHeap(sizeof(*op));
+  { op = allocHeapOrHalt(sizeof(*op));
 
     op->priority[OP_PREFIX]  = -1;
     op->priority[OP_INFIX]   = -1;
@@ -161,6 +172,8 @@ defOperator(Module m, atom_t name, int type, int priority)
     addHTable(m->operators, (void *)name, op);
   }
   UNLOCK();
+
+  return TRUE;
 }
 
 
@@ -288,7 +301,6 @@ atomToOperatorType(atom_t atom)
   else if (atom == ATOM_xfx)		return OP_XFX;
   else if (atom == ATOM_xfy)		return OP_XFY;
   else if (atom == ATOM_yfx)		return OP_YFX;
-  else if (atom == ATOM_yfy)		return OP_YFY;
   else if (atom == ATOM_yf)		return OP_YF;
   else if (atom == ATOM_xf)		return OP_XF;
 
@@ -301,7 +313,7 @@ operatorTypeToAtom(int type)
   { NULL_ATOM,
     ATOM_fx, ATOM_fy,
     ATOM_xf, ATOM_yf,
-    ATOM_xfx, ATOM_xfy, ATOM_yfx, ATOM_yfy
+    ATOM_xfx, ATOM_xfy, ATOM_yfx
   };
 
   return opnames[(type>>4)];
@@ -348,12 +360,7 @@ PRED_IMPL("op", 3, op, PL_FA_TRANSPARENT|PL_FA_ISO)
     return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_operator_specifier, type);
 
   if ( PL_get_atom(name, &nm) )
-  { if ( nm == ATOM_comma )
-    { protected:
-      return PL_error(NULL, 0, NULL, ERR_PERMISSION,
-		      ATOM_modify, ATOM_operator, name);
-    }
-    defOperator(m, nm, t, p);
+  { return defOperator(m, nm, t, p, FALSE);
   } else
   { term_t l = PL_copy_term_ref(name);
     term_t e = PL_new_term_ref();
@@ -361,9 +368,8 @@ PRED_IMPL("op", 3, op, PL_FA_TRANSPARENT|PL_FA_ISO)
     while( PL_get_list(l, e, l) )
     { if ( !PL_get_atom(e, &nm) )
 	return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_atom, e);
-      if ( nm == ATOM_comma )
-	goto protected;
-      defOperator(m, nm, t, p);
+      if ( !defOperator(m, nm, t, p, FALSE) )
+	return FALSE;
     }
     if ( !PL_get_nil(l) )
       return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_list, l);
@@ -386,7 +392,7 @@ addOpToBuffer(Buffer b, atom_t name, int type, int priority)
   opdef new;
 
   for(i = 0; i<mx; op++, i++)
-  { if ( op->name == name && op->type == type )
+  { if ( op->name == name && (op->type&OP_MASK) == (type&OP_MASK) )
       return;				/* got this one already */
   }
 
@@ -413,7 +419,7 @@ addOpsFromTable(Table t, atom_t name, int priority, int type, Buffer b)
 	assert(kind >= OP_PREFIX && kind <= OP_POSTFIX);
 
 	if ( op->priority[kind] < 0 ||
-	     op->type[kind] != type )
+	     (op->type[kind]&OP_MASK) != (type&OP_MASK) )
 	  continue;
 
 	if ( !priority ||
@@ -496,7 +502,7 @@ current_op(Module m, int inherit,
       } else
 	return FALSE;
 
-      e = allocHeap(sizeof(*e));
+      e = allocHeapOrHalt(sizeof(*e));
       b = &e->buffer;
       initBuffer(b);
       e->index = 0;
@@ -534,7 +540,13 @@ current_op(Module m, int inherit,
     if ( PL_unify_atom(name, match->name) &&
 	 PL_unify_integer(prec, match->priority) &&
 	 PL_unify_atom(type, operatorTypeToAtom(match->type)) )
+    { if ( e->index == mx )
+      { discardBuffer(&e->buffer);
+        freeHeap(e, sizeof(*e));
+	return TRUE;
+      }
       ForeignRedoPtr(e);
+    }
 
     PL_rewind_foreign_frame(fid);
   }
@@ -593,7 +605,8 @@ static const opdef operators[] = {
   OP(ATOM_ifthen,		 OP_XFY, 1050),	/* -> */
   OP(ATOM_softcut,		 OP_XFY, 1050),	/* *-> */
   OP(ATOM_divide,		 OP_YFX, 400),	/* / */
-  OP(ATOM_div,			 OP_YFX, 400),	/* // */
+  OP(ATOM_gdiv,			 OP_YFX, 400),	/* // */
+  OP(ATOM_div,			 OP_YFX, 400),	/* div */
   OP(ATOM_rdiv,			 OP_YFX, 400),	/* rdiv */
   OP(ATOM_and,			 OP_YFX, 500),	/* /\ */
   OP(ATOM_colon,		 OP_XFY, 600),	/* : */
@@ -620,9 +633,8 @@ static const opdef operators[] = {
   OP(ATOM_backslash,		 OP_FY,	 200),	/* \ */
   OP(ATOM_not_provable,		 OP_FY,	 900),	/* \+ */
   OP(ATOM_bitor,		 OP_YFX, 500),	/* \/ */
-  OP(ATOM_bw_xor,		 OP_YFX, 500),	/* >< */
   OP(ATOM_not_equals,		 OP_XFX, 700),	/* \= */
-  OP(ATOM_not_strickt_equals,	 OP_XFX, 700),	/* \== */
+  OP(ATOM_not_strict_equal,	 OP_XFX, 700),	/* \== */
   OP(ATOM_at_equals,		 OP_XFX, 700),	/* =@= */
   OP(ATOM_at_not_equals,	 OP_XFX, 700),	/* \=@= */
   OP(ATOM_hat,			 OP_XFY, 200),	/* ^ */
@@ -640,6 +652,7 @@ static const opdef operators[] = {
   OP(ATOM_module_transparent,	 OP_FX,	 1150),	/* module_transparent */
   OP(ATOM_multifile,		 OP_FX,	 1150),	/* multifile */
   OP(ATOM_meta_predicate,	 OP_FX,	 1150),	/* meta_predicate */
+  OP(ATOM_public,		 OP_FX,	 1150),	/* public */
   OP(ATOM_xor,			 OP_YFX, 400),	/* xor */
 
   OP(NULL_ATOM,			 0,	 0)
@@ -650,8 +663,10 @@ void
 initOperators(void)
 { const opdef *op;
 
+  MODULE_system->operators = newOperatorTable(32);
+
   for( op = operators; op->name; op++ )
-    defOperator(MODULE_system, op->name, op->type, op->priority);
+    defOperator(MODULE_system, op->name, op->type, op->priority, TRUE);
 }
 
 
