@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,10 +18,9 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
-/*#define O_SECURE 1*/
 /*#define O_DEBUG 1*/
 #include "pl-incl.h"
 #include "pl-dbref.h"
@@ -31,8 +29,9 @@
 #define WORDS_PER_PLINT (sizeof(int64_t)/sizeof(word))
 
 static RecordList lookupRecordList(word);
-static RecordList isCurrentRecordList(word);
-static void freeRecordRef(RecordRef r ARG_LD);
+static RecordList isCurrentRecordList(word, int must_be_non_empty);
+static void freeRecordRef(RecordRef r);
+static void unallocRecordList(RecordList rl);
 
 #define RECORDA 0
 #define RECORDZ 1
@@ -43,9 +42,30 @@ static void freeRecordRef(RecordRef r ARG_LD);
 #undef LD
 #define LD LOCAL_LD
 
+static void
+free_recordlist_symbol(Symbol s)
+{ RecordList l = s->value;
+
+  unallocRecordList(l);
+}
+
+
 void
 initRecords(void)
 { GD->recorded_db.record_lists = newHTable(8);
+  GD->recorded_db.record_lists->free_symbol = free_recordlist_symbol;
+}
+
+
+void
+cleanupRecords(void)
+{ Table t;
+
+  if ( (t=GD->recorded_db.record_lists) )
+  { GD->recorded_db.record_lists = NULL;
+    destroyHTable(t);
+    GD->recorded_db.head = GD->recorded_db.tail = NULL;
+  }
 }
 
 
@@ -59,14 +79,12 @@ lookupRecordList(word key)
   if ( (s = lookupHTable(GD->recorded_db.record_lists, (void *)key)) )
   { return s->value;
   } else
-  { GET_LD
-    RecordList l;
+  { RecordList l;
 
     if ( isAtom(key) )			/* can also be functor_t */
       PL_register_atom(key);
-    l = allocHeap(sizeof(*l));
+    l = allocHeapOrHalt(sizeof(*l));
     l->key = key;
-    l->type = RECORD_TYPE;
     l->references = 0;
     l->flags = 0;
     l->firstRecord = l->lastRecord = NULL;
@@ -85,11 +103,26 @@ lookupRecordList(word key)
 
 
 static RecordList
-isCurrentRecordList(word key)
+isCurrentRecordList(word key, int must_be_non_empty)
 { Symbol s;
 
   if ( (s = lookupHTable(GD->recorded_db.record_lists, (void *)key)) )
-    return s->value;
+  { RecordList rl = s->value;
+
+    if ( must_be_non_empty )
+    { RecordRef record;
+
+      LOCK();
+      for(record = rl->firstRecord; record; record = record->next)
+      { if ( false(record->record, R_ERASED) )
+	  break;
+      }
+      UNLOCK();
+      return record ? rl : NULL;
+    } else
+    { return rl;
+    }
+  }
 
   return NULL;
 }
@@ -99,7 +132,7 @@ isCurrentRecordList(word key)
 */
 
 static void
-cleanRecordList(RecordList rl ARG_LD)
+cleanRecordList(RecordList rl)
 { RecordRef *p;
   RecordRef r, prev=NULL;
 
@@ -108,7 +141,7 @@ cleanRecordList(RecordList rl ARG_LD)
     { *p = r->next;
       if ( r == rl->lastRecord )
 	rl->lastRecord = prev;
-      freeRecordRef(r PASS_LD);
+      freeRecordRef(r);
     } else
     { prev = r;
       p = &r->next;
@@ -117,6 +150,26 @@ cleanRecordList(RecordList rl ARG_LD)
 }
 
 
+/* unallocRecordList() is used when memory is cleaned for PL_cleanup().
+   We set R_NOLOCK to avoid needless update of the atom references in
+   freeRecord().
+*/
+
+static void
+unallocRecordList(RecordList rl)
+{ RecordRef r, n;
+
+  for(r = rl->firstRecord; r; r=n)
+  { n = r->next;
+
+    set(r->record, R_NOLOCK);
+    freeRecordRef(r);
+  }
+
+  freeHeap(rl, sizeof(*rl));
+}
+
+
 		 /*******************************
 		 *	    HEAP STORAGE	*
 		 *******************************/
@@ -148,10 +201,10 @@ typedef struct
 
 #define	PL_TYPE_VARIABLE	(1)	/* variable */
 #define PL_TYPE_ATOM		(2)	/* atom */
-#define PL_TYPE_INTEGER	  	(3)	/* big integer */
+#define PL_TYPE_INTEGER		(3)	/* big integer */
 #define PL_TYPE_TAGGED_INTEGER  (4)	/* tagged integer */
-#define PL_TYPE_FLOAT	  	(5)	/* double */
-#define PL_TYPE_STRING	  	(6)	/* string */
+#define PL_TYPE_FLOAT		(5)	/* double */
+#define PL_TYPE_STRING		(6)	/* string */
 #define PL_TYPE_COMPOUND	(7)	/* compound term */
 #define PL_TYPE_CONS		(8)	/* list-cell */
 
@@ -163,15 +216,15 @@ typedef struct
 #define PL_REC_CYCLE		(14)	/* cyclic reference */
 #define PL_REC_MPZ		(15)	/* GMP integer */
 
-#define addUnalignedBuf(b, ptr, type) \
-	do \
-	{ if ( (b)->top + sizeof(type) > (b)->max ) \
-	  { if ( !growBuffer((Buffer)b, sizeof(type)) ) \
-	      outOfCore(); \
-	  } \
-	  memcpy((b)->top, ptr, sizeof(type)); \
-	  (b)->top += sizeof(type); \
-	} while(0)
+static inline void
+addUnalignedBuf(TmpBuffer b, void *ptr, size_t bytes)
+{ if ( b->top + bytes > b->max )
+  { if ( !growBuffer((Buffer)b, bytes) )
+      outOfCore();
+  }
+  memcpy(b->top, ptr, bytes);
+  b->top += bytes;
+}
 
 static inline void
 addOpCode(CompileInfo info, int code)
@@ -272,14 +325,14 @@ addFloat(CompileInfo info, void *val)
   } else
   { addOpCode(info, PL_TYPE_FLOAT);
 
-    addUnalignedBuf(&info->code, val, double);
+    addUnalignedBuf(&info->code, val, sizeof(double));
   }
 }
 
 
 static inline void
 addWord(CompileInfo info, word w)
-{ addUnalignedBuf(&info->code, &w, word);
+{ addUnalignedBuf(&info->code, &w, sizeof(word));
 }
 
 
@@ -364,13 +417,13 @@ compile_term_to_heap(term_agenda *agenda, CompileInfo info ARG_LD)
 
 	continue;
       }
-  #if O_ATTVAR
+#if O_ATTVAR
       case TAG_ATTVAR:
       { intptr_t n = info->nvars++;
 	Word ap = valPAttVar(w);
 
 	if ( isEmptyBuffer(&info->code) )
-	{ addOpCode(info, PL_REC_ALLOCVAR); 	/* only an attributed var */
+	{ addOpCode(info, PL_REC_ALLOCVAR);	/* only an attributed var */
 	  info->size++;
 	}
 
@@ -386,7 +439,7 @@ compile_term_to_heap(term_agenda *agenda, CompileInfo info ARG_LD)
 	deRef(p);
 	goto again;
       }
-  #endif
+#endif
       case TAG_ATOM:
       { if ( storage(w) == STG_GLOBAL )	/* this is a variable */
 	{ intptr_t n = ((intptr_t)(w) >> 7);
@@ -419,12 +472,12 @@ compile_term_to_heap(term_agenda *agenda, CompileInfo info ARG_LD)
 	      addOpCode(info, PL_TYPE_INTEGER);
 	      addInt64(info, n.value.i);
 	      break;
-  #ifdef O_GMP
+#ifdef O_GMP
 	    case V_MPZ:
 	      addOpCode(info, PL_REC_MPZ);
 	      addMPZToBuffer((Buffer)&info->code, n.value.mpz);
 	      break;
-  #endif
+#endif
 	    default:
 	      assert(0);
 	  }
@@ -455,7 +508,7 @@ compile_term_to_heap(term_agenda *agenda, CompileInfo info ARG_LD)
 	int arity;
 	word functor;
 
-  #if O_CYCLIC
+#if O_CYCLIC
 	if ( isInteger(f->definition) )
 	{ addOpCode(info, PL_REC_CYCLE);
 	  addSizeInt(info, valInt(f->definition));
@@ -472,12 +525,12 @@ compile_term_to_heap(term_agenda *agenda, CompileInfo info ARG_LD)
 
 	  mark.term = f;
 	  mark.fdef = f->definition;
-	  pushSegStack(&LD->cycle.lstack, &mark);
-	  f->definition = (functor_t)consInt(info->size);
-					  /* overflow test */
-	  assert(valInt(f->definition) == (intptr_t)info->size);
+	  pushSegStack(&LD->cycle.lstack, mark, cycle_mark);
+	  f->definition = (functor_t)consUInt(info->size);
+				  /* overflow test (should not be possible) */
+	  DEBUG(CHK_SECURE, assert(valUInt(f->definition) == (uintptr_t)info->size));
 	}
-  #endif
+#endif
 
 	info->size += arity+1;
 	addFunctor(info, functor);
@@ -509,7 +562,7 @@ static void
 unvisit(ARG1_LD)
 { cycle_mark mark;
 
-  while( popSegStack(&LD->cycle.lstack, &mark) )
+  while( popSegStack(&LD->cycle.lstack, &mark, cycle_mark) )
   { mark.term->definition = mark.fdef;
   }
 }
@@ -538,15 +591,26 @@ restoreVars(compile_info *info)
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+compileTermToHeap__LD() is the core of the recorded database.
+
+Returns NULL if there is insufficient   memory.  Otherwise the result of
+the  allocation  function.   The   default    allocation   function   is
+PL_malloc_atomic_unmanaged().
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 Record
-compileTermToHeap__LD(term_t t, int flags ARG_LD)
+compileTermToHeap__LD(term_t t,
+		      void* (*allocate)(void *closure, size_t size),
+		      void* closure,
+		      int flags ARG_LD)
 { compile_info info;
   Record record;
   size_t size;
   size_t rsize = SIZERECORD(flags);
   term_agenda agenda;
 
-  SECURE(checkData(valTermRef(t)));
+  DEBUG(CHK_SECURE, checkData(valTermRef(t)));
 
   init_cycle(PASS_LD1);
   initBuffer(&info.code);
@@ -556,25 +620,32 @@ compileTermToHeap__LD(term_t t, int flags ARG_LD)
   info.external = (flags & R_EXTERNAL);
   info.lock = !(info.external || (flags&R_NOLOCK));
 
-  initTermAgenda(&agenda, valTermRef(t));
+  initTermAgenda(&agenda, 1, valTermRef(t));
   compile_term_to_heap(&agenda, &info PASS_LD);
   clearTermAgenda(&agenda);
   restoreVars(&info);
   unvisit(PASS_LD1);
 
   size = rsize + sizeOfBuffer(&info.code);
-  record = allocHeap(size);
+  if ( allocate )
+    record = (*allocate)(closure, size);
+  else
+    record = PL_malloc_atomic_unmanaged(size);
+
+  if ( record )
+  {
 #ifdef REC_MAGIC
-  record->magic = REC_MAGIC;
+    record->magic = REC_MAGIC;
 #endif
-  record->gsize = (unsigned int)info.size; /* only 28-bit */
-  record->nvars = info.nvars;
-  record->size  = (int)size;
-  record->flags = flags;
-  if ( flags & R_DUPLICATE )
-  { record->references = 1;
+    record->gsize = (unsigned int)info.size; /* only 28-bit */
+    record->nvars = info.nvars;
+    record->size  = (int)size;
+    record->flags = flags;
+    if ( flags & R_DUPLICATE )
+    { record->references = 1;
+    }
+    memcpy(addPointer(record, rsize), info.code.base, sizeOfBuffer(&info.code));
   }
-  memcpy(addPointer(record, rsize), info.code.base, sizeOfBuffer(&info.code));
   discardBuffer(&info.code);
 
   DEBUG(3, Sdprintf("--> record at %p\n", record));
@@ -617,7 +688,7 @@ PL_record_external(term_t t, size_t *len)
   int first = REC_HDR;
   term_agenda agenda;
 
-  SECURE(checkData(valTermRef(t)));
+  DEBUG(CHK_SECURE, checkData(valTermRef(t)));
   p = valTermRef(t);
   deRef(p);
 
@@ -638,7 +709,7 @@ PL_record_external(term_t t, size_t *len)
 
   ret_primitive:
     scode = (int)sizeOfBuffer(&info.code);
-    rec = allocHeap(scode);
+    rec = PL_malloc_atomic(scode);
     memcpy(rec, info.code.base, scode);
     discardBuffer(&info.code);
     *len = scode;
@@ -657,7 +728,7 @@ PL_record_external(term_t t, size_t *len)
   info.external = TRUE;
   info.lock = FALSE;
 
-  initTermAgenda(&agenda, p);
+  initTermAgenda(&agenda, 1, p);
   compile_term_to_heap(&agenda, &info PASS_LD);
   clearTermAgenda(&agenda);
   if ( info.nvars == 0 )
@@ -674,7 +745,7 @@ PL_record_external(term_t t, size_t *len)
     addUintBuffer((Buffer)&hdr, info.nvars);	/* Number of variables */
   shdr = (int)sizeOfBuffer(&hdr);
 
-  rec = allocHeap(shdr + scode);
+  rec = PL_malloc_atomic_unmanaged(shdr + scode);
   memcpy(rec, hdr.base, shdr);
   memcpy(rec+shdr, info.code.base, scode);
 
@@ -696,10 +767,10 @@ typedef struct
   const char   *base;			/* start of data */
   Word	       *vars;
   Word		gbase;			/* base of term on global stack */
-  Word 		gstore;			/* current storage location */
+  Word		gstore;			/* current storage location */
 					/* for se_record() */
-  uint  	nvars;			/* Variables seen */
-  TmpBuffer 	avars;			/* Values stored for attvars */
+  uint		nvars;			/* Variables seen */
+  TmpBuffer	avars;			/* Values stored for attvars */
 } copy_info, *CopyInfo;
 
 
@@ -717,7 +788,7 @@ other stacks as scratch-area.
   { Word *p; \
     uint i; \
     if ( (n) > MAX_ALLOCA_VARS ) \
-      info.vars = allocHeap(sizeof(Word) * (n)); \
+      info.vars = allocHeapOrHalt(sizeof(Word) * (n)); \
     else \
     { if ( !(info.vars = alloca(sizeof(Word) * (n))) ) \
 	fatalError("alloca() failed"); \
@@ -740,7 +811,7 @@ other stacks as scratch-area.
 #define fetchMultipleBuf(b, var, times, type) \
 		do \
 		{ memcpy(var, (b)->data, times*sizeof(type)); \
- 		  (b)->data +=  times*sizeof(type); \
+		  (b)->data +=  times*sizeof(type); \
 		} while(0)
 #define skipBuf(b, type) \
 		((b)->data += sizeof(type))
@@ -835,163 +906,171 @@ fetchChars(CopyInfo b, unsigned len, Word to)
 }
 
 
-static void
+static int
 copy_record(Word p, CopyInfo b ARG_LD)
-{ intptr_t tag;
+{ term_agenda agenda;
+  int is_compound = FALSE;
+  int tag;
 
-right_recursion:
-  switch( (tag = fetchOpCode(b)) )
-  { case PL_TYPE_VARIABLE:
-    { intptr_t n = fetchSizeInt(b);
+  do
+  {
+  right_recursion:
+    switch( (tag = fetchOpCode(b)) )
+    { case PL_TYPE_VARIABLE:
+      { intptr_t n = fetchSizeInt(b);
 
-      if ( b->vars[n] )
-      { if ( p > b->vars[n] )		/* ensure the reference is in the */
-	  *p = makeRef(b->vars[n]);	/* right direction! */
-	else
-	{ *p = *b->vars[n];		/* wrong way.  make sure b->vars[n] */
-	  *b->vars[n] = makeRef(p);	/* stays at the real variable */
-	  b->vars[n] = p;		/* NOTE: also links attvars! */
+	if ( b->vars[n] )
+	{ if ( p > b->vars[n] )		/* ensure the reference is in the */
+	    *p = makeRef(b->vars[n]);	/* right direction! */
+	  else
+	  { *p = *b->vars[n];		/* wrong way.  make sure b->vars[n] */
+	    *b->vars[n] = makeRef(p);	/* stays at the real variable */
+	    b->vars[n] = p;		/* NOTE: also links attvars! */
+	  }
+	} else
+	{ setVar(*p);
+	  b->vars[n] = p;
 	}
-      } else
-      {	setVar(*p);
-	b->vars[n] = p;
+
+	continue;
+      }
+      case PL_REC_ALLOCVAR:
+      { setVar(*b->gstore);
+	*p = makeRefG(b->gstore);
+	p = b->gstore++;
+	goto right_recursion;
       }
-
-      return;
-    }
-    case PL_REC_ALLOCVAR:
-    { setVar(*b->gstore);
-      *p = makeRefG(b->gstore);
-      p = b->gstore++;
-      goto right_recursion;
-    }
 #if O_ATTVAR
-    case PL_TYPE_ATTVAR:
-    { intptr_t n = fetchSizeInt(b);
+      case PL_TYPE_ATTVAR:
+      { intptr_t n = fetchSizeInt(b);
 
-      *p = consPtr(b->gstore, TAG_ATTVAR|STG_GLOBAL);
-      b->vars[n] = p;
-      p = b->gstore++;
-      goto right_recursion;
-    }
+	*p = consPtr(b->gstore, TAG_ATTVAR|STG_GLOBAL);
+	b->vars[n] = p;
+	p = b->gstore++;
+	goto right_recursion;
+      }
 #endif
-    case PL_TYPE_ATOM:
-    { *p = fetchWord(b);
+      case PL_TYPE_ATOM:
+      { *p = fetchWord(b);
+	continue;
+      }
+      case PL_TYPE_EXT_ATOM:
+      { fetchAtom(b, p);
+	PL_unregister_atom(*p);
+	continue;
+      }
+      case PL_TYPE_TAGGED_INTEGER:
+      { int64_t val = fetchInt64(b);
+	*p = consInt(val);
+	continue;
+      }
+      case PL_TYPE_INTEGER:
+      { size_t i;
+	union
+	{ int64_t i64;
+	  word    w[WORDS_PER_PLINT];
+	} val;
 
-      return;
-    }
-    case PL_TYPE_EXT_ATOM:
-    { fetchAtom(b, p);
-      PL_unregister_atom(*p);
-      return;
-    }
-    case PL_TYPE_TAGGED_INTEGER:
-    { int64_t val = fetchInt64(b);
+	val.i64 = fetchInt64(b);
 
-      *p = consInt(val);
-
-      return;
-    }
-    case PL_TYPE_INTEGER:
-    { size_t i;
-      union
-      { int64_t i64;
-	word    w[WORDS_PER_PLINT];
-      } val;
-
-      val.i64 = fetchInt64(b);
-
-      *p = consPtr(b->gstore, TAG_INTEGER|STG_GLOBAL);
-      *b->gstore++ = mkIndHdr(WORDS_PER_PLINT, TAG_INTEGER);
-      for(i=0; i<WORDS_PER_PLINT; i++)
-	*b->gstore++ = val.w[i];
-      *b->gstore++ = mkIndHdr(WORDS_PER_PLINT, TAG_INTEGER);
-
-      return;
-    }
+	*p = consPtr(b->gstore, TAG_INTEGER|STG_GLOBAL);
+	*b->gstore++ = mkIndHdr(WORDS_PER_PLINT, TAG_INTEGER);
+	for(i=0; i<WORDS_PER_PLINT; i++)
+	  *b->gstore++ = val.w[i];
+	*b->gstore++ = mkIndHdr(WORDS_PER_PLINT, TAG_INTEGER);
+	continue;
+      }
 #ifdef O_GMP
-    case PL_REC_MPZ:
-    { b->data = loadMPZFromCharp(b->data, p, &b->gstore);
-
-      return;
-    }
+      case PL_REC_MPZ:
+      { b->data = loadMPZFromCharp(b->data, p, &b->gstore);
+	continue;
+      }
 #endif
-    case PL_TYPE_FLOAT:
-    case PL_TYPE_EXT_FLOAT:
-    { *p = consPtr(b->gstore, TAG_FLOAT|STG_GLOBAL);
-      *b->gstore++ = mkIndHdr(WORDS_PER_DOUBLE, TAG_FLOAT);
-      if ( tag == PL_TYPE_FLOAT )
-	fetchFloat(b, b->gstore);
-      else
-	fetchExtFloat(b, b->gstore);
-      b->gstore += WORDS_PER_DOUBLE;
-      *b->gstore++ = mkIndHdr(WORDS_PER_DOUBLE, TAG_FLOAT);
+      case PL_TYPE_FLOAT:
+      case PL_TYPE_EXT_FLOAT:
+      { *p = consPtr(b->gstore, TAG_FLOAT|STG_GLOBAL);
+	*b->gstore++ = mkIndHdr(WORDS_PER_DOUBLE, TAG_FLOAT);
+	if ( tag == PL_TYPE_FLOAT )
+	  fetchFloat(b, b->gstore);
+	else
+	  fetchExtFloat(b, b->gstore);
+	b->gstore += WORDS_PER_DOUBLE;
+	*b->gstore++ = mkIndHdr(WORDS_PER_DOUBLE, TAG_FLOAT);
+	continue;
+      }
+      case PL_TYPE_STRING:
+      { unsigned len = fetchSizeInt(b);
+	int lw, pad;
+	word hdr;
 
-      return;
-    }
-    case PL_TYPE_STRING:
-    { unsigned len = fetchSizeInt(b);
-      int lw, pad;
-      word hdr;
-
-      lw = (len+sizeof(word))/sizeof(word); /* see globalNString() */
-      pad = (lw*sizeof(word) - len);
-      *p = consPtr(b->gstore, TAG_STRING|STG_GLOBAL);
-      *b->gstore++ = hdr = mkStrHdr(lw, pad);
-      b->gstore[lw-1] = 0L;		/* zero-padding */
-      fetchChars(b, len, b->gstore);
-      b->gstore += lw;
-      *b->gstore++ = hdr;
-
-      return;
-    }
+	lw = (len+sizeof(word))/sizeof(word); /* see globalNString() */
+	pad = (lw*sizeof(word) - len);
+	*p = consPtr(b->gstore, TAG_STRING|STG_GLOBAL);
+	*b->gstore++ = hdr = mkStrHdr(lw, pad);
+	b->gstore[lw-1] = 0L;		/* zero-padding */
+	fetchChars(b, len, b->gstore);
+	b->gstore += lw;
+	*b->gstore++ = hdr;
+	continue;
+      }
 #ifdef O_CYCLIC
-    case PL_REC_CYCLE:
-    { unsigned offset = fetchSizeInt(b);
-      Word ct = b->gbase+offset;
+      case PL_REC_CYCLE:
+      { unsigned offset = fetchSizeInt(b);
+	Word ct = b->gbase+offset;
 
-      *p = consPtr(ct, TAG_COMPOUND|STG_GLOBAL);
-
-      return;
-    }
+	*p = consPtr(ct, TAG_COMPOUND|STG_GLOBAL);
+	continue;
+      }
 #endif
-  { word fdef;
-    int arity;
-    case PL_TYPE_COMPOUND:
+    { word fdef;
+      int arity;
+      case PL_TYPE_COMPOUND:
 
-      fdef = fetchWord(b);
-      arity = arityFunctor(fdef);
+	fdef = fetchWord(b);
+	arity = arityFunctor(fdef);
 
-    compound:
-      *p = consPtr(b->gstore, TAG_COMPOUND|STG_GLOBAL);
-      *b->gstore++ = fdef;
-      p = b->gstore;
-      b->gstore += arity;
-      for(; --arity > 0; p++)
-	copy_record(p, b PASS_LD);
-      goto right_recursion;
-    case PL_TYPE_EXT_COMPOUND:
-    { atom_t name;
+      compound:
+	*p = consPtr(b->gstore, TAG_COMPOUND|STG_GLOBAL);
+	*b->gstore++ = fdef;
+	p = b->gstore;
+	b->gstore += arity;
+	if ( !is_compound )
+	{ is_compound = TRUE;
+	  initTermAgenda(&agenda, arity, p);
+	} else
+	{ if ( !pushWorkAgenda(&agenda, arity, p) )
+	    return MEMORY_OVERFLOW;
+	}
+	continue;
+      case PL_TYPE_EXT_COMPOUND:
+      { atom_t name;
 
-      arity = (int)fetchSizeInt(b);
-      fetchAtom(b, &name);
-      fdef = lookupFunctorDef(name, arity);
-
-      goto compound;
+	arity = (int)fetchSizeInt(b);
+	fetchAtom(b, &name);
+	fdef = lookupFunctorDef(name, arity);
+	goto compound;
+      }
     }
-  }
-    case PL_TYPE_CONS:
-    { *p = consPtr(b->gstore, TAG_COMPOUND|STG_GLOBAL);
-      *b->gstore++ = FUNCTOR_dot2;
-      p = b->gstore;
-      b->gstore += 2;
-      copy_record(p, b PASS_LD);
-      p++;
-      goto right_recursion;
+      case PL_TYPE_CONS:
+      { *p = consPtr(b->gstore, TAG_COMPOUND|STG_GLOBAL);
+	*b->gstore++ = FUNCTOR_dot2;
+	p = b->gstore;
+	b->gstore += 2;
+	if ( !is_compound )
+	{ is_compound = TRUE;
+	  initTermAgenda(&agenda, 2, p);
+	} else
+	{ if ( !pushWorkAgenda(&agenda, 2, p) )
+	    return MEMORY_OVERFLOW;
+	}
+	continue;
+      }
+      default:
+	assert(0);
     }
-    default:
-      assert(0);
-  }
+  } while ( is_compound && (p=nextTermAgendaNoDeRef(&agenda)) );
+
+  return TRUE;
 }
 
 
@@ -1014,11 +1093,13 @@ copyRecordToGlobal(term_t copy, Record r, int flags ARG_LD)
   gTop += r->gsize;
 
   INITCOPYVARS(b, r->nvars);
-  copy_record(valTermRef(copy), &b PASS_LD);
+  rc = copy_record(valTermRef(copy), &b PASS_LD);
   FREECOPYVARS(b, r->nvars);
+  if ( rc != TRUE )
+    return rc;
 
   assert(b.gstore == gTop);
-  SECURE(checkData(valTermRef(copy)));
+  DEBUG(CHK_SECURE, checkData(valTermRef(copy)));
 
   return TRUE;
 }
@@ -1057,79 +1138,78 @@ skipLong(CopyInfo b)
 
 static void
 scanAtomsRecord(CopyInfo b, void (*func)(atom_t a))
-{
-right_recursion:
+{ size_t work = 0;
 
-  switch( fetchOpCode(b) )
-  { case PL_TYPE_VARIABLE:
-    case PL_REC_CYCLE:
-    { skipSizeInt(b);
-      return;
-    }
-    case PL_REC_ALLOCVAR:
-      goto right_recursion;
+  do
+  { switch( fetchOpCode(b) )
+    { case PL_TYPE_VARIABLE:
+      case PL_REC_CYCLE:
+      { skipSizeInt(b);
+	continue;
+      }
+      case PL_REC_ALLOCVAR:
+	work++;
+	continue;
 #ifdef O_ATTVAR
-    case PL_TYPE_ATTVAR:
-    { skipSizeInt(b);
-      goto right_recursion;
-    }
+      case PL_TYPE_ATTVAR:
+      { skipSizeInt(b);
+	work++;
+	continue;
+      }
 #endif
-    case PL_TYPE_ATOM:
-    { atom_t a = fetchWord(b);
+      case PL_TYPE_ATOM:
+      { atom_t a = fetchWord(b);
 
-      (*func)(a);
-      return;
-    }
-    case PL_TYPE_EXT_ATOM:
-    { skipAtom(b);
-      return;
-    }
-    case PL_TYPE_TAGGED_INTEGER:
-    case PL_TYPE_INTEGER:
-    { skipLong(b);
-      return;
-    }
+	(*func)(a);
+	continue;
+      }
+      case PL_TYPE_EXT_ATOM:
+      { skipAtom(b);
+	continue;
+      }
+      case PL_TYPE_TAGGED_INTEGER:
+      case PL_TYPE_INTEGER:
+      { skipLong(b);
+	continue;
+      }
 #ifdef O_GMP
-    case PL_REC_MPZ:
-      b->data = skipMPZOnCharp(b->data);
-      return;
+      case PL_REC_MPZ:
+	b->data = skipMPZOnCharp(b->data);
+	continue;
 #endif
-    case PL_TYPE_FLOAT:
-    case PL_TYPE_EXT_FLOAT:
-    { skipBuf(b, double);
-      return;
-    }
-    case PL_TYPE_STRING:
-    { uint len = fetchSizeInt(b);
+      case PL_TYPE_FLOAT:
+      case PL_TYPE_EXT_FLOAT:
+      { skipBuf(b, double);
+	continue;
+      }
+      case PL_TYPE_STRING:
+      { uint len = fetchSizeInt(b);
+	b->data += len;
+	continue;
+      }
+      case PL_TYPE_COMPOUND:
+      { word fdef = fetchWord(b);
+	int arity;
 
-      b->data += len;
+	arity = arityFunctor(fdef);
+	work += arity;
+	continue;
+      }
+      case PL_TYPE_EXT_COMPOUND:
+      { intptr_t arity = fetchSizeInt(b);
 
-      return;
+	skipAtom(b);
+	work += arity;
+	continue;
+      }
+      case PL_TYPE_CONS:
+      { work += 2;
+	continue;
+      }
+      default:
+	assert(0);
     }
-    case PL_TYPE_COMPOUND:
-    { word fdef = fetchWord(b);
-      int arity;
-
-      arity = arityFunctor(fdef);
-      while(--arity > 0)
-	scanAtomsRecord(b, func);
-      goto right_recursion;
-    }
-    case PL_TYPE_EXT_COMPOUND:
-    { intptr_t arity = fetchSizeInt(b);
-
-      skipAtom(b);
-      while(--arity > 0)
-	scanAtomsRecord(b, func);
-      goto right_recursion;
-    }
-    case PL_TYPE_CONS:
-    { scanAtomsRecord(b, func);
-      goto right_recursion;
-    }
-    default:
-      assert(0);
-  }
+  } while ( work-- );
 }
 
 #endif /*O_ATOMGC*/
@@ -1385,7 +1465,7 @@ markAtomsRecord(Record record)
 
 
 bool
-freeRecord__LD(Record record ARG_LD)
+freeRecord(Record record)
 { if ( true(record, R_DUPLICATE) && --record->references > 0 )
     succeed;
 
@@ -1401,20 +1481,20 @@ freeRecord__LD(Record record ARG_LD)
   }
 #endif
 
-  freeHeap(record, record->size);
+  PL_free(record);
 
   succeed;
 }
 
 
 void
-unallocRecordRef(RecordRef r ARG_LD)
+unallocRecordRef(RecordRef r)
 { freeHeap(r, sizeof(*r));
 }
 
 
 static void
-freeRecordRef(RecordRef r ARG_LD)
+freeRecordRef(RecordRef r)
 { int reclaim_now = false(r->record, R_DBREF);
 
   freeRecord(r->record);
@@ -1471,7 +1551,7 @@ PL_recorded_external(const char *rec, term_t t)
   }
   assert(b.gstore == gTop);
 
-  SECURE(checkData(valTermRef(t)));
+  DEBUG(CHK_SECURE, checkData(valTermRef(t)));
 
   return TRUE;
 }
@@ -1479,9 +1559,7 @@ PL_recorded_external(const char *rec, term_t t)
 
 int
 PL_erase_external(char *rec)
-{ GET_LD
-  copy_info b;
-  uint scode;
+{ copy_info b;
   uchar m;
 
   b.base = b.data = rec;
@@ -1490,20 +1568,8 @@ PL_erase_external(char *rec)
   { Sdprintf("PL_erase_external(): incompatible version\n");
     fail;
   }
-  if (  m & (REC_INT|REC_ATOM) )
-  { if (  m & REC_INT )
-      skipLong(&b);
-    else
-      skipAtom(&b);
-  } else
-  { scode = fetchSizeInt(&b);
-    skipSizeInt(&b);			/* gsize */
-    if ( !(m & REC_GROUND) )
-      skipSizeInt(&b);			/* nvars */
-    b.data += scode;
-  }
 
-  freeHeap(rec, b.data-b.base);
+  PL_free(rec);
   return TRUE;
 }
 
@@ -1557,7 +1623,7 @@ PRED_IMPL("current_key", 1, current_key, PL_FA_NONDETERMINISTIC)
       { rl = GD->recorded_db.head;
 	break;
       } else if ( getKeyEx(A1, &k PASS_LD) &&
-		  isCurrentRecordList(k) )
+		  isCurrentRecordList(k, TRUE) )
 	succeed;
 
       fail;
@@ -1607,11 +1673,15 @@ record(term_t key, term_t term, term_t ref, int az)
   if ( ref && !PL_is_variable(ref) )
     return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_variable, ref);
 
-  copy = compileTermToHeap(term, 0);
-  r = allocHeap(sizeof(*r));
+  if ( !(copy = compileTermToHeap(term, 0)) )
+    return PL_no_memory();
+  r = allocHeapOrHalt(sizeof(*r));
   r->record = copy;
   if ( ref && !PL_unify_recref(ref, r) )
+  { PL_erase(copy);
+    freeHeap(r, sizeof(*r));
     return FALSE;
+  }
 
   LOCK();
   l = lookupRecordList(k);
@@ -1683,16 +1753,16 @@ PRED_IMPL("recorded", va, recorded, PL_FA_NONDETERMINISTIC)
 	}
 	return FALSE;
       }
-      if ( getKeyEx(key, &k PASS_LD) )
-      { if ( !(rl = isCurrentRecordList(k)) )
-	  fail;
-	varkey = FALSE;
-      } else if ( PL_is_variable(key) )
+      if ( PL_is_variable(key) )
       { if ( !(rl = GD->recorded_db.head) )
 	  fail;
 	varkey = TRUE;
+      } else if ( getKeyEx(key, &k PASS_LD) )
+      { if ( !(rl = isCurrentRecordList(k, FALSE)) )
+	  fail;
+	varkey = FALSE;
       } else
-      { fail;
+      { return FALSE;
       }
       LOCK();
       rl->references++;
@@ -1717,7 +1787,7 @@ PRED_IMPL("recorded", va, recorded, PL_FA_NONDETERMINISTIC)
 
 	LOCK();
 	if ( --rl->references == 0 && true(rl, RL_DIRTY) )
-	  cleanRecordList(rl PASS_LD);
+	  cleanRecordList(rl);
 	UNLOCK();
       }
     }
@@ -1748,7 +1818,7 @@ PRED_IMPL("recorded", va, recorded, PL_FA_NONDETERMINISTIC)
 	   (!ref || PL_unify_recref(ref, record)) )
       { PL_close_foreign_frame(fid);
 
-	if ( varkey && !unifyKey(key, rl->key) ) 	/* stack overflow */
+	if ( varkey && !unifyKey(key, rl->key) )	/* stack overflow */
 	{ UNLOCK();
 	  fail;
 	}
@@ -1758,7 +1828,7 @@ PRED_IMPL("recorded", va, recorded, PL_FA_NONDETERMINISTIC)
 	  ForeignRedoPtr(record->next);
 	} else
 	{ if ( --rl->references == 0 && true(rl, RL_DIRTY) )
-	    cleanRecordList(rl PASS_LD);
+	    cleanRecordList(rl);
 
 	  if ( varkey )
 	  { for( rl=rl->next; rl; rl=rl->next )
@@ -1779,7 +1849,7 @@ PRED_IMPL("recorded", va, recorded, PL_FA_NONDETERMINISTIC)
     }
 
     if ( --rl->references == 0 && true(rl, RL_DIRTY) )
-      cleanRecordList(rl PASS_LD);
+      cleanRecordList(rl);
 
     if ( varkey )
     { if ( rl->next )
@@ -1818,7 +1888,7 @@ PRED_IMPL("instance", 2, instance, 0)
 
   if ( type == DB_REF_CLAUSE )
   { Clause clause = ptr;
-    uintptr_t generation = environment_frame->generation;
+    gen_t generation = generationFrame(environment_frame);
 
     if ( true(clause, GOAL_CLAUSE) ||
 	 !visibleClause(clause, generation) )
@@ -1866,11 +1936,11 @@ PRED_IMPL("erase", 1, erase, 0)
   { Clause clause = ptr;
     Definition def = getProcDefinition(clause->procedure);
 
-    if ( !true(def, DYNAMIC) )
+    if ( !true(def, P_DYNAMIC) )
       return PL_error("erase", 1, NULL, ERR_PERMISSION,
 		      ATOM_clause, ATOM_erase, ref);
 
-    return retractClauseDefinition(def, clause PASS_LD);
+    return retractClauseDefinition(def, clause);
   } else
   { RecordRef record = ptr;
 
@@ -1887,7 +1957,7 @@ PRED_IMPL("erase", 1, erase, 0)
     { if ( !record->next )
 	l->lastRecord = NULL;
       l->firstRecord = record->next;
-      freeRecordRef(record PASS_LD);
+      freeRecordRef(record);
     } else
     { prev = l->firstRecord;
       r = prev->next;
@@ -1898,7 +1968,7 @@ PRED_IMPL("erase", 1, erase, 0)
 	    l->lastRecord = prev;
 	  }
 	  prev->next = r->next;
-	  freeRecordRef(r PASS_LD);
+	  freeRecordRef(r);
 	  goto ok;
 	}
       }
@@ -1913,68 +1983,6 @@ PRED_IMPL("erase", 1, erase, 0)
   return rval;
 }
 
-		 /*******************************
-		 *	     COMPLEXITY		*
-		 *******************************/
-
-static int
-count_term(Word t, int left ARG_LD)
-{ int count = 0;
-
-right_recursion:
-  deRef(t);
-
-  if ( --left < 0 )
-    return -1;
-  count++;
-
-  if ( isAttVar(*t) )
-  { Word p = valPAttVar(*t);
-
-    assert(onGlobalArea(p));
-    t = p;
-    goto right_recursion;
-  } else if ( isTerm(*t) )
-  { int arity = arityTerm(*t);
-    int me;
-
-    for(t = argTermP(*t, 0); arity-- > 0; count += me, t++ )
-    { if ( arity == 0 )
-	goto right_recursion;
-
-      me = count_term(t, left PASS_LD);
-      if ( me < 0 )
-	return me;
-      left -= me;
-      if ( left < 0 )
-	return -1;
-    }
-  }
-
-  return count+1;
-}
-
-
-#ifndef INT_MAX
-#define INT_MAX	    ((int)(((unsigned int)1<<(sizeof(int)*8-1))-1))
-#define INT_MIN     (-(INT_MIN)-1)
-#endif
-
-word
-pl_term_complexity(term_t t, term_t mx, term_t count)
-{ GET_LD
-  int c, m;
-
-  if ( !PL_get_integer(mx, &m) )
-    m = INT_MAX;
-
-  c = count_term(valTermRef(t), m PASS_LD);
-  if ( c < 0 || c > m )
-    fail;
-
-  return PL_unify_integer(count, c);
-}
-
 		 /*******************************
 		 *      PUBLISH PREDICATES	*
 		 *******************************/
