@@ -19,13 +19,14 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 /*#define O_DEBUG 1*/
 
 #define GLOBAL SO_LOCAL			/* allocate global variables here */
 #include "pl-incl.h"
+#include "os/pl-cstack.h"
 #include "pl-dbref.h"
 #include <sys/stat.h>
 #ifdef HAVE_UNISTD_H
@@ -51,27 +52,12 @@ static int allocStacks(size_t local, size_t global, size_t trail);
 static void initSignals(void);
 static void gcPolicy(Stack s, int policy);
 
-#undef I
-#define I TAGEX_INDIRECT
-
-const unsigned int tagtypeex[] =
-{
-	    /* var     attvar  int    float   atom   string    term     ref */
-/* static */	0,	0,	0,	0,	0,	0,	0,	0,
-/* heap */	0,	0,      I,	I,	0,	I,	0,	0,
-/* global */	0,	0,      I,	I,	0,	I,	0,	0,
-/* local */	0,	0,	0,	0,	0,	0,	0,	0
-};
-
-#undef I
-
-void
+int
 setupProlog(void)
 { GET_LD
   DEBUG(1, Sdprintf("Starting Heap Initialisation\n"));
 
   LD->critical = 0;
-  LD->aborted = ABORT_NONE;
   LD->signal.pending = 0;
 
   startCritical;
@@ -84,7 +70,9 @@ setupProlog(void)
 #if HAVE_SIGNAL
   DEBUG(1, Sdprintf("Prolog Signal Handling ...\n"));
   if ( truePrologFlag(PLFLAG_SIGNALS) )
-    initSignals();
+  { initSignals();
+    initBackTrace();
+  }
 #endif
   DEBUG(1, Sdprintf("Stacks ...\n"));
   if ( !initPrologStacks(GD->options.localSize,
@@ -126,13 +114,13 @@ setupProlog(void)
   initFiles();
   initIO();
   initCharConversion();
-  DEBUG(1, Sdprintf("Term ...\n"));
-  resetTerm();
   GD->io_initialised = TRUE;
 
-  endCritical;
+  if ( !endCritical )
+    return FALSE;
 
   DEBUG(1, Sdprintf("Heap Initialised\n"));
+  return TRUE;
 }
 
 
@@ -143,7 +131,9 @@ initPrologLocalData(ARG1_LD)
   depth_limit   = (uintptr_t)DEPTH_NO_LIMIT;
 #endif
 
+  LD->break_level = -1;
   LD->prolog_flag.write_attributes = PL_WRT_ATTVAR_IGNORE;
+
   updateAlerted(LD);
 }
 
@@ -182,7 +172,7 @@ provide very few.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static struct signame
-{ int 	      sig;
+{ int	      sig;
   const char *name;
   int	      flags;
 } signames[] =
@@ -277,7 +267,7 @@ static struct signame
   { -1,		NULL,     0}
 };
 
-static const char *
+const char *
 signal_name(int sig)
 { struct signame *sn = signames;
 
@@ -382,13 +372,15 @@ dispatch_signal(int sig, int sync)
     return;				/* what else?? */
   }
 
-  DEBUG(1, Sdprintf("Got signal %d in thread %d (=%d) %s\n",
-		    sig, LD->thread.info->pl_tid,
-		    pthread_self(),
-		    sync ? " (sync)" : " (async)"));
+  DEBUG(MSG_SIGNAL,
+	Sdprintf("Got signal %d in thread %d (=%d) %s\n",
+		 sig, LD->thread.info->pl_tid,
+		 pthread_self(),
+		 sync ? " (sync)" : " (async)"));
 #else
-  DEBUG(1, Sdprintf("Got signal %d %s\n",
-		    sig, sync ? " (sync)" : " (async)"));
+  DEBUG(MSG_SIGNAL,
+	Sdprintf("Got signal %d %s\n",
+		 sig, sync ? " (sync)" : " (async)"));
 #endif
 
   if ( true(sh, PLSIG_NOFRAME) && sh->handler )
@@ -430,8 +422,9 @@ dispatch_signal(int sig, int sync)
   LD->signal.current = sig;
   LD->signal.is_sync = sync;
 
-  DEBUG(1, Sdprintf("Handling signal %d, pred = %p, handler = %p\n",
-		    sig, sh->predicate, sh->handler));
+  DEBUG(MSG_SIGNAL,
+	Sdprintf("Handling signal %d, pred = %p, handler = %p\n",
+		 sig, sh->predicate, sh->handler));
 
   if ( sh->predicate )
   { term_t sigterm = PL_new_term_ref();
@@ -476,6 +469,10 @@ dispatch_signal(int sig, int sync)
   } else if ( sh->handler )
   { (*sh->handler)(sig);
 
+    DEBUG(MSG_SIGNAL,
+	  Sdprintf("Handler %p finished (pending=%lld)\n",
+		   sh->handler, LD->signal.pending));
+
     if ( exception_term && !sync )	/* handler: PL_raise_exception() */
     { LD->signal.exception = PL_record(exception_term);
       PL_raise(SIG_EXCEPTION);
@@ -572,15 +569,20 @@ unprepareSignal(int sig)
 }
 
 
+#ifdef SIGHUP
 static void
 hupHandler(int sig)
-{ PL_halt(2);
+{ (void)sig;
+
+  PL_halt(2);
 }
+#endif
 
 
 static void
 sig_exception_handler(int sig)
 { GET_LD
+  (void)sig;
 
   if ( LD && LD->signal.exception )
   { record_t ex = LD->signal.exception;
@@ -592,7 +594,7 @@ sig_exception_handler(int sig)
     PL_erase(ex);
     exception_term = exception_bin;
 
-    SECURE(checkData(valTermRef(exception_term)));
+    DEBUG(CHK_SECURE, checkData(valTermRef(exception_term)));
   }
 }
 
@@ -600,6 +602,7 @@ sig_exception_handler(int sig)
 static void
 agc_handler(int sig)
 { GET_LD
+  (void)sig;
 
   if ( GD->statistics.atoms >= GD->atoms.non_garbage + GD->atoms.margin &&
        !gc_status.blocked )
@@ -609,7 +612,9 @@ agc_handler(int sig)
 
 static void
 gc_handler(int sig)
-{ garbageCollect();
+{ (void)sig;
+
+  garbageCollect();
 }
 
 
@@ -617,6 +622,7 @@ static void
 free_clauses_handler(int sig)
 { GET_LD
   ClauseRef cref;
+  (void)sig;
 
   if ( (cref=LD->freed_clauses) )
   { LD->freed_clauses = NULL;
@@ -627,7 +633,9 @@ free_clauses_handler(int sig)
 
 static void
 abort_handler(int sig)
-{ abortProlog(ABORT_RAISE);
+{ (void)sig;
+
+  abortProlog();
 }
 
 
@@ -639,7 +647,7 @@ with EINTR and thus make them interruptable for thread-signals.
 #ifdef SIG_ALERT
 static void
 alert_handler(int sig)
-{
+{ (void)sig;
 }
 #endif
 
@@ -653,7 +661,13 @@ initSignals(void)
 #endif
 
   for( ; sn->name; sn++)
-  { if ( sn->flags )
+  {
+#ifdef HAVE_BOEHM_GC
+    if ( sn->sig == GC_get_suspend_signal() ||
+	 sn->sig == GC_get_thr_restart_signal() )
+      sn->flags = 0;
+#endif
+    if ( sn->flags )
     { SigHandler sh = prepareSignal(sn->sig);
       sh->flags |= sn->flags;
     }
@@ -708,10 +722,26 @@ resetSignals()
 
 void
 allSignalMask(sigset_t *set)
-{ sigfillset(set);
+{ static sigset_t allmask;
+  static int done = FALSE;
+
+  if ( !done )
+  { sigset_t tmp;
+
+    sigfillset(&tmp);
+    sigdelset(&tmp, SIGSTOP);
+    sigdelset(&tmp, SIGCONT);
+    sigdelset(&tmp, SIGQUIT);
+    sigdelset(&tmp, SIGSEGV);
+    sigdelset(&tmp, SIGBUS);
 #ifdef O_PROFILE
-  sigdelset(set, SIGPROF);
+    sigdelset(&tmp, SIGPROF);
 #endif
+    allmask = tmp;
+    done = TRUE;
+  }
+
+  *set = allmask;
 }
 
 
@@ -805,7 +835,7 @@ PL_signal(int sigandflags, handler_t func)
     SigHandler sh;
     int sig = (sigandflags & 0xffff);
 
-    if ( sig > MAXSIGNAL )
+    if ( sig >= MAXSIGNAL )
     { warning("PL_signal(): illegal signal number: %d", sig);
       return SIG_DFL;
     }
@@ -892,8 +922,8 @@ out:
 
 int
 endCritical__LD(ARG1_LD)
-{ if ( LD->aborted )
-    return abortProlog(ABORT_THROW);
+{ if ( exception_term )
+    return FALSE;
 
   return TRUE;
 }
@@ -1085,6 +1115,8 @@ initPrologStacks(size_t local, size_t global, size_t trail)
   base_addresses[STG_GLOBAL] = (uintptr_t)gBase;
   base_addresses[STG_TRAIL]  = (uintptr_t)tBase;
   *gBase++ = MARK_MASK;			/* see sweep_global_mark() */
+  gMax--;				/*  */
+  tMax--;
   emptyStacks();
 
   DEBUG(1, Sdprintf("base_addresses[STG_LOCAL] = %p\n",
@@ -1112,12 +1144,12 @@ emptyStack(Stack s)
 
 
 void
-emptyStacks()
+emptyStacks(void)
 { GET_LD
-  int i;
 
   environment_frame = NULL;
   fli_context       = NULL;
+  LD->query         = NULL;
 
   emptyStack((Stack)&LD->stacks.local);
   emptyStack((Stack)&LD->stacks.global);
@@ -1125,40 +1157,47 @@ emptyStacks()
   emptyStack((Stack)&LD->stacks.argument);
 
   LD->mark_bar          = gTop;
-  PL_open_foreign_frame();
-  exception_bin         = PL_new_term_ref();
-  exception_printed     = PL_new_term_ref();
-  LD->exception.tmp     = PL_new_term_ref();
-  LD->exception.pending = PL_new_term_ref();
-  LD->trim.dummy        = PL_new_term_ref();
+  if ( lTop && gTop )
+  { int i;
+
+    PL_open_foreign_frame();
+    exception_bin         = PL_new_term_ref();
+    exception_printed     = PL_new_term_ref();
+    LD->exception.tmp     = PL_new_term_ref();
+    LD->exception.pending = PL_new_term_ref();
+    LD->trim.dummy        = PL_new_term_ref();
 #ifdef O_ATTVAR
-  LD->attvar.head	= PL_new_term_ref();
-  LD->attvar.tail       = PL_new_term_ref();
-  DEBUG(3, Sdprintf("attvar.tail at %p\n", valTermRef(LD->attvar.tail)));
+    LD->attvar.head	= PL_new_term_ref();
+    LD->attvar.tail       = PL_new_term_ref();
+    DEBUG(3, Sdprintf("attvar.tail at %p\n", valTermRef(LD->attvar.tail)));
 #endif
 #ifdef O_GVAR
-  destroyGlobalVars();
+    destroyGlobalVars();
 #endif
-  for(i=0; i<TMP_PTR_SIZE; i++)
-    LD->tmp.h[i] = PL_new_term_ref();
-  LD->tmp.top = 0;
+    for(i=0; i<TMP_PTR_SIZE; i++)
+      LD->tmp.h[i] = PL_new_term_ref();
+    LD->tmp.top = 0;
+  }
 }
 
 
 		/********************************
-		*    	STACK ALLOCATION        *
+		*	STACK ALLOCATION        *
 		*********************************/
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Malloc/realloc/free based stack allocation
+init_stack() initializes the stack straucture. Params:
+
+  - name is the name of the stack (for diagnostic purposes)
+  - size is the allocated size
+  - limit is the maximum to which the stack is allowed to grow
+  - spare is the amount of spare stack we reserve
+  - gc indicates whether gc can collect data on the stack
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static void
-init_stack(Stack s, char *name,
-	   size_t size, size_t limit, size_t minfree, size_t spare)
-{ GET_LD
-
-  s->name 	= name;
+init_stack(Stack s, char *name, size_t size, size_t limit, size_t spare, int gc)
+{ s->name	= name;
   s->top	= s->base;
   s->size_limit	= limit;
   s->spare      = spare;
@@ -1166,8 +1205,7 @@ init_stack(Stack s, char *name,
   s->min_free   = 256*sizeof(word);
   s->max	= addPointer(s->base, size - spare);
   s->gced_size  = 0L;			/* size after last gc */
-  s->gc	        = ((s == (Stack) &LD->stacks.global ||
-		    s == (Stack) &LD->stacks.trail) ? TRUE : FALSE);
+  s->gc	        = gc;
   gcPolicy(s, GC_FAST_POLICY);
 }
 
@@ -1200,13 +1238,13 @@ allocStacks(size_t local, size_t global, size_t trail)
   lBase = (LocalFrame) addPointer(gBase, iglobal);
 
   init_stack((Stack)&LD->stacks.global,
-	     "global",   iglobal, global,  minglobal, 512*SIZEOF_VOIDP);
+	     "global",   iglobal, global, 512*SIZEOF_VOIDP, TRUE);
   init_stack((Stack)&LD->stacks.local,
-	     "local",    ilocal,  local,   minlocal, 512*SIZEOF_VOIDP);
+	     "local",    ilocal,  local,  512*SIZEOF_VOIDP, FALSE);
   init_stack((Stack)&LD->stacks.trail,
-	     "trail",    itrail,  trail,   mintrail, 256*SIZEOF_VOIDP);
+	     "trail",    itrail,  trail,  256*SIZEOF_VOIDP, TRUE);
   init_stack((Stack)&LD->stacks.argument,
-	     "argument", argument, argument, minargument, 0);
+	     "argument", argument, argument, 0, FALSE);
 
   LD->stacks.local.min_free = LOCAL_MARGIN;
 
@@ -1230,6 +1268,9 @@ stack_malloc(size_t size)
   if ( mem )
   { size_t *sp = mem;
     *sp++ = size;
+#ifdef SECURE_GC
+    memset(sp, 0xFB, size);
+#endif
 
     PL_LOCK(L_MISC);
     GD->statistics.stack_space += size;
@@ -1246,6 +1287,13 @@ stack_realloc(void *old, size_t size)
   size_t osize = *--sp;
   void *mem;
 
+#ifdef SECURE_GC
+  if ( (mem = stack_malloc(size)) )
+  { memcpy(mem, old, (size>osize?osize:size));
+    stack_free(old);
+    return mem;
+  }
+#else
   if ( (mem = realloc(sp, size+sizeof(size_t))) )
   { sp = mem;
     *sp++ = size;
@@ -1255,6 +1303,7 @@ stack_realloc(void *old, size_t size)
     PL_UNLOCK(L_MISC);
     return sp;
   }
+#endif
 
   return NULL;
 }
@@ -1268,16 +1317,13 @@ stack_free(void *mem)
   GD->statistics.stack_space -= osize;
   PL_UNLOCK(L_MISC);
 
+#ifdef SECURE_GC
+  memset(sp, 0xFB, osize+sizeof(size_t));
+#endif
   free(sp);
 }
 
 
-void
-resetStacks()
-{ emptyStacks();
-}
-
-
 int
 trim_stack(Stack s)
 { if ( s->spare < s->def_spare )
@@ -1285,7 +1331,8 @@ trim_stack(Stack s)
     ssize_t room = roomStackP(s);
 
     if ( room < reduce )
-    { DEBUG(0, Sdprintf("Only %d spare for %s-stack\n", room, s->name));
+    { DEBUG(MSG_SPARE_STACK,
+	    Sdprintf("Only %d spare for %s-stack\n", room, s->name));
       reduce = room;
     }
 
@@ -1380,9 +1427,10 @@ trimStacks(int resize ARG_LD)
     }
   }
 
-  SECURE({ scan_global(FALSE);
-	   checkStacks(NULL);
-	 });
+  DEBUG(CHK_SECURE,
+	{ scan_global(FALSE);
+	  checkStacks(NULL);
+	});
 }
 
 
@@ -1432,7 +1480,10 @@ freePrologLocalData(PL_local_data_t *ld)
   discardBuffer(&ld->fli._discardable_buffer);
 
   for(i=0; i<BUFFER_RING_SIZE; i++)
-    discardBuffer(&ld->fli._buffer_ring[i]);
+  { discardBuffer(&ld->fli._buffer_ring[i]);
+    initBuffer(&ld->fli._buffer_ring[i]);	/* Used by debug-print */
+						/* on shutdown */
+  }
 
   freeVarDefs(ld);
 
@@ -1441,6 +1492,9 @@ freePrologLocalData(PL_local_data_t *ld)
     destroyHTable(ld->gvar.nb_vars);
 #endif
 
+  if ( ld->bags.default_bag )
+    PL_free(ld->bags.default_bag);
+
 #ifdef O_CYCLIC
   clearSegStack(&ld->cycle.lstack);
   clearSegStack(&ld->cycle.vstack);
@@ -1449,7 +1503,15 @@ freePrologLocalData(PL_local_data_t *ld)
   freeArithLocalData(ld);
 #ifdef O_PLMT
   simpleMutexDelete(&ld->signal.sig_lock);
+  if ( ld->prolog_flag.table )
+  { PL_LOCK(L_PLFLAG);
+    destroyHTable(ld->prolog_flag.table);
+    PL_UNLOCK(L_PLFLAG);
+  }
 #endif
+
+  if ( ld->qlf.getstr_buffer )
+    free(ld->qlf.getstr_buffer);
 }
 
 
