@@ -19,7 +19,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 /*#define O_DEBUG 1*/
@@ -338,7 +338,7 @@ typedef struct
 { int		var_count;
   sha1_ctx	ctx[1];			/* The SHA1 Context */
   segstack	vars;
-  char		vars_first_chunk[256];
+  Word		vars_first_chunk[64];
 } sha1_state;
 
 typedef enum
@@ -351,7 +351,7 @@ typedef enum
 
 static int
 push_var(Word p, sha1_state *state)
-{ return pushSegStack(&state->vars, &p);
+{ return pushSegStack(&state->vars, p, Word);
 }
 
 
@@ -466,10 +466,10 @@ PRED_IMPL("variant_sha1", 2, variant_sha1, 0)
 	       sizeof(state.vars_first_chunk), state.vars_first_chunk);
   rc = variant_sha1(&agenda, &state PASS_LD);
   ac_clearTermAgenda(&agenda);
-  while(popSegStack(&state.vars, &p))
+  while(popSegStack(&state.vars, &p, Word))
     setVar(*p);
 
-  SECURE(checkData(valTermRef(A1)));
+  DEBUG(CHK_SECURE, checkData(valTermRef(A1)));
 
   switch( rc )
   { case E_ATTVAR:
