@@ -1,11 +1,9 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2011, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
 			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
@@ -20,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include "pl-incl.h"
@@ -88,10 +86,10 @@ PL_put_frame(term_t t, LocalFrame fr)
 static int
 PL_get_frame(term_t r, LocalFrame *fr)
 { GET_LD
-  long i;
+  intptr_t i;
   atom_t a;
 
-  if ( PL_get_long(r, &i) )
+  if ( PL_get_intptr(r, &i) )
   { LocalFrame f = ((LocalFrame)((Word)lBase + i));
 
     if ( !(f >= lBase && f < lTop) )
@@ -219,7 +217,6 @@ user to intercept and redefine the tracer.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 					/* Frame <-> Prolog integer */
-static LocalFrame	redoFrame(LocalFrame, Code *PC);
 static void		helpTrace(void);
 #ifdef O_INTERRUPT
 static void		helpInterrupt(void);
@@ -324,7 +321,6 @@ portPrompt(int port)
     case FAIL_PORT:	 return " Fail:  ";
     case EXIT_PORT:	 return " Exit:  ";
     case UNIFY_PORT:	 return " Unify: ";
-    case BREAK_PORT:	 return " Break: ";
     case EXCEPTION_PORT: return " Exception: ";
     case CUT_CALL_PORT:	 return " Cut call: ";
     case CUT_EXIT_PORT:	 return " Cut exit: ";
@@ -393,25 +389,21 @@ tracePort(LocalFrame frame, Choice bfr, int port, Code PC ARG_LD)
     }
   }
 
-  if ( port & BREAK_PORT )
-    goto ok;				/* always do break-points */
-
   if ( !debugstatus.tracing &&
-       (false(def, SPY_ME) || (port & CUT_PORT)) )
+       (false(def, SPY_ME) || (port & (CUT_PORT|REDO_PORT))) )
     return ACTION_CONTINUE;		/* not tracing and no spy-point */
   if ( debugstatus.skiplevel < levelFrame(frame) )
     return ACTION_CONTINUE;		/* skipped */
   if ( debugstatus.skiplevel == levelFrame(frame) &&
-       (port & (REDO_PORT|CUT_PORT)) )
-    return ACTION_CONTINUE;		/* redo or ! in skipped predicate */
+       (port & (REDO_PORT|CUT_PORT|UNIFY_PORT)) )
+    return ACTION_CONTINUE;		/* redo, unify or ! in skipped pred */
   if ( false(def, TRACE_ME) )
     return ACTION_CONTINUE;		/* non-traced predicate */
   if ( (!(debugstatus.visible & port)) )
     return ACTION_CONTINUE;		/* wrong port */
   if ( (true(def, HIDE_CHILDS) && !SYSTEM_MODE) &&
-       (port & (/*REDO_PORT|*/CUT_PORT)) )
+       (port & CUT_PORT) )
     return ACTION_CONTINUE;		/* redo or ! in system predicates */
-ok:
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Give a trace on the skipped goal for a redo.
@@ -425,7 +417,7 @@ Give a trace on the skipped goal for a redo.
 
       debugstatus.skiplevel = SKIP_REDO_IN_SKIP;
       SAVE_PTRS();
-      rc = tracePort(frame, bfr, REDO_PORT, pc2 PASS_LD);
+      rc = tracePort(fr, bfr, REDO_PORT, pc2 PASS_LD);
       RESTORE_PTRS();
       debugstatus.skiplevel = levelFrame(fr);
       set(fr, FR_SKIPPED);		/* cleared by "case 'c'" */
@@ -529,7 +521,7 @@ again:
 out:
   restoreWakeup(&wstate PASS_LD);
   if ( action == ACTION_ABORT )
-    abortProlog(ABORT_RAISE);
+    abortProlog();
 
   return action;
 }
@@ -578,7 +570,7 @@ setupFind(char *buf)
     FindData find;
 
     if ( !(find = LD->trace.find) )
-      find = LD->trace.find = allocHeap(sizeof(find_data));
+      find = LD->trace.find = allocHeapOrHalt(sizeof(find_data));
 
     if ( !PL_chars_to_term(s, t) )
     { PL_discard_foreign_frame(cid);
@@ -593,8 +585,12 @@ setupFind(char *buf)
     } else if ( PL_get_atom(t, &find->goal.name) )
     { find->type = TRACE_FIND_NAME;
     } else if ( PL_get_functor(t, &find->goal.term.functor) )
-    { find->type = TRACE_FIND_TERM;
-      find->goal.term.term = compileTermToHeap(t, 0);
+    { if ( (find->goal.term.term = compileTermToHeap(t, 0)) )
+      { find->type = TRACE_FIND_TERM;
+      } else
+      { Sfputs("ERROR: no memory to safe find target\n", Sdout);
+	fail;
+      }
     } else
     { Sfputs("[Illegal goal specification]\n", Sdout);
       fail;
@@ -663,20 +659,20 @@ traceAction(char *cmd, int port, LocalFrame frame, Choice bfr,
 
   switch( *s )
   { case 'a':	FeedBack("abort\n");
-    		return ACTION_ABORT;
+		return ACTION_ABORT;
     case 'b':	FeedBack("break\n");
 		pl_break();
 		return ACTION_AGAIN;
-    case '/': 	FeedBack("/");
-    		Sflush(Suser_output);
-    		if ( setupFind(&s[1]) )
+    case '/':	FeedBack("/");
+		Sflush(Suser_output);
+		if ( setupFind(&s[1]) )
 		{ clear(frame, FR_SKIPPED);
 		  return ACTION_CONTINUE;
 		}
 		return ACTION_AGAIN;
     case '.':   if ( LD->trace.find &&
 		     LD->trace.find->type != TRACE_FIND_NONE )
-      	        { FeedBack("repeat search\n");
+	        { FeedBack("repeat search\n");
 		  LD->trace.find->searching = TRUE;
 		  clear(frame, FR_SKIPPED);
 		  return ACTION_CONTINUE;
@@ -727,14 +723,14 @@ traceAction(char *cmd, int port, LocalFrame frame, Choice bfr,
 		setPrintOptions(ATOM_print);
 		return ACTION_AGAIN;
     case 'l':	FeedBack("leap\n");
-    		tracemode(FALSE, NULL);
+		tracemode(FALSE, NULL);
 		return ACTION_CONTINUE;
     case 'n':	FeedBack("no debug\n");
 		tracemode(FALSE, NULL);
-    		debugmode(DBG_OFF, NULL);
+		debugmode(DBG_OFF, NULL);
 		return ACTION_CONTINUE;
     case 'g':	FeedBack("goals\n");
-		backTrace(frame, num_arg == Default ? 5 : num_arg);
+		PL_backtrace(num_arg == Default ? 5 : num_arg, PL_BT_USER);
 		return ACTION_AGAIN;
     case 'A':	FeedBack("alternatives\n");
 		alternatives(bfr);
@@ -747,7 +743,7 @@ traceAction(char *cmd, int port, LocalFrame frame, Choice bfr,
 		}
 		return ACTION_AGAIN;
     case 'm':	FeedBack("Exception details");
-    	        if ( port & EXCEPTION_PORT )
+	        if ( port & EXCEPTION_PORT )
 		{ exceptionDetails();
 		} else
 		   Warn("No exception\n");
@@ -851,7 +847,7 @@ put_frame_goal(term_t goal, LocalFrame frame)
   }
 
   if ( def->module != MODULE_user &&
-       (false(def->module, SYSTEM) || SYSTEM_MODE))
+       (false(def->module, M_SYSTEM) || SYSTEM_MODE))
   { term_t a;
 
     if ( !(a=PL_new_term_ref()) )
@@ -878,7 +874,6 @@ static const portname portnames[] =
   { FAIL_PORT,	    ATOM_fail },
   { REDO_PORT,	    ATOM_redo },
   { UNIFY_PORT,	    ATOM_unify },
-  { BREAK_PORT,	    ATOM_break },
   { CUT_CALL_PORT,  ATOM_cut_call },
   { CUT_EXIT_PORT,  ATOM_cut_exit },
   { EXCEPTION_PORT, ATOM_exception },
@@ -907,10 +902,10 @@ writeFrameGoal(LocalFrame frame, Code PC, unsigned int flags)
     term_t pc   = PL_new_term_ref();
     const portname *pn = portnames;
 
-    if ( true(def, FOREIGN) )
+    if ( true(def, P_FOREIGN) )
       PL_put_atom(pc, ATOM_foreign);
     else if ( PC && frame->clause )
-      rc = PL_put_intptr(pc, PC-frame->clause->clause->codes);
+      rc = PL_put_intptr(pc, PC-frame->clause->value.clause->codes);
     else
       PL_put_nil(pc);
 
@@ -1068,51 +1063,71 @@ listGoal(LocalFrame frame)
 }
 
 
-void
-backTrace(LocalFrame frame, int depth)
-{ GET_LD
-  LocalFrame same_proc_frame = NULL;
-  Definition def = NULL;
-  int same_proc = 0;
-  int alien = FALSE;
-  Code PC = NULL;
+static void
+writeContextFrame(pl_context_t *ctx, int flags)
+{ if ( (flags&PL_BT_SAFE) )
+  { char buf[256];
 
-  if ( frame == NULL )
-     frame = environment_frame;
-
-  for(; depth > 0 && frame;
-        alien = (frame->parent == NULL),
-        PC = frame->programPointer,
-        frame = parentFrame(frame))
-  { if ( alien )
-      Sfputs("    <Alien goal>\n", Sdout);
-
-    if ( frame->predicate == def )
-    { if ( ++same_proc >= 10 )
-      { if ( same_proc == 10 )
-	  Sfputs("    ...\n    ...\n", Sdout);
-	same_proc_frame = frame;
-	continue;
-      }
-    } else
-    { if ( same_proc_frame != NULL )
-      { if ( isDebugFrame(same_proc_frame) || SYSTEM_MODE )
-        { writeFrameGoal(same_proc_frame, PC, WFG_BACKTRACE);
-	  depth--;
-	}
-	same_proc_frame = NULL;
-	same_proc = 0;
-      }
-      def = frame->predicate;
-    }
-
-    if ( isDebugFrame(frame) || SYSTEM_MODE)
-    { writeFrameGoal(frame, PC, WFG_BACKTRACE);
-      depth--;
-    }
+    PL_describe_context(ctx, buf, sizeof(buf));
+    Sdprintf("  %s\n", buf);
+  } else
+  { writeFrameGoal(ctx->fr, ctx->pc, WFG_BACKTRACE);
   }
 }
 
+
+void
+PL_backtrace(int depth, int flags)
+{ pl_context_t ctx;
+
+  if ( PL_get_context(&ctx, 0) )
+  { GET_LD
+    Definition def = NULL;
+    int same_proc = 0;
+    pl_context_t rctx;			/* recursive context */
+
+    if ( gc_status.active )
+    { flags |= PL_BT_SAFE;
+      flags &= ~PL_BT_USER;
+    }
+    if ( SYSTEM_MODE )
+      flags &= ~PL_BT_USER;
+
+    for(; depth > 0; PL_step_context(&ctx))
+    { LocalFrame frame;
+
+      if ( !(frame=ctx.fr) )
+	return;
+
+      if ( frame->predicate == def )
+      { if ( ++same_proc >= 10 )
+	{ if ( same_proc == 10 )
+	    Sdprintf("    ...\n    ...\n", Sdout);
+	  rctx = ctx;
+	  continue;
+	}
+      } else
+      { if ( same_proc >= 10 )
+	{ if ( isDebugFrame(rctx.fr) || !(flags&PL_BT_USER) )
+	  { writeContextFrame(&rctx, flags);
+	    depth--;
+	  }
+	  same_proc = 0;
+	}
+	def = frame->predicate;
+      }
+
+      if ( isDebugFrame(frame) || !(flags&PL_BT_USER) )
+      { writeContextFrame(&ctx, flags);
+	depth--;
+      }
+    }
+  } else
+  { Sdprintf("No stack??\n");
+  }
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Trace interception mechanism.  Whenever the tracer wants to perform some
 action   it   will   first   call   the    users'    Prolog    predicate
@@ -1137,7 +1152,7 @@ traceInterception(LocalFrame frame, Choice bfr, int port, Code PC)
 
   proc = _PL_predicate("prolog_trace_interception", 4, "user",
 		       &GD->procedures.prolog_trace_interception4);
-  if ( !getProcDefinition(proc)->definition.clauses )
+  if ( !getProcDefinition(proc)->impl.any )
     return rval;
 
   if ( !GD->bootsession && GD->debug_level == 0 )
@@ -1158,7 +1173,7 @@ traceInterception(LocalFrame frame, Choice bfr, int port, Code PC)
 
     switch(port)
     { case CALL_PORT:	   portname = ATOM_call;         break;
-      case REDO_PORT:	   portname = ATOM_redo;         break;
+      case REDO_PORT:	   portfunc = FUNCTOR_redo1;     break;
       case EXIT_PORT:	   portname = ATOM_exit;         break;
       case FAIL_PORT:	   portname = ATOM_fail;         break;
       case UNIFY_PORT:	   portname = ATOM_unify;	 break;
@@ -1168,7 +1183,6 @@ traceInterception(LocalFrame frame, Choice bfr, int port, Code PC)
 			      PL_TERM, LD->exception.pending) )
 	  goto out;
 	break;
-      case BREAK_PORT:     portfunc = FUNCTOR_break1;	 break;
       case CUT_CALL_PORT:  portfunc = FUNCTOR_cut_call1; break;
       case CUT_EXIT_PORT:  portfunc = FUNCTOR_cut_exit1; break;
       default:
@@ -1182,8 +1196,8 @@ traceInterception(LocalFrame frame, Choice bfr, int port, Code PC)
     } else if ( portfunc )
     { int pcn;
 
-      if ( PC && false(frame->predicate, FOREIGN) && frame->clause )
-	pcn = (int)(PC - frame->clause->clause->codes);
+      if ( PC && false(frame->predicate, P_FOREIGN) && frame->clause )
+	pcn = (int)(PC - frame->clause->value.clause->codes);
       else
 	pcn = 0;
 
@@ -1203,19 +1217,30 @@ traceInterception(LocalFrame frame, Choice bfr, int port, Code PC)
 
       if ( PL_get_atom(rarg, &a) )
       { if ( a == ATOM_continue )
-	  rval = ACTION_CONTINUE;
-	else if ( a == ATOM_nodebug )
+	{ rval = ACTION_CONTINUE;
+	} else if ( a == ATOM_nodebug )
 	{ rval = ACTION_CONTINUE;
 	  nodebug = TRUE;
 	} else if ( a == ATOM_fail )
-	  rval = ACTION_FAIL;
-	else if ( a == ATOM_retry )
-	  rval = ACTION_RETRY;
-	else if ( a == ATOM_ignore )
-	  rval = ACTION_IGNORE;
-	else if ( a == ATOM_abort )
-	  rval = ACTION_ABORT;
-	else
+	{ rval = ACTION_FAIL;
+	} else if ( a == ATOM_skip )
+	{ if ( !(port & CUT_PORT) )
+	  { LocalFrame fr;
+
+	    if ( PL_get_frame(argv+1, &fr) )
+	    { debugstatus.skiplevel = levelFrame(fr);
+	      set(fr, FR_SKIPPED);
+	    } else
+	      assert(0);
+	  }
+	  rval = ACTION_CONTINUE;
+	} else if ( a == ATOM_retry )
+	{ rval = ACTION_RETRY;
+	} else if ( a == ATOM_ignore )
+	{ rval = ACTION_IGNORE;
+	} else if ( a == ATOM_abort )
+	{ rval = ACTION_ABORT;
+	} else
 	  PL_warning("Unknown trace action: %s", stringAtom(a));
       } else if ( PL_is_functor(rarg, FUNCTOR_retry1) )
       { LocalFrame fr;
@@ -1242,53 +1267,145 @@ traceInterception(LocalFrame frame, Choice bfr, int port, Code PC)
   return rval;
 }
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Stores up to len bytes of a representation of the frame referenced by ref
-into buf. If ref is NULL, then the currently executing frame is used nextref
-gets the value of the parent of the frame
-Return value is 0: if there are no more frames (in which case nextref
-                   will be NULL)
-               >0: The number of bytes which are required to completely
-                   protray the frame
 
-Note that the text is returned as UTF-8, regardless of locale settings.
+		 /*******************************
+		 *	 SAFE STACK TRACE	*
+		 *******************************/
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+PL_get_context(pl_context_t *ctx, int tid)
+PL_step_context(pl_context_t *ctx)
+PL_describe_context(pl_context_t *ctx, char *buf, size_t len)
+
+These functions provide a public API  to   obtain  a trace of the Prolog
+stack in a fairly safe manner.
+
+    static void
+    dump_stack(void)
+    { pl_context_t ctx;
+
+      if ( PL_get_context(&ctx, 0) )
+      { int max = 5;
+
+	Sdprintf("Prolog stack:\n");
+
+	do
+	{ char buf[256];
+
+	  PL_describe_context(&ctx, buf, sizeof(buf));
+	  Sdprintf("  %s\n", buf);
+	} while ( max-- > 0 && PL_step_context(&ctx) );
+      } else
+	Sdprintf("No stack??\n");
+    }
+
+The second argument of PL_get_context() is a Prolog thread-id. Passing 0
+gets the context of the calling   thread. The current implementation can
+only deal with extracting the stack for  the calling thread, but the API
+is prepared to generalise this.
+
+See also PL_backtrace() and os/pl-cstack.c.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 int
-PL_walk_prolog_stack(void* ref, char* buf, size_t len, void** nextref)
+PL_get_context(pl_context_t *c, int thread_id)
 { GET_LD
-  LocalFrame fr;
+  (void)thread_id;
 
-  if (ref == NULL) /* Start of the chain - use current frame */
-  { LocalFrame topfr = environment_frame;
+  if ( !LD )
+    return FALSE;
 
-    if ( topfr->predicate->definition.function == pl_prolog_current_frame )
-      fr = parentFrame(topfr);
-    else
-      fr = topfr;
-  } else
-  { fr = (LocalFrame)ref;
+  c->ld = LD;
+  c->qf = LD->query;
+  if ( c->qf && c->qf->registers.fr )
+    c->fr = c->qf->registers.fr;
+  else
+    c->fr = environment_frame;
+  if ( c->qf && c->qf->registers.pc )
+    c->pc = c->qf->registers.pc;
+  else
+    c->pc = NULL;
+
+  return TRUE;
+}
+
+
+int
+PL_step_context(pl_context_t *c)
+{ if ( c->fr )
+  { GET_LD
+
+    if ( !onStack(local, c->fr) )
+      return FALSE;
+
+    if ( c->fr->parent )
+    { c->pc = c->fr->programPointer;
+      c->fr = c->fr->parent;
+    } else
+    { c->pc = NULL;
+      c->qf = queryOfFrame(c->fr);
+      c->fr = parentFrame(c->fr);
+    }
   }
 
-  if ( (*nextref = parentFrame(fr)) == NULL)
-    return 0;
+  return c->fr ? TRUE : FALSE;
+}
 
-  if ( fr->programPointer &&
-       false(fr->predicate, FOREIGN) &&
-       fr->parent &&
-       fr->parent->clause &&
-       fr->parent->predicate != PROCEDURE_dcall1->definition &&
-       fr->clause &&
-       fr->predicate != PROCEDURE_dc_call_prolog->definition)
-  { intptr_t pc = fr->programPointer - fr->parent->clause->clause->codes;
 
-    return snprintf(buf, len, "%s [PC=%ld] [Clause %d]",
-		    predicateName(fr->predicate),
-		    (long)pc,
-		    clauseNo(fr->predicate, fr->clause->clause));
-  } else
-  { return snprintf(buf, len, "%s <foreign>", predicateName(fr->predicate));
+int
+PL_describe_context(pl_context_t *c, char *buf, size_t len)
+{ LocalFrame fr;
+
+  buf[0] = 0;
+
+  if ( (fr=c->fr) )
+  { GET_LD
+    long level;
+    int printed;
+
+    if ( !onStack(local, fr) )
+      return snprintf(buf, len, "<invalid frame reference %p>", fr);
+
+    level = levelFrame(fr);
+    if ( !fr->predicate )
+      return snprintf(buf, len, "[%ld] <no predicate>", level);
+
+    printed = snprintf(buf, len, "[%ld] %s ", level, predicateName(fr->predicate));
+    len -= printed;
+    buf += printed;
+
+    if ( c->pc >= fr->predicate->codes &&
+	 c->pc < &fr->predicate->codes[fr->predicate->codes[-1]] )
+    { return printed+snprintf(buf, len, "[PC=%ld in supervisor]",
+			      (c->pc - fr->predicate->codes));
+    }
+
+    if ( false(fr->predicate, P_FOREIGN) )
+    { int clause_no = 0;
+      intptr_t pc = -1;
+
+      if ( fr->clause )
+      { Clause cl = fr->clause->value.clause;
+
+	if ( c->pc >= cl->codes && c->pc < &cl->codes[cl->code_size] )
+	  pc = c->pc - cl->codes;
+
+	if ( fr->predicate == PROCEDURE_dc_call_prolog->definition )
+	  return printed+snprintf(buf, len, "[PC=%ld in top query clause]",
+				  (long)pc);
+
+	clause_no = clauseNo(fr->predicate, cl);
+	return printed+snprintf(buf, len, "[PC=%ld in clause %d]",
+				(long)pc,
+				clause_no);
+      }
+      return printed+snprintf(buf, len, "<no clause>");
+    } else
+    { return printed+snprintf(buf, len, "<foreign>");
+    }
   }
+
+  return 0;
 }
 
 
@@ -1449,7 +1566,6 @@ interruptHandler(int sig)
 { GET_LD
   int c;
   int safe;
-  abort_type at = ABORT_RAISE;
 
   if ( !GD->initialised )
   { Sfprintf(Serror, "Interrupt during startup. Cannot continue\n");
@@ -1490,8 +1606,7 @@ again:
   if ( safe )
   { printMessage(ATOM_debug, PL_FUNCTOR, FUNCTOR_interrupt1, PL_ATOM, ATOM_begin);
   } else
-  { at = ABORT_THROW;
-    Sfprintf(Sdout, "\n%sAction (h for help) ? ", safe ? "" : "[forced] ");
+  { Sfprintf(Sdout, "\n%sAction (h for help) ? ", safe ? "" : "[forced] ");
     Sflush(Sdout);
   }
   ResetTty();                           /* clear pending input -- atoenne -- */
@@ -1500,11 +1615,17 @@ again:
   switch(c)
   { case 'a':	Sfputs("abort\n", Sdout);
 		unblockSignal(sig);
-		abortProlog(at);
+		abortProlog();
+		if ( !safe )
+		  PL_rethrow();
 		break;
     case 'b':	Sfputs("break\n", Sdout);
-		unblockSignal(sig);	/* into pl_break() itself */
-		pl_break();
+		if ( safe )
+		{ unblockSignal(sig);	/* into pl_break() itself */
+		  pl_break();
+		} else
+		{ Sfputs("Cannot break from forced interrupt\n", Sdout);
+		}
 		goto again;
     case 'c':	if ( safe )
 		{ printMessage(ATOM_debug, PL_FUNCTOR, FUNCTOR_interrupt1, PL_ATOM, ATOM_end);
@@ -1519,7 +1640,7 @@ again:
 		break;
 #ifdef O_DEBUGGER
     case 'g':	Sfputs("goals\n", Sdout);
-		backTrace(environment_frame, 5);
+		PL_backtrace(5, PL_BT_USER);
 		goto again;
 #endif /*O_DEBUGGER*/
     case 'h':
@@ -1555,7 +1676,7 @@ initTracer(void)
 
   debugstatus.visible      =
   debugstatus.leashing     = CALL_PORT|FAIL_PORT|REDO_PORT|EXIT_PORT|
-			     BREAK_PORT|EXCEPTION_PORT;
+			     EXCEPTION_PORT;
   debugstatus.showContext  = FALSE;
 
   resetTracer();
@@ -1567,6 +1688,17 @@ initTracer(void)
 
 #if O_DEBUGGER
 
+void
+suspendTrace(int suspend)
+{ GET_LD
+
+  if ( suspend )
+    debugstatus.suspendTrace++;
+  else
+    debugstatus.suspendTrace--;
+}
+
+
 int
 tracemode(int doit, int *old)
 { GET_LD
@@ -1602,21 +1734,27 @@ used to start tracing uncaught overflow exceptions.
 
 int
 trace_if_space(void)
-{ GET_LD;
+{ GET_LD
+  int trace;
 
-  if ( debugstatus.tracing == TRUE )
-    return debugstatus.tracing;
+#define minFreeStack(name, size) \
+	(spaceStack(name) > size*(int)sizeof(void*))
 
   if ( LD->outofstack )
-  { if ( spaceStack(local) > 10000 * sizeof(void*) &&
-	 spaceStack(global) > 10000 * sizeof(void*) &&
-	 spaceStack(trail) > 10000 * sizeof(void*) )
-      tracemode(TRUE, NULL);
+  { if ( minFreeStack(local,  50000) &&
+	 minFreeStack(global, 50000) &&
+	 minFreeStack(trail,  20000) )
+      trace = TRUE;
+    else
+      trace = FALSE;
   } else
-  { tracemode(TRUE, NULL);
+  { trace = TRUE;
   }
 
-  return debugstatus.tracing;
+  if ( trace )
+    tracemode(trace, NULL);
+
+  return trace;
 }
 
 
@@ -1660,27 +1798,20 @@ debugmode(debug_type doit, debug_type *old)
     { debugstatus.skiplevel = SKIP_VERY_DEEP;
       clearPrologFlagMask(PLFLAG_LASTCALL);
       if ( doit == DBG_ALL )
-      { LocalFrame fr = environment_frame;
+      { QueryFrame qf;
+
+	for(qf = LD->query; qf; qf = qf->parent)
+	  qf->debugSave = DBG_ON;
 
-	while( fr )
-	{ if ( fr->parent )
-	    fr = fr->parent;
-	  else
-	  { QueryFrame qf = queryOfFrame(fr);
-	    qf->debugSave = DBG_ON;
-	    fr = qf->saved_environment;
-	  }
-	}
 	doit = DBG_ON;
       }
-    } else
-    { setPrologFlagMask(PLFLAG_LASTCALL);
-    }
-    if ( doit )
       enlargeMinFreeStacks(8*1024*SIZEOF_VOIDP,
 			   8*1024*SIZEOF_VOIDP,
 			   8*1024*SIZEOF_VOIDP
 			   PASS_LD);
+    } else
+    { setPrologFlagMask(PLFLAG_LASTCALL);
+    }
     debugstatus.debugging = doit;
     updateAlerted(LD);
     printMessage(ATOM_silent,
@@ -1841,28 +1972,49 @@ pl_prolog_current_frame(term_t frame)
 { GET_LD
   LocalFrame fr = environment_frame;
 
-  if ( fr->predicate->definition.function == pl_prolog_current_frame )
+  if ( fr->predicate->impl.function == pl_prolog_current_frame )
     fr = parentFrame(fr);		/* thats me! */
 
   return PL_unify_frame(frame, fr);
 }
 
 
+/** prolog_current_choice(-Choice) is semidet.
+
+True when Choice refers to the most recent choice-point.
+*/
+
+static
+PRED_IMPL("prolog_current_choice", 1, prolog_current_choice, 0)
+{ PRED_LD
+  Choice ch = LD->choicepoints;
+
+  while(ch && ch->type == CHP_DEBUG)
+    ch = ch->parent;
+  if ( ch )
+    return PL_unify_choice(A1, ch);
+
+  return FALSE;
+}
+
+
 static int
-prolog_frame_attribute(term_t frame, term_t what,
-		       term_t value)
+prolog_frame_attribute(term_t frame, term_t what, term_t value)
 { GET_LD
   LocalFrame fr;
   atom_t key;
   int arity;
   term_t result = PL_new_term_ref();
+  Module m = NULL;
 
   if ( !PL_get_frame(frame, &fr) )
     return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_frame_reference, frame);
   if ( !fr )
-    fail;				/* frame == 'none' */
+    return FALSE;				/* frame == 'none' */
   if ( !PL_get_name_arity(what, &key, &arity) )
     return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_callable, what);
+  if ( !PL_strip_module(value, &m, value) )
+    return FALSE;
 
   set(fr, FR_WATCHED);			/* explicit call to do this? */
 
@@ -1875,11 +2027,11 @@ prolog_frame_attribute(term_t frame, term_t what,
     if ( argn < 1 )
       return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_natural, arg);
 
-    if ( true(fr->predicate, FOREIGN) || !fr->clause )
+    if ( true(fr->predicate, P_FOREIGN) || !fr->clause )
     { if ( argn > fr->predicate->functor->arity )
 	fail;
     } else
-    { if ( argn > fr->clause->clause->prolog_vars )
+    { if ( argn > fr->clause->value.clause->prolog_vars )
 	fail;
     }
 
@@ -1908,6 +2060,8 @@ prolog_frame_attribute(term_t frame, term_t what,
   { PL_put_integer(result, levelFrame(fr));
   } else if (key == ATOM_has_alternatives)
   { PL_put_atom(result, hasAlternativesFrame(fr) ? ATOM_true : ATOM_false);
+  } else if (key == ATOM_skipped)
+  { PL_put_atom(result, true(fr, FR_SKIPPED) ? ATOM_true : ATOM_false);
   } else if (key == ATOM_alternative)
   { LocalFrame alt;
 
@@ -1930,10 +2084,10 @@ prolog_frame_attribute(term_t frame, term_t what,
   } else if (key == ATOM_context_module)
   { PL_put_atom(result, contextModule(fr)->name);
   } else if (key == ATOM_clause)
-  { if ( false(fr->predicate, FOREIGN) &&
+  { if ( false(fr->predicate, P_FOREIGN) &&
 	 fr->clause &&
 	 fr->predicate != PROCEDURE_dc_call_prolog->definition )
-    { if ( !PL_unify_clref(result, fr->clause->clause) )
+    { if ( !PL_unify_clref(result, fr->clause->value.clause) )
 	return FALSE;
     } else
     { return FALSE;
@@ -1943,7 +2097,7 @@ prolog_frame_attribute(term_t frame, term_t what,
     term_t arg = PL_new_term_ref();
     Definition def = fr->predicate;
 
-    if ( def->module != MODULE_user )
+    if ( def->module != m )
     { if ( !PL_put_functor(result, FUNCTOR_colon2) )
 	return FALSE;
       _PL_get_arg(1, result, arg);
@@ -1985,7 +2139,7 @@ prolog_frame_attribute(term_t frame, term_t what,
       }
     }
   } else if ( key == ATOM_predicate_indicator )
-  { if ( !unify_definition(MODULE_user, result, fr->predicate, 0, GP_NAMEARITY) )
+  { if ( !unify_definition(m, result, fr->predicate, 0, GP_NAMEARITY) )
       return FALSE;
   } else if ( key == ATOM_parent_goal )
   { Procedure proc;
@@ -2024,10 +2178,10 @@ prolog_frame_attribute(term_t frame, term_t what,
   } else if ( key == ATOM_pc )
   { if ( fr->programPointer &&
 	 fr->parent &&
-	 false(fr->parent->predicate, FOREIGN) &&
+	 false(fr->parent->predicate, P_FOREIGN) &&
 	 fr->parent->clause &&
 	 fr->parent->predicate != PROCEDURE_dcall1->definition )
-    { intptr_t pc = fr->programPointer - fr->parent->clause->clause->codes;
+    { intptr_t pc = fr->programPointer - fr->parent->clause->value.clause->codes;
 
       PL_put_intptr(result, pc);
     } else
@@ -2071,10 +2225,10 @@ prolog_frame_attribute(term_t frame, term_t what,
 */
 
 static
-PRED_IMPL("prolog_frame_attribute", 3, prolog_frame_attribute, 0)
+PRED_IMPL("prolog_frame_attribute", 3, prolog_frame_attribute, PL_FA_TRANSPARENT)
 { int rc = prolog_frame_attribute(A1, A2, A3);
 
-  SECURE(scan_global(0));
+  DEBUG(CHK_SECURE, scan_global(0));
 
   return rc;
 }
@@ -2087,6 +2241,22 @@ PRED_IMPL("prolog_frame_attribute", 3, prolog_frame_attribute, 0)
 
 */
 
+static size_t
+in_clause_jump(Choice ch)
+{ Clause cl;
+
+  if ( ch->type == CHP_JUMP &&
+       false(ch->frame->predicate, P_FOREIGN) &&
+       ch->frame->clause &&
+       (cl=ch->frame->clause->value.clause) &&
+       ch->value.PC >= cl->codes &&
+       ch->value.PC < &cl->codes[cl->code_size] )
+    return ch->value.PC - cl->codes;
+
+  return (size_t)-1;
+}
+
+
 static
 PRED_IMPL("prolog_choice_attribute", 3, prolog_choice_attribute, 0)
 { PRED_LD
@@ -2098,8 +2268,12 @@ PRED_IMPL("prolog_choice_attribute", 3, prolog_choice_attribute, 0)
     fail;
 
   if ( key == ATOM_parent )
-  { if ( ch->parent )
-      return PL_unify_choice(A3, ch->parent);
+  { do
+    { ch = ch->parent;
+    } while(ch && ch->type == CHP_DEBUG);
+
+    if ( ch )
+      return PL_unify_choice(A3, ch);
     fail;
   } else if ( key == ATOM_frame )
   { return PL_unify_frame(A3, ch->frame);
@@ -2107,17 +2281,26 @@ PRED_IMPL("prolog_choice_attribute", 3, prolog_choice_attribute, 0)
   { static const atom_t types[] =
     { ATOM_jump,
       ATOM_clause,
-      ATOM_foreign,
       ATOM_top,
       ATOM_catch,
-      ATOM_debug,
-      ATOM_none
+      ATOM_debug
     };
 
-    return PL_unify_atom(A3, types[ch->type]);
+    if ( ch->type == CHP_JUMP &&
+	 in_clause_jump(ch) == (size_t)-1 )
+    { if ( ch->value.PC == SUPERVISOR(next_clause) )
+	return PL_unify_atom(A3, ATOM_clause);
+      if ( decode(ch->value.PC[0]) == I_FREDO )
+	return PL_unify_atom(A3, ATOM_foreign);
+      assert(0);
+      return FALSE;
+    } else
+      return PL_unify_atom(A3, types[ch->type]);
   } else if ( key == ATOM_pc )
-  { if ( ch->type == CHP_JUMP && ch->frame->clause )
-      return PL_unify_int64(A3, ch->value.PC - ch->frame->clause->clause->codes);
+  { size_t offset = in_clause_jump(ch);
+
+    if ( offset != (size_t)-1 )
+      return PL_unify_int64(A3, offset);
     return FALSE;
   } else
     return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_key, A2);
@@ -2135,7 +2318,7 @@ callEventHook(int ev, ...)
 { if ( !PROCEDURE_event_hook1 )
     PROCEDURE_event_hook1 = PL_predicate("prolog_event_hook", 1, "user");
 
-  if ( PROCEDURE_event_hook1->definition->definition.clauses )
+  if ( PROCEDURE_event_hook1->definition->impl.any )
   { GET_LD
     wakeup_state wstate;
     int rc;
@@ -2148,8 +2331,12 @@ callEventHook(int ev, ...)
 
     va_start(args, ev);
     switch(ev)
-    { case PLEV_ERASED_CLAUSE:
-      {	Clause cl = va_arg(args, Clause); 	/* object erased */
+    { case PLEV_ABORT:
+      { rc = PL_unify_atom(arg, ATOM_abort);
+	break;
+      }
+      case PLEV_ERASED_CLAUSE:
+      {	Clause cl = va_arg(args, Clause);	/* object erased */
 	term_t dbref = PL_new_term_ref();
 
 	rc = (  PL_unify_clref(dbref, cl) &&
@@ -2160,7 +2347,7 @@ callEventHook(int ev, ...)
 	break;
       }
       case PLEV_ERASED_RECORD:
-      {	RecordRef r = va_arg(args, RecordRef); 	/* object erased */
+      {	RecordRef r = va_arg(args, RecordRef);	/* object erased */
 	term_t dbref = PL_new_term_ref();
 
 	rc = (  PL_unify_recref(dbref, r) &&
@@ -2253,7 +2440,8 @@ callEventHook(int ev, ...)
 		 *******************************/
 
 BeginPredDefs(trace)
-  PRED_DEF("prolog_frame_attribute", 3, prolog_frame_attribute, 0)
+  PRED_DEF("prolog_current_choice", 1, prolog_current_choice, 0)
+  PRED_DEF("prolog_frame_attribute", 3, prolog_frame_attribute, PL_FA_TRANSPARENT)
   PRED_DEF("prolog_choice_attribute", 3, prolog_choice_attribute, 0)
   PRED_DEF("prolog_skip_frame", 1, prolog_skip_frame, PL_FA_NOTRACE)
   PRED_DEF("prolog_skip_level", 2, prolog_skip_level, PL_FA_NOTRACE)
