@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        J.Wielemaker@uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2008, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,10 +18,9 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
-/*#define O_SECURE 1*/
 /*#define O_DEBUG 1*/
 #include "pl-incl.h"
 #include "pl-inline.h"
@@ -77,7 +75,7 @@ count(code c, Code PC)
       if ( !counting[c].vartimesptr )
       { int bytes = sizeof(int)*MAXVAR;
 
-	counting[c].vartimesptr = allocHeap(bytes);
+	counting[c].vartimesptr = allocHeapOrHalt(bytes);
 	memset(counting[c].vartimesptr, 0, bytes);
       }
       counting[c].vartimesptr[v]++;
@@ -190,10 +188,10 @@ DbgPrintInstruction(LocalFrame FR, Code PC)
 	    if ( offset >= 0 && offset < (int)FR->predicate->codes[-1] )
 	      relto = FR->predicate->codes;
 	  } else if ( FR->clause )
-	  { relto = FR->clause->clause->codes;
+	  { relto = FR->clause->value.clause->codes;
 	  }
 
-	  Sdprintf("\t%s\n", codeTable[decode(*PC)].name);
+	  Sdprintf("\t%4ld %s\n", (long)(PC-relto), codeTable[decode(*PC)].name);
 	});
 }
 
@@ -340,7 +338,7 @@ open_foreign_frame(ARG1_LD)
   lTop = (LocalFrame)(fr+1);
   fr->size = 0;
   Mark(fr->mark);
-  SECURE(assert(fr>fli_context));
+  DEBUG(CHK_SECURE, assert(fr>fli_context));
   fr->parent = fli_context;
   fr->magic = FLI_MAGIC;
   fli_context = fr;
@@ -400,27 +398,21 @@ fid_t
 PL_open_signal_foreign_frame(int sync)
 { GET_LD
   FliFrame fr;
-  size_t margin = sizeof(struct localFrame);
+  size_t minspace = sizeof(struct localFrame) + MINFOREIGNSIZE*sizeof(word);
+  size_t margin   = sync ? 0 : MAXARITY*sizeof(word);
 
-  if ( sync )
-    margin += MINFOREIGNSIZE*sizeof(word);
-  else
-    margin += MAXARITY*sizeof(word);
-
-  if ( (char*)lTop + margin > (char*)lMax )
+  if ( (char*)lTop + minspace + margin  > (char*)lMax )
   { if ( sync )
     { int rc;
 
-      if ( (rc=ensureLocalSpace(margin, ALLOW_SHIFT)) != TRUE )
-	return FALSE;
+      if ( (rc=ensureLocalSpace(minspace, ALLOW_SHIFT)) != TRUE )
+	return 0;
     } else
-    { return FALSE;
+    { return 0;
     }
   }
 
-  fr = (FliFrame) lTop;
-  lTop = addPointer(fr, margin);
-
+  fr = addPointer(lTop, margin);
   fr->magic = FLI_MAGIC;
   fr->size = 0;
   Mark(fr->mark);
@@ -513,12 +505,12 @@ static void
 discardForeignFrame(LocalFrame fr ARG_LD)
 { Definition def = fr->predicate;
   int argc       = def->functor->arity;
-  Func function  = def->definition.function;
+  Func function  = def->impl.function;
   struct foreign_context context;
   fid_t fid;
 
-  DEBUG(5, Sdprintf("\tCut %s, context = 0x%lx\n",
-		    predicateName(def), context));
+  DEBUG(5, Sdprintf("\tCut %s, context = %p\n",
+		    predicateName(def), fr->clause));
 
   context.context = (word)fr->clause;
   context.control = FRG_CUTTED;
@@ -562,7 +554,7 @@ unify_finished(term_t catcher, enum finished reason)
   { functor_t f = (reason == FINISH_EXCEPT ? FUNCTOR_exception1
 					   : FUNCTOR_external_exception1);
 
-    SECURE(checkData(valTermRef(exception_bin)));
+    DEBUG(CHK_SECURE, checkData(valTermRef(exception_bin)));
 
     return PL_unify_term(catcher,
 			 PL_FUNCTOR, f,
@@ -608,34 +600,16 @@ callCleanupHandler(LocalFrame fr, enum finished reason ARG_LD)
     { term_t clean;
       term_t ex = 0;
       int rval;
-      int set_env = (reason == FINISH_CUT ||
-		     reason == FINISH_EXCEPT ||
-		     reason == FINISH_EXTERNAL_EXCEPT);
-      term_t esave;
       wakeup_state wstate;
 
       fr = (LocalFrame)valTermRef(fref);
-
-      if ( set_env )
-      { esave = consTermRef(environment_frame);
-	environment_frame = fr;
-      } else
-      { esave = 0;			/* keep compiler happy */
-	assert(environment_frame == fr);
-      }
-
       clean = consTermRef(argFrameP(fr, 3));
-      startCritical;
       if ( saveWakeup(&wstate, FALSE PASS_LD) )
       { rval = callProlog(contextModule(fr), clean, PL_Q_CATCH_EXCEPTION, &ex);
 	restoreWakeup(&wstate PASS_LD);
       } else
       { rval = FALSE;
       }
-      endCritical;
-
-      if ( set_env )
-	environment_frame = (LocalFrame)valTermRef(esave);
 
       if ( !rval && ex && !exception_term )
 	PL_raise_exception(ex);
@@ -752,7 +726,7 @@ __do_undo(mark *m ARG_LD)
 
   tTop = mt;
   if ( LD->frozen_bar > m->globaltop )
-  { SECURE(assert(gTop >= LD->frozen_bar));
+  { DEBUG(CHK_SECURE, assert(gTop >= LD->frozen_bar));
     gTop = LD->frozen_bar;
   } else
   { gTop = m->globaltop;
@@ -775,58 +749,26 @@ do_undo(mark *m)
 		 *	    PROCEDURES		*
 		 *******************************/
 
-#ifdef _MSC_VER				/* Windows MSVC version */
-
-static inline int
-MSB(unsigned int i)
-{ unsigned long mask = i;
-  unsigned long index;
-
-  _BitScanReverse(&index, mask);
-  return index;
-}
-
-#ifndef MemoryBarrier
-#define MemoryBarrier() (void)0
-#endif
-
-#elif defined(__GNUC__)			/* GCC version */
-
-#define MSB(i) (31 - __builtin_clz(i))
-#define MemoryBarrier() __sync_synchronize()
-
-#else					/* Other */
-
-static inline int
-MSB(unsigned int i)
-{ int j = 0;
-
-  if (i >= 0x10000) {i >>= 16; j += 16;}
-  if (i >=   0x100) {i >>=  8; j +=  8;}
-  if (i >=    0x10) {i >>=  4; j +=  4;}
-  if (i >=     0x4) {i >>=  2; j +=  2;}
-  if (i >=     0x2) j++;
-
-  return j;
-}
-#define MemoryBarrier() (void)0
-
-#endif
+/* Note that we use PL_malloc_uncollectable() here because the pointer in
+   our block is not the real memory pointer.
+*/
 
 static Definition
 localDefinition(Definition def ARG_LD)
 { unsigned int tid = LD->thread.info->pl_tid;
   size_t idx = MSB(tid);
-  LocalDefinitions v = def->definition.local;
+  LocalDefinitions v = def->impl.local;
 
   if ( !v->blocks[idx] )
   { LOCKDYNDEF(def);
     if ( !v->blocks[idx] )
     { size_t bs = (size_t)1<<idx;
-      Definition *newblock = allocHeap(bs*sizeof(Definition));
+      Definition *newblock;
+
+      if ( !(newblock=PL_malloc_uncollectable(bs*sizeof(Definition))) )
+	outOfCore();
 
       memset(newblock, 0, bs*sizeof(Definition));
-
       v->blocks[idx] = newblock-bs;
     }
     UNLOCKDYNDEF(def);
@@ -842,7 +784,7 @@ localDefinition(Definition def ARG_LD)
 void
 destroyLocalDefinition(Definition def, unsigned int tid)
 { size_t idx = MSB(tid);
-  LocalDefinitions v = def->definition.local;
+  LocalDefinitions v = def->impl.local;
   Definition local;
 
   local = v->blocks[idx][tid];
@@ -866,7 +808,7 @@ getProcDefinition__LD(Definition def ARG_LD)
 
 static inline Definition
 getProcDefinedDefinition(Definition def ARG_LD)
-{ if ( !def->definition.clauses && false(def, PROC_DEFINED) )
+{ if ( !def->impl.any && false(def, PROC_DEFINED) )
     def = trapUndefined(def PASS_LD);
 
 #ifdef O_PLMT
@@ -1073,7 +1015,7 @@ static Code
 findCatchExit()
 { if ( !GD->exceptions.catch_exit_address )
   { Definition catch3 = PROCEDURE_catch3->definition;
-    Clause cl = catch3->definition.clauses->clause;
+    Clause cl = catch3->impl.clauses.first_clause->value.clause;
     Code Exit = &cl->codes[cl->code_size-1];
     assert(*Exit == encode(I_EXIT));
 
@@ -1096,7 +1038,7 @@ reference, so we can deal with relocation of the local stack.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static term_t
-findCatcher(LocalFrame fr, term_t ex ARG_LD)
+findCatcher(LocalFrame fr, Choice ch, term_t ex ARG_LD)
 { Definition catch3  = PROCEDURE_catch3->definition;
 
   for(; fr; fr = fr->parent)
@@ -1106,7 +1048,9 @@ findCatcher(LocalFrame fr, term_t ex ARG_LD)
     if ( fr->predicate != catch3 )
       continue;
     if ( true(fr, FR_CATCHED) )
-      continue;
+      continue;				/* thrown from recover */
+    if ( (void*)fr > (void*)ch )
+      continue;				/* call-port of catch/3 */
 
     tref = consTermRef(fr);
     rc = PL_unify(consTermRef(argFrameP(fr, 1)), ex);
@@ -1184,10 +1128,10 @@ static inline int
 slotsInFrame(LocalFrame fr, Code PC)
 { Definition def = fr->predicate;
 
-  if ( !PC || true(def, FOREIGN) || !fr->clause )
+  if ( !PC || true(def, P_FOREIGN) || !fr->clause )
     return def->functor->arity;
 
-  return fr->clause->clause->prolog_vars;
+  return fr->clause->value.clause->prolog_vars;
 }
 
 
@@ -1202,23 +1146,33 @@ non-debug code. We walk up the stack to   find  a debug frame. If we are
 already in the box of this frame, we   have  an internal retry of called
 system predicate, which we should not trace. If we are outside the `box'
 however, we must trace the toplevel visible predicate.
+
+FR_INBOX is maintained when the debugger is   active.  It is set on CALL
+and REDO and reset on EXIT. So, if we find this set, we are dealing with
+an internal retry and otherwise we have an external retry.
+
+The cht argument is one of CHP_CLAUSE   or CHP_JUMP, indicating the type
+of redo.  We always show redo for an external redo.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static LocalFrame
-dbgRedoFrame(LocalFrame fr ARG_LD)
-{ DEBUG(1, Sdprintf("REDO on %s\n", predicateName(fr->predicate)));
+dbgRedoFrame(LocalFrame fr, choice_type cht ARG_LD)
+{ DEBUG(MSG_TRACE, Sdprintf("REDO on [%d] %s\n",
+			    (int)levelFrame(fr), predicateName(fr->predicate)));
 
   if ( SYSTEM_MODE )
-    return fr;
-  if ( isDebugFrame(fr) )
-    return true(fr->predicate, HIDE_CHILDS) ? NULL : fr;
+    return fr;				/* system mode; debug everything */
+  if ( isDebugFrame(fr) && false(fr->predicate, HIDE_CHILDS) )
+    return fr;				/* normal user code */
   for( ; fr && !isDebugFrame(fr); fr = fr->parent)
-    ;
-  DEBUG(1, if ( fr )
-	Sdprintf("REDO user frame of %s\n",
-		 predicateName(fr->predicate)));
+    ;					/* find top of hidden children */
+  DEBUG(MSG_TRACE, if ( fr )
+	Sdprintf("REDO user frame of [%d] %s%s\n",
+		 (int)levelFrame(fr),
+		 predicateName(fr->predicate),
+		 true(fr, FR_INBOX) ? " (inbox)" : ""));
   if ( fr && false(fr, FR_INBOX) )
-  { set(fr, FR_INBOX);
+  { set(fr, FR_INBOX);			/* External retry */
     return fr;
   }
 
@@ -1229,7 +1183,7 @@ dbgRedoFrame(LocalFrame fr ARG_LD)
 
 static int
 exception_hook(LocalFrame fr, term_t catchfr_ref ARG_LD)
-{ if ( PROCEDURE_exception_hook4->definition->definition.clauses )
+{ if ( PROCEDURE_exception_hook4->definition->impl.clauses.first_clause )
   { if ( !LD->exception.in_hook )
     { wakeup_state wstate;
       qid_t qid;
@@ -1386,8 +1340,12 @@ choice_type last_choice;
 #define FRAME_FAILED		GO(frame_failed)
 #define CLAUSE_FAILED		GO(clause_failed)
 #define BODY_FAILED		GO(body_failed)
+#ifdef O_DEBUG
 #define THROW_EXCEPTION		do { throwed_from_line = __LINE__; \
 				     goto b_throw; } while(0)
+#else
+#define THROW_EXCEPTION		goto b_throw
+#endif
 
 #ifdef O_PROFILE
 #define Profile(g) if ( unlikely(LD->profile.active) ) g
@@ -1404,7 +1362,7 @@ choice_type last_choice;
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static void
-leaveFrame(LocalFrame fr ARG_LD)
+leaveFrame(LocalFrame fr)
 { Definition def = fr->predicate;
 
   fr->clause = NULL;
@@ -1420,7 +1378,7 @@ discardFrame(LocalFrame fr ARG_LD)
 		    loffset(fr),
 		    predicateName(fr->predicate)));
 
-  if ( true(def, FOREIGN) )
+  if ( true(def, P_FOREIGN) )
   { if ( fr->clause )
     { discardForeignFrame(fr PASS_LD);
       fr->clause = NULL;
@@ -1461,67 +1419,70 @@ discardChoicesAfter() discards all choicepoints created  after fr, while
 calling possible hooks on the frames.   It return the oldest choicepoint
 created after fr was created or NULL if this doesn't exist.
 
-(*) Note that we cannot put discardFrame() in the first loop. See C_CUT.
-(**) Running GC goes wrong because discardFrame() sets clause to NULL.
-     If we have a frame where a variable is a ref to the parent frame
-     and where the var in the parent points to the global stack and
-     the clause of this parent is NULL, the variable is marked, but
-     not sweeped.
-
-     What we would need is a way to discard the choice and the frames
-     that can be reached from it, but *not* the frame that can be
-     reached from older choice-points.  How can we do this?
-
-     Please note this isn't too bad: cleanup handlers are supposed to
-     do small jobs.
+GC interaction is tricky here. See also   C_CUT.  The loop needs to call
+the  cleanup  handlers  and  call   discardFrame().  The  latter  resets
+LocalFrame->clause to NULL. This means  that   these  frames  may not be
+visible to GC. In older versions, this was achieved using two loops: one
+for the cleanup and one to discard the  frames, but if Undo() is called,
+the resets may corrupt the datastructures,   which makes the second loop
+fail. We now moved discardFrame() back into the primary loop and set BFR
+before calling frameFinished() such that the discarded frames are really
+invisible to GC.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static Choice
 discardChoicesAfter(LocalFrame fr, enum finished reason ARG_LD)
 { if ( (LocalFrame)BFR > fr )
-  { for(;;)
-    { Choice me = BFR;
-      LocalFrame fr2;
+  { Choice me;
+
+    for(me = BFR; ; me=me->parent)
+    { LocalFrame fr2;
+      LocalFrame delto;
+
+      if ( me->parent && me->parent->frame > fr )
+	delto = me->parent->frame;
+      else
+	delto = fr;
 
       DEBUG(3, Sdprintf("Discarding %s\n", chp_chars(me)));
       for(fr2 = me->frame;
-	  fr2 && fr2->clause && fr2 > fr;
+	  fr2 > delto;
 	  fr2 = fr2->parent)
-      { if ( true(fr2, FR_WATCHED) )
-	{ if ( reason == FINISH_EXCEPT ||
+      { assert(fr2->clause || true(fr2->predicate, P_FOREIGN));
+
+	if ( true(fr2, FR_WATCHED) )
+	{ char *lSave = (char*)lBase;
+
+	  if ( reason == FINISH_EXCEPT ||
 	       reason == FINISH_EXTERNAL_EXCEPT )
 	    Undo(me->mark);
-	  blockGC(ALLOW_SHIFT PASS_LD);	/* (**) */
+	  BFR = me->parent;
 	  frameFinished(fr2, reason PASS_LD);
-	  unblockGC(ALLOW_SHIFT PASS_LD);
-	  if ( me != BFR )		/* shifted */
-	  { intptr_t offset = (char*)BFR - (char*)me;
+	  if ( lSave != (char*)lBase )	/* shifted */
+	  { intptr_t offset = (char*)lBase - lSave;
 
 	    me  = addPointer(me, offset);
+	    me->parent = BFR;		/* not updated because BFR=me->parent */
 	    fr  = addPointer(fr, offset);
 	    fr2 = addPointer(fr2, offset);
+	    delto = addPointer(delto, offset);
+	    fr2->parent = addPointer(fr2->parent, offset);
 	  }
 #if 0					/* What to do if we have multiple */
 	  if ( exception_term )		/* handlers and multiple exceptions? */
 	    break;
 #endif
 	}
+
+	discardFrame(fr2 PASS_LD);
       }
 
-      for(fr2 = me->frame;		/* (*) */
-	  fr2 && fr2->clause && fr2 > fr;
-	  fr2 = fr2->parent)
-      { discardFrame(fr2 PASS_LD);
-      }
-
-      BFR = me->parent;
-      if ( (LocalFrame)BFR <= fr )
+      if ( (LocalFrame)me->parent <= fr )
       { DiscardMark(me->mark);
+	BFR = me->parent;
 	return me;
       }
     }
-
-    DEBUG(3, Sdprintf(" --> BFR = #%ld\n", loffset(BFR)));
   }
 
   return NULL;
@@ -1572,7 +1533,7 @@ static Choice
 newChoice(choice_type type, LocalFrame fr ARG_LD)
 { Choice ch = (Choice)lTop;
 
-  SECURE(assert(ch+1 <= (Choice)lMax));
+  DEBUG(CHK_SECURE, assert(ch+1 <= (Choice)lMax));
   lTop = (LocalFrame)(ch+1);
 
   ch->type = type;
@@ -1609,13 +1570,14 @@ PL_open_query(Module ctx, int flags, Procedure proc, term_t args)
   size_t lneeded;
   static int top_initialized = FALSE;
   static struct clause clause;
-  static struct clause_ref cref = {&clause};
+  static struct clause_ref cref;
 
   if ( !top_initialized )
   { clause.procedure = PROCEDURE_dc_call_prolog;
-    clause.generation.erased = ~0L;
+    clause.generation.erased = ~(gen_t)0;
     clause.code_size = 1;
     clause.codes[0] = encode(I_EXITQUERY);
+    cref.value.clause = &clause;
 
     top_initialized = TRUE;
   }
@@ -1631,9 +1593,9 @@ PL_open_query(Module ctx, int flags, Procedure proc, term_t args)
 	     }
 	     Sdprintf(")\n");
 	   });
-  SECURE(checkStacks(NULL));
+  DEBUG(CHK_SECURE, checkStacks(NULL));
   assert((void*)fli_context > (void*)environment_frame);
-  assert((void*)lTop >= (void*)(fli_context+1));
+  assert((Word)lTop >= refFliP(fli_context, fli_context->size));
 
 					/* resolve can call-back */
   def = getProcDefinedDefinition(proc->definition PASS_LD);
@@ -1655,16 +1617,18 @@ PL_open_query(Module ctx, int flags, Procedure proc, term_t args)
 					/* should be struct alignment, */
 					/* but for now, I think this */
 					/* is always the same */
+  qf = (QueryFrame)lTop;
 #ifdef JMPBUF_ALIGNMENT
-  while ( (uintptr_t)lTop % JMPBUF_ALIGNMENT )
-    lTop = addPointer(lTop, sizeof(word));
+  while ( (uintptr_t)qf % JMPBUF_ALIGNMENT )
+    qf = addPointer(qf, sizeof(word));
 #endif
+  qf->saved_ltop = lTop;
 
-  qf	             = (QueryFrame) lTop;
 					/* fill top-frame */
   top	             = &qf->top_frame;
   top->parent        = NULL;
   top->predicate     = PROCEDURE_dc_call_prolog->definition;
+  top->programPointer= NULL;
   top->clause        = &cref;
 #ifdef O_PROFILE
   if ( LD->profile.active )
@@ -1673,9 +1637,11 @@ PL_open_query(Module ctx, int flags, Procedure proc, term_t args)
     top->prof_node = NULL;
 #endif
   if ( environment_frame )
-    setNextFrameFlags(top, environment_frame);
-  else
-    top->flags	     = 0;
+  { setNextFrameFlags(top, environment_frame);
+  } else
+  { top->flags	     = 0;
+    top->level	     = 0;
+  }
   fr                 = &qf->frame;
   fr->parent         = top;
   setNextFrameFlags(fr, top);
@@ -1683,7 +1649,7 @@ PL_open_query(Module ctx, int flags, Procedure proc, term_t args)
   fr->programPointer = clause.codes;
   arity		     = def->functor->arity;
 
-  SECURE(checkStacks(NULL));
+  DEBUG(CHK_SECURE, checkStacks(NULL));
   assert((uintptr_t)fli_context > (uintptr_t)environment_frame);
   assert((uintptr_t)lTop >= (uintptr_t)(fli_context+1));
 
@@ -1717,7 +1683,7 @@ PL_open_query(Module ctx, int flags, Procedure proc, term_t args)
   DEBUG(3, Sdprintf("Level = %d\n", levelFrame(fr)));
   if ( true(qf, PL_Q_NODEBUG) )
   { set(fr, FR_HIDE_CHILDS);
-    debugstatus.suspendTrace++;
+    suspendTrace(TRUE);
     qf->debugSave = debugstatus.debugging;
     debugstatus.debugging = DBG_OFF;
     qf->flags_saved = (LD->prolog_flag.mask.flags & NDEBUG_SAVE_FLAGS);
@@ -1739,15 +1705,7 @@ PL_open_query(Module ctx, int flags, Procedure proc, term_t args)
   fr->prof_node = NULL;			/* true? */
 #endif
   Mark(qf->choice.mark);
-
-  if ( true(def, FOREIGN) )
-  { fr->clause = NULL;			/* initial context */
-  } else
-  { fr->clause = def->definition.clauses;
-  }
-#ifdef O_LOGICAL_UPDATE
-  fr->generation = GD->generation;
-#endif
+  setGenerationFrame(fr, GD->generation);
 					/* context module */
   if ( true(def, P_TRANSPARENT) )
   { if ( ctx )
@@ -1759,12 +1717,12 @@ PL_open_query(Module ctx, int flags, Procedure proc, term_t args)
   }
 
 					/* publish environment */
-  PL_LOCK(L_AGC);			/* see restore_after_query() */
+  PL_LOCK(L_STOPTHEWORLD);		/* see restore_after_query() */
   LD->choicepoints  = &qf->choice;
   environment_frame = fr;
   qf->parent = LD->query;
   LD->query = qf;
-  PL_UNLOCK(L_AGC);
+  PL_UNLOCK(L_STOPTHEWORLD);
 
   DEBUG(2, Sdprintf("QID=%d\n", QidFromQuery(qf)));
   updateAlerted(LD);
@@ -1803,15 +1761,15 @@ restore_after_query(QueryFrame qf)
 
   DiscardMark(qf->choice.mark);
 
-  PL_LOCK(L_AGC);		/* see Tests/thread/test_agc_callback.pl */
+  PL_LOCK(L_STOPTHEWORLD);	/* see Tests/thread/test_agc_callback.pl */
   LD->query         = qf->parent;
   LD->choicepoints  = qf->saved_bfr;
   environment_frame = qf->saved_environment;
-  PL_UNLOCK(L_AGC);
+  PL_UNLOCK(L_STOPTHEWORLD);
   aTop		    = qf->aSave;
-  lTop		    = (LocalFrame)qf;
+  lTop		    = qf->saved_ltop;
   if ( true(qf, PL_Q_NODEBUG) )
-  { debugstatus.suspendTrace--;
+  { suspendTrace(FALSE);
     debugstatus.debugging = qf->debugSave;
     LD->prolog_flag.mask.flags &= (~NDEBUG_SAVE_FLAGS);
     LD->prolog_flag.mask.flags |= qf->flags_saved;
@@ -1821,7 +1779,7 @@ restore_after_query(QueryFrame qf)
 #endif /*O_LIMIT_DEPTH*/
   }
   updateAlerted(LD);
-  SECURE(checkStacks(NULL));
+  DEBUG(CHK_SECURE, checkStacks(NULL));
 }
 
 
@@ -1830,7 +1788,7 @@ PL_cut_query(qid_t qid)
 { GET_LD
   QueryFrame qf = QueryFromQid(qid);
 
-  SECURE(assert(qf->magic == QID_MAGIC));
+  DEBUG(CHK_SECURE, assert(qf->magic == QID_MAGIC));
   if ( qf->foreign_frame )
     PL_close_foreign_frame(qf->foreign_frame);
 
@@ -1850,7 +1808,7 @@ PL_close_query(qid_t qid)
   { GET_LD
     QueryFrame qf = QueryFromQid(qid);
 
-    SECURE(assert(qf->magic == QID_MAGIC));
+    DEBUG(CHK_SECURE, assert(qf->magic == QID_MAGIC));
     if ( qf->foreign_frame )
       PL_close_foreign_frame(qf->foreign_frame);
 
@@ -1931,24 +1889,24 @@ typedef enum
 #define TRUST_CLAUSE(cref) \
 	umode = uread; \
 	CL    = cref; \
-	lTop  = (LocalFrame)(ARGP + cref->clause->variables); \
- 	ENSURE_LOCAL_SPACE(LOCAL_MARGIN, THROW_EXCEPTION); \
+	lTop  = (LocalFrame)(ARGP + cref->value.clause->variables); \
+	ENSURE_LOCAL_SPACE(LOCAL_MARGIN, THROW_EXCEPTION); \
 	if ( debugstatus.debugging ) \
 	  newChoice(CHP_DEBUG, FR PASS_LD); \
-	PC    = cref->clause->codes; \
+	PC    = cref->value.clause->codes; \
 	NEXT_INSTRUCTION;
 #define TRY_CLAUSE(cref, cond, altpc) \
 	umode = uread; \
 	CL    = cref; \
-	lTop  = (LocalFrame)(ARGP + cref->clause->variables); \
+	lTop  = (LocalFrame)(ARGP + cref->value.clause->variables); \
 	ENSURE_LOCAL_SPACE(LOCAL_MARGIN, THROW_EXCEPTION); \
 	if ( cond ) \
 	{ Choice ch = newChoice(CHP_JUMP, FR PASS_LD); \
- 	  ch->value.PC = altpc; \
+	  ch->value.PC = altpc; \
 	} else if ( debugstatus.debugging ) \
 	{ newChoice(CHP_DEBUG, FR PASS_LD); \
 	} \
-	PC    = cref->clause->codes; \
+	PC    = cref->value.clause->codes; \
 	NEXT_INSTRUCTION;
 
 
@@ -1964,7 +1922,9 @@ PL_next_solution(qid_t qid)
   Definition DEF = NULL;		/* definition of current procedure */
   unify_mode umode = uread;		/* Unification mode */
   exception_frame throw_env;		/* PL_thow() environment */
-  int	     throwed_from_line;		/* Debugging: line we came from */
+#ifdef O_DEBUG
+  int	     throwed_from_line=0;	/* Debugging: line we came from */
+#endif
 #define	     CL (FR->clause)		/* clause of current frame */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -1984,13 +1944,13 @@ pl-comp.c
 #define VMI(Name,f,na,a)	Name ## _LBL: \
 				  count(Name, PC); \
 				  START_PROF(Name, #Name);
-#define VMI_GOTO(n)		{ END_PROF(); \
-				  goto n ## _LBL; \
-				}
-#define NEXT_INSTRUCTION	{ END_PROF(); \
-				  DbgPrintInstruction(FR, PC); \
-				  goto *(void *)((intptr_t)(*PC++)); \
-				}
+#define VMI_GOTO(n)		do { END_PROF(); \
+				       goto n ## _LBL; \
+				   } while(0)
+#define NEXT_INSTRUCTION	do { END_PROF(); \
+				     DbgPrintInstruction(FR, PC); \
+				     goto *(void *)((intptr_t)(*PC++)); \
+				   } while(0)
 #ifndef ASM_NOP
 #define ASM_NOP asm("nop")
 #endif
@@ -2000,8 +1960,14 @@ pl-comp.c
 
 code thiscode;
 
+#if __GNUC__
+#define UNUSED_LABEL __attribute__ ((unused))
+#else
+#define UNUSED_LABEL
+#endif
+
 #define VMI(Name,f,na,a)	case Name: \
-				  case_ ## Name:
+				  case_ ## Name: UNUSED_LABEL \
 				  count(Name, PC); \
 				  START_PROF(Name, #Name);
 #define VMI_GOTO(n)		{ END_PROF(); \
@@ -2032,7 +1998,7 @@ depart_continue() to do the normal thing or to the backtrack point.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
   QF  = QueryFromQid(qid);
-  SECURE(assert(QF->magic == QID_MAGIC));
+  DEBUG(CHK_SECURE, assert(QF->magic == QID_MAGIC));
   if ( true(QF, PL_Q_DETERMINISTIC) )	/* last one succeeded */
   { fid_t fid = QF->foreign_frame;
     QF->foreign_frame = 0;
@@ -2098,9 +2064,7 @@ registers  should  hold  valid  data  and  the  machine stacks should be
 initialised properly.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#if VMCODE_IS_ADDRESS
-  NEXT_INSTRUCTION;
-#else
+#if !VMCODE_IS_ADDRESS			/* no goto *ptr; use a switch */
 next_instruction:
   thiscode = *PC++;
 #ifdef O_DEBUGGER
@@ -2165,7 +2129,10 @@ retry:					MARK(RETRY);
     }
   }
   Sdprintf("[Could not find retry-point]\n");
-  abortProlog(ABORT_THROW);		/* dubious */
+  SAVE_REGISTERS(qid);
+  abortProlog();				/* What else? */
+  LOAD_REGISTERS(qid);
+  THROW_EXCEPTION;
 
 do_retry:
   if ( rframe0 != rframe )
@@ -2176,8 +2143,10 @@ do_retry:
 	   predicateName(rframe->predicate));
 
   discardChoicesAfter(rframe, FINISH_CUT PASS_LD);
+  rframe->clause = NULL;
   environment_frame = FR = rframe;
   DEF = FR->predicate;
+  clear(FR, FR_SKIPPED);
   Undo(m);
   exception_term = 0;
 
@@ -2218,19 +2187,16 @@ START_PROF(P_SHALLOW_BACKTRACK, "P_SHALLOW_BACKTRACK");
 
       NEXT_INSTRUCTION;
     } else if ( ch->type == CHP_CLAUSE )
-    { ClauseRef next;
-
-      ARGP = argFrameP(FR, 0);
-      if ( !(CL = findClause(ch->value.clause, ARGP, FR, DEF, &next PASS_LD)) )
-	FRAME_FAILED;			/* should not happen */
-      PC = CL->clause->codes;
+    { ARGP = argFrameP(FR, 0);
+      if ( !(CL = nextClause(&ch->value.clause, ARGP, FR, DEF)) )
+	FRAME_FAILED;		/* can happen if scan-ahead was too short */
+      PC = CL->value.clause->codes;
       umode = uread;
 
-      if ( ch == (Choice)argFrameP(FR, CL->clause->variables) )
-      { DiscardMark(ch->mark);
-	if ( next )
-	{ ch->value.clause = next;
-	  Mark(ch->mark);
+      if ( ch == (Choice)argFrameP(FR, CL->value.clause->variables) )
+      { DiscardMark(ch->mark);		/* is this needed? */
+	if ( ch->value.clause.cref )
+	{ Mark(ch->mark);
 	  lTop = (LocalFrame)(ch+1);
 	  NEXT_INSTRUCTION;
 	} else if ( unlikely(debugstatus.debugging) )
@@ -2243,15 +2209,18 @@ START_PROF(P_SHALLOW_BACKTRACK, "P_SHALLOW_BACKTRACK");
 	BFR = ch->parent;
 	lTop = (LocalFrame)ch;
 	NEXT_INSTRUCTION;
-      } else
-      { DiscardMark(ch->mark);
+      } else				/* Choice point needs to move */
+      { struct clause_choice chp;
+
+        DiscardMark(ch->mark);
 	BFR = ch->parent;
-	lTop = (LocalFrame)argFrameP(FR, CL->clause->variables);
+	chp = ch->value.clause;
+	lTop = (LocalFrame)argFrameP(FR, CL->value.clause->variables);
 	ENSURE_LOCAL_SPACE(LOCAL_MARGIN, THROW_EXCEPTION);
 
-	if ( next )
+	if ( chp.cref )
 	{ ch = newChoice(CHP_CLAUSE, FR PASS_LD);
-	  ch->value.clause = next;
+	  ch->value.clause = chp;
 	} else if ( unlikely(debugstatus.debugging) )
 	{ ch = newChoice(CHP_DEBUG, FR PASS_LD);
 	}
@@ -2270,13 +2239,11 @@ START_PROF(P_DEEP_BACKTRACK, "P_DEEP_BACKTRACK");
   Choice ch0 = BFR;
 #endif
   Choice ch;
-  LocalFrame fr0;
 
   DEBUG(3, Sdprintf("BACKTRACKING\n"));
 
 next_choice:
   ch = BFR;
-  fr0 = FR;
 					/* leave older frames */
   for(; (void *)FR > (void *)ch; FR = FR->parent)
   {
@@ -2296,12 +2263,13 @@ next_choice:
 	SAVE_REGISTERS(qid);
 	rc = tracePort(FR, BFR, FAIL_PORT, NULL PASS_LD);
 	LOAD_REGISTERS(qid);
+	ch = BFR;			/* can be shifted */
 
 	switch( rc )
 	{ case ACTION_RETRY:
 	    environment_frame = FR;
 	    DEF = FR->predicate;
-	    clear(FR, FR_CATCHED);
+	    clear(FR, FR_CATCHED|FR_SKIPPED);
 	    goto retry_continue;
 	    case ACTION_ABORT:
 	      THROW_EXCEPTION;
@@ -2313,7 +2281,7 @@ next_choice:
     }
 #endif
 
-    leaveFrame(FR PASS_LD);
+    leaveFrame(FR);
     if ( true(FR, FR_WATCHED) )
     { environment_frame = FR;
       lTop = (LocalFrame)argFrameP(FR, FR->predicate->functor->arity);
@@ -2341,6 +2309,31 @@ next_choice:
       DEBUG(3, Sdprintf("    REDO #%ld: Jump in %s\n",
 			loffset(FR),
 			predicateName(DEF)));
+#ifdef O_DEBUGGER
+      if ( debugstatus.debugging && !debugstatus.suspendTrace  )
+      { LocalFrame fr = dbgRedoFrame(FR, CHP_JUMP PASS_LD);
+
+	if ( fr )
+	{ int action;
+
+	  SAVE_REGISTERS(qid);
+	  action = tracePort(fr, BFR, REDO_PORT, ch->value.PC PASS_LD);
+	  LOAD_REGISTERS(qid);
+	  ch = BFR;			/* can be shifted */
+
+	  switch( action )
+	  { case ACTION_FAIL:
+	      FRAME_FAILED;
+	    case ACTION_IGNORE:
+	      VMI_GOTO(I_EXIT);
+	    case ACTION_RETRY:
+	      goto retry_continue;
+	    case ACTION_ABORT:
+	      THROW_EXCEPTION;
+	  }
+	}
+      }
+#endif
       PC   = ch->value.PC;
       DiscardMark(ch->mark);
       BFR  = ch->parent;
@@ -2349,8 +2342,8 @@ next_choice:
       ARGP = argFrameP(lTop, 0);
       NEXT_INSTRUCTION;
     case CHP_CLAUSE:			/* try next clause */
-    { ClauseRef next;
-      Clause clause;
+    { Clause clause;
+      struct clause_choice chp;
 
       DEBUG(3, Sdprintf("    REDO #%ld: Clause in %s\n",
 			loffset(FR),
@@ -2358,12 +2351,13 @@ next_choice:
       ARGP = argFrameP(FR, 0);
       DiscardMark(ch->mark);
       BFR = ch->parent;
-      if ( !(CL = findClause(ch->value.clause, ARGP, FR, DEF, &next PASS_LD)) )
-	goto next_choice;		/* should not happen */
+      if ( !(CL = nextClause(&ch->value.clause, ARGP, FR, DEF)) )
+	goto next_choice;	/* Can happen of look-ahead was too short */
+      chp = ch->value.clause;
 
 #ifdef O_DEBUGGER
       if ( debugstatus.debugging && !debugstatus.suspendTrace  )
-      { LocalFrame fr = dbgRedoFrame(FR PASS_LD);
+      { LocalFrame fr = dbgRedoFrame(FR, CHP_CLAUSE PASS_LD);
 
 	if ( fr )
 	{ int action;
@@ -2371,6 +2365,7 @@ next_choice:
 	  SAVE_REGISTERS(qid);
 	  action = tracePort(fr, BFR, REDO_PORT, NULL PASS_LD);
 	  LOAD_REGISTERS(qid);
+	  ch = BFR;			/* can be shifted */
 
 	  switch( action )
 	  { case ACTION_FAIL:
@@ -2387,15 +2382,15 @@ next_choice:
 #endif
 
       umode  = uread;
-      clause = CL->clause;
+      clause = CL->value.clause;
       PC     = clause->codes;
       Profile(profRedo(ch->prof_node PASS_LD));
       lTop   = (LocalFrame)argFrameP(FR, clause->variables);
       ENSURE_LOCAL_SPACE(LOCAL_MARGIN, THROW_EXCEPTION);
 
-      if ( next )
+      if ( chp.cref )
       { ch = newChoice(CHP_CLAUSE, FR PASS_LD);
-	ch->value.clause = next;
+	ch->value.clause = chp;
       } else if ( unlikely(debugstatus.debugging) )
       { newChoice(CHP_DEBUG, FR PASS_LD);
       }
@@ -2432,10 +2427,14 @@ next_choice:
 	{ SAVE_REGISTERS(qid);
 	  callCleanupHandler(ch->frame, FINISH_FAIL PASS_LD);
 	  LOAD_REGISTERS(qid);
+	} else
+	{ set(ch->frame, FR_CATCHED);
 	}
 	ch = BFR;			/* can be shifted */
 	if ( exception_term )
 	  THROW_EXCEPTION;
+      } else
+      { set(ch->frame, FR_CATCHED);
       }
       /*FALLTHROUGH*/
     case CHP_DEBUG:			/* Just for debugging purposes */
