@@ -1,12 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2010, University of Amsterdam
-			      Vu University Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -20,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 #include <math.h>
@@ -42,11 +40,6 @@
 #define HAVE_FPCLASSIFY 1
 #endif
 
-typedef struct visited
-{ Word address;				/* we have done this address */
-  struct visited *next;			/* next already visited */
-} visited;
-
 typedef struct
 { int   flags;				/* PL_WRT_* flags */
   int   max_depth;			/* depth limit */
@@ -54,7 +47,9 @@ typedef struct
   atom_t spacing;			/* Where to insert spaces */
   Module module;			/* Module for operators */
   IOSTREAM *out;			/* stream to write to */
-  visited *visited;			/* visited (attributed-) variables */
+  term_t portray_goal;			/* call/2 activated portray hook */
+  term_t write_options;			/* original write options */
+  term_t prec_opt;			/* term in write options with prec */
 } write_options;
 
 static bool	writeTerm2(term_t term, int prec,
@@ -63,34 +58,9 @@ static bool	writeTerm(term_t t, int prec,
 			  write_options *options) WUNUSED;
 static bool	writeArgTerm(term_t t, int prec,
 			     write_options *options, bool arg) WUNUSED;
-
-static Word
-address_of(term_t t)
-{ GET_LD
-  Word adr = valTermRef(t);
-
-  deRef(adr);
-  switch(tag(*adr))
-  { case TAG_ATTVAR:
-      return adr;
-    case TAG_COMPOUND:
-      return valPtr(*adr);
-    default:
-      return NULL;			/* non-recursive structure */
-  }
-}
-
-
-static int
-has_visited(visited *v, Word addr)
-{ for( ; v; v=v->next )
-  { if ( v->address == addr )
-      succeed;
-  }
-
-  fail;
-}
-
+static int	PutToken(const char *s, IOSTREAM *stream);
+static int	writeAtom(atom_t a, write_options *options);
+static int	callPortray(term_t arg, int prec, write_options *options);
 
 char *
 varName(term_t t, char *name)
@@ -108,6 +78,96 @@ varName(term_t t, char *name)
 }
 
 
+static int
+atomIsVarName(atom_t a)
+{ Atom atom = atomValue(a);
+
+  if ( false(atom->type, PL_BLOB_TEXT) || atom->length == 0 )
+    fail;
+  if ( isUCSAtom(atom) )
+  { pl_wchar_t *w = (pl_wchar_t*)atom->name;
+    size_t len = atom->length / sizeof(pl_wchar_t);
+
+    return atom_varnameW(w, len);
+  } else
+  { const char *s = atom->name;
+    size_t len = atom->length;
+
+    if ( isUpper(*s) || *s == '_' )
+    { for(s++; --len > 0; s++)
+      { if ( !isAlpha(*s) )
+	  return FALSE;
+      }
+
+      return TRUE;
+    }
+
+    return FALSE;
+  }
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Return:	TRUE:  processes
+	FALSE: not processed
+	-1:    error
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static int
+writeNumberVar(term_t t, write_options *options ARG_LD)
+{ Word p = valTermRef(t);
+  Functor f;
+
+  deRef(p);
+  if ( !isTerm(*p) )
+    return FALSE;
+  f = valueTerm(*p);
+
+  if ( f->definition != FUNCTOR_isovar1 )
+    return FALSE;
+
+  if ( LD->var_names.numbervars_frame )
+  { FliFrame fr = (FliFrame)valTermRef(LD->var_names.numbervars_frame);
+
+    assert(fr->magic == FLI_MAGIC);
+    if ( fr->mark.globaltop > (Word)f )
+      return FALSE;			/* older $VAR term */
+  }
+
+  p = &f->arguments[0];
+  deRef(p);
+  if ( isInteger(*p) )
+  { int64_t n = valInteger(*p);
+    char buf[32];			/* Max is H354745078340568300 */
+
+    if ( n < 0 )
+    { sprintf(buf, "S_" INT64_FORMAT, -n);
+    } else
+    { int i = (int)(n % 26);
+      int64_t j = n / 26;
+
+      if ( j == 0 )
+      { buf[0] = i+'A';
+	buf[1] = EOS;
+      } else
+      { sprintf(buf, "%c" INT64_FORMAT, i+'A', j);
+      }
+    }
+
+    return PutToken(buf, options->out) ? TRUE : -1;
+  }
+
+  if ( isAtom(*p) && atomIsVarName(*p) )
+  { write_options o2 = *options;
+    clear(&o2, PL_WRT_QUOTED);
+
+    return writeAtom(*p, &o2) ? TRUE : -1;
+  }
+
+  return FALSE;
+}
+
+
 #define AT_LOWER	0
 #define AT_QUOTE	1
 #define AT_FULLSTOP	2
@@ -194,6 +254,15 @@ PutComma(write_options *options)
 }
 
 
+static bool
+PutBar(write_options *options)
+{ if ( options->spacing == ATOM_next_argument )
+    return PutString("| ", options->out);
+  else
+    return PutString("|", options->out);
+}
+
+
 static bool
 PutStringN(const char *str, size_t length, IOSTREAM *s)
 { size_t i;
@@ -209,10 +278,17 @@ PutStringN(const char *str, size_t length, IOSTREAM *s)
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-PutOpenToken() inserts a space in the output stream if the last-written
+PutOpenToken() inserts a space in the  output stream if the last-written
 and given character require a space to ensure a token-break.
+
+The C_* flags denote special cases handled  using a flag. The first flag
+is 0x200000, which is above the Unicode range.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#define C_PREFIX_SIGN		0x00200000	/* +/- as prefix op */
+#define C_PREFIX_OP		0x00400000	/* any prefix op */
+#define C_MASK			0xffe00000
+
 #define isquote(c) ((c) == '\'' || (c) == '"')
 
 static bool
@@ -220,14 +296,24 @@ needSpace(int c, IOSTREAM *s)
 { if ( c == EOF )
   { s->lastc = EOF;
     return FALSE;
-  } else if ( s->lastc != EOF &&
-	      ((isAlphaW(s->lastc) && isAlphaW(c)) ||
-	       (isSymbolW(s->lastc) && isSymbolW(c)) ||
-	       (s->lastc != '(' && !isBlank(s->lastc) && c == '(') ||
-	       (c == '\'' && (isDigit(s->lastc))) ||
-	       (isquote(c) && s->lastc == c)) )
-  { return TRUE;
   }
+  if ( s->lastc == EOF )
+    return FALSE;
+
+  if ( (s->lastc&C_PREFIX_SIGN) && (isDigit(c) || isSymbolW(c)) )
+    return TRUE;
+  if ( (s->lastc&C_PREFIX_OP) && c == '(' )
+    return TRUE;				/* avoid op(...) */
+
+  s->lastc &= ~C_MASK;
+
+  if ( ((isAlphaW(s->lastc) && isAlphaW(c)) ||
+	(isSymbolW(s->lastc) && isSymbolW(c)) ||
+	(s->lastc != '(' && !isBlank(s->lastc) && c == '(') ||
+	(c == '\'' && (isDigit(s->lastc))) ||
+	(isquote(c) && s->lastc == c)
+       ) )
+    return TRUE;
 
   return FALSE;
 }
@@ -391,48 +477,40 @@ writeAttVar(term_t av, write_options *options)
   } else if ( (options->flags & PL_WRT_ATTVAR_WRITE) )
   { fid_t fid;
     term_t a;
-    visited v;
 
     if ( !(fid = PL_open_foreign_frame()) )
       return FALSE;
 
-    v.address = address_of(av);
-    if ( has_visited(options->visited, v.address) )
-      succeed;
-    v.next = options->visited;
-    options->visited = &v;
     Sputcode('{', options->out);
     a = PL_new_term_ref();
     PL_get_attr__LD(av, a PASS_LD);
     if ( !writeTerm(a, 1200, options) )
-      goto error;
+      return FALSE;
     Sputcode('}', options->out);
-    PL_discard_foreign_frame(fid);
+    PL_close_foreign_frame(fid);
 
-    options->visited = v.next;
-    succeed;
-
-  error:
-    options->visited = v.next;
-    fail;
+    return TRUE;
   } else if ( (options->flags & PL_WRT_ATTVAR_PORTRAY) &&
 	      GD->cleaning <= CLN_PROLOG )
-  { fid_t fid;
-    predicate_t pred;
+  { predicate_t pred;
     IOSTREAM *old;
-
-    if ( !(fid = PL_open_foreign_frame()) )
-      return FALSE;
+    wakeup_state wstate;
+    int rc;
 
     pred = _PL_predicate("portray_attvar", 1, "$attvar",
 			 &GD->procedures.portray_attvar1);
 
+    if ( !saveWakeup(&wstate, TRUE PASS_LD) )
+      return FALSE;
     old = Scurout;
     Scurout = options->out;
-    PL_call_predicate(NULL, PL_Q_NODEBUG, pred, av);
+    rc = PL_call_predicate(NULL, PL_Q_NODEBUG|PL_Q_PASS_EXCEPTION, pred, av);
+    if ( rc != TRUE && !PL_exception(0) )
+      rc = TRUE;
     Scurout = old;
+    restoreWakeup(&wstate PASS_LD);
 
-    PL_discard_foreign_frame(fid);
+    return rc;
   }
 
   succeed;
@@ -468,24 +546,14 @@ writeAtom(atom_t a, write_options *options)
   { GET_LD
     int rc;
     fid_t fid;
-    predicate_t pred;
-    IOSTREAM *old;
     term_t av;
 
     if ( !(fid = PL_open_foreign_frame()) )
       return FALSE;
     av = PL_new_term_ref();
     PL_put_atom(av, a);
-
-    pred = _PL_predicate("portray", 1, "user",
-			 &GD->procedures.portray);
-
-    old = Scurout;
-    Scurout = options->out;
-    rc = PL_call_predicate(NULL, PL_Q_NODEBUG, pred, av);
-    Scurout = old;
-
-    PL_discard_foreign_frame(fid);
+    rc = callPortray(av, 1200, options);
+    PL_close_foreign_frame(fid);
     if ( rc == TRUE )
       return TRUE;
   }
@@ -858,21 +926,31 @@ writePrimitive(term_t t, write_options *options)
 }
 
 
-word
-pl_nl1(term_t stream)
+static int
+pl_nl__LD(term_t stream ARG_LD)
 { IOSTREAM *s;
 
-  if ( getOutputStream(stream, &s) )
+  if ( getTextOutputStream(stream, &s) )
   { Sputcode('\n', s);
     return streamStatus(s);
   }
 
-  fail;
+  return FALSE;
 }
 
-word
-pl_nl()
-{ return pl_nl1(0);
+
+static
+PRED_IMPL("nl", 1, nl, PL_FA_ISO)
+{ PRED_LD
+
+  return pl_nl__LD(A1 PASS_LD);
+}
+
+static
+PRED_IMPL("nl", 0, nl, PL_FA_ISO)
+{ PRED_LD
+
+  return pl_nl__LD(0 PASS_LD);
 }
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -880,25 +958,77 @@ Call user:portray/1 if defined.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static int
-callPortray(term_t arg, write_options *options)
-{ predicate_t portray;
+put_write_options(term_t opts_in, write_options *options)
+{ GET_LD
+  term_t newlist = PL_new_term_ref();
+  term_t precopt = PL_new_term_ref();
+  fid_t fid = PL_open_foreign_frame();
+  term_t head = PL_new_term_ref();
+  term_t tail = PL_copy_term_ref(opts_in);
+  term_t newhead = PL_new_term_ref();
+  term_t newtail = PL_copy_term_ref(newlist);
+  int rc = TRUE;
+
+  while(rc && PL_get_list(tail, head, tail))
+  { if ( !PL_is_functor(head, FUNCTOR_priority1) )
+      rc = ( PL_unify_list(newtail, newhead, newtail) &&
+	     PL_unify(newhead, head) );
+  }
+
+  if ( rc )
+  { rc = ( PL_unify_list(newtail, head, newtail) &&
+	   PL_unify_functor(head, FUNCTOR_priority1) &&
+	   PL_get_arg(1, head, precopt) &&
+	   PL_unify_nil(newtail) );
+  }
+  if ( rc )
+  { options->write_options = newlist;
+    options->prec_opt = precopt;
+  }
+
+  PL_close_foreign_frame(fid);
+  return rc;
+}
+
+
+
+
+static int
+callPortray(term_t arg, int prec, write_options *options)
+{ predicate_t pred;
 
   if ( GD->cleaning > CLN_PROLOG )
     fail;				/* avoid dangerous callbacks */
 
-  portray = _PL_predicate("portray", 1, "user", &GD->procedures.portray);
+  if ( options->portray_goal )
+  { pred = _PL_predicate("call", 3, "user", &GD->procedures.call3);
+  } else
+  { pred = _PL_predicate("portray", 1, "user", &GD->procedures.portray);
+    if ( !pred->definition->impl.any )
+      return FALSE;
+  }
 
-  if ( portray->definition->definition.clauses )
   { GET_LD
     wakeup_state wstate;
     IOSTREAM *old = Scurout;
     int rval;
+    term_t av;
 
     if ( !saveWakeup(&wstate, TRUE PASS_LD) )
       return FALSE;
     Scurout = options->out;
+    if ( options->portray_goal )
+    { av = PL_new_term_refs(3);
+
+      PL_put_term(av+0, options->portray_goal);
+      PL_put_term(av+1, arg);
+      PL_unify_integer(options->prec_opt, prec);
+      PL_put_term(av+2, options->write_options);
+    } else
+    { av = arg;
+    }
     rval = PL_call_predicate(NULL, PL_Q_NODEBUG|PL_Q_PASS_EXCEPTION,
-			     portray, arg);
+			     pred, av);
     if ( !rval && PL_exception(0) )
       rval = -1;
     Scurout = old;
@@ -927,19 +1057,8 @@ writeArgTerm(term_t t, int prec, write_options *options, bool arg)
   }
 
   if ( ++options->depth > options->max_depth && options->max_depth )
+  { PutOpenToken('.', options->out);
     rval = PutString("...", options->out);
-  else if ( PL_is_compound(t) )
-  { visited v;
-
-    v.address = address_of(t);
-    if ( has_visited(options->visited, v.address) )
-    { rval = PutString("**", options->out);
-    } else
-    { v.next = options->visited;
-      options->visited = &v;
-      rval = writeTerm2(t, prec, options, arg);
-      options->visited = v.next;
-    }
   } else
   { rval = writeTerm2(t, prec, options, arg);
   }
@@ -958,7 +1077,7 @@ writeTerm(term_t t, int prec, write_options *options)
 }
 
 static bool
-writeList2(term_t list, write_options *options, int cyclic)
+writeList(term_t list, write_options *options)
 { GET_LD
   term_t head = PL_new_term_ref();
   term_t l    = PL_copy_term_ref(list);
@@ -978,19 +1097,6 @@ writeList2(term_t list, write_options *options, int cyclic)
       break;
     }
 
-					/* cycle detection */
-    { Word addr =  address_of(l);
-
-      if ( has_visited(options->visited, addr) )
-      { return PutString("|**]", options->out);
-      } else if ( cyclic )
-      { visited *v = alloca(sizeof(*v));
-	v->address = addr;
-	v->next = options->visited;
-	options->visited = v;
-      }
-    }
-
     TRY(PutComma(options));
   }
 
@@ -998,22 +1104,6 @@ writeList2(term_t list, write_options *options, int cyclic)
 }
 
 
-static bool
-writeList(term_t list, write_options *options)
-{ GET_LD
-  visited *v = options->visited;
-  Word tail;
-  int rc;
-
-  skip_list(valTermRef(list), &tail PASS_LD);
-  rc = writeList2(list, options, isList(*tail));
-  options->visited = v;
-
-  return rc;
-}
-
-
-
 static bool
 writeTerm2(term_t t, int prec, write_options *options, bool arg)
 { GET_LD
@@ -1025,7 +1115,7 @@ writeTerm2(term_t t, int prec, write_options *options, bool arg)
 
   if ( !PL_is_variable(t) &&
        true(options, PL_WRT_PORTRAY) )
-  { switch( callPortray(t, options) )
+  { switch( callPortray(t, prec, options) )
     { case TRUE:
 	return TRUE;
       case FALSE:
@@ -1048,37 +1138,15 @@ writeTerm2(term_t t, int prec, write_options *options, bool arg)
   if ( !PL_get_name_arity(t, &functor, &arity) )
   { return writePrimitive(t, options);
   } else
-  { if ( arity == 1 &&
-	 functor == ATOM_isovar &&			/* $VAR/1 */
-	 true(options, PL_WRT_NUMBERVARS) )
-    { int n;
-      atom_t a;
-      term_t arg = PL_new_term_ref();
-
-      _PL_get_arg(1, t, arg);
-      if ( PL_get_integer(arg, &n) && n >= 0 )
-      { int i = n % 26;
-	int j = n / 26;
-	char buf[16];
-
-	if ( j == 0 )
-	{ buf[0] = i+'A';
-	  buf[1] = EOS;
-	} else
-	{ sprintf(buf, "%c%d", i+'A', j);
-	}
-
-	return PutToken(buf, out);
-      }
-      if ( PL_get_atom(arg, &a) )
-      { write_options o2 = *options;
-	clear(&o2, PL_WRT_QUOTED);
-
-	return writeAtom(a, &o2);
+  { if ( true(options, PL_WRT_NUMBERVARS) )
+    { switch( writeNumberVar(t, options PASS_LD) )
+      { case -1:
+	  return FALSE;
+	case TRUE:
+	  return TRUE;
       }
     }
 
-
     if ( false(options, PL_WRT_IGNOREOPS) )
     { term_t arg = PL_new_term_ref();
 
@@ -1102,24 +1170,20 @@ writeTerm2(term_t t, int prec, write_options *options, bool arg)
 
 	  _PL_get_arg(1, t, arg);
 	  if ( embrace )
-	  { TRY(PutOpenBrace(out));
-	  }
+	    TRY(PutOpenBrace(out));
 	  TRY(writeAtom(functor, options));
 
 				/* +/-(Number) : avoid parsing as number */
-	  if ( (functor == ATOM_minus || functor == ATOM_plus) &&
-	       PL_is_number(arg) )
-	  { TRY(Putc('(', out));
-	    TRY(writeTerm(arg, 999, options));
-	    TRY(Putc(')', out));
-	  } else
-	  { TRY(writeTerm(arg,
-			  op_type == OP_FX ? op_pri-1 : op_pri,
-			  options));
-	  }
+	  options->out->lastc |= C_PREFIX_OP;
+	  if ( functor == ATOM_minus || functor == ATOM_plus )
+	    options->out->lastc |= C_PREFIX_SIGN;
+
+	  TRY(writeTerm(arg,
+			op_type == OP_FX ? op_pri-1 : op_pri,
+			options));
+
 	  if ( embrace )
-	  { TRY(PutCloseBrace(out));
-	  }
+	   TRY(PutCloseBrace(out));
 
 	  succeed;
 	}
@@ -1162,6 +1226,8 @@ writeTerm2(term_t t, int prec, write_options *options, bool arg)
 			options));
 	  if ( functor == ATOM_comma )
 	  { TRY(PutComma(options));
+	  } else if ( functor == ATOM_bar )
+	  { TRY(PutBar(options));
 	  } else
 	  { switch(writeAtom(functor, options))
 	    { case FALSE:
@@ -1180,6 +1246,10 @@ writeTerm2(term_t t, int prec, write_options *options, bool arg)
 	}
       }
     }
+
+    if ( (options->flags&PL_WRT_LIST) && arity == 2 && functor == ATOM_dot )
+      return writeList(t, options);
+
 					/* functor(<args> ...) */
     { term_t a = PL_new_term_ref();
 
@@ -1197,6 +1267,88 @@ writeTerm2(term_t t, int prec, write_options *options, bool arg)
 }
 
 
+		 /*******************************
+		 *	  CYCLE HANDLING	*
+		 *******************************/
+
+static int
+reunify_acyclic_substitutions(term_t substitutions, term_t cycles,
+			      write_options *options)
+{ GET_LD
+  term_t s_tail, c_tail, s_head, c_head, var, value;
+  intptr_t count = 0;
+
+  if ( !(s_tail = PL_copy_term_ref(substitutions)) ||
+       !(c_tail = PL_copy_term_ref(cycles)) ||
+       !(s_head = PL_new_term_ref()) ||
+       !(c_head = PL_new_term_ref()) ||
+       !(var    = PL_new_term_ref()) ||
+       !(value  = PL_new_term_ref()) )
+    return FALSE;
+
+  while(PL_get_list(s_tail, s_head, s_tail))
+  { _PL_get_arg(1, s_head, var);
+    _PL_get_arg(2, s_head, value);
+    if ( PL_var_occurs_in(var, value) )
+    { if ( (options->flags&PL_WRT_NUMBERVARS) )
+      { if ( !PL_unify_term(var,
+			    PL_FUNCTOR, FUNCTOR_isovar1,
+			      PL_INTPTR, --count) )
+	  return FALSE;
+      }
+
+      if ( !PL_unify_list(c_tail, c_head, c_tail) ||
+	   !PL_unify(c_head, s_head) )
+	return FALSE;
+    } else
+    { if ( !PL_unify(var, value) )
+	return FALSE;
+    }
+  }
+
+  return PL_unify_nil(c_tail);
+}
+
+
+static int
+writeTopTerm(term_t term, int prec, write_options *options)
+{ GET_LD
+  int rc;
+
+  if ( (!(options->flags&PL_WRT_NO_CYCLES) && options->max_depth) ||
+       PL_is_acyclic(term) )
+  { rc = writeTerm(term, prec, options);
+  } else
+  { fid_t fid;
+    term_t template, substitutions, cycles, at_term;
+
+    if ( options->flags & PL_WRT_NO_CYCLES )
+      return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_cyclic_term, term);
+
+    if ( !(fid = PL_open_foreign_frame()) ||
+	 !(template = PL_new_term_ref()) ||
+	 !(substitutions = PL_new_term_ref()) ||
+	 !(cycles = PL_new_term_ref()) ||
+	 !(at_term = PL_new_term_ref()) ||
+	 !PL_factorize_term(term, template, substitutions) ||
+	 !reunify_acyclic_substitutions(substitutions, cycles, options) ||
+	 !PL_unify_term(at_term,
+			PL_FUNCTOR, FUNCTOR_xpceref2,
+			  PL_TERM, template,
+			  PL_TERM, cycles) )
+      return FALSE;
+    rc = writeTerm(at_term, prec, options);
+    PL_discard_foreign_frame(fid);
+  }
+
+  return rc;
+}
+
+
+		 /*******************************
+		 *	      TOPLEVEL		*
+		 *******************************/
+
 int
 writeAttributeMask(atom_t a)
 { if ( a == ATOM_ignore )
@@ -1228,6 +1380,7 @@ static const opt_spec write_term_options[] =
   { ATOM_ignore_ops,	    OPT_BOOL },
   { ATOM_numbervars,        OPT_BOOL },
   { ATOM_portray,           OPT_BOOL },
+  { ATOM_portray_goal,      OPT_TERM },
   { ATOM_character_escapes, OPT_BOOL },
   { ATOM_max_depth,	    OPT_INT  },
   { ATOM_module,	    OPT_ATOM },
@@ -1237,7 +1390,8 @@ static const opt_spec write_term_options[] =
   { ATOM_partial,	    OPT_BOOL },
   { ATOM_spacing,	    OPT_ATOM },
   { ATOM_blobs,		    OPT_ATOM },
-  { NULL_ATOM,	     	    0 }
+  { ATOM_cycles,	    OPT_BOOL },
+  { NULL_ATOM,		    0 }
 };
 
 word
@@ -1247,6 +1401,7 @@ pl_write_term3(term_t stream, term_t term, term_t opts)
   bool ignore_ops = FALSE;
   bool numbervars = -1;			/* not set */
   bool portray    = FALSE;
+  term_t gportray = 0;
   bool bqstring   = truePrologFlag(PLFLAG_BACKQUOTED_STRING);
   bool charescape = -1;			/* not set */
   atom_t mname    = ATOM_user;
@@ -1254,6 +1409,7 @@ pl_write_term3(term_t stream, term_t term, term_t opts)
   atom_t blobs    = ATOM_nil;
   int  priority   = 1200;
   bool partial    = FALSE;
+  bool cycles     = TRUE;
   IOSTREAM *s;
   write_options options;
   int rc;
@@ -1262,10 +1418,10 @@ pl_write_term3(term_t stream, term_t term, term_t opts)
   options.spacing = ATOM_standard;
 
   if ( !scan_options(opts, 0, ATOM_write_option, write_term_options,
-		     &quoted, &ignore_ops, &numbervars, &portray,
+		     &quoted, &ignore_ops, &numbervars, &portray, &gportray,
 		     &charescape, &options.max_depth, &mname,
 		     &bqstring, &attr, &priority, &partial, &options.spacing,
-		     &blobs) )
+		     &blobs, &cycles) )
     fail;
 
   if ( attr == ATOM_nil )
@@ -1304,13 +1460,17 @@ pl_write_term3(term_t stream, term_t term, term_t opts)
     }
   }
 
-  if ( !getOutputStream(stream, &s) )
-    fail;
-
   options.module = lookupModule(mname);
   if ( charescape == TRUE ||
-       (charescape == -1 && true(options.module, CHARESCAPE)) )
+       (charescape == -1 && true(options.module, M_CHARESCAPE)) )
     options.flags |= PL_WRT_CHARESCAPES;
+  if ( gportray )
+  { options.portray_goal = gportray;
+    if ( !put_write_options(opts, &options) ||
+	 !PL_qualify(options.portray_goal, options.portray_goal) )
+      return FALSE;
+    portray = TRUE;
+  }
   if ( numbervars == -1 )
     numbervars = (portray ? TRUE : FALSE);
 
@@ -1319,16 +1479,19 @@ pl_write_term3(term_t stream, term_t term, term_t opts)
   if ( numbervars ) options.flags |= PL_WRT_NUMBERVARS;
   if ( portray )    options.flags |= PL_WRT_PORTRAY;
   if ( bqstring )   options.flags |= PL_WRT_BACKQUOTED_STRING;
+  if ( !cycles )    options.flags |= PL_WRT_NO_CYCLES;
 
+  if ( !getTextOutputStream(stream, &s) )
+    fail;
   options.out = s;
   if ( !partial )
     PutOpenToken(EOF, s);		/* reset this */
   if ( (options.flags & PL_WRT_QUOTED) && !(s->flags&SIO_REPPL) )
   { s->flags |= SIO_REPPL;
-    rc = writeTerm(term, priority, &options);
+    rc = writeTopTerm(term, priority, &options);
     s->flags &= ~SIO_REPPL;
   } else
-  { rc = writeTerm(term, priority, &options);
+  { rc = writeTopTerm(term, priority, &options);
   }
 
   return streamStatus(s) && rc;
@@ -1351,7 +1514,7 @@ PL_write_term(IOSTREAM *s, term_t term, int precedence, int flags)
   options.module    = MODULE_user;
 
   PutOpenToken(EOF, s);			/* reset this */
-  return writeTerm(term, precedence, &options);
+  return writeTopTerm(term, precedence, &options);
 }
 
 
@@ -1360,7 +1523,7 @@ do_write2(term_t stream, term_t term, int flags)
 { GET_LD
   IOSTREAM *s;
 
-  if ( getOutputStream(stream, &s) )
+  if ( getTextOutputStream(stream, &s) )
   { write_options options;
     int rc;
 
@@ -1368,13 +1531,15 @@ do_write2(term_t stream, term_t term, int flags)
     options.flags     = flags;
     options.out	      = s;
     options.module    = MODULE_user;
-    if ( options.module && true(options.module, CHARESCAPE) )
+    if ( options.module && true(options.module, M_CHARESCAPE) )
       options.flags |= PL_WRT_CHARESCAPES;
     if ( truePrologFlag(PLFLAG_BACKQUOTED_STRING) )
       options.flags |= PL_WRT_BACKQUOTED_STRING;
 
     PutOpenToken(EOF, s);		/* reset this */
-    rc = writeTerm(term, 1200, &options);
+    rc = writeTopTerm(term, 1200, &options);
+    if ( rc && (flags&PL_WRT_NEWLINE) )
+      rc = Putc('\n', s);
 
     return streamStatus(s) && rc;
   }
@@ -1402,19 +1567,22 @@ pl_print2(term_t stream, term_t term)
 word
 pl_write_canonical2(term_t stream, term_t term)
 { GET_LD
-  fid_t fid;
   nv_options options;
   word rc;
 
-  if ( !(fid = PL_open_foreign_frame()) )
-    return FALSE;
+  BEGIN_NUMBERVARS();
+
   options.functor = FUNCTOR_isovar1;
   options.on_attvar = AV_SKIP;
-  options.singletons = TRUE;
-  numberVars(term, &options, 0 PASS_LD);
-  rc = do_write2(stream, term,
-		 PL_WRT_QUOTED|PL_WRT_IGNOREOPS|PL_WRT_NUMBERVARS);
-  PL_discard_foreign_frame(fid);
+  options.singletons = PL_is_acyclic(term);
+  options.numbered_check = FALSE;
+
+  rc = ( numberVars(term, &options, 0 PASS_LD) >= 0 &&
+	 do_write2(stream, term,
+		   PL_WRT_QUOTED|PL_WRT_IGNOREOPS|PL_WRT_NUMBERVARS)
+       );
+
+  END_NUMBERVARS();
 
   return rc;
 }
@@ -1439,14 +1607,9 @@ pl_write_canonical(term_t term)
 { return pl_write_canonical2(0, term);
 }
 
-word					/* for debugging purposes! */
+word
 pl_writeln(term_t term)
-{ if ( PL_write_term(Serror, term, 1200,
-		     PL_WRT_QUOTED|PL_WRT_NUMBERVARS) &&
-       Sdprintf("\n") >= 0 )
-    succeed;
-
-  fail;
+{ return do_write2(0, term, PL_WRT_NUMBERVARS|PL_WRT_NEWLINE);
 }
 
 
@@ -1468,12 +1631,146 @@ PRED_IMPL("$put_token", 2, put_token, 0)
   fail;
 }
 
+/** '$put_quoted_codes'(+Stream, +Quote, +Codes, +Options)
+
+Emit Codes using the escaped character  syntax,   but  does not emit the
+start and end-code itself. Options is  currently ignored. It is intended
+to provide additional preferences, so as using \uXXXX, \UXXXXXXXX, etc.
+*/
+
+static
+PRED_IMPL("$put_quoted", 4, put_quoted_codes, 0)
+{ IOSTREAM *out;
+  pl_wchar_t *w;
+  size_t i, len;
+  int quote;
+  int flags = PL_WRT_CHARESCAPES;
+  int rc = TRUE;
+
+  if ( !PL_get_stream_handle(A1, &out) ||
+       !PL_get_char_ex(A2, &quote, FALSE) ||
+       !PL_get_wchars(A3, &len, &w, CVT_LIST|CVT_STRING|CVT_EXCEPTION) )
+    return FALSE;
+
+  for(i=0; rc && i<len; i++)
+    rc = putQuoted(w[i], quote, flags, out);
+
+  if ( rc )
+    rc = PL_release_stream(out);
+
+  return rc;
+}
+
+
+		 /*******************************
+		 *	   PRINT LENGTH		*
+		 *******************************/
+
+typedef struct limit_size_stream
+{ IOSTREAM	*stream;		/* Limited stream */
+  int64_t	 length;		/* Max size */
+} limit_size_stream;
+
+static ssize_t
+Swrite_lss(void *handle, char *buf, size_t size)
+{ limit_size_stream *lss = handle;
+  (void)buf;
+
+  if ( lss->stream->position->charno > lss->length )
+    return -1;
+
+  return size;
+}
+
+static int
+Sclose_lss(void *handle)
+{ (void)handle;
+
+  return 0;
+}
+
+static const IOFUNCTIONS lss_functions =
+{ NULL,
+  Swrite_lss,
+  NULL,
+  Sclose_lss
+};
+
+/** write_length(+Term, -Len, +Options) is det.
+
+(*) Avoid error on max_length in iso mode.  It might be nicer to get the
+option processing out of pl_write_term3(), so we can take control of the
+whole lot here more easily.
+*/
+
+static
+PRED_IMPL("write_length", 3, write_length, 0)
+{ PRED_LD
+  limit_size_stream lss;
+  int sflags = SIO_NBUF|SIO_RECORDPOS|SIO_OUTPUT|SIO_TEXT;
+  IOSTREAM *s;
+  term_t options = PL_copy_term_ref(A3);
+  term_t head = PL_new_term_ref();
+  char buf[100];
+
+  lss.length = PLMAXINT;
+  while(PL_get_list(options, head, options))
+  { atom_t name;
+    int arity;
+
+    if ( PL_get_name_arity(head, &name, &arity) &&
+	 name == ATOM_max_length && arity == 1 )
+    { term_t a = PL_new_term_ref();
+
+      _PL_get_arg(1, head, a);
+      if ( !PL_get_int64_ex(a, &lss.length) )
+	return FALSE;
+    }
+  }
+
+  if ( (s = Snew(&lss, sflags, (IOFUNCTIONS *)&lss_functions)) )
+  { int64_t len;
+    int rc;
+    pl_features_t oldmask = LD->prolog_flag.mask; /* (*) */
+
+    lss.stream = s;
+    s->encoding = ENC_UTF8;
+    Ssetbuffer(s, buf, sizeof(buf));
+    s->flags |= SIO_USERBUF;
+
+    clearPrologFlagMask(PLFLAG_ISO);
+    pushOutputContext();
+    Scurout = s;
+    rc = pl_write_term3(0, A1, A3);
+    popOutputContext();
+    LD->prolog_flag.mask = oldmask;
+
+    if ( rc && s->position->charno <= lss.length )
+    { len = s->position->charno;
+    } else
+    { len = -1;
+      if ( s->position->charno > lss.length )
+	PL_clear_exception();
+    }
+
+    Sclose(s);
+    if ( len >= 0 )
+      return PL_unify_int64(A2, len);
+  }
+
+  return FALSE;
+}
+
 
 		 /*******************************
 		 *      PUBLISH PREDICATES	*
 		 *******************************/
 
 BeginPredDefs(write)
+  PRED_DEF("nl", 0, nl, PL_FA_ISO)
+  PRED_DEF("nl", 1, nl, PL_FA_ISO)
   PRED_DEF("$put_token", 2, put_token, 0)
+  PRED_DEF("$put_quoted", 4, put_quoted_codes, 0)
+  PRED_DEF("write_length", 3, write_length, 0)
 EndPredDefs
 
