@@ -19,69 +19,123 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-This is the source-file for  swipl-ld,   the  SWI-Prolog linker for embedded
-applications. See swipl-ld(1) for details.  Feel   free  to comment and send
-contributions.
+/*
+  Preprocessor symbols that may be defined during the compilation
+  of this file:
 
-The    file    pl-extend.c,    copied    by    the    installation    to
-<home>/include/stub.c  contains  a  minimal  skeleton  for  creating  an
-embedded application.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+  #ifdef __WINDOWS__
+    This file (swipl-ld.c) is being compiled by a Windows compiler
+    (either MSC or gcc/MinGW).  The resulting swipl-ld* executable
+    will thus run under Windows.
+  #endif
+
+  #ifdef HOST_TOOLCHAIN_MSC
+    The generated swipl-ld* executable will use the Microsoft C/C++
+    compiler to produce shared objects and executables.
+  #endif
+
+  #ifdef HOST_TOOLCHAIN_MINGW
+    The generated swipl-ld* executable will use the MinGW compilers
+    to produce shared objects and executables.
+  #endif
+
+  #ifdef HOST_OS_WINDOWS
+    The generated swipl-ld* executable will produce shared objects
+    and executable that will run under Windows.
+  #endif
+ */
+
+#ifndef __WINDOWS__
+#if defined(_MSC_VER) || defined(__MINGW32__)
+#define __WINDOWS__ 1
+#endif
+#endif
 
 #define UNQUOTED_PREFIX "\1"
 
+#include "pl-incl.h"
+
 #ifdef __WINDOWS__
-#if (_MSC_VER >= 1400)			/* VC8 */
-#ifdef WIN64
-#include <config/win64.h>
-#else
-#include <config/win32.h>
-#endif
-#endif
-
-#if (_MSC_VER < 1300)			/* <= VC6 */
-typedef long intptr_t;
-typedef unsigned long uintptr_t;
-#endif
-
 #include <process.h>
 #include <io.h>
+
 #if (_MSC_VER < 1400)
 #define off_t intptr_t
 #endif
+
+#ifdef _MSC_VER
+#define HOST_TOOLCHAIN_MSC
+
 #define popen _popen
 #define pclose _pclose
+
 #define O_WRONLY _O_WRONLY
 #define O_RDONLY _O_RDONLY
 #define O_CREAT _O_CREAT
 #define O_TRUNC _O_TRUNC
 #define O_BINARY _O_BINARY
+#endif
+
+#endif /*__WINDOWS__*/
+
+#if defined(__WINDOWS__)
+#define HOST_OS_WINDOWS 1
+#define DEF_PROG_PL "swipl.exe"
+#define PROG_OUT "a.exe"
+#else
+#define DEF_PROG_PL "swipl"
+#define PROG_OUT "a.out"
+#endif
 
 #ifndef PROG_PL
-#define PROG_PL "swipl.exe"
+#define PROG_PL DEF_PROG_PL
 #endif
-#define PROG_LD "link.exe"
-#define SO_LD "link.exe"
+
+#if defined(HOST_TOOLCHAIN_MSC)
+
 #define PROG_CC "cl.exe /MD"
 #define PROG_CXX "cl.exe /MD /GX"
-#define PROG_OUT "plout.exe"
+/* PROG_CPP is defined in config.h: we need to redefine it. */
+#undef PROG_CPP
 #define PROG_CPP "cl.exe -P"
-#define LIB_PL_DEBUG "libswiplD.lib"
-#define EXT_OBJ "obj"
-#define OPT_DEBUG "/DEBUG"
-#else /*__WINDOWS__*/
-#include "pl-incl.h"
 
-#ifndef PROG_PL
-#define PROG_PL "swipl"
+#define PROG_LD "link.exe"
+/* SO_LD is defined in config.h: we need to redefine it. */
+#undef SO_LD
+#define SO_LD "link.exe"
+
+#define EXT_OBJ "obj"
+
+#define LIB_PL_DEBUG "swiplD.lib"
+#define OPT_DEBUG "/DEBUG"
+
+#elif defined(HOST_TOOLCHAIN_MINGW)
+
+#define PROG_CC    "gcc"
+#define PROG_CXX   "g++"
+#define PROG_LD    PROG_CC
+#define EXT_OBJ    "obj"
+#define OPT_DEBUG  "-g"
+#define SO_LDFLAGS "-shared"
+#undef SO_LD
+#define SO_LD	   PROG_LD
+
+#else /*Native*/
+
+#define PROG_CC C_CC
+#define PROG_CXX C_CC " -x c++"
+/* PROG_CPP is defined in config.h */
+
+#define PROG_LD C_CC
+
+/* SO_LD is defined in config.h. */
+#ifndef PROG_LD
+#define PROG_LD C_CC
 #endif
-#define PROG_CC "cc"
-#define PROG_CXX "c++"
-#define PROG_OUT "a.out"
+
 #define EXT_OBJ "o"
 #define OPT_DEBUG "-g"
 
@@ -89,7 +143,8 @@ typedef unsigned long uintptr_t;
 #define SO_LDFLAGS "-shared"
 #endif
 
-#endif /*__WINDOWS__*/
+#endif /*Toolchain selection*/
+
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -134,7 +189,7 @@ typedef unsigned long uintptr_t;
 #define streq(s, q)     (strcmp((s), (q)) == 0)
 #endif
 #define strprefix(s, p) (strncmp((s), (p), strlen(p)) == 0)
-#ifdef __WINDOWS__
+#if defined(__WINDOWS__)
 #define strfeq(s, q)	(stricmp((s), (q)) == 0)
 #else
 #define strfeq(s, q)	streq(s, q)
@@ -183,7 +238,8 @@ static arglist lastlibs;		/* libs that must be at the end */
 static arglist libdirs;			/* -L library directories */
 static arglist includedirs;		/* -I include directories */
 
-static char *pllib;    			/* -lswipl, libswipl.lib, ... */
+static char *pllib;			/* -lswipl, swipl.lib, ... */
+static char *pllibs = "";		/* Requirements to link to pllib */
 
 static char *pl;			/* Prolog executable */
 static char *cc;			/* CC executable */
@@ -205,9 +261,13 @@ static char *ctmp;			/* base executable */
 static char *pltmp;			/* base saved state */
 static char *out;			/* final output */
 static int  opt_o=FALSE;		/* -o out given */
+static int  opt_E=FALSE;		/* -E given */
 
-static int nostate = FALSE;		/* do not make a state */
+static int build_defaults = FALSE;	/* don't ask Prolog for parameters*/
+
+static int nostate = TRUE;		/* do not make a state */
 static int nolink = FALSE;		/* do not link */
+static int nolibswipl = FALSE;		/* do not link with -lswipl */
 static int shared = FALSE;		/* -shared: make a shared-object/DLL */
 static char *soext;			/* extension of shared object */
 static int embed_shared = FALSE;	/* -dll/-embed-shared: embed Prolog */
@@ -522,7 +582,7 @@ typedef struct
 
 static extdef extdefs[] =
 { { EXT_OBJ,	&ofiles },
-#ifdef __WINDOWS__
+#if defined(HOST_TOOLCHAIN_MSC)
   { "lib",	&libs },
 #else
   { "a",	&libs },
@@ -563,7 +623,9 @@ dispatchFile(const char *name)
 
     for( ; d->extension; d++ )
     { if ( strfeq(d->extension, ext) )
-      { appendArgList(d->list, name);
+      { if ( d->list == &plfiles )
+	  nostate = FALSE;
+	appendArgList(d->list, name);
 	return TRUE;
       }
     }
@@ -587,10 +649,9 @@ usage()
 	  "usage: %s -help\n"
 	  "       %s [options] inputfile ...\n"
 	  "       %s -shared -o out inputfile ...\n"
-#ifdef __WINDOWS__
+#if defined(HOST_OS_WINDOWS)
 	  "       %s -dll -o out inputfile ...\n"
 #endif
-	  "       %s -E cppargument ...\n"
 	  "\n"
 	  "options:\n"
 	  "       -o out           define output file\n"
@@ -606,10 +667,15 @@ usage()
 	  "       -c++ compiler    compiler for C++ source files\n"
 	  "\n"
 	  "       -c               only compile C/C++ files, do not link\n"
+	  "       -S               emit assembler, do not link\n"
+	  "       -E               only run preprocessor, do not link\n"
+          "       -build-defaults  use default parameters, don't ask Prolog\n"
 	  "       -nostate         just relink the kernel\n"
+	  "       -state           add a Prolog saved state\n"
+	  "       -nolibswipl      do not link with -lswipl\n"
 	  "       -shared          create target for load_foreign_library/2\n"
 	  "       -embed-shared    embed Prolog in a shared object/DLL\n"
-#ifdef __WINDOWS__
+#if defined(HOST_OS_WINDOWS)
 	  "       -dll             synonym for -embed-shared\n"
 #endif
 	  "       -fpic            compile small position-independent code\n"
@@ -634,7 +700,12 @@ usage()
 	  "       -Iincludedir     Include directory (C/C++)\n"
 	  "       -Llibdir         Library directory (C/C++ link)\n"
 	  "       -llib            library (C/C++)\n",
-	plld, plld, plld, plld);
+	plld,
+        plld,
+#if defined(HOST_OS_WINDOWS)
+        plld,
+#endif
+        plld);
 
   exit(1);
 }
@@ -664,17 +735,22 @@ parseOptions(int argc, char **argv)
     { nolink++;
       appendArgList(&coptions, opt);
       appendArgList(&cppoptions, opt);
+    } else if ( streq(opt, "-E") )		/* -E */
+    { nolink++;
+      opt_E = TRUE;
+      appendArgList(&coptions, opt);
+      appendArgList(&cppoptions, opt);
     } else if ( streq(opt, "-g") )		/* -g */
     { appendArgList(&coptions, OPT_DEBUG);
       appendArgList(&cppoptions, OPT_DEBUG);
-#ifdef __WINDOWS__					/* MSVC DEBUG OPTIONS */
+#if defined(HOST_TOOLCHAIN_MSC)			/* MSVC DEBUG OPTIONS */
       appendArgList(&coptions, "/ZI");
       appendArgList(&coptions, "/Od");
       appendArgList(&cppoptions, "/ZI");
       appendArgList(&cppoptions, "/Od");
 #endif
       appendArgList(&ldoptions, OPT_DEBUG);
-#ifdef __WINDOWS__
+#ifdef LIB_PL_DEBUG
       pllib = LIB_PL_DEBUG;
 #endif
     } else if ( strprefix(opt, "-pg") )		/* -pg* */
@@ -683,7 +759,7 @@ parseOptions(int argc, char **argv)
     } else if ( streq(opt, "-g3") )		/* -g3 */
     { appendArgList(&coptions, opt);
       appendArgList(&cppoptions, opt);
-    } else if ( strprefix(opt, "gdwarf-") ) 	/* -gdwarf-* */
+    } else if ( strprefix(opt, "gdwarf-") )	/* -gdwarf-* */
     { appendArgList(&coptions, opt);
       appendArgList(&cppoptions, opt);
     } else if ( strprefix(opt, "-O") )		/* -O* */
@@ -694,12 +770,18 @@ parseOptions(int argc, char **argv)
     } else if ( strprefix(opt, "-W") )		/* -W* */
     { appendArgList(&coptions, opt);
       appendArgList(&cppoptions, opt);
-    } else if ( streq(opt, "-nostate") ) 	/* -nostate */
+    } else if ( streq(opt, "-build-defaults") )	/* -build-defaults */
+    { build_defaults = TRUE;
+    } else if ( streq(opt, "-nostate") )	/* -nostate */
     { nostate = TRUE;
+    } else if ( streq(opt, "-state") )		/* -state */
+    { nostate = FALSE;
+    } else if ( streq(opt, "-nolibswipl") )	/* -nolibswipl */
+    { nolibswipl = TRUE;
     } else if ( streq(opt, "-dll") ||		/* -dll */
 		streq(opt, "-embed-shared") )   /* -embed-shared */
     { embed_shared = TRUE;
-#ifdef __WINDOWS__
+#if defined(HOST_TOOLCHAIN_MSC)
       appendArgList(&ldoptions, "/DLL");
 #else
 #ifdef SO_pic
@@ -738,39 +820,39 @@ parseOptions(int argc, char **argv)
       appendArgList(&coptions, SO_PIC);
       appendArgList(&cppoptions, SO_PIC);
 #endif
-    } else if ( streq(opt, "-o") ) 		/* -o out */
+    } else if ( streq(opt, "-o") )		/* -o out */
     { if ( argc > 1 )
       { out = argv[1];
 	opt_o = TRUE;
 	argc--, argv++;
       } else
 	usage();
-    } else if ( streq(opt, "-goal") ) 		/* -goal goal */
+    } else if ( streq(opt, "-goal") )		/* -goal goal */
     { if ( argc > 1 )
       { plgoal = argv[1];
 	argc--, argv++;
       } else
 	usage();
-    } else if ( streq(opt, "-toplevel") ) 	/* -toplevel goal */
+    } else if ( streq(opt, "-toplevel") )	/* -toplevel goal */
     { if ( argc > 1 )
       { pltoplevel = argv[1];
 	argc--, argv++;
       } else
 	usage();
-    } else if ( streq(opt, "-initfile") ) 	/* -initfile goal */
+    } else if ( streq(opt, "-initfile") )	/* -initfile goal */
     { if ( argc > 1 )
       { plinitfile = argv[1];
 	argc--, argv++;
       } else
 	usage();
-    } else if ( streq(opt, "-F") ) 		/* -F base */
+    } else if ( streq(opt, "-F") )		/* -F base */
     { if ( argc > 1 )
       { plsysinit = argv[1];
 	argc--, argv++;
       } else
 	usage();
-    } else if ( streq(opt, "-class") ) 		/* -class runtime,kernel,
-    							  development */
+    } else if ( streq(opt, "-class") )		/* -class runtime,kernel,
+							  development */
     { if ( argc > 1 )
       { plclass = argv[1];
 	if ( !streq(plclass, "runtime") &&
@@ -781,25 +863,25 @@ parseOptions(int argc, char **argv)
 	argc--, argv++;
       } else
 	usage();
-    } else if ( streq(opt, "-pl") ) 		/* -pl prolog */
+    } else if ( streq(opt, "-pl") )		/* -pl prolog */
     { if ( argc > 1 )
       { pl = argv[1];
 	argc--, argv++;
       } else
 	usage();
-    } else if ( streq(opt, "-cc") ) 		/* -cc compiler */
+    } else if ( streq(opt, "-cc") )		/* -cc compiler */
     { if ( argc > 1 )
       { cc = argv[1];
 	argc--, argv++;
       } else
 	usage();
-    } else if ( streq(opt, "-c++") ) 		/* -c++ compiler */
+    } else if ( streq(opt, "-c++") )		/* -c++ compiler */
     { if ( argc > 1 )
       { cxx = argv[1];
 	argc--, argv++;
       } else
 	usage();
-    } else if ( streq(opt, "-ld") ) 		/* -ld linker */
+    } else if ( streq(opt, "-ld") )		/* -ld linker */
     { if ( argc > 1 )
       { ld = argv[1];
 	argc--, argv++;
@@ -865,20 +947,13 @@ tmpPath(char **store, const char *base)
 }
 
 
-#ifndef SO_LD
-#define SO_LD cc
-#endif
-#ifndef PROG_LD
-#define PROG_LD cc
-#endif
-
 static void
 fillDefaultOptions()
 { char tmp[1024];
   char *defcxx = PROG_CXX;
 
   defaultProgram(&cc,  PROG_CC);
-  if ( streq(cc, "gcc") )
+  if ( streq(cc, "gcc") )			/* TBD: MINGW */
     defcxx = "g++";
   defaultProgram(&cxx, defcxx);
 
@@ -889,7 +964,7 @@ fillDefaultOptions()
       ld = cxx;
   }
 
-#ifdef __WINDOWS__
+#if defined(HOST_TOOLCHAIN_MSC)
   if ( strcmp(LIB_PL_DEBUG,pllib) == 0 )
     ensureOption(&coptions, "/MDd");
   else ensureOption(&coptions, "/MD");
@@ -906,9 +981,9 @@ fillDefaultOptions()
   free(ctmp);
   ctmp = strdup(tmp);
 #endif
-#if defined(__WINDOWS__) || defined(__CYGWIN__)
+#if defined(HOST_OS_WINDOWS)
 /* Saved states have the .exe extension under Windows */
-  replaceExtension(pltmp, embed_shared ? "dll" : "exe", tmp);
+  replaceExtension(pltmp, "exe", tmp);
   free(pltmp);
   pltmp = strdup(tmp);
 #endif
@@ -916,11 +991,9 @@ fillDefaultOptions()
   { fprintf(stderr, "%s: \"-o out\" required for linking shared object\n", plld);
     exit(1);
   }
-#if defined(__CYGWIN__)
-  /* Tack correct suffix to out, otherwise cat in createOutput fails. */
+#if defined(HOST_OS_WINDOWS)
   if ( out && !nolink )
   { replaceExtension(out, shared || embed_shared ? "dll" : "exe", tmp);
-    /* Dont call free(out), it could be an argv member. */
     out = strdup(tmp);
   }
 #endif
@@ -978,12 +1051,8 @@ getPrologOptions()
 	  defaultPath(&plbase, v);
 	else if ( streq(name, "PLARCH") )
 	  defaultPath(&plarch, v);
-#ifdef __CYGWIN__
 	else if ( streq(name, "PLLIBS") )	/* Always required. */
-#else
-	else if ( streq(name, "PLLIBS") && !shared )
-#endif
-	  addOptionString(v);
+	  pllibs = strdup(v);
 	else if ( streq(name, "PLLIB") )
 	  defaultProgram(&pllib, v);
 	else if ( streq(name, "PLLDFLAGS") )
@@ -1013,7 +1082,7 @@ getPrologOptions()
 
     pclose(fd);
 
-#ifdef __WINDOWS__
+#if defined(__WINDOWS__) && defined(HOST_OS_WINDOWS)
     sprintf(buf, "%s/bin/%s", plbase, PROG_PL);
 #else
     sprintf(buf, "%s/bin/%s/%s", plbase, plarch, PROG_PL);
@@ -1109,8 +1178,9 @@ compileFile(const char *compiler, arglist *options, const char *cfile)
       strcpy(ext, EXT_OBJ);
   }
 
-  prependArgList(args, "-c");
-#ifdef __WINDOWS__
+  if ( !opt_E )
+    prependArgList(args, "-c");
+#if defined(HOST_OS_WINDOWS)
   appendArgList(args, "-D__WINDOWS__");
   appendArgList(args, "-D_WINDOWS");
 #endif
@@ -1119,8 +1189,10 @@ compileFile(const char *compiler, arglist *options, const char *cfile)
     appendArgList(args, "-D__SWI_EMBEDDED__");
   concatArgList(args, "-I", &includedirs);
 
-  appendArgList(args, "-o");
-  appendArgList(args, ofile);
+  if ( opt_o || !opt_E )
+  { appendArgList(args, "-o");
+    appendArgList(args, ofile);
+  }
   appendArgList(args, cfile);
 
   callprog(compiler, args);
@@ -1141,7 +1213,7 @@ compileObjectFiles()
     compileFile(cxx, &cppoptions, cppfiles.list[n]);
 }
 
-#ifdef __WINDOWS__
+#if defined(HOST_TOOLCHAIN_MSC)
 char *
 os_path(char *out, const char *in)
 { for(; *in; in++)
@@ -1182,36 +1254,42 @@ void
 linkBaseExecutable()
 { char *cout = out;
 
-#ifndef __WINDOWS__				/* bit of a hack ... */
+#if !defined(HOST_OS_WINDOWS)			/* bit of a hack ... */
   if ( embed_shared )
   { linkSharedObject();
     return;
   }
 #endif
 
-#ifdef __WINDOWS__
+#if defined(HOST_TOOLCHAIN_MSC)
 { char tmp[MAXPATHLEN];
   sprintf(tmp, "/out:%s", cout);
   prependArgList(&ldoptions, tmp);
 }
   concatArgList(&ldoptions, "", &ofiles);	/* object files */
   exportlibdirs();
-  appendArgList(&ldoptions, pllib);		/* -lpl */
+  if ( !nolibswipl )
+  { appendArgList(&ldoptions, pllib);		/* -lswipl */
+    addOptionString(pllibs);
+  }
   concatArgList(&ldoptions, "", &libs);		/* libraries */
   concatArgList(&ldoptions, "", &lastlibs);	/* libraries */
-#else /*__WINDOWS__*/
+#else /* !defined(HOST_TOOLCHAIN_MSC) */
   prependArgList(&ldoptions, cout);
   prependArgList(&ldoptions, "-o");		/* -o ctmp */
   concatArgList(&ldoptions, "", &ofiles);	/* object files */
   concatArgList(&ldoptions, "-L", &libdirs);    /* library directories */
-  appendArgList(&ldoptions, pllib);		/* -lpl */
+  if ( !nolibswipl )
+  { appendArgList(&ldoptions, pllib);		/* -lswipl */
+    addOptionString(pllibs);
+  }
   concatArgList(&ldoptions, "", &libs);		/* libraries */
   concatArgList(&ldoptions, "", &lastlibs);	/* libraries */
-#endif
+#endif /* !defined(HOST_TOOLCHAIN_MSC) */
 
   if ( !nostate )
   {
-#ifdef __WINDOWS__
+#if defined(HOST_TOOLCHAIN_MSC)
     if ( !embed_shared )
     { char buf[MAXPATHLEN];
       appendArgList(&tmpfiles, replaceExtension(cout, "exp", buf));
@@ -1238,7 +1316,7 @@ linkSharedObject()
   { soout = replaceExtension(out, soext, soname);
   }
 
-#ifdef __WINDOWS__
+#if defined(HOST_TOOLCHAIN_MSC)
   prependArgList(&ldoptions, "/dll");
 { char tmp[MAXPATHLEN];
   sprintf(tmp, "/out:%s", soout);
@@ -1246,16 +1324,22 @@ linkSharedObject()
 }
   concatArgList(&ldoptions, "", &ofiles);	/* object files */
   exportlibdirs();
-  appendArgList(&ldoptions, pllib);		/* libpl.lib */
+  if ( !nolibswipl )
+  { appendArgList(&ldoptions, pllib);		/* swipl.lib */
+    addOptionString(pllibs);
+  }
   concatArgList(&ldoptions, "", &libs);		/* libraries */
   concatArgList(&ldoptions, "", &lastlibs);	/* libraries */
-#else /*__WINDOWS__*/
+#else /* !defined(HOST_TOOLCHAIN_MSC) */
 #ifdef __CYGWIN__
   prependArgList(&ldoptions, SO_LDFLAGS);
   prependArgList(&ldoptions, soout);
   prependArgList(&ldoptions, "-o");		/* -o ctmp */
   concatArgList(&ldoptions, "", &ofiles);	/* object files */
-  appendArgList(&ldoptions, pllib);		/* -lpl */
+  if ( !nolibswipl )
+  { appendArgList(&ldoptions, pllib);		/* -lswipl */
+    addOptionString(pllibs);
+  }
   concatArgList(&ldoptions, "-L", &libdirs);    /* library directories */
   concatArgList(&ldoptions, "", &libs);		/* libraries */
   concatArgList(&ldoptions, "", &lastlibs);	/* libraries */
@@ -1273,18 +1357,18 @@ linkSharedObject()
 #endif /*SO_FORMAT_LDFLAGS*/
   concatArgList(&ldoptions, "", &ofiles);	/* object files */
   concatArgList(&ldoptions, "-L", &libdirs);    /* library directories */
-#ifndef O_SHARED_KERNEL
-  if ( embed_shared )
-#endif
-  { appendArgList(&ldoptions, pllib);		/* -lpl */
-  }
   concatArgList(&ldoptions, "", &libs);		/* libraries */
+#ifdef O_SHARED_KERNEL
+  if ( !nolibswipl )
+#endif
+  { appendArgList(&ldoptions, pllib);		/* -lswipl */
+  }
   concatArgList(&ldoptions, "", &lastlibs);	/* libraries */
 #ifdef __BEOS__
   appendArgList(&ldoptions, plexe);		/* last is executable */
 #endif
 #endif /*__CYGWIN__*/
-#endif /*__WINDOWS__*/
+#endif /* !defined(HOST_TOOLCHAIN_MSC) */
 
   callprog(ld, &ldoptions);
 }
@@ -1381,11 +1465,11 @@ createSavedState()
 static void
 copy_fd(int i, int o)
 { char buf[CPBUFSIZE];
-  int n;
+  ssize_t n;
 
   while( (n=read(i, buf, sizeof(buf))) > 0 )
   { while( n > 0 )
-    { int n2;
+    { ssize_t n2;
 
       if ( (n2 = write(o, buf, n)) > 0 )
       { n -= n2;
@@ -1403,7 +1487,7 @@ copy_fd(int i, int o)
 }
 
 
-#ifdef __WINDOWS__
+#if defined(HOST_TOOLCHAIN_MSC)
 void
 saveExportLib()
 { char ibuf[MAXPATHLEN];
@@ -1425,7 +1509,7 @@ saveExportLib()
     }
   }
 }
-#endif /*__WINDOWS__*/
+#endif /* defined(HOST_TOOLCHAIN_MSC) */
 
 
 void
@@ -1434,7 +1518,7 @@ createOutput()
 
   if ( verbose )
   {
-#ifdef __WINDOWS__
+#if defined(HOST_TOOLCHAIN_MSC)
     printf("\tcopy /b %s+%s %s\n", out, pltmp, out);
 #else
     printf("\tcat %s >> %s\n", pltmp, out);
@@ -1538,19 +1622,6 @@ main(int argc, char **argv)
     special = 2;
   else
     special = 0;
-					/* swipl-ld [-pl x] -E ...: behave as cpp */
-  if ( argc-special > 0 && streq(argv[special], "-E") )
-  { arglist cppoptions;
-    int i;
-
-    memset(&cppoptions, 0, sizeof(cppoptions));
-    for(i=special+1 ; i < argc; i++)
-      appendArgList(&cppoptions, argv[i]);
-
-    callprog(PROG_CPP, &cppoptions);
-
-    return 0;
-  }
 					  /* swipl-ld [-pl x] -v: verbose */
   if ( argc-special == 1 && streq(argv[special], "-v") )
   { arglist coptions;
@@ -1567,7 +1638,36 @@ main(int argc, char **argv)
 
   parseOptions(argc, argv);
   defaultProgram(&pl, PROG_PL);
-  getPrologOptions();
+
+  if ( build_defaults )
+  { nostate = TRUE;			/* not needed and Prolog won't run */
+    defaultProgram(&cc, C_CC);
+#ifdef PLBASE
+    defaultPath(&plbase, PLBASE);
+#else
+    defaultPath(&plbase, PLHOME);
+#endif
+    defaultPath(&plarch, PLARCH);
+    defaultProgram(&pllib, C_PLLIB);
+    addOptionString(C_LIBS);
+    appendArgList(&ldoptions, C_LDFLAGS);
+    appendArgList(&coptions, C_CFLAGS);
+    appendArgList(&cppoptions, C_CFLAGS);
+#ifdef SO_EXT
+    soext = strdup(SO_EXT);
+#endif
+#ifdef O_PLMT
+    ensureOption(&coptions, "-D_REENTRANT");
+    ensureOption(&cppoptions, "-D_REENTRANT");
+#ifdef _THREAD_SAFE			/* FreeBSD */
+    ensureOption(&coptions, "-D_THREAD_SAFE");
+    ensureOption(&cppoptions, "-D_THREAD_SAFE");
+#endif
+#endif
+  } else
+  { getPrologOptions();
+  }
+
   fillDefaultOptions();
 
   if ( show_version )
