@@ -1,11 +1,10 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker and Anjo Anjewierden
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2009, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -19,7 +18,7 @@
 
     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
 %:- set_prolog_flag(optimise, true).
@@ -97,6 +96,12 @@ syntax(char-2) :-
 	52 == 0'\x34.
 syntax(char-3) :-
 	"\\" =:= 0'\\.
+syntax(char-4) :-
+	1-48 == 1-0'0.
+syntax(cannot_start_term-1) :-
+	catch(term_to_atom(_T, 'p(]'), E, true),
+	E = error(syntax_error(cannot_start_term), _).
+
 
 :- op(100, yf, af).
 
@@ -104,8 +109,9 @@ syntax(string-1) :-
 	'\c ' == ''.
 syntax(string-2) :-
 	'x\c y' == xy.
-syntax(quote-1) :-			% not really clear this is ok
-	'\x' == x.
+syntax(quote-1) :-
+	catch(atom_to_term('\'\\x\'', _, _), E, true),
+	E = error(syntax_error(undefined_char_escape(x)), _).
 syntax(quote-2) :-
 	'\x61' == a.
 syntax(quote-3) :-
@@ -147,10 +153,18 @@ syntax(latin-1) :-
 		 *	       WRITE		*
 		 *******************************/
 
+:- op(200, fx, op_fa).
+
 write_test(q-1) :-
-	term_to_atom(-(0), X), X == '-(0)'.
+	T = -(0),
+	term_to_atom(T, X),
+	term_to_atom(T2, X),
+	T == T2.
 write_test(q-2) :-
-	term_to_atom(+(0), X), X == '+(0)'.
+	T = +(0),
+	term_to_atom(T, X),
+	term_to_atom(T2, X),
+	T == T2.
 write_test(q-3) :-
 	term_to_atom(+(a), X), X == '+a'.
 write_test(q-4) :-
@@ -160,12 +174,21 @@ write_test(q-5) :-
 write_test(q-6) :-
 	term_to_atom('*/*', X), X == '*/*'.
 write_test(q-7) :-
-	term_to_atom(p((0|a)), X), X == 'p((0 \'|\' a))'.
+	term_to_atom(p((0|a)), X), X == 'p((0|a))'.
 write_test(q-8) :-
-	term_to_atom(p((a|b)), X), X == 'p((a\'|\'b))'.
+	term_to_atom(p((a|b)), X), X == 'p((a|b))'.
+write_test(op-1) :-
+	term_to_atom(-((a,b)), X), X == '- (a,b)'.
+write_test(op-2) :-
+	term_to_atom(op_fa((a,b)), X), X == 'op_fa (a,b)'.
+write_test(op-3) :-
+	term_to_atom(dynamic((a,b)), X), X == 'dynamic a,b'.
 write_test(c-1) :-
 	T = [a,b,c|T],
-	term_to_atom(T, X), X == '[a,b,c|**]'.
+	term_to_atom(T, X),
+	term_to_atom(@(T2,S2), X),
+	maplist(call, S2),
+	T2 =@= T.
 
 
 		 /*******************************
@@ -360,8 +383,23 @@ arithmetic(arith-12) :-
 arithmetic(arith-13) :-
 	1.0 =:= sin(pi/2).
 arithmetic(sign-1) :-
-	-1 =:= sign(-1),   0 =:= sign(0),   1 =:= sign(1),
-	-1 =:= sign(-1.5), 0 =:= sign(0.0), 1 =:= sign(pi).
+	-1   is sign(-1),   0   is sign(0),   1   is sign(1).
+arithmetic(sign-2) :-
+	-1.0 is sign(-1.5), 0.0 is sign(0.0), 1.0 is sign(pi).
+arithmetic(sign-3) :-
+	0.0 is sign(-0.0).
+arithmetic(copysign-1) :-
+	-1 is copysign(1, -0.0),
+	 1 is copysign(1, 0.0).
+arithmetic(copysign-2) :-
+	-2.0 is copysign(2.0, -0.0),
+	2.0 is copysign(2.0, 0.0).
+arithmetic(copysign-3) :-
+	1  is copysign(1, 1),
+	-1 is copysign(1, -1).
+arithmetic(abs-1) :-
+	A is abs(-0.0),
+	A == 0.0.
 arithmetic(floor-1) :-
 	0 is floor(0.0),
 	0 is floor(0.9),
@@ -444,49 +482,6 @@ floattest(float-8) :-
 	E = error(evaluation_error(float_overflow), _).
 
 
-		 /*******************************
-		 *	 PROLOG FUNCTIONS	*
-		 *******************************/
-
-:- arithmetic_function(ten/0).
-:- arithmetic_function(twice/1).
-:- arithmetic_function(mean/2).
-:- arithmetic_function(euler/0).
-:- arithmetic_function(fail/0).
-:- arithmetic_function(except/0).
-
-ten(10).
-twice(X, R) :-
-	R is X * 2.
-mean(X1, X2, R) :-
-	R is (X1 + X2)/2.
-
-euler(2.71828).
-
-fail(_) :- fail.
-
-except(_) :-
-	throw(error(foobar)).
-
-arithmetic_functions(func-1) :-
-	A is ten, A =:= 10.
-arithmetic_functions(func-2) :-
-	A is twice(5), A =:= 10.
-arithmetic_functions(func-3) :-
-	A is mean(0, 20), A =:= 10.
-arithmetic_functions(func-4) :-
-        Exp = 6*euler*7*1,		% test functions corrupting stack
-        EE is Exp,
-	EE =:= 6*euler*7*1.
-arithmetic_functions(fail-1) :-
-	catch(_ is fail, E, true),
-	compound(E), E = error(E2, _),
-	compound(E2), E2 = failure_error(_).
-arithmetic_functions(except-1) :-
-	catch(_ is except, E, true),
-	E == error(foobar).
-
-
 		 /*******************************
 		 *     UNBOUNDED ARITHMETIC	*
 		 *******************************/
@@ -520,13 +515,6 @@ ratp(Count, In, Out) :-
         T is In + (In rdiv 2),
 	ratp(Count0, T, Out).
 
-:- arithmetic_function(idiv/2).
-
-idiv(Dd,Dr,Iq):-
-        Q is Dd/Dr,
-        rational(Q,Qt,Qn),
-        Iq is Qt//Qn.
-
 dec(X, Y) :-
 	Y is X - 1.
 
@@ -541,6 +529,9 @@ gmp(neg-1) :-				% check conversion of PLMININT
 	-A =:= 9223372036854775808.
 gmp(neg-2) :-
 	A is -(1<<63+1), A == -9223372036854775809.
+gmp(neg-promote) :-
+	A is 0 - -9223372036854775808,
+	A =:= 9223372036854775808.
 gmp(abs-1) :-
 	A is -9223372036854775808,
 	abs(A) =:= 9223372036854775808.
@@ -647,6 +638,14 @@ gmp(rational-2) :-
 gmp(rational-3) :-
 	dec(6 rdiv 5, X),
 	X == 1 rdiv 5.
+gmp(rational-4) :-
+	X is 5 rdiv 3,
+	X == (5 rdiv 3).
+gmp(rational-5) :-
+	5 rdiv 3 is 5 rdiv 3.
+gmp(rational-6) :-
+	5 rdiv X is 5 rdiv 3,
+	X == 3.
 gmp(rationalize-1) :-
 	A is rationalize(0.0), A == 0,
 	B is rationalize(0.1), B == 1 rdiv 10,
@@ -730,9 +729,6 @@ gmp(fmtf-1) :-
 	ratp(999, 1, X),
 	format(atom(S), '~5f', [X]),
 	sub_atom(S, _, _, 0, '935376.65824').
-gmp(idiv-1) :-
-	Qi is idiv(3 rdiv 2,2 rdiv 5),
-	Qi == 3.
 gmp(random) :-
 	A is random((1<<200)-((1<<200)-20)),
 	A < 20.
@@ -794,6 +790,8 @@ wchars(cmp-2) :-
 
 foo:hello(world).
 
+ten(10).
+
 meta(call-1) :-
 	call(ten(X)),
 	X == 10.
@@ -828,7 +826,8 @@ meta(call-11) :-
 	forall(A, true),
 	flag(a, 3, Old).
 meta(call-12) :-
-	catch(call(1), E, true),
+	G = 1,				% avoid in-line expansion
+	catch(call(G), E, true),
 	error(E, type_error(callable, _)).
 meta(apply-1) :-
 	apply(=, [a,a]).
@@ -889,7 +888,8 @@ cleanup(clean-7) :-
 	catch(call_cleanup(fail, throw(b)), E, true),
 	E == b.
 cleanup(clean-8) :-			% check handling of CHP_TOP
-	'$break'(call_cleanup(true, true)).
+					% notrace/1 does a call-back via C
+	notrace(call_cleanup(true, true)).
 cleanup(clean-9) :-
 	retractall(clean_rval(_)),
 	call_cleanup(bagof(x, cleanup_1, _Xs), Reason,
@@ -927,24 +927,24 @@ dl_fail(N) :-
 	NN is N - 1,
 	dl_fail(NN).
 
-:- arithmetic_function(fac/1).
-
 depth_limit(depth-1) :-
-	call_with_depth_limit(dl_det(1), 10, 1),
+	G = dl_det(1),
+	call_with_depth_limit(G, 10, 1),
 	deterministic(true).
 depth_limit(depth-2) :-
-	call_with_depth_limit(dl_det(10), 10, 10).
+	G = dl_det(10),
+	call_with_depth_limit(G, 10, 10).
 depth_limit(depth-3) :-
-	call_with_depth_limit(dl_det(10), 9, depth_limit_exceeded).
+	G = dl_det(10),
+	call_with_depth_limit(G, 9, depth_limit_exceeded).
 depth_limit(ndet-1) :-
+	G = dl_ndet(5),
 	findall(X,
-		call_with_depth_limit(dl_ndet(5), 10, X),
+		call_with_depth_limit(G, 10, X),
 		L),
 	L = [5, depth_limit_exceeded].
 depth_limit(fail-1) :-
 	\+ call_with_depth_limit(dl_fail(2), 10, _).
-depth_limit(arith-1) :-
-	call_with_depth_limit(_A is fac(10), 8, depth_limit_exceeded).
 
 
 		 /*******************************
@@ -1081,18 +1081,17 @@ sets(setof-2) :-
 		   compound(_A1, _B1)-[1, 2]]
 	).
 sets(vars-1) :-
-	'$e_free_variables'(A^satisfy(B^C^(setof(D:E,
-						 (country(E), area(E, D)),
-						 C),
-					   aggregate(max, C, B),
-					   in(B, A),
-					   {place(A)})),
-			    Free),
-	Free == v(D, E).
+	'$free_variable_set'(X^(m:Y^hello(X,Y)), G, V),
+	G == m:hello(X,Y),
+	V == v.
 sets(bagof-1) :-
 	List = [_,_,_],
 	bagof(X, member(X, List), Xs),
 	Xs == List.
+sets(bagof-2) :-
+	Goal = member(_, [1,2]),
+	bagof(Goal, Goal, Xs),
+	Xs == [member(1, [1,2]), member(2, [1,2])].
 sets(setof-3001) :-
 	List = [_,_,_],
 	setof(X, member(X, List), Xs),
@@ -1183,6 +1182,10 @@ atom_handling(current-1) :-
 	member(foobar, Atoms),
 	length(Atoms, L),
 	L > 100.			% else something is wrong!
+atom_handling(complete-1) :-
+	'$atom_completions'(stat, List),
+	length(List, Len),
+	Len > 2.
 
 
 		 /*******************************
@@ -1437,11 +1440,6 @@ flag(arith-1) :-
 	flag(f, V, V+1),
 	flag(f, NV, Old),
 	NV == 1.
-flag(arith-2) :-
-	flag(f, Old, 100),
-	flag(f, V, mean(V, 0)),
-	flag(f, NV, Old),
-	NV == 50.
 
 
 		 /*******************************
@@ -1637,9 +1635,11 @@ gc(gc-3) :-
 gc(gc-4) :-
 	catch(_,_,garbage_collect).
 gc(gc-5) :-
-	catch(25,_,garbage_collect).
+	G = 25,
+	catch(G,_,garbage_collect).
 gc(gc-6) :-
-	catch(1.25,_,garbage_collect).
+	G = 1.25,
+	catch(G,_,garbage_collect).
 gc(agc-1) :-
 	garbage_collect_atoms.
 gc(agc-2) :-		% not if concurrent: this is too simple.  There
@@ -1930,8 +1930,10 @@ copy_term(av-3) :-
 	A = f(Z),
 	Y == Z.
 copy_term(av-4) :-
-	freeze(X, true),
-	freeze(X, Done = true),
+	G1 = true,
+	G2 = (Done = true),
+	freeze(X, G1),
+	freeze(X, G2),
 	copy_term(X, Y),
 	X = ok,
 	Done == true,
@@ -1993,18 +1995,18 @@ term_hash(simple-5) :-
 term_hash(compound-1) :-
 	term_hash(hello(world), X),
 	memberchk(X, [ 2512014,		% little endian
-		       13811310		% big endian	FIXME: Incorrect
+		       4285241		% big endian
 		     ]).
 term_hash(compound-2) :-
 	A = x(a),
 	term_hash(hello(A, A), X),
 	memberchk(X, [ 6171734,		% little endian
-		       13137004		% big endian	FIXME: Incorrect
+		       13034251		% big endian
 		     ]).
 term_hash(compound-3) :-
 	term_hash(hello(x(a), x(a)), X),
 	memberchk(X, [ 6171734,		% little endian
-		       13137004		% big endian	FIXME: Incorrect
+		       13034251		% big endian
 		     ]).
 
 
@@ -2117,13 +2119,10 @@ popen(cat-2) :-
 	->  Cmd = 'cmd /c rem true'
 	;   Cmd = true
 	),
-	absolute_file_name(swi('library/MANUAL'), Manual),
-	open(Manual, read, Fd),
 	open(pipe(Cmd), write, Pipe),
-	catch(copy_stream_data(Fd, Pipe),
+	catch(forall(between(1, 10000, _), format(Pipe, '0123456789~n', [])),
 	      E,
 	      true),
-	close(Fd),
 	catch(close(Pipe), _, true),	% ???
 	(   var(E)
 	->  format(user_error, 'No exception?~n', []),
@@ -2266,7 +2265,7 @@ file(absfile-2) :-			% canoniseDir() caching issues
 	delete_directory(Y),
 	atom_concat(_, YF, Abs2).
 file(ext-2) :-
-	\+ file_name_extension(foo, _, 'bar.pl'). 	% Bug#69
+	\+ file_name_extension(foo, _, 'bar.pl').	% Bug#69
 file(open-1) :-
 	catch(open(foobar, read, _, [lock(qqq)]), E, true),
 	E =@= error(domain_error(lock, qqq), context(system:open/4, _)).
@@ -2452,7 +2451,7 @@ thread(status-1) :-
 	thread_join(Id2, _).
 thread(create_error-1) :-
 	catch(thread_create(true, _, [local(a)]), E, true),
-	E = error(domain_error(thread_option, local(a)), _).
+	E = error(type_error(integer, a), _).
 
 
 		 /*******************************
